

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_FEATURESEXTREMAFEATURES_H
#define ADORE_FEATURESEXTREMAFEATURES_H

#include <adrcore/features/featurefilter.h>
#include <adrcore/types/featurepoint.h>
#include <adrcore/features/channelextrema.h>
#include <adrcore/imgprocessing/channel.h>
#include <adrcore/search/grid2d.h>
#include <adrcore/util/recommendations.h>
//#include <adrcore/util/printimage.h>

namespace adore
{

	namespace core
	{

		/**
		 * Finds extrema in an image using ChannelExtrema and contructs FeaturePoints
		 * consisting of subpixel (x, y) and the feature strength.
		 * A minimum distance between the features can be set.
		 * @tparam T the image type in which extrema shall be found
		 * @tparam CallbackT the type of the callback function which is called with each constructed feature
		 * @tparam ComparatorT a comparator which determins the type of extrema found
		 *  
		*/
		template <class T, class CallbackT, class ComparatorT = greater_equal<T> >
		class ExtremaFeatures
		{
			protected:
				uint fid;
				const float dxMax;
				const float dyMax;
				float minDistance;
				float minimumFraction;
				int maxPoints;
				FeatureGrid candidates;
				ChannelExtrema<T, ExtremaFeatures<T, CallbackT, ComparatorT>, ComparatorT> extrema;
			public:
				/**
				 * Constructor
				 * @param ID the feature id given to newly constructed features
				 * @param minDist the minimum distance of feature points
				 * @param minStrength the minimum feature strength
				 * @param minFraction the minimum fraction thr strength of a feature must have, relative to the strength of the strongest feature
				 * @param maxPts the maximum number of features returned. The strongest features are preferred.
				 */
				ExtremaFeatures( uint ID, float minDist = 10.0F, T minStrength = EFDefaults<T>::minThresh, float minFraction = 0.00001, int maxPts = INT_MAX ):
						fid( ID ),
						dxMax( 0.5 ),
						dyMax( 0.5 ),
						minDistance( minDist ),
						minimumFraction( minFraction ),
						maxPoints( maxPts ),
						candidates( false ),//don't delete points
						extrema( minStrength )
				{
					;
				}


				/**
				 * Destructor
				 */
				~ExtremaFeatures()
				{
					;
				}

				/**
				 * set the minimum distance allowed for feature points.
				 * @param theValue new distance.
				 */
				void setMinDistance( float theValue )
				{
					minDistance = theValue;
				}

				/**
				 * get the current minimum distance allowed for feature points.
				 * @return current minimum distance.
				 */
				float getMinDistance() const
				{
					return minDistance;
				}

				/**
				 * callback for ChannelExtrema
				 * This constructs the FeaturePoints and adds them to the local collection of candidates.
				 * @param img the image
				 * @param T a pointer to the local maximum
				 */
				inline void operator()( const Channel<T> &img, T *localMaximum )
				{
					uint x, y;
					img.getCoors( localMaximum, x, y );
					FeaturePoint *pt = new FeaturePoint( ( double ) x,( double ) y, fid );
					pt->setStrength( double ( *localMaximum ) );//TODO - this is approximately
					candidates.add( pt );

					//test the correct sequence
//#ifndef NDEBUG
//					printf( "%d ponts in candidates\n", candidates.getSize() );
//					FeatureFilter::FeatureGrid::Iterator it = candidates.getBegin();
//					FeatureFilter::FeatureGrid::Iterator endIt = candidates.getEnd();
//					FeaturePoint *pt1, *pt2;
//
//					pt2 = *it; ++it;
//					while ( it != endIt )
//					{
//						pt1 = pt2;
//						pt2 = *it;
//						assert( pt1->getStrength() > pt2->getStrength() );
//						++it;
//					}
//					pressKey();
//#endif


				}

				/**
				 * initiates the feature calculation.
				 * @param img the image whose maxima shall be computed
				 * @param callback a callback called for each constructed feature
				 */
				void getFeatures( Channel<T> &img, CallbackT &callback );

		};

		template<class T, class CallbackT, class ComparatorT>
		void ExtremaFeatures<T, CallbackT, ComparatorT>::getFeatures( Channel<T> &img, CallbackT &callback )
		{
			candidates.init( 0, 0, img.getWidth(), img.getHeight() );

			extrema( img, *this );

			DBGOUT( DBGNORMAL, "ExtremaFeatures: found %d maxima\n", candidates.getSize() );

			filterFeatures( candidates, minDistance, minimumFraction, maxPoints );

			DBGOUT( DBGNORMAL, "ExtremaFeatures: after dist/str filtering %d maxima\n", candidates.getSize() );

			adjustFeatures( candidates, img );

			//determin subpixel posistion
			FeatureGrid::Iterator it = candidates.getBegin();
			FeatureGrid::Iterator endIt = candidates.getEnd();

			while ( it != endIt )
			{
				callback( *it );
				++it;
			}
		}




	}

}

#endif
