

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_IMGPROCESSINGDRAWINGELLI_H
#define ADORE_IMGPROCESSINGDRAWINGELLI_H

#include <adrcore/types/aarray.h>
#include <adrcore/types/view.h>
#include <adrcore/types/nview.h>
#include <adrcore/util/mathutil.h>
#include <adrcore/util/calctype.h>
#include <adrcore/util/templatelimits.h>
#include <adrcore/imgprocessing/bilinearinterpolator.h>

namespace adore
{

	namespace core
	{

		inline pair<float, float> getCovarEllipseDim( float mat[2][2] )
		{
			return pair<float, float>( 2.0F * sqrtf( mat[0][0] ), 2.0F * sqrtf( mat[1][1] ) );
		}

		template<class typeT>
		void covarianceEllipse( Channel<typeT>& chan, float mat[2][2], int posx, int posy, typeT color, float scale = 5 )
		{
			float sigma1Sqr = mat[0][0];
			float sigma2Sqr = mat[1][1];

			assert( sigma1Sqr >= 0 );//no irrational sigmas
			assert( sigma2Sqr >= 0 );

			float sigma1 = sqrtf( sigma1Sqr );
			float sigma2 = sqrtf( sigma2Sqr );

			if ( sigma1Sqr < FLT_EPSILON ) //if sigma1Sqr is 0 or nearly 0 draw a vertical line
			{
				drawLine( chan, posx, ( uint )( posy - sigma2 * scale ), posx, ( uint )( posy + sigma2 * scale ), color );
				return;
			}
			else if ( sigma2Sqr < FLT_EPSILON ) //if sigma2Sqr is 0 or nearly 0 draw a horizontal line
			{
				drawLine( chan, ( uint )( posx - sigma1 * scale ), posy, ( uint )( posx + sigma1 * scale ), posy, color );
				return;
			}

			float cov = mat[0][1];
			float p1sqr = sigma1Sqr;
			float p2sqr = sigma2Sqr;
			float sinAlpha = 0.0F;
			float cosAlpha = 1.0F;

			if ( !( ( fabs( sigma1Sqr - sigma2Sqr ) < FLT_EPSILON ) || ( fabs( cov ) < FLT_EPSILON ) ) )
			{
				float rho = cov / ( sigma1 * sigma2 );
				float alpha = 0.5F * atan2f( 2 * cov, sigma1Sqr - sigma2Sqr );

				sinAlpha = sinf( alpha );
				cosAlpha = cosf( alpha );
				printf( "alpha: %f\n", alpha );
				float cc = cosAlpha * cosAlpha;
				float ss = sinAlpha * sinAlpha;
				float divident = ( 1.0 - rho * rho );
				float rsc = 2 * rho * sinAlpha * cosAlpha / ( sigma1 * sigma2 );

				p1sqr = divident / ( cc / sigma1Sqr - rsc + ss / sigma2Sqr );
				p2sqr = divident / ( ss / sigma1Sqr + rsc + cc / sigma2Sqr );
			}

			//first point: phi == 0
			float rPhi = sqrtf( p1sqr );
			uint prevX = ( uint )( posx + rPhi * cosAlpha * scale );
			uint prevY = ( uint )( posy + rPhi * sinAlpha * scale );

			//usual case: draw a covariance ellipse limited by sigma1 and sigma2
			const float phiStep = 0.03;
			for ( float phi = phiStep; phi < 2.0 * M_PI; phi += phiStep )
			{
				float sinPhi = sin( phi );
				float cosPhi = cos( phi );

				float rPhi = 1.0 / sqrtf( ( cosPhi * cosPhi ) / p1sqr + ( sinPhi * sinPhi ) / p2sqr );

				//x1Phi and x2Phi are the 2D coordinates of the ellipse
				float x1Phi = rPhi * ( cosAlpha * cosPhi - sinAlpha * sinPhi );
				float x2Phi = rPhi * ( sinAlpha * cosPhi + cosAlpha * sinPhi );

				uint drawX = ( uint )( posx + x1Phi * scale );
				uint drawY = ( uint )( posy + x2Phi * scale );

				if ( ( drawX >= 0 ) && ( drawX < chan.getWidth() ) && ( drawY >= 0 ) && ( drawY < chan.getHeight() ) )
				{
					drawLine( chan, prevX, prevY, drawX, drawY, color );
				}
				prevX = drawX;
				prevY = drawY;
			}
		}

		template<class covarT>
        inline pair<float, float> getCovarEllipseDim( const AArray<covarT, 2>& covar )
		{
			return pair<float, float>( 2.0F * sqrtf( covar( 0, 0 ) ), 2.0F * sqrtf( covar( 1, 1 ) ) );
		}

		template<class typeT, class covarT>
        void covarianceEllipse( Channel<typeT>& chan, const AArray<covarT, 2>& covar, int posx, int posy, typeT color, float scale = 5 )
		{
			float mat[2][2];
			mat[0][0] = covar( 0, 0 );
			mat[0][1] = covar( 0, 1 );
			mat[1][0] = covar( 1, 0 );
			mat[1][1] = covar( 1, 1 );
			covarianceEllipse( chan, mat, posx, posy, color, scale );
		}

	}
}

#endif
