

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_IMGPROCESSINGEDGESAT_H
#define ADORE_IMGPROCESSINGEDGESAT_H

#include <adrcore/types/aarray.h>
#include <adrcore/imgprocessing/channel.h>

namespace adore
{

	namespace core
	{

		/**
		 * This is for an interest point detector of the kind:
		 * 1) canny edges
		 * 2) dilation by one
		 * 3) for each edge-pixel: local hough (this algo)
		 * 4) measure the cornerness (e.g. by applying harris) and put a threhld on it.
		     
		*/

		template<class inT>
		class EdgesAt
		{

				typedef AArray<float, 2> T_angles;

			public:

				typedef typename vector<int>::iterator iterator;
				typedef typename vector<int>::const_iterator const_iterator;

			public:
				EdgesAt( const Channel<inT> &_chan, int _hood = 6,
				         int _needed = 8, int _numAngleBins = 12 ):
						chan( _chan ),
						hood( _hood ),
						width( hood + hood + 1 ),
						needed( _needed ),
						numAngleBins( _numAngleBins ),
//						minAngles( Range( -hood, hood ), Range( -hood, hood ) ),
//						maxAngles( Range( -hood, hood ), Range( -hood, hood ) ),
						minAngles( width, width ),
						maxAngles( width, width ),
						PI_2( 0.5 * M_PI ),
						binStep( M_PI / float( numAngleBins ) ),
						thresh( 0 ),
						bins( numAngleBins )
				{
					initAngles();
				}

				~EdgesAt()
				{
				}

				int operator()( int x, int y );

				inT getThresh() const
				{
					return thresh;
				}

				void setThresh( const inT t )
				{
					thresh = t;
				}

				iterator begin()
				{
					return bins.begin();
				}

				iterator end()
				{
					return bins.end();
				}

				const_iterator begin() const
				{
					return bins.begin();
				}

				const_iterator end() const
				{
					return bins.end();
				}

			private:
				void initAngles();
				void operatorHelper(
				        typename Channel<inT>::Iterator2D &chanIt,
				        const typename Channel<inT>::Iterator2D &chanStop,
				        typename T_angles::iterator &maxIt,
				        typename T_angles::iterator &minIt );

				const Channel<inT> &chan;
				const int hood;
				const int width;
				const int needed;
				const int numAngleBins;
				T_angles minAngles;
				T_angles maxAngles;

				const float PI_2;
				const float binStep;

				inT thresh; //the corner strength that is needed.
				vector<int> bins;
		};

		template <class inT>
		int EdgesAt<inT>::operator()( int x, int y )
		{
			fill( bins.begin(), bins.end(), 0 );
//			bzero( bins, numAngleBins * sizeof( int ) );

//			printf( "edges at: (%d, %d), thresh: %d\n", x, y, thresh );
//			printf( "numAngleBins: %d\n", numAngleBins );
//			printf( "hood: %d\n", hood );
//			printf( "minAngles: %d, %d\n", minAngles.rows(), minAngles.cols() );
//			cout << minAngles << endl;
//			printf( "maxAngles: %d, %d\n", maxAngles.rows(), maxAngles.cols() );
//			cout << maxAngles << endl;

			typename Channel<inT>::Iterator2D chanIt1 = chan.getIterator( y-hood, x-hood, width );
			typename T_angles::iterator maxIt = maxAngles.begin();
			typename T_angles::iterator minIt = minAngles.begin();

			typename Channel<inT>::Iterator2D chanStop = chan.getIterator( y,  x, width );

			operatorHelper( chanIt1, chanStop, maxIt, minIt );

			//skip point (0,0)
			++chanIt1;
			++maxIt;
			++minIt;
			chanStop = chan.getIterator( y+hood,  x+hood, width );

			operatorHelper( chanIt1, chanStop, maxIt, minIt );

			int count = 0;
//			printf( "bins: [ " );
			for ( int i = 0; i < numAngleBins; ++i )
			{
//				printf( "%d ", bins[i] );
				if ( bins[i] >= needed ) ++count;
			}
//			printf( "]\n" );

			return count;
		}

		template <class inT>
		void EdgesAt<inT>::operatorHelper(
		        typename Channel<inT>::Iterator2D &chanIt,
		        const typename Channel<inT>::Iterator2D &chanStop,
		        typename T_angles::iterator &maxIt,
		        typename T_angles::iterator &minIt )
		{
			for ( ; chanIt != chanStop; ++chanIt, ++maxIt, ++minIt )
			{
				if ( *chanIt > thresh )
				{
					float min = *minIt;
					float max = *maxIt;
					if ( min < -PI_2 + 0.5 * binStep )
					{
						min += M_PI;
						max += M_PI;
					}
					int minBinIdx = ( int( ( ( PI_2 + min ) / binStep ) - 0.5 ) + numAngleBins ) % numAngleBins;
					int maxBinIdx = ( int( ( ( PI_2 + max ) / binStep ) - 0.5 ) + numAngleBins ) % numAngleBins;

					if ( minBinIdx > maxBinIdx )
					{
						for ( int k = minBinIdx; k < numAngleBins; ++k ) ++bins[k];
						for ( int k = 0; k <= maxBinIdx; ++k ) ++bins[k];
					}
					else
					{
						for ( int k = minBinIdx; k <= maxBinIdx; ++k ) ++bins[k];
					}
				}
			}
		}

		template <class inT>
		void EdgesAt<inT>::initAngles()
		{
			for ( int y = -hood; y <= hood; ++y )
			{
				for ( int x = -hood; x <= hood; ++x )
				{
					float min, max;
					int testX = x;
					int testY = y;
					if ( x < 0 )
					{
						//only the orientation is needed, so mirror points left of the y-axis at the origin
						testX = -testX;
						testY = -testY;
					}
					anglesCoveredByPixelAt( testX, testY, min, max );
					minAngles( y + hood, x + hood ) = min;
					maxAngles( y + hood, x + hood ) = max;
//					printf( "min, max: %f, %f\n", min, max );
//					pressKey();
				}
			}
		}

	}

}

#endif
