

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_IMGPROCESSINGHARRISFAST_H
#define ADORE_IMGPROCESSINGHARRISFAST_H

#include <adrcore/imgprocessing/imgfunctions.h>
#include <adrcore/types/aarray.h>
#include <adrcore/imgprocessing/multiiterator.h>
#include "harriscommon.h"

namespace adore
{

	namespace core
	{

		/**
		 * this is called harrisfast, because you don't need to precompute the sobel images
		 * the harris value is calculated from an 3x3 image patch, which you
		 * can't do easily otherwise with adore
		 
		*/
		template <class inT>
		class HarrisFast
		{
				typedef typename signedT< typename calcT<inT>::type >::type calcT;
			public:
				typedef typename harrisOutT<inT>::ouT ouT;


				HarrisFast ( float _scale = 1.0 ) :
						k ( 0.04 ),
						scale ( _scale )
				{
				}

				~HarrisFast()
				{
					;
				}

				void process ( Channel<inT> *in, Channel<ouT> *out );

				void process ( ChannelSet<inT> &in, Channel<ouT> *out );

			protected:

				void addG ( calcT a, calcT b, calcT c, calcT d, calcT &gxx, calcT &gyy, calcT &gxy )
				{
					calcT xx = b - a + d - c;
					calcT yy = c - a + d - b;
					gxy += xx * yy;
					gxx += xx * xx;
					gyy += yy * yy;
				}

				const float k;
				const float scale;
		};


		template <class inT>
		void HarrisFast<inT>::process ( Channel<inT> *in, Channel<ouT> *out )
		{
			UsingROI r1 ( in );
			UsingROI r2 ( out );
			in->minROIDistance ( 1 );
			out->fill ( 0 );
			int rX = in->getROI().getX();
			int rY = in->getROI().getY();
			int rW = in->getROI().getWidth();
			int rH = in->getROI().getHeight();
			out->setCenterdROI ( rW, rH );

			typename Channel<inT>::Iterator2D inIt0 = in->getIterator ( rY - 1, rX - 1, rW );
			typename Channel<inT>::Iterator2D inIt1 = in->getIterator ( rY - 1, rX, rW );
			typename Channel<inT>::Iterator2D inIt2 = in->getIterator ( rY - 1, rX + 1, rW );
			typename Channel<inT>::Iterator2D inIt3 = in->getIterator ( rY, rX - 1, rW );
			typename Channel<inT>::Iterator2D inIt4 = in->getIterator ( rY, rX, rW );
			typename Channel<inT>::Iterator2D inIt5 = in->getIterator ( rY, rX + 1, rW );
			typename Channel<inT>::Iterator2D inIt6 = in->getIterator ( rY + 1, rX - 1, rW );
			typename Channel<inT>::Iterator2D inIt7 = in->getIterator ( rY + 1, rX, rW );
			typename Channel<inT>::Iterator2D inIt8 = in->getIterator ( rY + 1, rX + 1, rW );

			typename Channel<ouT>::Iterator2D outIt = out->getIteratorBegin();
			typename Channel<ouT>::Iterator2D outItEnd = out->getIteratorEnd();

//			uint x, y;
			for ( ; outIt != outItEnd; ++outIt,
			                ++inIt0, ++inIt1, ++inIt2, ++inIt3, ++inIt4, ++inIt5, ++inIt6, ++inIt7, ++inIt8 )
			{
//				out->getCoors( outIt.getPointer(), x, y );

				calcT i0 = calcT ( *inIt0 );
				calcT i1 = calcT ( *inIt1 );
				calcT i2 = calcT ( *inIt2 );
				calcT i3 = calcT ( *inIt3 );
				calcT i4 = calcT ( *inIt4 );
				calcT i5 = calcT ( *inIt5 );
				calcT i6 = calcT ( *inIt6 );
				calcT i7 = calcT ( *inIt7 );
				calcT i8 = calcT ( *inIt8 );
//				printf( "%d, %d, %d\n%d, %d, %d\n%d, %d, %d\n", i0, i1, i2, i3, i4, i5, i6, i7, i8 );

				calcT gxx = 0, gyy = 0, gxy = 0;
				addG ( i0, i1, i3, i4, gxx, gyy, gxy );
				addG ( i1, i2, i4, i5, gxx, gyy, gxy );
				addG ( i3, i4, i6, i7, gxx, gyy, gxy );
				addG ( i4, i5, i7, i8, gxx, gyy, gxy );

				const float shift = 0.01; //to stay in a sane range
				float fgxx = shift * scale * gxx;
				float fgyy = shift * scale * gyy;
				float fgxy = shift * scale * gxy;

//				printf( "(%d, %d) - fgxx: %f, fgyy: %f, fgxy: %f\n", x, y, fgxx, fgyy, fgxy );

				float trace = fgxx + fgyy;
//				printf( "%f\n", trace );
				trace *= trace;
//				printf( "%f\n", trace );
				float result = k * trace;
//				printf( "%f\n", result );
//				printf( "%f\n", fgxx * fgyy - fgxy * fgxy - result );
				*outIt = ouT ( fgxx * fgyy - fgxy * fgxy - result );

//				if ( *outIt > 0 ) printf( "!!!!!!!!!!!!!!!!!!" );
//				printf( "%f, %f, %f -> %d\n", fgxx, fgyy, fgxy, *outIt );//pressKey();
			}
		}

		template <>
		inline void HarrisFast<float>::process ( Channel<float> *, Channel<float> * )
		{
			abort();
		}

		template <>
		inline void HarrisFast<double>::process ( Channel<double> *, Channel<double> * )
		{
			abort();
		}

		template <class inT>
		void HarrisFast<inT>::process ( ChannelSet<inT> &in, Channel<ouT> *out )
		{
			UsingROI r1 ( in[0] );
			UsingROI r2 ( out );
			in[0]->minROIDistance ( 1 );
			out->fill ( 0 );
			int rX = in[0]->getROI().getX();
			int rY = in[0]->getROI().getY();
			int rW = in[0]->getROI().getWidth();
			int rH = in[0]->getROI().getHeight();
			out->setCenterdROI ( rW, rH );

			typedef MultiIterator <typename Channel<inT>::Iterator2D> Iter;
			int n = in.size();
			Iter inIt0 ( n ), inIt1 ( n ), inIt2 ( n ), inIt3 ( n ), inIt4 ( n ), inIt5 ( n ), inIt6 ( n ), inIt7 ( n ), inIt8 ( n );
			for ( int i = 0; i < n; ++i )
			{
				inIt0.add ( in[i]->getIterator ( rY - 1, rX - 1, rW ) );
				inIt1.add ( in[i]->getIterator ( rY - 1, rX, rW ) );
				inIt2.add ( in[i]->getIterator ( rY - 1, rX + 1, rW ) );
				inIt3.add ( in[i]->getIterator ( rY, rX - 1, rW ) );
				inIt4.add ( in[i]->getIterator ( rY, rX, rW ) );
				inIt5.add ( in[i]->getIterator ( rY, rX + 1, rW ) );
				inIt6.add ( in[i]->getIterator ( rY + 1, rX - 1, rW ) );
				inIt7.add ( in[i]->getIterator ( rY + 1, rX, rW ) );
				inIt8.add ( in[i]->getIterator ( rY + 1, rX + 1, rW ) );
			}

			typename Channel<ouT>::Iterator2D outIt = out->getIteratorBegin();
			typename Channel<ouT>::Iterator2D outItEnd = out->getIteratorEnd();

//			uint x, y;
			for ( ; outIt != outItEnd; ++outIt,
			                ++inIt0, ++inIt1, ++inIt2, ++inIt3, ++inIt4, ++inIt5, ++inIt6, ++inIt7, ++inIt8 )
			{
//				out->getCoors( outIt.getPointer(), x, y );

				calcT gxx = 0, gyy = 0, gxy = 0;
				for ( int i = 0; i < n; ++i )
				{
					calcT i0 = calcT ( inIt0[i] );
					calcT i1 = calcT ( inIt1[i] );
					calcT i2 = calcT ( inIt2[i] );
					calcT i3 = calcT ( inIt3[i] );
					calcT i4 = calcT ( inIt4[i] );
					calcT i5 = calcT ( inIt5[i] );
					calcT i6 = calcT ( inIt6[i] );
					calcT i7 = calcT ( inIt7[i] );
					calcT i8 = calcT ( inIt8[i] );
//					printf( "%d, %d, %d\n%d, %d, %d\n%d, %d, %d\n", i0, i1, i2, i3, i4, i5, i6, i7, i8 );

					addG ( i0, i1, i3, i4, gxx, gyy, gxy );
					addG ( i1, i2, i4, i5, gxx, gyy, gxy );
					addG ( i3, i4, i6, i7, gxx, gyy, gxy );
					addG ( i4, i5, i7, i8, gxx, gyy, gxy );
				}

				const float shift = 0.01 / float ( n ); //to stay in a sane range
				float fgxx = shift * scale * gxx;
				float fgyy = shift * scale * gyy;
				float fgxy = shift * scale * gxy;

//				printf( "(%d, %d) - fgxx: %f, fgyy: %f, fgxy: %f\n", x, y, fgxx, fgyy, fgxy );

				float trace = fgxx + fgyy;
//				printf( "%f\n", trace );
				trace *= trace;
//				printf( "%f\n", trace );
				float result = k * trace;
//				printf( "%f\n", result );
//				printf( "%f\n", fgxx * fgyy - fgxy * fgxy - result );
				*outIt = ouT ( fgxx * fgyy - fgxy * fgxy - result );

//				if ( *outIt > 0 ) printf( "!!!!!!!!!!!!!!!!!!" );
//				printf( "%f, %f, %f -> %d\n", fgxx, fgyy, fgxy, *outIt );//pressKey();
			}
		}

		template <>
		inline void HarrisFast<float>::process ( ChannelSet<float> &, Channel<float> * )
		{
			abort();
		}

		template <>
		inline void HarrisFast<double>::process ( ChannelSet<double> &, Channel<double> * )
		{
			abort();
		}

	}
}

#endif
