

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_IMGPROCESSINGIMAGEITERATOR_H
#define ADORE_IMGPROCESSINGIMAGEITERATOR_H

#include <adrcore/util/debug.h>
#include <adrcore/imgprocessing/imgconstants.h>
#include <iterator>

namespace adore
{

	namespace core
	{

		/**
		 
		*/
		template <class T> class Image2DIterator;

		template <class T> class Image1DIterator
		{
				friend class Image2DIterator<T>;

			public:
				typedef std::random_access_iterator_tag iterator_category;
				typedef int difference_type;
				typedef T value_type;
				typedef T* pointer;
				typedef T& reference;

			public:

				Image1DIterator( T* _start = NULL, const int _inc = 1, const int bpp = 1 ):
						start( _start ),
						inc( _inc ),
						it( _start ),
						pixelBytes( bpp )
				{
					assert_msg( abs( inc ) >= pixelBytes / ( int )sizeof( T ), "the increment must equal or exceed the bpp (%d < %d)\n", inc, pixelBytes );
//					printf( "%p, numBytes: %d, incStep: %d\n", start,pixelBytes, inc );
				}

				Image1DIterator( const Image1DIterator<T> &i ):
						start( i.start ),
						inc( i.inc ),
						it( i.it ),
						pixelBytes( i.pixelBytes )
				{
				}

				void init( T* _start, const int _inc )
				{
					start = _start;
					inc = _inc;
					it = _start;
				}

				void jumpLine( const int widthStep )
				{
					start += widthStep;
				}

				void reset()
				{
					it = start;
				}

				void nextLine( const int widthStep )
				{
					jumpLine( widthStep );
					reset();
				}

				const Image1DIterator<T> operator+( int i ) const
				{
					Image1DIterator it( *this );
					it += i;
					return it;
				}

				const Image1DIterator<T> operator-( int i ) const
				{
					Image1DIterator it( *this );
					it -= i;
					return it;
				}

				void operator++()
				{
//					printf( "incrementing 1D %d\n", inc );
					it += inc;
				}

				void operator--()
				{
//					printf( "incrementing 1D %d\n", inc );
					it -= inc;
				}

				void operator+=( int a )
				{
//					printf( "incrementing 1D %d\n", inc );
					it += ( a * inc );
				}

				void operator-=( int a )
				{
//					printf( "incrementing 1D %d\n", inc );
					it -= ( a * inc );
				}

				T& operator*() const
				{
					return *it;
				}

				bool operator!=( const Image1DIterator<T> &i ) const
				{
//					printf( "%p - %p\n", it, i.it );
					return it != i.it;
				}

				bool operator==( const Image1DIterator<T> &i ) const
				{
					return it == i.it;
				}

				void operator()( const Image1DIterator<T> &i )
				{
					memcpy( it, i.it, pixelBytes );
				}

				void operator=( const Image1DIterator<T> &i )
				{
					start = i.start;
					inc = i.inc;
					it = i.it;
					pixelBytes = i.pixelBytes;
				}

				/**
				 * distance between iterators
				 */
				void operator-( const Image1DIterator<T> &i )
				{
					return ( it - i.it ) / inc;
				}

				int getLoc() const
				{
					return ( it - start ) / inc;
				}

				T* getPointer() const
				{
					return it;
				}

			protected:
				T* start;
				int inc;
				T* it;
				int pixelBytes;
		};

		template <class T>
		class Image2DIterator
		{
			public:
				typedef std::random_access_iterator_tag iterator_category;
				typedef int difference_type;
				typedef T value_type;
				typedef T* pointer;
				typedef T& reference;

			public:

				Image2DIterator( const Image1DIterator<T> &_start,
				                 const Image1DIterator<T> &_stop, const int _wStep ):
						it1DStart( _start ),
						it1DStop( _stop ),
						wStep( _wStep )
				{
//					printf( "ctor %p - %p\n", it1DStart.it, it1DStop.it );
				}

				Image2DIterator( const Image2DIterator<T> &i ):
						it1DStart( i.it1DStart ),
						it1DStop( i.it1DStop ),
						wStep( i.wStep )
				{
//					printf( "copy ctor\n" );
				}

				void operator++()
				{
					++it1DStart;
					if ( it1DStart == it1DStop )
					{
//						printf( "nextLine 2D %d\n", incY );
						it1DStart.nextLine( wStep );
						it1DStop.nextLine( wStep );
					}
				}

				bool operator!=( const Image2DIterator<T> &i ) const
				{
					return it1DStart != i.it1DStart;
				}

				bool operator==( const Image2DIterator<T> &i ) const
				{
					return it1DStart == i.it1DStart;
				}

				void operator=( const Image2DIterator<T> &i )
				{
					it1DStart = i.it1DStart;
					it1DStop = i.it1DStop;
					wStep = i.wStep;
//					printf( "set\n" );
				}

				T& operator*() const
				{
					return *( it1DStart.it );
				}

				void operator()( Image2DIterator<T> &i )
				{
					it1DStart( i.it1DStart );
				}

				T *getPointer( const int x, const int y )
				{
					return  getRelPointer( x, y );
				}

				T *getPointer( int coronaIdx )
				{
					return  getRelPointer( CORONA[coronaIdx][0], CORONA[coronaIdx][1] );
				}

				Image2DIterator<T> getIterator( int x, int y, uint width ) const
				{
					T *start = getRelPointer( x, y );
					Image1DIterator<T> it1( start, it1DStart.inc, it1DStart.pixelBytes );

					T *stop = getRelPointer( x + width, y );
					Image1DIterator<T> it2( stop, it1DStart.inc, it1DStart.pixelBytes );

					Image2DIterator<T> it( it1, it2, wStep );
					return  it;
				}

				T* getPointer() const
				{
					return it1DStart.getPointer();
				}

			protected:
				Image1DIterator<T> it1DStart;
				Image1DIterator<T> it1DStop;
				int wStep;

				inline T *getRelPointer( const int &x, const int &y ) const
				{
					return it1DStart.it + ( x * it1DStart.inc ) * it1DStart.inc + wStep * y;
				}
		};


	}

}

#endif
