

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_IMGPROCESSINGIMGFUNCTIONS_H
#define ADORE_IMGPROCESSINGIMGFUNCTIONS_H

#include <adrcore/types/aarray.h>
#include <adrcore/types/rect.h>

namespace adore
{

	namespace core
	{

		/**
		 
		*/

		void getBicubicFactors( const float x, const float y,
		                        float Rx[4], float Ry[4] );

		bool subCoorQuadratic( float &subX, float &subY, const float r[8] );

		/**
		 * The iterator must be:
		 * (-1, -1) (0, -1) (1, -1)
		 * (-1,  0) (0,  0) (1,  0)
		 * (-1,  1) (0,  1) (1,  1)
		 */
		template<class Iterator>
		inline bool subCoorQuadratic( Iterator rect, float &subX, float &subY )
		{
			float a = *rect; ++rect;
			float b = *rect; ++rect;
			float c = *rect; ++rect;
			float d = *rect; ++rect;
			float sub = *rect; ++rect; //DC-part
			float e = *rect; ++rect;
			float f = *rect; ++rect;
			float g = *rect; ++rect;
			float h = *rect;
//			printf( "%f, %f, %f\n%f, %f, %f\n%f, %f, %f\n", a,b,c,d,sub,e,f,g,h );
			float r[8] = {a - sub, b - sub, c - sub, d - sub, e - sub, f - sub, g - sub, h - sub};

			return subCoorQuadratic( subX, subY, r );
		}

		void anglesCoveredByPixelAt( int x, int y, float &min, float &max );

		/**
		    * calculates the coordinates of an offset given a width step
		    * Be careful: If you use this for images, know this assumes one
		    * channel. You want to divide the x-coordinate through the number
		    * of channels to fix this (given your wStep is numChans * imgWidth).
		    */
		inline void offsetToCoors( int offset, int wStep, int &x, int &y )
		{
			div_t d = div( offset, wStep );
			y = d.quot;
			x = d.rem;
		}

		template<class IteratorT>
		void getCenterOfGravity( int wStep, int numChans,
		                         IteratorT start,
		                         IteratorT stop,
		                         float &outX, float &outY )
		{
			outX = outY = 0;
			if ( start == stop ) return;

			int x = 0, y = 0;

			int count = 0;
			for ( ; start != stop; ++start, ++count )
			{
				offsetToCoors( *start, wStep, x, y );
				outX += x / numChans;
				outY += y;
			}
			outX /= count;
			outY /= count;
		}

		/**
		 * returns the bounding Rect of the area to which (0, 0)-(w, h) is projected
		 * affMat has to be a 2x2 Matrix
		 */
		template<class T>
		Rect<T> getAreaOfTrafo( const AArray<T, 2> &affMat, T w, T h )
		{
			T minX = 0.0F, minY = 0.0F, maxX = 0.0F, maxY = 0.0F;

			//(w, 0)
			T x1 = w * affMat( 0, 0 );
			T y1 = w * affMat( 1, 0 );
			if ( x1 < minX ) minX = x1;
			if ( x1 > maxX ) maxX = x1;
			if ( y1 < minY ) minY = y1;
			if ( y1 > maxY ) maxY = y1;

			//(0, h)
			T x2 = h * affMat( 0, 1 );
			T y2 = h * affMat( 1, 1 );
			if ( x2 < minX ) minX = x2;
			if ( x2 > maxX ) maxX = x2;
			if ( y2 < minY ) minY = y2;
			if ( y2 > maxY ) maxY = y2;

			//(w, h)
			x1 += x2;
			y1 += y2;
			if ( x1 < minX ) minX = x1;
			if ( x1 > maxX ) maxX = x1;
			if ( y1 < minY ) minY = y1;
			if ( y1 > maxY ) maxY = y1;

			return Rect<T>( minX, minY, maxX - minX, maxY - minY );
		}

		/**
		 * this computes the difference of sampled values a and b given
		 * their neighbouring pixels using the method of Birchfield.
		 * This method is robust to sampling artefacts
		 */
		inline int interpolatedDifference( int aLeft, int aMid, int aRight, int bLeft, int bMid, int bRight )
		{
			//compute everything using a factor of 2

			int aMin, aMax, bMin, bMax;

			int aL = ( aLeft + aMid );
			int aR = ( aRight + aMid );
			int bL = ( bLeft + bMid );
			int bR = ( bRight + bMid );
			aMid <<= 1;
			bMid <<= 1;

			minMax3( aL, aR, aMid, aMin, aMax );
			minMax3( bL, bR, bMid, bMin, bMax );

			int dAB = max3( 0, aMid - bMax, bMin - aMid );
			int dBA = max3( 0, bMid - aMax, aMin - bMid );

                        return std::min( dBA, dAB ) >> 1;
		}
	}
}

#endif
