

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_CORESTRETCHNSKEWREMOVER_H
#define ADORE_CORESTRETCHNSKEWREMOVER_H

#include <adrcore/imgprocessing/channel.h>
#include <adrcore/imgprocessing/imgfunctions.h>
#include <adrcore/imgprocessing/channelfunctions.h>
#include <adrcore/imgprocessing/derivatives.h>
#include <adrcore/util/mathutil.h>
#include <adrcore/util/calctype.h>
#include <adrcore/util/printmat.h>
#include <adrcore/util/gaussians.h>
namespace adore
{

	namespace core
	{

		/**
			 
		*/
		template<class inT>
		class StretchNSkewRemover
		{

			private:

				typedef typename signedT< typename calcT<inT>::type >::type calcT;

			public:

				typedef inT img_in_type;
				typedef img_in_type img_out_type;

			private:

				Channel<img_in_type> *chan;
				AMatrixf invAffine;
				AMatrixf bestInvAffine;
				AMatrixf weights;
				AMatrixf chol2nd; //space for second moment and cholesky matrix

			public:
				/**
				 * @param patchsize the size of the imagepatch. This is to save only a few computations,
				 * because if the size doesn't fit, the internal params get changed appropriately anyway.
				 */
				StretchNSkewRemover( uint patchsize = 7 ):invAffine( 2,2 ), bestInvAffine( 2, 2 ), chol2nd( 2, 2 )
				{
					setPatchSize( patchsize );
				}


				/**
				 * x, y are the coordinates of the center pixel in Channel in.
				 * out is the channel where the affinely transformed patch is copied to.
				 * the ROI of the Channel out determines the size of the patch and must be odd and square.
				 */
				bool process( Channel<img_in_type> *in, const float x, const float y, ImagePlane<img_out_type> *out )
				{
					//the dimensions must be odd
					assert( isOdd( out->getROI().getHeight() ) );
					//and the same
					assert( out->getROI().getHeight() == out->getROI().getWidth() );
					assert( !isNullArray( weights ) );

					//because we work with hoods
					uint height = out->getROI().getHeight();
					uint hood = height >> 1;

					//get new weights if patch size has been changed
					assert( height == uint( weights.rows() + 1 ) );
//					if ( height != uint( weights.rows() + 1 ) )
//					{
//						weights.resize( Gaussian[height - 1].rows(), Gaussian[height - 1].cols() );
//						weights = Gaussian[height - 1];
//					}

					//firstly, we assume identity
					setIdentity( invAffine );
					setIdentity( bestInvAffine );

					double diff = 1000.0, diffOld = 1000.0;
//					int count = 0;
//					char imgName[32];

					do
					{
						diffOld = diff;

						//get a patch
						//template<class typeT, class T_numType, class IteratorT>
						printf( "warping with\n" );
						printMat( invAffine );
						warpAffine( *in, x, y, invAffine, out->begin(), hood, hood );

//						sprintf( imgName, "aff%03d.png", count );
//						out->save( imgName );
//						count++;

						adore::core::calc2ndMomentMat( out, weights, 1.0F, chol2nd );
						printf( "second mom\n" );
						printMat( chol2nd );

						int retVal = cholesky( chol2nd, chol2nd );
						if ( retVal != 0 )
						{
							DBGOUT( DBGNORMAL, "StretchNSkewRemover: cholesky failed\n" );
							return false;
						}
						printf( "not normed\n" );
						printMat( chol2nd );

						normalizeDet2( chol2nd );
						printf( "normed\n" );
						printMat( chol2nd );

						//calculate stop criterion
						diff = fabs( ( chol2nd( 0, 0 ) - 1.0 ) + ( chol2nd( 1, 1 ) - 1.0 ) +
						             chol2nd( 1, 0 ) + chol2nd( 0, 1 ) );

						if ( diff < diffOld )
						{
							bestInvAffine = invAffine;
							invAffine = matMul_AT( chol2nd, invAffine );

							normalizeDet2( invAffine );//to prevent rounding errors
						}
						printf( "diff: %f, diffOld: %f\n", diff, diffOld );
					}
					while ( diff < diffOld );

					return true;
				}

				const AMatrixf &getMatrix()
				{
					return bestInvAffine;
				}

				void setPatchSize( int width )
				{
					assert( isOdd( width ) );
//					weights.reference( Gaussian[width - 1] );
                    weights.resize(Gaussian[width - 1].rows(), Gaussian[width - 1].cols());
                    weights = Gaussian[width - 1];
				}


			private:

				/**
				 * normalizes the det of a 2x2 matrix to 1
				 */
				void normalizeDet2( AMatrixf &m )
				{
					double d = det2( m );
					d = sqrt( d );
					assert( d > FLT_EPSILON );
					d = 1.0 / d;
					m *= d;
				}

				void invertTri2( AMatrix &m )
				{
					double d = 1.0/( m( 0,0 )*m( 1,1 ) );
					m( 1,0 ) = -m( 0,1 );
					m( 0,1 ) = 0;
					m *= d;
				}

				void get2ndDerivates( typename Channel<img_out_type>::Iterator2D &it,
				                      calcT &gxx, calcT &gyy, calcT &gxy )
				{
					calcT center2 = 2 * calcT( *it );
					gxx = calcT( *it.getPointer( -1, 0 ) + *it.getPointer( 1, 0 ) - center2 );
					gyy = calcT( *it.getPointer( 0, -1 ) + *it.getPointer( 0, 1 ) - center2 );
					gxy = calcT( 0.25 * ( ( *it.getPointer( 1, 1 ) - *it.getPointer( -1, 1 ) ) -
					                      ( *it.getPointer( 1, -1 ) - *it.getPointer( -1, -1 ) ) ) );
				}

				void get1stDerivates( typename Channel<img_out_type>::Iterator2D &it, calcT &gx, calcT &gy )
				{
					gx = calcT( *it.getPointer( 1, 0 ) - *it.getPointer( -1, 0 ) );
					gy = calcT( *it.getPointer( 0, 1 ) - *it.getPointer( 0, -1 ) );
				}

				AMatrix calc2ndMomentMat( Channel<img_out_type> *out )
				{
					UsingROI ur( out );
					out->minROIDistance( 1 );

					assert( out->getROI().getWidth() == ( uint )weights.cols() );
					assert( out->getROI().getHeight() == ( uint )weights.rows() );

					typename Channel<img_out_type>::Iterator2D it    = out->getIteratorBegin();
					typename Channel<img_out_type>::Iterator2D itEnd = out->getIteratorEnd();
					typename AMatrix::iterator wit = weights.begin();

					uint x, y;
					double A = .0, B = .0, C = .0;
					calcT gx = 0, gy = 0;
					for ( ; it != itEnd; ++it, ++wit )
					{
						out->getCoors( it.getPointer(), x, y );
						get1stDerivates( it, gx, gy );
						double w = *wit;

//						printf( "at(%d, %d): %d, %d (w = %f)\n", x, y, gx, gy, w );

						A += w * gx * gx;
						B += w * gy * gy;
						C += w * gx * gy;
					}
					AMatrix mat2nd( 2, 2 );
					mat2nd = A, C, C, B;
					return mat2nd;
				}

		};

	}

}

#endif
