

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_COREUPSAMPLER_H
#define ADORE_COREUPSAMPLER_H

#include <adrcore/imgprocessing/channelprocessor.h>
#include <adrcore/imgprocessing/bilinearinterpolator.h>

namespace adore
{

	namespace core
	{

		template<class inT>
		class UpSamplerSurf : public ChannelProcessor<1, 1, inT, inT>
		{
			public:
				UpSamplerSurf() :ChannelProcessor<1, 1, inT, inT>(){}
				
				/// bilinear interpolate for special case: width => width*2 and height => height*2.
				/// ROIs are processed
				/// left and bottom row/column are simply duplicated.
				void process ( Channel<inT> *in, Channel<inT> *out )
				{
					int xIn  = in->getROI().getX();
					int yIn  = in->getROI().getY();
					int xOut = out->getROI().getX();
					int yOut = out->getROI().getY();

					int height = in->getROI().getHeight();
					int width = in->getROI().getWidth();
					int height2 = height * 2;
					int width2 = width * 2;

					assert((int)out->getROI().getWidth() == width2); /// TODO: make some fallback
					assert((int)out->getROI().getHeight() == height2);
					
					Image1DIterator<inT> lineS = in->getXIterator( yIn, xIn );
					Image1DIterator<inT> lineD = out->getXIterator( yOut, xOut );
					Image1DIterator<inT> itS, lineS2, itS2, itD;

					uint stepSrc = in->getWStep();
					uint stepDst = out->getWStep(), stepDst2 = stepDst * 2;

					/// copy data to second image.
					for( int y = 0; y < height; ++y ) {
						itS = lineS;
						itD = lineD;

						for( int x = 0; x < width; ++x, ++itS, itD += 2 )
							*itD = *itS;

						lineS.nextLine( stepSrc );
						lineD.nextLine( stepDst2 );
					}

					/// linear interpolate odd rows
					lineD = out->getXIterator( yOut, xOut );
					inT lastVal = 0;
					for( int y = 0; y < height; ++y ) {
						itD = lineD;
						++itD;

						itS = lineD;
						lastVal = *itS;
						itS += 2;

						for( int x = 0; x < width - 1; ++x, itS += 2, itD += 2) {
							*itD = ( *itS + lastVal ) / 2;
							lastVal = *itS;
						}
						/// copy last column
						*itD = lastVal;

						lineD.nextLine( stepDst2 );
					}

					/// linear interpolate all columns
					lineS = out->getXIterator( yOut, xOut );
					lineD = out->getXIterator( yOut + 1, xOut );
					lineS2 = out->getXIterator( yOut + 2, xOut );
					for( int y = 1; y < height; ++y ) {
						itS = lineS; itD = lineD; itS2 = lineS2;

						for ( int x = 0; x < width2; ++x, ++itS, ++itD, ++itS2 )
							*itD = ( *itS + *itS2 ) / 2;

						lineS.nextLine( stepDst2 );
						lineD.nextLine( stepDst2 );
						lineS2.nextLine( stepDst2 );
					}

					/// copy last row.
					itS = out->getXIterator( yOut - 2 + height2, xOut );
					itD = out->getXIterator( yOut - 1 + height2, xOut );
					for ( int x = 0; x < width2; ++x, ++itS, ++itD )
						*itD = *itS;

				}
		};

		template <typename inT> class LinearInterpolator
		{
			private:
				int coor1, coor2;
				double u, u1;

			public:
				LinearInterpolator() : coor1(0.0), coor2(0.0), u(1.0), u1(0.0) {}

				void init(double coor) {
					coor1 = floor(coor);
					coor2 = ceil(coor);
					u = coor - coor1;
					u1 = 1.0 - u;
				}

				inT valueAt_interpY(const Channel<inT> & in, int x) {
					inT one = *( in.getBase() + in.getOffset( x, coor1 ) );
					inT another = *( in.getBase() + in.getOffset( x, coor2 ) );
					return inT( u1 * one + u * another );
				}
				inT valueAt_interpX(const Channel<inT> & in, int y) {
					inT one = *( in.getBase() + in.getOffset( coor1, y ) );
					inT another = *( in.getBase() + in.getOffset( coor2, y ) );
					return inT( u1 * one + u * another );
				}
		};

		template <typename inT> class BiLinearInterpolatorLine
		{
			private:
				int coor1, coor2;
				double u, u1;

			public:
				BiLinearInterpolatorLine() : coor1(0.0), coor2(0.0), u(1.0), u1(0.0) {}

				void init(double coor) {
					coor1 = floor(coor);
					coor2 = ceil(coor);
					u = coor - coor1;
					u1 = 1.0 - u;
				}

				inT valueAt_lineY(const Channel<inT> & in, double x) {
					int x1 = floor(x);
					int x2 = ceil(x);
					double v = x - x1;
					double v1 = 1.0 - v;

					inT nw = *( in.getBase() + in.getOffset( x1, coor1 ) );
					inT ne = *( in.getBase() + in.getOffset( x2, coor1 ) );
					inT sw = *( in.getBase() + in.getOffset( x1, coor2 ) );
					inT se = *( in.getBase() + in.getOffset( x2, coor2 ) );
					return inT( v1 * u1 * nw + v * u1 * ne + v1 * u * sw + v * u * se );
				}
		};

		template<class inT>
		class UpSampler : public ChannelProcessor<1, 1, inT, inT>
		{
			public:
				UpSampler() :ChannelProcessor<1, 1, inT, inT>() {}

				void process ( Channel<inT> * in, Channel<inT> * out )
				{
					assert ( in );
					assert ( out );
					assert ( in != out );

					int xIn  = in->getROI().getX();
					int yIn  = in->getROI().getY();
					int xOut = out->getROI().getX();
					int yOut = out->getROI().getY();

					int heightIn = in->getROI().getHeight();
					int widthIn = in->getROI().getWidth();
					int heightOut = out->getROI().getHeight();
					int widthOut = out->getROI().getWidth();

					assert( (widthOut % widthIn == 0) && (widthOut / widthIn == 2) ); /// use ReSampler(first).
					assert( (heightOut % heightIn == 0) && (heightOut / heightIn == 2) );

					int numStepsX = widthIn;
					int numStepsY = heightIn;

					Image1DIterator<inT> itS, itD;

					LinearInterpolator<inT> linInterp;
					BiLinearInterpolatorLine<inT> bilInterp;

					/// copy data to second image.
					for(int y = 0; y < numStepsY; ++y) {

						itS = in->getXIterator( yIn + y, xIn );
						itD = out->getXIterator( yOut + y * 2, xOut, 2 );

						bilInterp.init(0.75 + yIn + y - 1);

						for(int x = 0; x < numStepsX; ++x, ++itS, ++itD) {

							if(y == 0) {
								if(x == 0)
									*itD = *itS;
								else {
									linInterp.init(0.75 + xIn + x - 1);
									*itD = linInterp.valueAt_interpX(*in, yIn);
								}
								continue;
							}

							if(x == 0) {
								linInterp.init(0.75 + yIn + y - 1);
								*itD = linInterp.valueAt_interpY(*in, xIn);
								continue;
							}

							*itD = bilInterp.valueAt_lineY(*in, 0.75 + xIn + x - 1);
						}
					}

					/// linear interpolate odd rows
					inT lastVal = 0;
					for( int y = 0; y < numStepsY; ++y ) {

						itS = out->getXIterator( yOut + y * 2, xOut, 2 );
						itD = out->getXIterator( yOut + y * 2, xOut + 1, 2 );
						lastVal = *itS;
						++itS;

						for( int x = 0; x < numStepsX - 1; ++x, ++itS, ++itD) {
							*itD = ( *itS + lastVal ) * 0.5;
							lastVal = *itS;
						}
						/// copy last column
						if(y == 0) {
							*itD = *( in->getBase() + in->getOffset( xIn + widthIn - 1, yIn ) );
						} else {
							linInterp.init(0.75 + yIn + y - 1);
							*itD = linInterp.valueAt_interpY(*in, xIn + widthIn - 1);
						}

					}

					/// linear interpolate all columns
					Image1DIterator<inT> itS2;
					for( int y = 0; y < numStepsY - 1; ++y ) {

						itS = out->getXIterator( yOut + y * 2, xOut, 1 );
						itD = out->getXIterator( yOut + y * 2 + 1, xOut, 1 );
						itS2 = out->getXIterator( yOut + y * 2 + 2, xOut, 1 );

						for ( int x = 0; x < widthOut; ++x, ++itS, ++itD, ++itS2 )
							*itD = ( *itS + *itS2 ) * 0.5;

					}

					/// copy last row.
					itS = in->getXIterator( yIn + heightIn - 1, xIn );
					itD = out->getXIterator( yOut + heightOut - 1, xOut, 2 );
					for ( int x = 0; x < numStepsX; ++x, ++itS, ++itD ) {
						if(x == 0) *itD = *itS;
						else {
							linInterp.init(0.75 + xIn + x - 1);
							*itD = linInterp.valueAt_interpX(*in, yIn + heightIn - 1);
						}
					}
					/// linear interpolate last row.
					itS = out->getXIterator( yOut + heightOut - 1, xOut, 2 );
					itD = out->getXIterator( yOut + heightOut - 1, xOut + 1, 2 );
					lastVal = *itS;
					++itS;

					for( int x = 0; x < numStepsX - 1; ++x, ++itS, ++itD) {
						*itD = ( *itS + lastVal ) * 0.5;
						lastVal = *itS;
					}
					/// copy last point
					*itD = *( in->getBase() + in->getOffset( xIn + widthIn - 1, yIn + heightIn - 1 ) );
				}
		};

	}

}

#endif
