

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include "filterutil.h"
#include <adrcore/nview/featurepointfilter2.h>
#include <adrcore/nview/featurepointfilterfmat.h>
#include <adrcore/viewmatch/kdtreematcher.h>
#include <adrcore/viewmatch/epilinematcher.h>
#include <adrcore/imgprocessing/drawing.h>

namespace adore
{

	namespace core
	{

		void filter ( NView & nview, NVIRANSACClient * ranser )
		{
			//remove the outliers
			int kick = 0;

			for ( NVIRANSACClient::iterator it = ranser->begin(); it != ranser->end(); ++it )
			{
				int where = *it;
				//remove all between where and kick, because those indices are not registered as inliers

				for ( int k = kick; k < where; ++k )
				{
					nview.removePointCorrAnd3D ( ranser->getPointCorr ( k ) );
				}

				kick = where + 1;
			}

			for ( int k = kick; k < ( int ) ranser->getNumCandidates(); ++k )
				nview.removePointCorrAnd3D ( ranser->getPointCorr ( k ) );
		}


		void filter ( NView &nview, NVIRANSACClient *ranser, const RansacParams &rp )
		{
			ADRSAC ransac;
			ransac ( ranser, rp );
			filter( nview, ranser );
		}

		void filter ( NView &nview, NVIRANSACClient *ranser, const RansacParams &rp, ViewKnitter &knit )
		{
			assert ( ranser );
			NView::ViewColl vc;//TODO fix NVRANSACClient

			for ( ViewKnitter::iterator it = knit.begin ( 2 ); it != knit.end ( 2 ); ++it )
			{
				vc.clear();
				vc.insert ( ( * ( *it ).get<2>() ) [0] );
				vc.insert ( ( * ( *it ).get<2>() ) [1] );
				ranser->init ( nview.getPointCorrsBegin(), nview.getNumPointCorrs(), vc.getBegin() );
				assert ( ranser->isOkay() );
				filter ( nview, ranser, rp );
			}
		}

		/**
			deletes all Points from a View v that are not member in any coor of nv
		*/
		void filterNonCorr ( ViewBase *v, NView *nv )
		{
			ViewBase::PointsIterator it = v->getPointsBegin();

			while ( it != v->getPointsEnd() )
			{
				ViewBase::PointsIterator oldIt = it;
				++it;
				FeaturePoint *pt = *oldIt;

				if ( pt && !nv->hasFeaturePoint ( pt ) )
				{
					nv->removeFeaturePoint ( pt );
					v->remove ( pt );//deletes
				}
			}
		}

//		encapsulates debug output
		void filterNonCorr ( ViewBase *lv, ViewBase *rv, NView *nv )
		{
			DBGOUT ( DBGNORMAL, "filter2: lView before filtering: %d points\n", lv->getNumAllFeaturePoints() );

			filterNonCorr ( lv, nv );

			DBGOUT ( DBGNORMAL, "filter2: lView after filtering: %d points\n", lv->getNumAllFeaturePoints() );
			DBGOUT ( DBGNORMAL, "filter2: rView before filtering: %d points\n", rv->getNumAllFeaturePoints() );

			filterNonCorr ( rv, nv );

			DBGOUT ( DBGNORMAL, "filter2: rView after filtering: %d points\n", rv->getNumAllFeaturePoints() );

#ifndef NDEBUG
			ImagePlane<uint8_t> *img = createViewsImage ( dynamic_cast<View<uint8_t> *> ( lv ), dynamic_cast<View<uint8_t> *> ( rv ), nv );
			img->save ( "pairout.png" );
			delete img;
#endif
		}

		/**
			deltes all features from the views that are not compatible with
			an estimated epipolar geometry
			returns true on success;
		*/
		bool filter2KDTreeRansac ( ViewBase *lView, ViewBase *rView, int countThresh )
		{
			assert ( lView->isNormalized() );
			assert ( rView->isNormalized() );

			KDTreeMatcher matcher ( KDTREE_SINGLE_BEST );
			matcher.setCountThresh ( countThresh );
			NView nv;
			matcher.match ( lView, rView, &nv );

			float thresh = getNormalizedPixel ( getOptThreshFM(), lView );
			FilterFMat client ( thresh );
			FeaturePointFilter2 fpfilter ( &client, NULL );
			fpfilter.modify ( nv );

			//check each point for existence in a corr
			filterNonCorr ( lView, rView, &nv );

			return true;
		}

		bool filter2KDTreeEpi ( AMatrix fmatrix, ViewBase *lView, ViewBase *rView, float distThresh, int countThresh )
		{
			assert ( lView->isNormalized() == rView->isNormalized() );

			KDTreeMatcher matcher ( KDTREE_SINGLE_BEST );
			matcher.setCountThresh ( countThresh );

			NView nv;
			NViewInfoFM nvi;
			nvi.setFMatrix ( fmatrix );
			nv.setInfo ( &nvi );

			matcher.match ( lView, rView, &nv );

			FeaturePointFilterFMat fpfilter ( distThresh );
			fpfilter.modify ( nv );

			//check each point for existence in a corr
			filterNonCorr ( lView, rView, &nv );

			return true;
		}

		bool filter2EpiAlone ( AMatrix fmatrix, ViewBase *lView, ViewBase *rView, float distThresh, int countThresh )
		{
			assert ( false );//epilinematcher paramters missing
			assert ( lView->isNormalized() == rView->isNormalized() );

			EpiLineMatcher matcher ( fmatrix, distThresh, 0.15f );
			matcher.setCountThresh ( countThresh );
			NView nv;
			matcher.match ( lView, rView, &nv );

			//check each point for existence in a corr
			filterNonCorr ( lView, rView, &nv );

			return true;
		}

		/** gric */

		static double getGricFactor ( int type )
		{
			switch ( type )
			{
				case MODEL_TYPE_H:
					return 4;
				case MODEL_TYPE_F:
					return 2;
				default: assert ( false );
			}
			return 0;
		}

		double getGricThresh ( double sigma, int type )
		{
			return getGricFactor ( type ) * sigma;
		}

		/**
		 * sigma is the threshold from which the ransac threshold had been computed
		 * do not use the ransac threshold directly!!!
		 */
		static double finalizeGricFactor ( int type, double error, double sigma2, uint size, uint in )
		{
			double r2V, dX, kX;
			double lnr2V, maxRo;
			switch ( type )
			{
				case MODEL_TYPE_H:
					r2V = 4; dX = 2; kX = 8;
					break;
				case MODEL_TYPE_F:
					r2V = 4; dX = 3; kX = 7;
					break;
				default:
					assert ( false );
					throw;
			}
			lnr2V = log ( r2V );
			maxRo = 2 * ( r2V - dX );

			double inlierError = error - sigma2 * maxRo * ( size - in );
			double modifiedError = inlierError / sigma2 + maxRo * ( size - in );
			return modifiedError + ( size * dX * lnr2V ) + kX * log ( r2V * size );
		}

		NVIRANSACClient * assessRansacClients ( NView & nview, const ClientList & data )
		{
			double error = numeric_limits<double>::max();
			NVIRANSACClient * the_one = NULL;
			ADRSAC ransac;

			for ( uint i = 0; i < data.size(); ++i )
			{
				NVIRANSACClient * client = data.at ( i ).first;
				int m_type = client->getModelType();
				double f = getGricFactor ( m_type );
				const double sigma2 = client->getThreshold();
				client->setThreshold ( sigma2 * f ); // GRIC threshold

				ransac ( client, data.at ( i ).second );

				double err = finalizeGricFactor ( m_type, client->getError(), sigma2,
				                                  client->getNumCandidates(), client->getNumIn() );
				cout << "*********** GRIC error: " << err << endl;
				if ( error > err )
				{
					the_one = client;
					error = err;
				}
				client->setThreshold( sigma2 ); // very important!
			}

			return the_one;
		}

	}
}
