

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_BABUNDLEADJUSTER_H
#define ADORE_BABUNDLEADJUSTER_H

#include <adrcore/types/camera.h>
#include <adrcore/parameterization/simplecamparam.h>
#include <adrcore/parameterization/zfixedpointparam.h>
#include <adrcore/minimization/lmsparse.h>
#include <vector>

namespace adore
{
	namespace core
	{

		/**
		 * This class is a wrapper around the sparse LM algo.
		 * It essentially reformats the input, solves it and reformats it back again.
		 * missing 2D points should be no problem.
		 *  
		*/
		template<class MinFuncT, class ErrorFuncT, class CamParamT = SimpleCamParam, class P3DParamT = ZFixedPointParam>
		class BundleAdjuster
		{
			protected:

				bool lastCamFixed;

				LMSparse<MinFuncT, ErrorFuncT> lm;

				vector<AVector> lineCams;
				vector<AVector> linePts3D;
				vector<int> missing;
				int maxIterations;
			public:

				/**
				 * Constructor.
				 */
				BundleAdjuster(int max_it = 10000) :
					lastCamFixed ( true ),
					missing ( 0 ),
					maxIterations(max_it)
				{

				}


				/**
				 * Destructor.
				 */
				~BundleAdjuster()
				{
				}

				/**
				 * Adjusts the cameras and the 3D-points to find a maximum likelihood estimation that minimizes
				 * the reprojection error according to the given 2D-points.
				 * the 2D points are captured inside the error func
				 * @param cams The cameras
				 * @param pts3D The 3D-points
				 * @param pts The 2D-points
				 */
				void modify ( vector<Camera *> &cams,
                              vector<AVector3D*> &pts3D,
				              MinFuncT &minFunc,
				              ErrorFuncT &errorFunc,
				              CamParamT &cParam = SimpleCamParam(),
				              P3DParamT &pParam = ZFixedPointParam(), double thresh = FLT_EPSILON );


				/**
				 * should the last cam assumed to be the identity cam?
				 * @param fixed last cam fixed to identity or not.
				 */
				void setLastCamIsIdentity ( bool fixed )
				{
					lastCamFixed = fixed;
				}

				/**
				 * Is the last cam assumed to be the identity cam?
				 * @return true, if the last cam is fixed to identity.
				 */
				bool getLastCamIsIdentity()
				{
					return lastCamFixed;
				}

				/**
				 * returns the error of the last call of this algorithm
				 * @return the error
				 */
				double getLastError() const
				{
					return lm.getLastError();
				}

				/**
				 * returns the number of iterations needed by the last call of this algorithm.
				 * @return the number of iterations
				 */
				int getLastIterations() const
				{
					return lm.getLastIterations();
				}

				/**
				 * returns the error before starting of minimization
				 * @return the error
				 */
				double getInitialError() const
				{
					return lm.getInitialError();
				}

				void setMissingIndices ( const vector< int >& theValue )
				{
					missing = theValue;
				}

				int getMaxIterations() const { return maxIterations; }

				void setMaxIterations(int max_it) { maxIterations = max_it; }


			protected:

				void create ( vector<AVector> & vec, uint num, uint dim )
				{
					vec.resize ( num );
					for ( vector<AVector>::iterator it = vec.begin(); it != vec.end(); ++it )
					{
						( *it ).resize( dim );
					}
				}

		};


		template<class MinFuncT, class ErrorFuncT, class CamParamT, class P3DParamT>
		void BundleAdjuster<MinFuncT, ErrorFuncT, CamParamT, P3DParamT>::modify (
		        vector<Camera *> &cams,
                vector<AVector3D*> &pts3D,
		        MinFuncT &minFunc,
		        ErrorFuncT &errorFunc,
		        CamParamT &cParam,
		        P3DParamT &pParam, double thresh )

		{
			if ( cams.size() < 1 || pts3D.size() < 1 ) return;
			const int outDim = 2;

			if ( cams.size() != lineCams.size() || pts3D.size() != linePts3D.size() )
			{
				DBGOUT ( DBGNORMAL, "BundleAdjuster: reinitializing the lm-algo\n" );
//				lm.init( cams.size(), 12, lastCamFixed ? 1 : 0, pts3D.size(), 3, 2 );
				lm.init ( cams.size(), cParam.getSize(), lastCamFixed ? 1 : 0, pts3D.size(), pParam.getSize(), outDim );
			}
//			else
//			{
//				printf( "Vecs: %d/%d; %d/%d\n", cams.size(), lineCams.size(), pts3D.size(), linePts3D.size() );
//			}


//			int numPts2D = ( cams.size()  + ( lastCamFixed ? 1 : 0 ) ) * pts3D.size();

			create ( lineCams, cams.size(), cParam.getSize() );
			create ( linePts3D, pts3D.size(), pParam.getSize() );

			cParam.toParam ( cams, lineCams );
			pParam.toParam ( pts3D, linePts3D );

			lm.solve ( lineCams, linePts3D, minFunc, missing, errorFunc,
			           DefaultSparseInvCovar,
					   thresh, maxIterations );

			DBGOUT( DBGNORMAL, "BundleAdjuster: Residual: %.20f - %.20f after %d iterations\n", lm.getInitialError(), lm.getLastError(), lm.getLastIterations() );

			cParam.fromParam ( lineCams, cams );
			pParam.fromParam ( linePts3D, pts3D );

		}

	}

}

#endif
