

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_MINIMIZATIONLM2_H
#define ADORE_MINIMIZATIONLM2_H

#include <adrcore/util/debug.h>
#include <adrcore/util/mathutil.h>
#include <adrcore/util/matmul2.h>
#include <adrcore/util/templatelimits.h>
#include <cfloat>
#include <adrcore/types/aarray.h>
#include <adrcore/minimization/lmconstants.h>
#include <adrcore/minimization/covarsquarer2.h>
#include <adrcore/minimization/defaulterrorfunc2.h>
#include <adrcore/util/solvesvd.h>

#include <adrcore/util/printmat.h>

#include <iostream>
using namespace std;
namespace adore
{
	namespace core
	{
		/**
		 * A general purpose Levenberg-Marquard Implementation.
		 * Use this to adjust the initial estimated parameters of a function.
		 *
		 * @param FuncT the function type 1D->1D
		 *
		 *  
		*/
		template<class FuncT, class JacobiT, class ErrorT, class InvCovarT>
		class LM2 : public CovarSquarer2<InvCovarT>
		{
				typedef typename FuncT::T_numtype NumberT;

			public:

				static const AArray<typename FuncT::T_numtype, 7> DefaultInvCovar;

			public:
				/**
				 * Constructor
				 */
				LM2():
						useUserLambda( false ),
						userLambda( 0.001 )
				{
					;
				}

				/**
				 * Destructor.
				 */
				~LM2()
				{
					;
				}

				/**
				 * Initialize the internal variables tor fit the desired function.
				 * This is called by the constructor when it's both parameters are greater zero.
				 *
				 */
//				void init( int domainDim, int rangeDim );

				/**
				 * Solve the parameter estimation problem.
				 * @param params initial estimate, this gets adjusted, so the result is here, too.
				 * @param func the function whose parameters are searched.
				 * @param jacobiFunc the Funtion that returns a matrix of first partial derivatives.
				 * @param invCovar the invers of a covariance matrix. If it is a column-vector, it is assumed to hold the diagonal elements.
				 * @param errorFunc error function (val - goal). possibly attenuates the error-vector according to a cost-function (p. 622).
				 * @param eps if norm(error)&lt;eps or the differents of two subsequent error-norms is less than eps, then the algo stops.
				 * @param maxIt the algorithm stops after maxIt iterations.
				 */
				void solve( AArray<NumberT, 1> &params,
				            const FuncT &func,
				            const int rangeDim,
				            const JacobiT &jacobiFunc,
				            const ErrorT &errorFunc = ZeroErrorFunc2<NumberT>(),
				            const InvCovarT &invCovar = DefaultInvCovar,
				            NumberT eps = epsilon<NumberT>(),
				            int maxIt = 1000 );

				/**
				 * returns the error of the last call of this algorithm
				 * @return the error
				 */
				double getInitialError() const
				{
					return intialError;
				}

				/**
				 * returns the error of the last call of this algorithm
				 * @return the error
				 */
				double getLastError() const
				{
					return lastError;
				}

				/**
				 * returns the number of iterations needed by the last call of this algorithm.
				 * @return the number of iterations
				 */
				int getLastIterations() const
				{
					return lastIterations;
				}

				void setInitialLambda( double theValue )
				{
					userLambda = theValue;
				}


				double getInitialLambda() const
				{
					return userLambda;
				}

				void useInitialLambda( bool theValue )
				{
					useUserLambda = theValue;
				}

				bool isInitialLambdaUsed() const
				{
					return useUserLambda;
				}

			private:

				double lastError;
				double intialError;
				int lastIterations;
				bool useUserLambda;
				double userLambda;

		};

		template<class FuncT, class JacobiT, class ErrorT, class InvCovarT>
		void LM2 <FuncT, JacobiT, ErrorT, InvCovarT>::solve(
		        AArray<NumberT, 1> &params,
		        const FuncT &func,
		        const int rangeDim,
		        const JacobiT &jacobiFunc,
		        const ErrorT &errorFunc,
		        const InvCovarT &invCovar,
		        NumberT eps,
		        int maxIt )
		{
			NumberT err1 = 0.0;
			NumberT err2 = 0.0;
			NumberT mu = 0.0;
			NumberT nu = 2.0;
			NumberT rho = 2.0;
			NumberT tau = 10e-3;
			int count = 0;
			bool stop = false;

			int domainDim = params.rows();

			SVDSolveStorage<NumberT, 1> store ( domainDim, domainDim );

			AArray<NumberT, 2> jacobiSqr( domainDim, domainDim );
			AArray<NumberT, 1> errNew( rangeDim );

			AArray<NumberT, 1> newParams( domainDim );
			AArray<NumberT, 1> Delta( domainDim );

			AArray<NumberT, 2> jacobi( rangeDim, domainDim );
			AArray<NumberT, 1> val( rangeDim );
			AArray<NumberT, 1> err( rangeDim );
			AArray<NumberT, 1> g( domainDim );

			jacobiFunc( params, jacobi );

			//calculate current error
			func( params, val );

			errorFunc( val, err );

			intialError = lastError = err1 = getSqrNorm( err );
			lastIterations = 0;

			/*    printf("Jacobi\n");
			    whereUnInited(jacobi);
			    printf("err\n");
			    whereUnInited(err);*/

			matMul2_AT( jacobi, err, g );

			if ( getMaxNorm( g ) < eps ) return;

			if ( useUserLambda )
			{
				mu = userLambda;
			}
			else
			{
				//compute JTJ to initialize lambda
				matMul2_AT( jacobi, jacobi, jacobiSqr );
				//labmda is initialized as the average of the JTJ * something small, compute it
				for ( int diag = 0; diag < (int)jacobiSqr.rows(); ++diag )
				{
					mu = MAX( mu, jacobiSqr( diag, diag ) );
				}
				mu *= tau;
			}

			do
			{
				//printf("err1: %f\n", err1);
				++count;

				do //LM2-iteration
				{
					jacobiFunc( params, jacobi );

					covarProduct( jacobi, invCovar, jacobiSqr );

					matAddDiag( jacobiSqr, mu );
					//cout << "left:\n" << jacobiSqr << endl;
					//cout << "err:\n" << err << endl;
					matMul2_AT( jacobi, err, g, NumberT(-1.0) );

					solveSVD ( jacobiSqr, g, Delta, store );
					//cout << "Delta:\n" << Delta << endl;
					//pressKey();

					if ( getSqrNorm( Delta ) <= eps * getSqrNorm( params ) )
					{
						stop = true;
					}
					else
					{
						//calculate new error
						add2( params, Delta, newParams );

						func( newParams, val );

						errorFunc( val, errNew );

						err2 = getSqrNorm( errNew );

						NumberT divisor = 0;
						for ( int row = Delta.rows() - 1; row >= 0; --row )
						{
							divisor += Delta( row ) * ( mu * Delta( row ) + g( row ) );
						}

						if ( divisor <= 0 )
						{
							printf( "divisor is: %f\n", divisor );
							printf( "mu is: %f\n", mu );
							printf( "eps is: %f\n", eps );
							printf( "Delta is; \n" );
							std::cout << Delta << std::endl;
							printf( "g is; \n" );
							std::cout << g << std::endl;
							printf( "params is; \n" );
							std::cout << params << std::endl;
							assert( divisor > 0 );
						}
						rho = ( err1 - err2 )/divisor;

						if ( rho > 0 )
						{
							params = newParams;
							err = errNew;
							assert( err2 < err1 );
							err1 = err2;

							//adjust mu (formely known as lambda)
							if ( getMaxNorm( g ) < eps ) stop = true;
							double magic = ( 2 * rho - 1 );

							if ( mu == 0 )
							{
								mu = 0.000001;
							}
							else
							{
								mu = mu * std::max( 1.0/3.0, 1.0 - magic * magic * magic );
								nu = 2;
							}
						}
						else
						{
							mu = mu * nu;
							nu = 2 * nu;
						}
					}

				}
				while ( rho < 0 && !stop );

			}
			while ( count < maxIt && !stop );
			//DBGOUT(DBGALL, "LM2: err: %f, count: %d of %d\n", err1, count, maxIt);
			lastError = err1;
			lastIterations = count;
		}

		template<class FuncT, class JacobiT, class ErrorT, class InvCovarT>
		const AArray<typename FuncT::T_numtype, 7> LM2 <FuncT, JacobiT, ErrorT, InvCovarT>::DefaultInvCovar;

	}
}

#endif
