

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_MINIMIZATIONLMSPARSE_H
#define ADORE_MINIMIZATIONLMSPARSE_H

#include <adrcore/util/debug.h>
#include <adrcore/util/mathutil.h>
#include <adrcore/util/solvesvd.h>
#include <adrcore/util/ldlt.h>
#include <adrcore/minimization/lmconstants.h>
#include <adrcore/minimization/covarsquarer.h>
#include <adrcore/minimization/sparsedefaulterrorfunc.h>
#include <cfloat>
#include <vector>

namespace adore
{
	namespace core
	{

		using namespace std;


		/**
		 * This is a Levenberg-Marquard minimizatin algorithm, as described in Hertley-Zisserman, p.613
		 * It is for a special kind of Problem:
		 *
		 * Given:
		 * - a function f:
		 * f(a1, a2, ..., am, b1, b2, ..., bn) = (x11, ..., x1m, x21, ..., x2m, xn1, ..., xnm)
		 * dxij/dak = 0, j != k
		 * dxij/dbk = 0, i != k
		 * - a desired value of the xij-vector
		 * - an initial estimate for the aj and bi.
		 *
		 * Result:
		 * aj and bj that minimize ||(f(aj, bi) - (xij))||
		 *
		 *  
		*/
		template<class FuncT, class ErrorT,
		class InvCovarT = typename InvCovarUnused<typename FuncT::T_numtype>::type>
		class LMSparse : private CovarSquarer<InvCovarT>
		{

				typedef typename FuncT::T_numtype numberT;
				typedef vector<AMatrix> MatColl;
				typedef vector<AVector> VecColl;
				typedef typename MatColl::iterator MatIterator;
				typedef typename VecColl::iterator VecIterator;

			public:
				/**
				 * Constructor.
				 * @param aSize number of paramters of type a.
				 * @param aDim The dimension of the parameters of type a, e.g. 12 would be a good choice for camera matrices.
				 * @param aFixed number of Parameters that are fixed and only known by the function whose parameters have to be adjusted.
				 * This is to allow a fixed camera (namely the identity cam) in BA.
				 * @param bSize number of parameters of type b.
				 * @param bDim The dimension of the parameters of type b, e.g. 3 for 3D-points.
				 * @param resultDim the dimension of the result, e.g. 2 for inhomogeneous 2D-points.
				 */
				LMSparse ( int aSize = -1, int aDim = -1, int aFixed = -1, int bSize = -1, int bDim = -1, int resultDim = -1 );

				/**
				 * Destructor.
				 */
				~LMSparse();

				/**
				 * initializes the internal variables.
				 * @param aSize number of paramters of type a.
				 * @param aDim The dimension of the parameters of type a, e.g. 12 would be a good choice for camera matrices.
				 * @param aFixed number of Parameters that are fixed and only known by the function whose parameters have to be adjusted.
				 * This is to allow a fixed camera (namely the identity cam) in BA.
				 * @param bSize number of parameters of type b.
				 * @param bDim The dimension of the parameters of type b, e.g. 3 for 3D-points.
				 * @param resultDim the dimension of the result, e.g. 2 for inhomogeneous 2D-points.
				 */
				void init ( int aSize, int aDim, int aFixed, int bSize, int bDim, int resultDim );

				/**
				 * An implenemtation of levenberg-marquard for sparse data of the particular form (Notice the first index of a):
				 * Parameter: (a2, ..., am, b1, ..., bn)
				 * Goal-vector: (x11, ... , x1m, x21, ... , x2m, ... , xn1, ..., xnm) with
				 * xij only depend on aj and bi (e.g. a:Cameras, b:3D-Points, xij image point depends on Camera j and 3D-Point i)
				 * aj, bi and xij may be vectors
				 *
				 * Result is in params
				 * @param a The vector containing CvMat* for every (a1, ..., am)
				 * @param b The vector containing CvMat* for every (b1, ..., bn)
				 * @param func the function whose parameters are searched.
				 * @param invCovar the invers of a covariance matrix. If it is a column-vector, it is assumed to hold the diagonal elements.
				 * @param errorFunc calculate the error goal - val, probably attenuates the error-vector according to a cost-function (p. 622).
				 * @param eps if norm(error)&lt;eps or the differens of two subsequent error-norms is less than eps, the algo stops.
				 * @param maxIt the algorithm stops after maxIt iterations.
				 */
				void solve ( VecColl &a, VecColl &b,
				             FuncT &func,
				             vector<int> &missingIndizes = vector<int> ( 0 ),
				             ErrorT &errorFunc = SparseZeroErrorFunc(),
				             vector<InvCovarT> &invCovar = DefaultSparseInvCovar,
				             double eps = FLT_EPSILON, int maxIt = 10000 );

				/**
				 * returns the error after the last call of this algorithm
				 * @return the error
				 */
				double getLastError() const
				{
					return lastError;
				}

				/**
				 * returns the number of iterations needed by the last call of this algorithm.
				 * @return the number of iterations
				 */
				int getLastIterations() const
				{
					return lastIterations;
				}

				/**
				 * returns the error before starting of minimization
				 * @return the error
				 */
				double getInitialError() const
				{
					return initialError;
				}

			private:

				bool created;
				double initialError;
				double lastError;
				int lastIterations;

				int n; //number of 3D-Points
				int m; //number of Cameras
				int mInc; //number of Cameras + 1, this represents the const first Camera

				MatColl Aij;//derivative matrices dxij/daj
				MatColl Bij;//derivative matrices dxij/dbj
				VecColl newA;
				VecColl newB;
				VecColl DeltaA;
				VecColl DeltaB;
				VecColl err;
				VecColl tempA;
				VecColl tempB;
				VecColl errNew;
				VecColl val;
				MatColl U;
				MatColl V;
				VecColl erra;
				VecColl errb;
				MatColl Wij;
				MatColl Yij;

				//for jacobi function
				VecColl value;

				AMatrix S;
				AVector SFactorizeSpace;
				AVector e;
				AVector deltaAMat;
				std::vector<int> minSIdx;
				std::vector<int> maxSIdx;

				//temporary transposes
				AMatrix WijTrans;
				AMatrix YijTrans;
				AMatrix VTrans;

			private:

				/**
				 * The Jacobian matrix is calculated here.
				 */
				void jacobiFunc ( VecColl &a, VecColl &b,
				                  FuncT &func,
				                  vector<AMatrix> &Aij, vector<AMatrix> &Bij );

				//LM Step, computes Delta and adds it to params to get newParams.
				void basicLMStep ( vector<AMatrix> &Aij, vector<AMatrix>&Bij,
				                   VecColl &erra, VecColl &errb,
				                   double mu,
				                   VecColl &DeltaA, VecColl &DeltaB,
				                   vector<InvCovarT> &invCovar );

				//some utility functions
				int ijToIdx ( int i, int j )
				{
					return i * mInc + j;
				}

				void computeJTErrorA ( vector<AMatrix> &Aij, VecColl &err,
				                       VecColl &erra, InvCovarT *invCovar );
				void computeJTErrorB ( vector<AMatrix> &Bij, VecColl &err,
				                       VecColl &errb, VecColl *invCovar );

				void createMats ( int aSize, int aDim, int aFixed, int bSize, int bDim, int resultDim );

				//mat = mat + lambda * I
				void lambdafy ( AMatrix &mat, double lambda )
				{
					for ( int i = 0; i < (int)mat.rows(); ++i ) mat ( i, i ) += lambda;
				}

				template<class what>
				void lmSparseSub ( vector<what> &m1, vector<what> &m2, vector<what> &result )
				{
					assert ( m1.size() == m2.size() && m2.size() == result.size() );
					for ( uint i = 0; i < m1.size(); ++i )
					{
						if ( isNullArray( m1[i] ) || isNullArray( m2[i] ) ) result[i] = 0;
						else result[i] = m1[i] - m2[i];
					}
				}

				template<class what>
				void lmSparseAdd ( vector<what> &m1, vector<what> &m2, vector<what> &result )
				{
					assert ( m1.size() == m2.size() && m2.size() == result.size() );
					for ( uint i = 0; i < m1.size(); ++i ) result[i] = m1[i] + m2[i];
				}

				template<class what>
				void lmSparseCopy ( vector<what> &m1, vector<what> &m2 )
				{
					assert ( m1.size() == m2.size() );
					copy( m1.begin(), m1.end(), m2.begin() );
				}

				double lmSparseNorm ( VecColl &m )
				{
					double sqrsum = 0.0;
					for ( uint i = 0; i < m.size(); ++i ) sqrsum += getSqrNorm ( m[i] );
					return sqrsum;
				}

				double lmMaxNorm ( VecColl &m )
				{
					double result = 0.0;
					for ( uint i = 0; i < m.size(); ++i ) result = max ( result, getMaxNorm ( m[i] ) );
					return result;
				}

				void lmCreateMat ( MatColl &m, int size, int rows, int cols )
				{
					m.resize( size );
					for ( MatIterator it = m.begin(); it != m.end(); ++it )
					{
						( *it ).resize( rows, cols );
					}
				}

				void lmCreateVec ( VecColl &m, int size, int rows )
				{
					m.resize( size );
					for ( VecIterator it = m.begin(); it != m.end(); ++it )
					{
						( *it ).resize( rows );
					}
				}

				template<class what>
				void lmSparseScaleAdd ( const vector<what> &sum1, const double scale,
				                        const vector<what> &sum2, vector<what> &dest )
				{
					for ( uint i = 0; i < sum2.size(); ++i ) dest[i] = scale * sum1[i] + sum2[i];
				}

				void computeJTErrorA ( vector<AMatrix> &Aij, VecColl &err,
				                       VecColl &erra, vector<InvCovarT> &invCovar )
				{
					for ( int j = 0; j < m; ++j )
					{
						erra[j] = 0;
						for ( int i = 0; i < n; ++i )
						{
							int idx = ijToIdx ( i,j );
							InvCovarT invCov = invCovar[idx];
							erra[j] = erra[j] + this->covarProduct ( Aij[idx], invCov, err[idx] );
						}
					}
				}

				void computeJTErrorB ( vector<AMatrix> &Bij, VecColl &err,
				                       VecColl &errb, vector<InvCovarT> &invCovar )
				{
					for ( int i = 0; i < n; ++i )
					{
						errb[i] = 0;
						for ( int j = 0; j < mInc; j++ )
						{
							int idx = ijToIdx ( i,j );
							InvCovarT invCov = invCovar[idx];

							errb[i] = errb[i] + this->covarProduct ( Bij[idx], invCov, err[idx] );
						}
					}
				}


		};


		template<class FuncT, class ErrorT, class InvCovarT>
		LMSparse<FuncT, ErrorT, InvCovarT>::LMSparse ( int aSize, int aDim, int aFixed, int bSize, int bDim, int resultDim ) :
				created ( false ),
				initialError ( 0 ),
				lastError ( 0 ),
				lastIterations ( 0 ),
				S(),
				SFactorizeSpace(),
				e(),
				deltaAMat()
		{
			init ( aSize, aDim, aFixed, bSize, bDim, resultDim );
		}

		template<class FuncT, class ErrorT, class InvCovarT>
		LMSparse<FuncT, ErrorT, InvCovarT>::~LMSparse()
		{
		}

		template<class FuncT, class ErrorT, class InvCovarT>
		void LMSparse<FuncT, ErrorT, InvCovarT>::init ( int aSize, int aDim, int aFixed, int bSize, int bDim, int resultDim )
		{
			if ( aSize > 0 && aDim > 0 && aFixed >= 0 && bSize > 0 && bDim > 0 && resultDim > 0 )
			{
				//printf("LMsparse parameter: aSize: %d, aDim: %d, aFixed: %d, bSize: %d, bDim: %d, resultDim: %d\n", aSize, aDim, aFixed, bSize, bDim, resultDim);
				createMats ( aSize, aDim, aFixed, bSize, bDim, resultDim );
			}
// 			else
// 			{
// 				DBGOUT(DBGNORMAL, "LMsparse parameter: aSize: %d, aDim: %d, aFixed: %d, bSize: %d, bDim: %d, resultDim: %d\n", aSize, aDim, aFixed, bSize, bDim, resultDim);
// 				error("conditions not met\n");
// 			}
		}

		template<class FuncT, class ErrorT, class InvCovarT>
		void LMSparse<FuncT, ErrorT, InvCovarT>::createMats ( int aSize, int aDim, int aFixed, int bSize, int bDim, int resultDim )
		{
			m = aSize;
			mInc = aSize + aFixed;
			n = bSize;
			int resultSize = ( mInc ) * n;//first a is unknown/const, only the LMSparseFunction knowns it

			lmCreateMat ( Aij, mInc*n, resultDim, aDim );
			lmCreateMat ( Bij, mInc*n, resultDim, bDim );
			lmCreateVec ( newA, m, aDim );
			lmCreateVec ( newB, n, bDim );
			lmCreateVec ( DeltaA, m, aDim );
			lmCreateVec ( DeltaB, n, bDim );
			lmCreateVec ( err, resultSize, resultDim );
			lmCreateVec ( tempA, m, aDim );
			lmCreateVec ( tempB, n, bDim );
			lmCreateVec ( errNew, resultSize, resultDim );
			lmCreateVec ( val, resultSize, resultDim );
			lmCreateMat ( U, m, aDim, aDim );
			lmCreateMat ( V, n, bDim, bDim );
			lmCreateVec ( erra, m, aDim );
			lmCreateVec ( errb, n, bDim );
			lmCreateMat ( Wij, resultSize, aDim, bDim );
			lmCreateMat ( Yij, resultSize, aDim, bDim );

			WijTrans.resize( bDim, aDim );
			YijTrans.resize( bDim, aDim );
			VTrans.resize( bDim, bDim );

			lmCreateVec ( value, resultSize, resultDim );

			S.resize ( m * aDim, m * aDim );
			SFactorizeSpace.resize ( S.rows() );
			e.resize ( m * aDim );
			deltaAMat.resize ( m * aDim );
			minSIdx.resize( m * aDim );
			std::fill( minSIdx.begin(), minSIdx.end(), 0 );
			maxSIdx.resize( m * aDim );
			std::fill( maxSIdx.begin(), maxSIdx.end(), m * aDim );

			S = 0;
			SFactorizeSpace = 0;
			e = 0;
			deltaAMat = 0;

			created = true;
		}

		template<class FuncT, class ErrorT, class InvCovarT>
		void LMSparse<FuncT, ErrorT, InvCovarT>::solve ( VecColl &a, VecColl &b,
		                FuncT &func,
		                vector<int> &missingIndizes,
		                ErrorT &errorFunc,
		                vector<InvCovarT> &invCovar,
		                double eps, int maxIt )
		{
			assert_msg ( created, "You must call init fist\n" );
			//test whether parameters are okay or not
			assert ( newA.size() == a.size() );
			assert ( newB.size() == b.size() );
			assert_msg ( isNullArray( a[0] ) || a[0].rows() == newA[0].rows(), "%zd != %zd", a[0].rows(), newA[0].rows() );
			assert_msg ( isNullArray( b[0] ) || b[0].rows() == newB[0].rows(), "%zd != %zd", b[0].rows(), newB[0].rows() );

			//this is for the dummy covar to behave properly
			if ( invCovar.size() < ( uint ) ( mInc * n ) ) invCovar.resize ( mInc * n );

			double err1 = 0.0;
			double err2 = 0.0;
			double mu = 0.001;
			double nu = 2.0;
			double rho = 2.0;
			int count = 0;
			bool stop = false;

			jacobiFunc ( a, b, func, Aij, Bij );

			//zero the missing indizes
			for ( uint i = 0; i < missingIndizes.size(); ++i )
			{
				setNullArray( Aij[missingIndizes[i]] );
				setNullArray( Bij[missingIndizes[i]] );
			}

			//calculate current error
			func ( a, b, val );
			errorFunc ( val, err );

			lastError = initialError = err1 = lmSparseNorm ( err );
			lastIterations = 0;

			computeJTErrorA ( Aij, err, erra, invCovar );
			computeJTErrorB ( Bij, err, errb, invCovar );

			if ( max ( lmMaxNorm ( erra ), lmMaxNorm ( errb ) ) < eps ) return;

			do
			{
				do //LM-iteration
				{
					++count;
// 					++ ( *o );//progressobserver

// 					printf("current val:\n");
// 					for(uint z = 0; z < a.size(); ++z){printf("a[%d]:\n", z); printMat(*a[z]);}
// 					for(uint z = 0; z < b.size(); ++z){printf("b[%d]:\n", z); printMat(*b[z]);}
// 					for(uint z = 0; z < val.size(); ++z){printf("\n"); printMats(*val[z], *err[z]);}
// 					printf("err1: %.30f - err2: %.30f\n", err1, err2);
// 					printf("lambda: %f\n", mu);
// 					getchar();


					jacobiFunc ( a, b, func, Aij, Bij );
					//zero the missing indizes
					for ( uint i = 0; i < missingIndizes.size(); ++i )
					{
						Aij[missingIndizes[i]].resize( 0, 0 );
						Bij[missingIndizes[i]].resize( 0, 0 );
					}

					computeJTErrorA ( Aij, err, erra, invCovar );
					computeJTErrorB ( Bij, err, errb, invCovar );

					basicLMStep ( Aij, Bij, erra, errb, mu, DeltaA, DeltaB, invCovar );

					assert_msg ( lmSparseNorm ( DeltaA ) >= 0, "sparseNorm = %f\n", lmSparseNorm ( DeltaA ) );
					assert_msg ( lmSparseNorm ( DeltaB ) >= 0, "sparseNorm = %f\n", lmSparseNorm ( DeltaB ) );

					if ( lmSparseNorm ( DeltaA ) + lmSparseNorm ( DeltaB ) <= eps * ( lmSparseNorm ( a ) + lmSparseNorm ( b ) ) )
					{
						stop = true;
					}
					else
					{
						lmSparseAdd ( a, DeltaA, newA );
						lmSparseAdd ( b, DeltaB, newB );
						func ( newA, newB, val );
						errorFunc ( val, errNew );
						err2 = lmSparseNorm ( errNew );

						lmSparseScaleAdd ( DeltaA, mu, erra, tempA );
						lmSparseScaleAdd ( DeltaB, mu, errb, tempB );

						double divisor = 0.0;
						for ( uint i = 0; i < DeltaA.size(); ++i )
						{
							divisor += dot ( DeltaA[i], tempA[i] );
						}
						for ( uint i = 0; i < DeltaB.size(); ++i )
						{
							divisor += dot ( DeltaB[i], tempB[i] );
						}
						rho = ( err1 - err2 ) /divisor;

// 						printf("divisor: %f, rho: %f, err1: %f, err2: %f\n", divisor, rho, err1, err2);
						if ( rho > 0 )
						{
							lmSparseCopy ( newA, a );
							lmSparseCopy ( newB, b );
							lmSparseCopy ( errNew, err );

							assert ( err2 < err1 );
							err1 = err2;

							if ( max ( lmMaxNorm ( erra ), lmMaxNorm ( errb ) ) < eps )
							{
								stop = true;
							}

							double magic = ( 2 * rho - 1 );
							//printf("mu: %f, magic: %f\n", mu, magic);
							mu = mu * MAX ( 1.0/3.0, 1.0 - magic * magic * magic );
							nu = 2;
						}
						else
						{
							mu = mu * nu;
							nu =  2 * nu;
						}

					}
				}
				while ( rho < 0 && !stop );

			}
			while ( count < maxIt && !stop );
// 			DBGOUT(DBGALL, "LMSparse: err: %f, count: %d of %d\n", err1, count, maxIt);
			lastError = err1;
			lastIterations = count;
		}

		template<class FuncT, class ErrorT, class InvCovarT>
		void LMSparse<FuncT, ErrorT, InvCovarT>::
		jacobiFunc ( VecColl &a, VecColl &b,
		             FuncT &func, vector<AMatrix> &Aij, vector<AMatrix> &Bij )
		{
			func ( a, b, value );

			double delta = 0.000001;

			//derivative Matrices (Jacobian-sub-matrices)
			for ( int i = 0; i < n; ++i )
			{
				AVector tmpB = b[i]; //yes, a copy
				for ( int j = 0; j < m; j++ )
				{
					AVector tmpA = a[j]; //yes, a copy

					int idx = ijToIdx ( i,j );
					AVector& val1xij = value[idx]; //a reference!

					//build jacobian Aij as ratios
					int outDim = val1xij.rows();//a point
					int inDim = a[j].rows();//should be 16, is a cam

					for ( int out = 0; out < outDim; ++out )
						for ( int in = 0; in < inDim; ++in )
						{
							delta = max ( fabs ( ( a[j] ) ( in ) * 0.0001 ), 0.000001 ); //magic numbers :)
							tmpA ( in ) = ( a[j] ) ( in ) + delta;
							AVector val2xij = func ( tmpA, tmpB, j );//sparse function evaluation
							Aij[idx]( out, in ) = ( val2xij ( out ) - val1xij ( out ) ) /delta;//quotient of differences
							tmpA ( in ) = a[j]( in );
						}

					//build jacobian Bij as ratios
					//outdim stays the same. is a point again.
					inDim = b[i].rows();//should be 3 or 4, is a point3D
					for ( int out = 0; out < outDim; ++out )
						for ( int in = 0; in < inDim; ++in )
						{
							delta = max ( fabs ( b[i]( in ) * 0.0001 ), 0.000001 ); //magic numbers :)
							tmpB ( in ) = b[i]( in ) + delta;
							AVector val2xij = func ( tmpA, tmpB, j );//sparse function evaluation
							Bij[idx]( out, in ) = ( val2xij ( out ) - val1xij ( out ) ) /delta;//quotient of differences
							tmpB ( in ) = b[i]( in );
						}
				}
			}
			//the values for the fixed a's are trailing
			//Now compute the appropriate Bi0
			if ( m != mInc )
				for ( int i = 0; i < n; ++i )
				{
					AVector tmpB = b[i]; //copy
					for ( int j = m; j < mInc; j++ ) //this is important!
					{
						int idx = ijToIdx ( i,j );
						AVector& val1xij = value[idx];//a reference!
						int outDim = val1xij.rows();
						int inDim = b[i].rows();//should be 3 or 4, is a point3D
						for ( int out = 0; out < outDim; ++out )
							for ( int in = 0; in < inDim; ++in )
							{
								delta = max ( fabs ( ( b[i] ) ( in ) * 0.0001 ), 0.000001 ); //magic numbers :)
								tmpB ( in ) = ( b[i] ) ( in ) + delta;
								AVector val2xij = func ( tmpB, j - m );//sparse function evaluation, j-m adresses the camera known by the function
								Bij[idx]( out, in ) = ( val2xij ( out ) - val1xij ( out ) ) /delta;//quotient of differences
								tmpB ( in ) = b[i]( in );
							}

					}
				}

		}

		//LM Step, computes Delta and adds it to params to get newParams.
		template<class FuncT, class ErrorT, class InvCovarT>
		void LMSparse<FuncT, ErrorT, InvCovarT>::basicLMStep ( vector<AMatrix> &Aij, vector<AMatrix>&Bij,
		                VecColl &erra, VecColl &errb,
		                double mu,
		                VecColl &DeltaA, VecColl &DeltaB,
		                vector<InvCovarT> &invCovar )
		{

			//Uj and erraj
			for ( int j = 0; j < m; ++j )
			{
				U[j] = 0;
				for ( int i = 0; i < n; ++i )
				{
					int idx = ijToIdx ( i,j );
					InvCovarT invCov = invCovar[idx];

					U[j] = U[j] + covarProduct ( Aij[idx], invCov, Aij[idx] );
				}
				lambdafy ( U[j], mu );
			}

			//Vi and errbi
			for ( int i = 0; i < n; ++i )
			{
				V[i] = 0;
				for ( int j = 0; j < mInc; j++ )
				{
					int idx = ijToIdx ( i,j );
					InvCovarT invCov = invCovar[idx];

					V[i] = V[i] + covarProduct ( Bij[idx], invCov, Bij[idx] );
				}
				lambdafy ( V[i], mu );
			}

			//Wij and Yij
			for ( int i = 0; i < n; ++i )
			{
				for ( int j = 0; j < m; j++ )
				{
					int idx = ijToIdx ( i,j );
					InvCovarT invCov = invCovar[idx];

					Wij[idx] = covarProduct ( Aij[idx], invCov, Bij[idx] );

					//instead inverting V*, LU-Decomposition is used to get Yij
					trans( V[i], VTrans );
					trans( Wij[idx], WijTrans );
// 					solveSVD ( VTrans, WijTrans, YijTrans );//TODO:LU is sufficient
// 					trans(YijTrans, Yij[idx]);
					/*bool ok = */solveLUinPlace( VTrans, WijTrans );
					trans( WijTrans, Yij[idx] );
				}
			}

			//now compute dalta_a and delta_b, first delat_a
// 			printf("S\n");
			//S is m x m of a[j]-Rows x a[j]-Rows
			int x = 0;
			int y = 0;
			int width = DeltaA[0].rows();
			int height = DeltaA[0].rows();
			S = 0;

			y = 0;
			for ( int j = 0; j < m; ++j )
			{
				x = 0;
				for ( int k = 0; k < m; ++k )
				{
					AMatrix ajMask = S ( Range ( y, y + height - 1 ), Range ( x, x + width - 1 ) );//reference

					for ( int i = 0; i < n; ++i )
					{
						int idxY = ijToIdx ( i,j );
						int idxW = ijToIdx ( i,k );

						AMatrix temp( ajMask.rows(), ajMask.cols());
						matMul2_BT ( Yij[idxY], Wij[idxW], temp );
						ajMask = ajMask - temp;
					}
					x += width; //move masking rectangle
				}
				y += height;
			}
// 			printf("vor diag\n");
// 			printMat(S);
// 			printf("S2\n");
			//add Uj to the diagonal blocks
			x = y = 0;
			for ( int diag = 0; diag < m; ++diag )
			{
				AArray<numberT, 2> a = S ( Range ( y, y + height - 1 ), Range ( x, x + width - 1 ) );
				a = a + U[diag];

				x += width; //move masking rectangle
				y += height;
			}
			//after S is computed, compute e, ej has a[j].rows() rows
			x = y = 0;
			for ( int j = 0; j < m; ++j )
			{
				AVector ejMask = e ( Range ( y, y + height - 1 ) );
				ejMask = erra[j];
				for ( int i = 0; i < n; ++i )
				{
					int idx = ijToIdx ( i,j );
					AVector temp( ejMask.rows() );
					matMul2( Yij[idx], errb[i], temp );
					ejMask = ejMask - temp;
				}
				y += height;
			}

			//core::util::ADORE_SOLVE(S, e, deltaAMat, CV_LU);

// 			printf("solving\n");
			factorizeLDLT ( S, SFactorizeSpace );
			solveLDLTc ( S, e, deltaAMat, SFactorizeSpace, minSIdx, maxSIdx );
// 			pressKey();

			//now compute delta_b by back-substitution
// 			printf("deltaB\n");
			x = 0;

			for ( int i = 0; i < n; ++i )
			{
				y = 0;
				AVector rightB = errb[i];
				for ( int j = 0; j < m; ++j )
				{
					int idx = ijToIdx ( i,j );
					AVector subDeltaA = deltaAMat ( Range ( y, y + height - 1 ) );

					AVector temp( rightB.rows() );
					matMul2_AT ( Wij[idx], subDeltaA, temp );
					rightB = rightB - temp;
					y += height;
				}
				DeltaB[i] = solveSVD ( V[i], rightB );
			}

			//update the parameter estimate for a
// 			printf("deltaA\n");
// 			printMat(deltaAMat);
// 			pressKey();
			x = y = 0;
			for ( uint j = 0; j < DeltaA.size(); ++j )
			{
				DeltaA[j] = deltaAMat ( Range ( y, y + height - 1 ) );
				y += height;
			}
		}


	}//ns

}

#endif
