

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_MINIMIZATIONLMSPARSESIMPLE_H
#define ADORE_MINIMIZATIONLMSPARSESIMPLE_H

#include <adrcore/util/debug.h>
#include <adrcore/util/mathutil.h>
#include <adrcore/util/ldlt.h>
#include <adrcore/minimization/lmconstants.h>
#include <adrcore/minimization/sparsedefaulterrorfunc.h>
#include <adrcore/minimization/baminfunctionsimple.h>
#include <cfloat>
#include <vector>
#include <adrcore/util/matmul2.h>
#include <adrcore/util/solvesvd.h>

namespace adore
{
	namespace core
	{

		using namespace std;


		/**
		 * THIS IS THE SIMPLIFIED VERSION FOR Projective Bundle Adjustment only
		 * This is a Levenberg-Marquard minimizatin algorithm, as described in Hertley-Zisserman, p.613
		 * It is for a special kind of Problem:
		 *
		 * Given:
		 * - a function f:
		 * f(a1, a2, ..., am, b1, b2, ..., bn) = (x11, ..., x1m, x21, ..., x2m, xn1, ..., xnm)
		 * dxij/dak = 0, j != k
		 * dxij/dbk = 0, i != k
		 * - a desired value of the xij-vector
		 * - an initial estimate for the aj and bi.
		 *
		 * Result:
		 * aj and bj that minimize ||(f(aj, bi) - (xij))||
		 *
		 *  
		*/
		template<class ErrorT, class MinFunctionT>
		class LMSparseSimple
		{
				typedef MinFunctionT FuncT;
				typedef typename FuncT::T_numtype numberT;
				typedef vector<AMatrix> MatColl;
				typedef vector<AVector> VecColl;
				typedef typename MatColl::iterator MatIterator;
				typedef typename VecColl::iterator VecIterator;

				enum { CamDim = MinFunctionT::CamDim };

			public:
				/**
				 * Constructor.
				 * @param camSize number of paramters of type a.
				 * @param camDim The dimension of the parameters of type a, e.g. 12 would be a good choice for camera matrices.
				 * @param camFixed number of Parameters that are fixed and only known by the function whose parameters have to be adjusted.
				 * This is to allow a fixed camera (namely the identity cam) in BA.
				 * @param pointSize number of parameters of type b.
				 * @param pointDim The dimension of the parameters of type b, e.g. 3 for 3D-points.
				 * @param resultDim the dimension of the result, e.g. 2 for inhomogeneous 2D-points.
				 */
				LMSparseSimple ( int camSize = -1, int pointSize = -1 );

				/**
				 * initializes the internal variables.
				 * @param camSize number of paramters of type a.
				 * @param camDim The dimension of the parameters of type a, e.g. 12 would be a good choice for camera matrices.
				 * @param camFixed number of Parameters that are fixed and only known by the function whose parameters have to be adjusted.
				 * This is to allow a fixed camera (namely the identity cam) in BA.
				 * @param pointSize number of parameters of type b.
				 * @param pointDim The dimension of the parameters of type b, e.g. 3 for 3D-points.
				 * @param resultDim the dimension of the result, e.g. 2 for inhomogeneous 2D-points.
				 */
				void init ( int camSize, int pointSize );

				/**
				 * An implenemtation of levenberg-marquard for sparse data of the particular form (Notice the first index of a):
				 * Parameter: (a2, ..., am, b1, ..., bn)
				 * Goal-vector: (x11, ... , x1m, x21, ... , x2m, ... , xn1, ..., xnm) with
				 * xij only depend on aj and bi (e.g. a:Cameras, b:3D-Points, xij image point depends on Camera j and 3D-Point i)
				 * aj, bi and xij may be vectors
				 *
				 * Result is in params
				 * @param a The vector containing CvMat* for every (a1, ..., am)
				 * @param b The vector containing CvMat* for every (b1, ..., bn)
				 * @param invCovar the invers of a covariance matrix. If it is a column-vector, it is assumed to hold the diagonal elements.
				 * @param errorFunc calculate the error goal - val, probably attenuates the error-vector according to a cost-function (p. 622).
				 * @param eps if norm(error)&lt;eps or the differens of two subsequent error-norms is less than eps, the algo stops.
				 * @param maxIt the algorithm stops after maxIt iterations.
				 */
				void solve ( VecColl &a, VecColl &b,
				             vector<int> &missingIndizes = vector<int> ( 0 ),
				             ErrorT &errorFunc = SparseZeroErrorFunc(),
				             double eps = FLT_EPSILON, int maxIt = 10000 );

				/**
				 * returns the error after the last call of this algorithm
				 * @return the error
				 */
				double getLastError() const
				{
					return lastError;
				}

				/**
				 * returns the number of iterations needed by the last call of this algorithm.
				 * @return the number of iterations
				 */
				int getLastIterations() const
				{
					return lastIterations;
				}

				/**
				 * returns the error before starting of minimization
				 * @return the error
				 */
				double getInitialError() const
				{
					return initialError;
				}

			private:

				const int camDim;
				const int pointDim;
				const int resultDim;

				bool created;
				double initialError;
				double lastError;
				int lastIterations;

				int n; //number of 3D-Points
				int m; //number of Cameras

				MatColl Aij;//derivative matrices dxij/daj
				MatColl Bij;//derivative matrices dxij/dbj
				VecColl newA;
				VecColl newB;
				VecColl DeltaA;
				VecColl DeltaB;
				VecColl err;
				VecColl tempA;
				VecColl tempB;
				VecColl errNew;
				VecColl val;
				MatColl U;
				MatColl V;
				VecColl erra;
				VecColl errb;
				MatColl Wij;
				MatColl Yij;

				AMatrix S;
				AVector SFactorizeSpace;
				AVector e;
				AVector deltaAMat;
				std::vector<int> minSIdx;
				std::vector<int> maxSIdx;
				AArray<int, 2> SStructure;

				//temporary transposes
				AMatrix WijTrans;
				AMatrix YijTrans;
				AMatrix VTrans;

				FuncT func;
				SVDSolveStorage<numberT, 2> store;

				vector<bool> missing;
			private:

				/**
				 * The Jacobian matrix is calculated here.
				 */
				void jacobiFunc ( VecColl &a, VecColl &b, vector<AMatrix> &Aij, vector<AMatrix> &Bij );

				//LM Step, computes Delta and adds it to params to get newParams.
				void basicLMStep ( vector<AMatrix> &Aij, vector<AMatrix>&Bij,
				                   VecColl &erra, VecColl &errb,
				                   double mu,
				                   VecColl &DeltaA, VecColl &DeltaB );

				//some utility functions
				inline int ijToIdx ( const int i, const int j )
				{
					return i * m + j;
				}
				inline void idxToij ( const int idx, int &i, int &j )
				{
					div_t d = div( idx, m );
					i = d.quot;
					j = d.rem;
				}

				void createMats ( int camSize, int pointSize );

				//mat = mat + lambda * I
				void lambdafy ( AMatrix &mat, double lambda )
				{
					for ( int i = 0; i < (int)mat.rows(); ++i ) mat ( i, i ) += lambda;
				}

				template<class what>
				void lmSparseSub ( vector<what> &m1, vector<what> &m2, vector<what> &result )
				{
					assert ( m1.size() == m2.size() && m2.size() == result.size() );
					for ( uint i = 0; i < m1.size(); ++i )
					{
						if ( isNullArray ( m1[i] ) || isNullArray ( m2[i] ) ) result[i] = 0;
						else result[i] = m1[i] - m2[i];
					}
				}

				template<class what>
				void lmSparseAdd ( vector<what> &m1, vector<what> &m2, vector<what> &result )
				{
					assert ( m1.size() == m2.size() && m2.size() == result.size() );
					for ( uint i = 0; i < m1.size(); ++i ) result[i] = m1[i] + m2[i];
				}

				template<class what>
				void lmSparseCopy ( vector<what> &m1, vector<what> &m2 )
				{
					assert ( m1.size() == m2.size() );
					copy ( m1.begin(), m1.end(), m2.begin() );
				}

				double lmSparseNorm ( VecColl &m )
				{
					double sqrsum = 0.0;
					for ( uint i = 0; i < m.size(); ++i ) sqrsum += getSqrNorm ( m[i] );
					return sqrsum;
				}

				double lmMaxNorm ( VecColl &m )
				{
					double result = 0.0;
					for ( uint i = 0; i < m.size(); ++i ) result = max ( result, getMaxNorm ( m[i] ) );
					return result;
				}

				void lmCreateMat ( MatColl &m, int size, int rows, int cols )
				{
					m.resize ( size );
					for ( MatIterator it = m.begin(); it != m.end(); ++it )
					{
						( *it ).resize ( rows, cols );
					}
				}

				void lmCreateVec ( VecColl &m, int size, int rows )
				{
					m.resize ( size );
					for ( VecIterator it = m.begin(); it != m.end(); ++it )
					{
						( *it ).resize ( rows );
					}
				}

				template<class what>
				void lmSparseScaleAdd ( const vector<what> &sum1, const double scale,
				                        const vector<what> &sum2, vector<what> &dest )
				{
					for ( uint i = 0; i < sum2.size(); ++i ) dest[i] = scale * sum1[i] + sum2[i];
				}

				inline void computeJTErrorA ( vector<AMatrix> &Aij, VecColl &err,
				                              VecColl &erra )
				{
					for ( int j = 0; j < m; ++j )
					{
						erra[j] = 0;
						for ( int i = 0; i < n; ++i )
						{
							int idx = ijToIdx ( i,j );
							matMulAdd2_AT ( Aij[idx], err[idx], erra[j] );
						}
					}
				}

				inline void computeJTErrorB ( vector<AMatrix> &Bij, VecColl &err,
				                              VecColl &errb )
				{
					for ( int i = 0; i < n; ++i )
					{
						errb[i] = 0;
						for ( int j = 0; j < m; j++ )
						{
							int idx = ijToIdx ( i,j );
							matMulAdd2_AT ( Bij[idx], err[idx], errb[i] );
						}
					}
				}

				inline bool isMissing( int idx )
				{
					return missing[idx];
				}
				inline void setMissing( int idx )
				{
					missing[idx] = true;
				}

				void examineStructure()
				{
// 					printf("examining structure\n");
					//calculate banded structure
					//use S as book-keeping space
					SStructure = 0;
					for ( int k = 0; k < m; ++k )
					{
						for ( int j = 0; j < m; ++j )
						{
							for ( int i = 0; i < n; ++i )
							{
								if ( !( isMissing( ijToIdx( i, j ) ) || isMissing( ijToIdx( i, k ) ) ) )
								{
									SStructure( j, k ) = 1;
								}
							}
						}
					}
// 					printMat(DBGNORMAL, SStructure);

					for ( int sRow = 0; sRow < m; ++sRow )
					{
						for ( int sCol = 0; sCol <= sRow; ++sCol )
						{
							if ( SStructure( sRow, sCol ) == 1 )
							{
								vector<int>::iterator bit1 = minSIdx.begin() + sRow * camDim;
								vector<int>::iterator bit2 = bit1 + camDim;
								int val = sCol * camDim;
								std::fill( bit1, bit2, val );

// 								printf("%d - %d\n", sCol, val);
								break;
							}
						}
					}
					for ( int sCol = m - 1; sCol >= 0; --sCol )
					{
						for ( int sRow = m - 1; sRow >= sCol; --sRow )
						{
							if ( SStructure( sRow, sCol ) == 1 )
							{
								vector<int>::iterator bit1 = maxSIdx.begin() + sCol * camDim;
								vector<int>::iterator bit2 = bit1 + camDim;
								int val = ( sRow + 1 ) * camDim;
								std::fill( bit1, bit2, val );

// 								printf("%d - %d\n", (sRow + 1), val);
								break;
							}
						}
					}
				}

		};


		template< class ErrorT, class MinFunctionT>
		LMSparseSimple<ErrorT, MinFunctionT>::LMSparseSimple ( int camSize, int pointSize ) :
				camDim ( CamDim ),
				pointDim ( 3 ),
				resultDim ( 2 ),
				created ( false ),
				initialError ( 0 ),
				lastError ( 0 ),
				lastIterations ( 0 ),
				S(),
				SFactorizeSpace(),
				e(),
				deltaAMat()
		{
			init ( camSize, pointSize );
		}

		template< class ErrorT, class MinFunctionT>
		void LMSparseSimple<ErrorT, MinFunctionT>::init ( int camSize, int pointSize )
		{
			if ( camSize > 0 && pointSize > 0 )
			{
				//printf("LMsparse parameter: camSize: %d, camDim: %d, camFixed: %d, pointSize: %d, pointDim: %d, resultDim: %d\n", camSize, camDim, camFixed, pointSize, pointDim, resultDim);
				createMats ( camSize, pointSize );
			}
// 			else
// 			{
// 				DBGOUT(DBGNORMAL, "LMsparse parameter: camSize: %d, camDim: %d, camFixed: %d, pointSize: %d, pointDim: %d, resultDim: %d\n", camSize, camDim, camFixed, pointSize, pointDim, resultDim);
// 				error("conditions not met\n");
// 			}
		}

		template< class ErrorT, class MinFunctionT>
		void LMSparseSimple<ErrorT, MinFunctionT>::createMats ( int camSize, int pointSize )
		{
			m = camSize;
			n = pointSize;
			int resultSize = m * n;//first a is unknown/const, only the LMSparseSimpleFunction knowns it

			lmCreateMat ( Aij, m*n, resultDim, camDim );
			lmCreateMat ( Bij, m*n, resultDim, pointDim );
			lmCreateVec ( newA, m, camDim );
			lmCreateVec ( newB, n, pointDim );
			lmCreateVec ( DeltaA, m, camDim );
			lmCreateVec ( DeltaB, n, pointDim );
			lmCreateVec ( err, resultSize, resultDim );
			lmCreateVec ( tempA, m, camDim );
			lmCreateVec ( tempB, n, pointDim );
			lmCreateVec ( errNew, resultSize, resultDim );
			lmCreateVec ( val, resultSize, resultDim );
			lmCreateMat ( U, m, camDim, camDim );
			lmCreateMat ( V, n, pointDim, pointDim );
			lmCreateVec ( erra, m, camDim );
			lmCreateVec ( errb, n, pointDim );
			lmCreateMat ( Wij, resultSize, camDim, pointDim );
			lmCreateMat ( Yij, resultSize, camDim, pointDim );

			WijTrans.resize( pointDim, camDim );
			YijTrans.resize( pointDim, camDim );
			VTrans.resize( pointDim, pointDim );
			store.init ( pointDim, pointDim, camDim );

			int sDim = m * camDim;
			S.resize ( sDim, sDim );
			SStructure.resize ( m, m );
			//for factorization, S.rows() - 1 would suffice, but we use it also in solve:
			SFactorizeSpace.resize ( sDim );
			e.resize ( sDim );
			deltaAMat.resize ( sDim );
			minSIdx.resize( sDim );
			std::fill( minSIdx.begin(), minSIdx.end(), 0 );
			maxSIdx.resize( sDim );
			std::fill( maxSIdx.begin(), maxSIdx.end(), sDim );

			S = 0;
			SFactorizeSpace = 0;
			e = 0;
			deltaAMat = 0;

			missing.resize( m*n );

			created = true;
		}

		template< class ErrorT, class MinFunctionT>
		void LMSparseSimple<ErrorT, MinFunctionT>::solve ( VecColl &a, VecColl &b,
		                                     vector<int> &missingIndizes,
		                                     ErrorT &errorFunc,
		                                     double eps, int maxIt )
		{
			assert_msg ( created, "You must call init fist\n" );
			//test whether parameters are okay or not
			assert ( newA.size() == a.size() );
			assert ( newB.size() == b.size() );
			assert_msg ( isNullArray ( a[0] ) || a[0].rows() == newA[0].rows(), "%zd != %zd", a[0].rows(), newA[0].rows() );
			assert_msg ( isNullArray ( b[0] ) || b[0].rows() == newB[0].rows(), "%zd != %zd", b[0].rows(), newB[0].rows() );

			double err1 = 0.0;
			double err2 = 0.0;
			double mu = 0.001;
			double nu = 2.0;
			double rho = 2.0;
			int count = 0;
			bool stop = false;
			const double deltaEps = eps;

			std::fill( missing.begin(), missing.end(), false );

			for ( uint idx = 0; idx < missingIndizes.size(); ++idx )
			{
				setMissing( missingIndizes[idx] );
				setNullArray ( Aij[missingIndizes[idx]] );
				setNullArray ( Bij[missingIndizes[idx]] );
			}

			examineStructure();

			jacobiFunc ( a, b, Aij, Bij );

			//calculate current error
			func ( a, b, val );
// 			errorFunc ( val, err, a, b );
			errorFunc ( val, err );

			lastError = initialError = err1 = lmSparseNorm ( err );
			lastIterations = 0;

			computeJTErrorA ( Aij, err, erra );

			computeJTErrorB ( Bij, err, errb );

			if ( max ( lmMaxNorm ( erra ), lmMaxNorm ( errb ) ) < eps ) return;

			do
			{
				do //LM-iteration
				{
					++count;

// 					printf("current val:\n");
// 					for(uint z = 0; z < a.size(); ++z){printf("a[%d]:\n", z); printMat(a[z], 2);}
// 					for(uint z = 0; z < b.size(); ++z){printf("b[%d]:\n", z); printMat(b[z], 2);}
// 					for(uint z = 0; z < val.size(); ++z){printf("\n"); printMats(*val[z], *err[z]);}
// 					printf("err1: %.30f - err2: %.30f\n", err1, err2);
// 					printf("lambda: %f, num missing: %d\n", mu, missingIndizes.size());
// 					getchar();


					jacobiFunc ( a, b, Aij, Bij );
// 					for(uint z = 0; z < Aij.size(); ++z){printf("Aij[%d]:\n", z); printMat(Aij[z], 2);}
// 					for(uint z = 0; z < Bij.size(); ++z){printf("Bij[%d]:\n", z); printMat(Bij[z], 2);}

					computeJTErrorA ( Aij, err, erra );
					computeJTErrorB ( Bij, err, errb );

// 					for(uint z = 0; z < erra.size(); ++z){printf("errA[%d]:\n", z); printMat(erra[z], 2);}
// 					for(uint z = 0; z < errb.size(); ++z){printf("errB[%d]:\n", z); printMat(errb[z], 2);}

					basicLMStep ( Aij, Bij, erra, errb, mu, DeltaA, DeltaB );

					assert_msg ( lmSparseNorm ( DeltaA ) >= 0, "sparseNorm = %f\n", lmSparseNorm ( DeltaA ) );
					assert_msg ( lmSparseNorm ( DeltaB ) >= 0, "sparseNorm = %f\n", lmSparseNorm ( DeltaB ) );

					if ( lmSparseNorm ( DeltaA ) + lmSparseNorm ( DeltaB ) <= deltaEps * ( lmSparseNorm ( a ) + lmSparseNorm ( b ) ) )
					{
						DBGOUT( DBGNORMAL, "LM SPARSE SIMPLE stops because small delta\n" );
						stop = true;
					}
					else
					{
						lmSparseAdd ( a, DeltaA, newA );
						lmSparseAdd ( b, DeltaB, newB );
						func ( newA, newB, val );
// 						errorFunc ( val, errNew, a, b );
						errorFunc ( val, errNew );
						err2 = lmSparseNorm ( errNew );

						lmSparseScaleAdd ( DeltaA, mu, erra, tempA );
						lmSparseScaleAdd ( DeltaB, mu, errb, tempB );

						double divisor = 0.0;
						for ( uint i = 0; i < DeltaA.size(); ++i )
						{
							divisor += dot ( DeltaA[i], tempA[i] );
						}
						for ( uint i = 0; i < DeltaB.size(); ++i )
						{
							divisor += dot ( DeltaB[i], tempB[i] );
						}
						rho = ( err1 - err2 ) / divisor;

// 						printf("divisor: %f, rho: %f, err1: %f, err2: %f\n", divisor, rho, err1, err2);
						if ( rho > 0 )
						{
							lmSparseCopy ( newA, a );
							lmSparseCopy ( newB, b );
							lmSparseCopy ( errNew, err );

							assert ( err2 < err1 );
							err1 = err2;

							if ( max ( lmMaxNorm ( erra ), lmMaxNorm ( errb ) ) < eps )
							{
								DBGOUT( DBGNORMAL, "LM SPARSE SIMPLE stops because small error\n" );
								stop = true;
							}

							double magic = ( 2 * rho - 1 );
							//printf("mu: %f, magic: %f\n", mu, magic);
							mu = mu * MAX ( 1.0/3.0, 1.0 - magic * magic * magic );
							nu = 2;
						}
						else
						{
							mu = mu * nu;
							nu =  2 * nu;
						}

					}
				}
				while ( rho < 0 && !stop );

			}
			while ( count < maxIt && !stop );
// 			DBGOUT(DBGALL, "LMSparseSimple: err: %f, count: %d of %d\n", err1, count, maxIt);
			lastError = err1;
			lastIterations = count;
		}

		template< class ErrorT, class MinFunctionT>
		void LMSparseSimple<ErrorT, MinFunctionT>::
		jacobiFunc ( VecColl &a, VecColl &b, vector<AMatrix> &Aij, vector<AMatrix> &Bij )
		{
			//derivative Matrices (Jacobian-sub-matrices)
			for ( int i = 0; i < n; ++i )
			{
				for ( int j = 0; j < m; j++ )
				{
					int idx = ijToIdx ( i,j );
					if ( !isMissing( idx ) ) func.jacobi( a[j], b[i], Aij[idx], Bij[idx] );
				}
			}

		}

		//LM Step, computes Delta and adds it to params to get newParams.
		template< class ErrorT, class MinFunctionT>
		void LMSparseSimple<ErrorT, MinFunctionT>::basicLMStep ( vector<AMatrix> &Aij, vector<AMatrix>&Bij,
		                VecColl &erra, VecColl &errb,
		                double mu,
		                VecColl &DeltaA, VecColl &DeltaB )
		{

			//Uj and erraj
			for ( int j = 0; j < m; ++j )
			{
				U[j] = 0;
				for ( int i = 0; i < n; ++i )
				{
					int idx = ijToIdx ( i,j );
					if ( !isMissing( idx ) ) matMulAdd2_AT ( Aij[idx], Aij[idx], U[j] );
				}
				lambdafy ( U[j], mu );
			}

			//Vi and errbi
			for ( int i = 0; i < n; ++i )
			{
				V[i] = 0;
				for ( int j = 0; j < m; j++ )
				{
					int idx = ijToIdx ( i,j );
					if ( !isMissing( idx ) ) matMulAdd2_AT ( Bij[idx], Bij[idx], V[i] );
				}
				lambdafy ( V[i], mu );
			}

			//Wij and Yij
			for ( int i = 0; i < n; ++i )
			{
				for ( int j = 0; j < m; j++ )
				{
					int idx = ijToIdx ( i,j );
					if ( isMissing( idx ) )
					{
						Wij[idx] = 0;
						Yij[idx] = 0;
					}
					else
					{
						matMul2_AT ( Aij[idx], Bij[idx], Wij[idx] );

						//instead inverting V*, LU-Decomposition is used to get Yij
						trans ( V[i], VTrans );
						trans ( Wij[idx], WijTrans );
						// 				solveSVD ( VTrans, WijTrans, YijTrans, store );//TODO:LU is sufficient
						//				  trans(YijTrans, Yij[idx]);
						/*bool ok = */solveLUinPlace( VTrans, WijTrans );
						trans( WijTrans, Yij[idx] );
					}
				}
			}

			//now compute dalta_a and delta_b, first delat_a
// 			printf("S\n");
			//S is m x m of a[j]-Rows x a[j]-Rows
			int x = 0;
			int y = 0;
			int width = DeltaA[0].rows();
			int height = DeltaA[0].rows();
			S = 0;

			y = 0;
			for ( int j = 0; j < m; ++j )
			{
				x = 0;
				for ( int k = 0; k < m; ++k )
				{
					AMatrix ajMask = S ( Range ( y, y + height - 1 ), Range ( x, x + width - 1 ) );//NICHT MEHR reference

					for ( int i = 0; i < n; ++i )
					{
						int idxY = ijToIdx ( i,j );
						int idxW = ijToIdx ( i,k );

						if ( !( isMissing( idxY ) || isMissing( idxW ) ) )
						{
							matMulSub2_BT ( Yij[idxY], Wij[idxW], ajMask );
							S ( Range ( y, y + height - 1 ), Range ( x, x + width - 1 ) ) = ajMask;
						}
					}
					x += width; //move masking rectangle
				}
				y += height;
			}
// 			printf("vor diag\n");
// 			printMat(S);
// 			printf("S2\n");
			//add Uj to the diagonal blocks
			x = y = 0;
			for ( int diag = 0; diag < m; ++diag )
			{
				AArray<numberT, 2> a = S ( Range ( y, y + height - 1 ), Range ( x, x + width - 1 ) );
				add2( a, U[diag] );
				S ( Range ( y, y + height - 1 ), Range ( x, x + width - 1 ) ) = a;

				x += width; //move masking rectangle
				y += height;
			}
			//after S is computed, compute e, ej has a[j].rows() rows
			x = y = 0;
			for ( int j = 0; j < m; ++j )
			{
				AVector ejMask = e ( Range ( y, y + height - 1 ) );
				ejMask = erra[j];
				for ( int i = 0; i < n; ++i )
				{
					int idx = ijToIdx ( i,j );
					if ( !isMissing( idx ) ) matMulSub2 ( Yij[idx], errb[i], ejMask );
				}
				e ( Range ( y, y + height - 1 ) ) = ejMask;
				y += height;
			}

// 			printf("solving\n");
			bool success = factorizeLDLT ( S, SFactorizeSpace );
// 			assert(success);
			ADORE_UNUSED( success );

			//solve S * delta = e, resulting in deltaAMat
			solveLDLTc ( S, e, deltaAMat, SFactorizeSpace, minSIdx, maxSIdx );
// 			pressKey();

			//now compute delta_b by back-substitution
// 			printf("deltaB\n");
			x = 0;

			for ( int i = 0; i < n; ++i )
			{
				y = 0;
                                AVector rightB = errb[i];
				for ( int j = 0; j < m; ++j )
				{
					int idx = ijToIdx ( i,j );
					AVector subDeltaA = deltaAMat ( Range ( y, y + height - 1 ) );

					if ( !isMissing( idx ) ) matMulSub2_AT ( Wij[idx], subDeltaA, rightB );
					y += height;
				}
// 				DeltaB[i] = solveSVD ( V[i], rightB );
				DeltaB[i] = rightB;
				solveLUinPlace( V[i], DeltaB[i] );
			}

			//update the parameter estimate for a
// 			printf("deltaA\n");
// 			printMat(deltaAMat);
// 			pressKey();
			x = y = 0;
			for ( uint j = 0; j < DeltaA.size(); ++j )
			{
				DeltaA[j] = deltaAMat ( Range ( y, y + height - 1 ) );
				y += height;
			}
		}


	}//ns

}

#endif
