

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_AUTOCALIBRATIONAUTOCALIBRATORCOMMON_H
#define ADORE_AUTOCALIBRATIONAUTOCALIBRATORCOMMON_H

#include <adrcore/types/camera.h>
#include <adrcore/util/mathutil.h>
#include <adrcore/util/solvesvd.h>
#include <adrcore/util/floatshift.h>
#include <adrcore/types/nview.h>

namespace adore
{
	namespace core
	{
		class Camera;
		class Vector3D;
		class NView;
		class LinearQCalculatorDummy;

		enum
		{
			CALIB_Q_ERROR,
			CALIB_Q_UNIQUE,
			CALIB_Q_PENCIL
		};

		/**
		 
		*/
		class AutocalibratorCommon
		{
			public:

				AutocalibratorCommon() :
						H ( 4, 4 ),
						Hinv ( 4, 4 ),
						tmpCam ( 3, 4 ),
						store4 ( 4, 4, SVD_VT ),
						tmpPoint( 4 ),
						QPT( 4, 3 )
				{}

				/**
				 * This function calculates the metric reconstruction using the precomputed dual quadric.
				 * Make sure to call one of the dualQuadric...()-functions first.
				 */
				int computeMetricReconstruction ( const AMatrix &Q, NView::CameraIterator camIt, const int numCams, NView::Point3DIterator ptIt, const int numPts );

				int computeMetricCamReconstruction ( const AMatrix &H, NView::CameraIterator camIt, const int numCams );

				int computeMetricPointReconstruction ( const AMatrix &Hinv, NView::Point3DIterator ptIt, const int numPts );

				int decomposeQ ( const AMatrix &Q, AMatrix &H );

				void enforceRank3SVD ( AMatrix &Q );

				/**
					counts the number of not-supporting 3D points.
				*/
				int countInvisible( const AMatrix &Q, NView &nv );

				void computePQPT( const AMatrix &Q, const AMatrix &P, AMatrix &w )
				{
					matMul2_BT( Q, P, QPT );
					matMul2( P, QPT, w );
				}

			private:

				AMatrix H;
				AMatrix Hinv;
				AMatrix tmpCam;
				SVDStoraged store4;
				AVector tmpPoint;
				AMatrix QPT;//temp matrix for computePQPT
		};

		/**
			debug
			prints PQPT to stdout
		*/
		void printQEquation ( AMatrix &Q, NView::CameraIterator camIt, int numCams );

		/**
		 * calculate Q linearly
		 * The last parameter provides workspace
		 * */
		template<class LinQCalc> bool calculateQ( NView & nview, AMatrix & Q, AutocalibratorCommon & ac )
		{
			AMatrix Q2 ( 4, 4 );
			LinQCalc lqc( nview.getNumCameras() );

			int result = lqc.calcDualQuadric ( nview.getCamerasBegin(), nview.getNumCameras(), Q, Q2 );

//			printQEquation ( Q, nview.getCamerasBegin(), nview.getNumCameras() );

			if( Q(0, 0) <= 0)
			{
				DBGOUT ( DBGNORMAL, "squared focal length is negative. no solutions.\n" );
				return false;
			}

			if ( result == CALIB_Q_PENCIL )
			{
				DBGOUT ( DBGNORMAL, "searching 4 solutions\n" );

				AMatrix QCand( 4, 4 );
				double l[4];

				int num = lqc.getDet0Solutions ( Q, Q2, l );

				int bestIdx = -1;
				int bestVote = INT_MAX;

				for ( int i = 0; i < num; ++i )
				{
					QCand = Q + ( Q2 * l[i] );
					assert(QCand(0, 0) > 0);

					normalize( QCand, 1.0 );

					int vote = ac.countInvisible( QCand, nview );
					if ( vote == 0 )
					{
						Q = QCand;
						return true;
					}
					else if ( vote < bestVote )
					{
						bestIdx = i;
						bestVote = vote;
					}
				}

				if(bestIdx >= 0) {
					int num3d = nview.getNumPoints3D();
					if(num3d / 2 < bestVote)
						return false;

					Q += Q2 * l[bestIdx];
					return true;
				}
			}
			else if ( result == CALIB_Q_UNIQUE && nview.getNumCameras() > 2)
			{
				return true;
			}

			return false;
		}

		template<> inline bool calculateQ<LinearQCalculatorDummy>( NView & nview, AMatrix & Q, AutocalibratorCommon & ac )
		{
			DBGOUT( DBGNORMAL, "calculateQ uses LinearQCalculatorDummy\n" );
			Q = 	1, 0, 0, 0,
			     0, 1, 0, 0,
			     0, 0, 1, 0,
				 0, 0, 0, 0;
			return true;
		}

	}

}

#endif
