

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include "cameracalculator.h"
#include <float.h>
#include <adrcore/nview/fmatrixcalculator.h>
#include <adrcore/nview/ttensorcalculator.h>
#include <adrcore/util/printmat.h>
#include <adrcore/minimization/defaulterrorfunc2.h>

namespace adore
{
	namespace core
	{

		CameraCalculator::CameraCalculator():
				store( 0, 12, SVD_VT ),
				cameraMinFunc()
		{
		}

		CameraCalculator::CameraCalculator( const CameraCalculator & ):
				store( 0, 12, SVD_VT ),
				cameraMinFunc()
		{
		}

		CameraCalculator::~CameraCalculator()
		{
		}

//		pts2D are row-vectors, pts3D are column-vectors
		int CameraCalculator::calculateCamera( const vector<AVector3D*> & pts3D,
											   const vector<AVector2D*> & pts2D, AMatrix & camMat )
		{
			AMatrix camEq( 2 * pts2D.size(), 12 );

			int row = 0;
			for ( uint i = 0; i < pts2D.size(); ++i )
			{
				camEq( row, 0 ) =  0;
				camEq( row, 1 ) =  0;
				camEq( row, 2 ) =  0;
				camEq( row, 3 ) =  0;

				const AVector3D& pt3d = *pts3D[i];
				const AVector2D& pt2d = *pts2D[i];
				camEq( row, 4 ) =  -pt2d( 2 ) * pt3d( 0 );
				camEq( row, 5 ) =  -pt2d( 2 ) * pt3d( 1 );
				camEq( row, 6 ) =  -pt2d( 2 ) * pt3d( 2 );
				camEq( row, 7 ) =  -pt2d( 2 ) * pt3d( 3 );

				camEq( row, 8 )  =  pt2d( 1 ) * pt3d( 0 );
				camEq( row, 9 )  =  pt2d( 1 ) * pt3d( 1 );
				camEq( row, 10 ) =  pt2d( 1 ) * pt3d( 2 );
				camEq( row, 11 ) =  pt2d( 1 ) * pt3d( 3 );

				++row;

				camEq( row, 0 ) =   pt2d( 2 ) * pt3d( 0 );
				camEq( row, 1 ) =   pt2d( 2 ) * pt3d( 1 );
				camEq( row, 2 ) =   pt2d( 2 ) * pt3d( 2 );
				camEq( row, 3 ) =   pt2d( 2 ) * pt3d( 3 );

				camEq( row, 4 ) =  0;
				camEq( row, 5 ) =  0;
				camEq( row, 6 ) =  0;
				camEq( row, 7 ) =  0;

				camEq( row, 8 )  =  -pt2d( 0 ) * pt3d( 0 );
				camEq( row, 9 )  =  -pt2d( 0 ) * pt3d( 1 );
				camEq( row, 10 ) =  -pt2d( 0 ) * pt3d( 2 );
				camEq( row, 11 ) =  -pt2d( 0 ) * pt3d( 3 );

				++row;
			}

//			printMat( camEq );
//			pressKey();

			int retVal = calculateCamera( camEq, camMat );

#ifndef NDEBUG
//			AVector dummyPt2D( 3 );
//			double scale, var;
//			for ( uint i = 0; i < pts2D.size(); ++i )
//			{
//				matMul2( camMat, pts3D[i], dummyPt2D );
//				scaleEqualityTest( dummyPt2D, pts2D[i], scale, var );
//				std::cout << pts3D[i] << "   " << pts2D[i] << "\n";
//				printf( "CameraCalculator: 2d point and its reprojection\n" );
//				printMats( dummyPt2D, pts2D[i] );
//				printf( "CameraCalculator: scale: %f, var: %f\n", scale, var );
//			}
#endif

			return retVal;
		}



		int CameraCalculator::calculateCamera( const AMatrix &camEq, AMatrix &camMat )
		{
			store.changeRowsWorkspaceOnly( camEq.rows() );
			store.repairD();
			svdVT( camEq, store );
			AVector &singulars12 = store.D;
			AMatrix &V_T12 = store.VT;
//			printf( "svd\n" );
//			printMat( singulars12 );
//			printMat( V_T12 );
//			getchar();

			//test, if something went wrong
			if ( singulars12( 10 ) <= FLT_EPSILON )
			{
				DBGOUT( DBGALL, "DLT in CameraCalculator went wrong\n" );
				return -1;
			}
			else
			{
				camMat( 0, 0 ) =  V_T12( 11, 0 );
				camMat( 0, 1 ) =  V_T12( 11, 1 );
				camMat( 0, 2 ) =  V_T12( 11, 2 );
				camMat( 0, 3 ) =  V_T12( 11, 3 );

				camMat( 1, 0 ) =  V_T12( 11, 4 );
				camMat( 1, 1 ) =  V_T12( 11, 5 );
				camMat( 1, 2 ) =  V_T12( 11, 6 );
				camMat( 1, 3 ) =  V_T12( 11, 7 );

				camMat( 2, 0 ) =  V_T12( 11, 8 );
				camMat( 2, 1 ) =  V_T12( 11, 9 );
				camMat( 2, 2 ) =  V_T12( 11, 10 );
				camMat( 2, 3 ) =  V_T12( 11, 11 );
			}

			//normalizeExponents(camMat);//OPT

			return 0;
		}

		void CameraCalculator::minimizeError( const vector<AVector3D*>  &pts3D, const vector<AVector2D*> & pts2D, AMatrix & camMat )
		{
			assert( pts3D.size() == pts2D.size() );
//			printf( "CameraCalculator:: minimize error 1\n" );

			AVector linCam( 12 );
			int idx = 0;
			for ( int row = 0; row < 3; ++row )
				for ( int col = 0; col < 4; ++col )
				{
					linCam( idx++ ) = camMat( row, col );
				}
//			printMat( linCam );
//			pressKey();

			AVector goal( 2 * pts3D.size() );
			idx = 0;
			for ( uint i = 0; i < pts2D.size(); ++i )
			{
				const AVector2D& pt2d = *pts2D[i];
				double w = pt2d( 2 );
				goal( idx++ ) = pt2d( 0 ) / w;
				goal( idx++ ) = pt2d( 1 ) / w;
			}
//			printMat( goal );
//			pressKey();

			cameraMinFunc.set3DPoints( &pts3D );
			CameraMinJacobi jacobi;
			jacobi.set3DPoints( &pts3D );
			DefaultErrorFunc2<double> errFunc( goal );
			LM2<CameraMinFunc, CameraMinJacobi, DefaultErrorFunc2<double>, InvCovarUnused<double>::type > lm;
			lm.solve( linCam, cameraMinFunc, goal.size(), jacobi, errFunc );
//			printf( "camera minimization errors: %.20f - %.20f (%d)\n", lm.getInitialError(), lm.getLastError(), lm.getLastIterations() );

			idx = 0;
			for ( int row = 0; row < 3; ++row )
				for ( int col = 0; col < 4; ++col )
				{
					camMat( row, col ) = linCam( idx++ );
				}

//			printf( "minimize error 2\n" );
		}

		void CameraCalculator::modify( NView &nview )
		{
			if ( nview.isNViewSet() )
			{
				int numviews = nview.getNumViews();
				switch ( numviews )
				{
					case 2:
						modify2( nview );
						DBGOUT( DBGNORMAL, "CameraCalculator:modify2: %d cameras calculated\n", nview.getNumCameras() );
						return;
					case 3:
						modify3( nview );
						DBGOUT( DBGNORMAL, "CameraCalculator:modify3: %d cameras calculated\n", nview.getNumCameras() );
						return;
					default:
						error( "wrong nviewset\n" );
						break;
				}
			}

			vector<ViewBase *> views;
			NView::ViewIterator itV = nview.getViewsBegin();
			for ( ; itV != nview.getViewsEnd(); ++itV )
			{
				ViewBase *v = *itV;
				if ( !nview.hasCamera( v ) ) views.push_back( v );
			}
			DBGOUT( DBGNORMAL, "calculating %zd cameras\n", views.size() );
			modify( nview, views );
		}

		void CameraCalculator::modify( NView &nview, vector<ViewBase *> &views )
		{
			vector<AVector2D *> pts2D;
			vector<AVector3D *> pts3D;

			vector<ViewBase *>::iterator itV = views.begin();
			for ( ; itV != views.end(); ++itV, pts2D.clear() )
			{
				ViewBase *v = *itV;
				assert_msg( !nview.hasCamera( v ), "this is only for views without cameras\n" );

				//get all correspondences that have a point in this particular view
				nview.getPointCorrs( v, pts2D );

				vector<AVector2D *>::iterator it2D = pts2D.begin();
				for ( ; it2D != pts2D.end(); ++it2D )
				{
					AVector3D *pt = nview.getPoint3D( ( FeaturePoint * )( *it2D ) );
					assert( pt );

					pts3D.push_back( pt );
				}

				AMatrix & camMat = nview.getCamera( v )->getProjMat();

				calculateCamera( pts3D, pts2D, camMat );
				minimizeError( pts3D, pts2D, camMat );
			}
		}


		void CameraCalculator::modify2( NView &nview, FMatrixCalculator *_fcalc )
		{
			DBGOUT( DBGALL, "Calculating cameras from Tensor\n" );
			assert( nview.getNumViews() == 2 );
			const int numViews = 2;

			AVector epipole1( 3 );

			NViewInfo *nvi = nview.getInfo();
			if ( !nvi )
			{
				DBGOUT( DBGALL, "No Tensor in Cameracalculator\n" );
				return;
			}

			FMatrixCalculator *fcalc = _fcalc;
			if ( !_fcalc ) fcalc = new FMatrixCalculator();

			vector<ViewBase *> views( nview.getNumViews() );
			nview.getViews( views.begin() );

			vector<Camera *> cams( numViews );
			for ( int i = 0; i < numViews; ++i )
			{
				cams[i] = nview.getCamera( views[i], true );//true: create them on demand
//				printf( "cam %d is %p\n", i, cams[i] );
			}

			setIdentity( cams[0]->getProjMat() ); //cam1 is the canonical first camera

			AMatrix fmatrix = nvi->getFMatrix();
			transInPlace( fmatrix );
			fcalc->calculateEpipolarPoint( fmatrix, epipole1 );
//			printf( "epi1:\n" );
//			printMat( epipole1 );
			transInPlace( fmatrix );
			fcalc->calculateCamMatrix( fmatrix, epipole1, cams[1]->getProjMat() );
//			printf( "cam1:\n" );
//			printMats( cams[0]->getProjMat(), cams[1]->getProjMat() );

#ifndef NDEBUG
			double scale, variance;
			AMatrix debugMat( 3, 3 );
			debugMat = fcalc->calculateFundamentalMatrix1Cam( cams[1]->getProjMat() );
			DBGOUT( DBGALL, "debugMat1:\n" );
			printMat( DBGALL, debugMat );
			debugMat = fcalc->calculateFundamentalMatrix2Cams( cams[0]->getProjMat(), cams[1]->getProjMat() );
			DBGOUT( DBGALL, "debugMat2:\n" );
			printMat( DBGALL, debugMat );
			scaleEqualityTest( fmatrix.begin(), debugMat.begin(), 9, scale, variance );
			DBGOUT( DBGALL, "ScaleEQ-Test FMATRIX: scale: %f, var: %f\n", scale, variance );
#endif

			if ( !_fcalc ) delete fcalc;
		}

		void CameraCalculator::modify3( NView &nview, TTensorCalculator *_tcalc )
		{
			DBGOUT( DBGALL, "Calculating cameras from Tensor\n" );
			assert( nview.getNumViews() == 3 );
			const int numViews = 3;

			NViewInfo *nvi = nview.getInfo();
			if ( !nvi || !nvi->isOkay() )
			{
				DBGOUT( DBGALL, "No Tensor in Cameracalculator\n" );
				return;
			}

			TTensorCalculator *tcalc = _tcalc;
			if ( !_tcalc ) tcalc = new TTensorCalculator();

			vector<ViewBase *> views( nview.getNumViews() );
			nview.getViews( views.begin() );

			vector<Camera *> cams( numViews );
			for ( int i = 0; i < numViews; ++i )
			{
				cams[i] = nview.getCamera( views[i], true );//true: create them on demand
//				printf( "cam %d is %p\n", i, cams[i] );
			}

			setIdentity( cams[0]->getProjMat() ); //cam1 is the canonical first camera

			DBGOUT( DBGALL, "cam 2 and 3\n" );

			ATensor3 t = nvi->getTTensor();
			AVector epipole1 = tcalc->calculateEpipole( t, true );
			AVector epipole2 = tcalc->calculateEpipole( t, false );
//			printMat(epipole1, 25); cout << endl;
//			printMat(epipole2, 25); cout << endl;
			cams[1]->getProjMat() = tcalc->calculateCameraMatrix( t, epipole1, epipole2, tcalc->CAMERA_P1 );
			cams[2]->getProjMat() = tcalc->calculateCameraMatrix( t, epipole1, epipole2, tcalc->CAMERA_P2 );
//			printMat(cams[1]->getProjMat(), 25); cout << endl;
//			printMat(cams[2]->getProjMat(), 25); cout << endl;

#ifndef NDEBUG
			DBGOUT( DBGALL, "CamCalc: scale test\n" );
			printMats( DBGALL, epipole1, epipole2 );
			printMats( DBGALL, cams[1]->getProjMat(), cams[2]->getProjMat() );
			ATensor3 debugTensor( 3, 3, 3 );
			tcalc->calculateTrifocalTensorBase3CamsT( cams[0]->getProjMat(), cams[1]->getProjMat(), cams[2]->getProjMat(), debugTensor );
			COUT( DBGALL, t );
			COUT( DBGALL, debugTensor );
			double scale, variance;
			scaleEqualityTest( t.begin(), debugTensor.begin(), 27, scale, variance );
			DBGOUT( DBGALL, "ScaleEQ-Test TTENSOR: scale: %f, var: %f\n", scale, variance );
#endif

			if ( !_tcalc ) delete tcalc;
		}

		AMatrix CameraCalculator::mulW( const AMatrix & V, bool trans )
		{
			double s1 = -1;
			double s2 =  1;
			if ( trans )
			{
				s1 =  1;
				s2 = -1;
			}

			AMatrix result( 3, 3 );
			result( 0, 0 ) = s1 * V( 1, 0 );
			result( 1, 0 ) = s2 * V( 0, 0 );
			result( 2, 0 ) =      V( 2, 0 );
			result( 0, 1 ) = s1 * V( 1, 1 );
			result( 1, 1 ) = s2 * V( 0, 1 );
			result( 2, 1 ) =      V( 2, 1 );
			result( 0, 2 ) = s1 * V( 1, 2 );
			result( 1, 2 ) = s2 * V( 0, 2 );
			result( 2, 2 ) =      V( 2, 2 );

			return result;
		}

	}
}
