

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include "camerapointduality.h"
#include <adrcore/util/debug.h>
#include <adrcore/util/printmat.h>
#include <adrcore/util/matmul2.h>
#include <adrcore/util/numericalanalysis.h>

namespace adore
{

	namespace core
	{

		CameraPointDuality::CameraPointDuality() :
				canonicalX1 ( 4 ),
				canonicalX2 ( 4 ),
				canonicalX3 ( 4 ),
				canonicalX4 ( 4 ),
				canonicalX5 ( 4 ),
				store3 ( 3, 3, SVD_VT ),
				store4 ( 6, 4, SVD_VT ),
				store10_12 ( 10, 12, SVD_VT ),
				store3_5 ( 3, 5, SVD_VT ),
				fmatrixT ( 3, 3 ),
				fshuffle ( 3, 3 ),
				ratioMat ( 4, 4 ),
				M ( 10, 12 ),//from here including 6-point-algo-variables
				Q ( 4, 4 ),
				tmpPt ( 3 ),
				tmpCam ( 3, 4 ),
				tmpA4x4 ( 4, 4 ),
				tmpB4x4 ( 4, 4 ),
				cam1 ( 3, 4 ),
				cam2 ( 3, 4 ),
				w ( 3, 5 ),
				designMat ( 6, 4 )
		{

			canonicalX5 ( 0 ) =  1;
			canonicalX5 ( 1 ) =  1;
			canonicalX5 ( 2 ) =  1;
			canonicalX5 ( 3 ) =  1;

			canonicalX1 ( 0 ) =  1;
			canonicalX1 ( 1 ) =  0;
			canonicalX1 ( 2 ) =  0;
			canonicalX1 ( 3 ) =  0;

			canonicalX2 ( 0 ) =  0;
			canonicalX2 ( 1 ) =  1;
			canonicalX2 ( 2 ) =  0;
			canonicalX2 ( 3 ) =  0;

			canonicalX3 ( 0 ) =  0;
			canonicalX3 ( 1 ) =  0;
			canonicalX3 ( 2 ) =  1;
			canonicalX3 ( 3 ) =  0;

			canonicalX4 ( 0 ) =  0;
			canonicalX4 ( 1 ) =  0;
			canonicalX4 ( 2 ) =  0;
			canonicalX4 ( 3 ) =  1;
		}


		CameraPointDuality::~CameraPointDuality()
		{
		}

		AVector CameraPointDuality::cameraToPoint ( const AMatrix &cam ) const
		{
			AVector pt ( 4 );
			pt ( 0 ) = cam ( 0, 0 );
			pt ( 1 ) = cam ( 1, 1 );
			pt ( 2 ) = cam ( 2, 2 );
			pt ( 3 ) = cam ( 0, 3 );
			return pt;
		}

		AMatrix CameraPointDuality::pointToCamera ( const AVector &pt ) const
		{
			AMatrix cam;
			cam = 0;
			cam ( 0, 0 ) = pt ( 0 );
			cam ( 1, 1 ) = pt ( 1 );
			cam ( 2, 2 ) = pt ( 2 );
			cam ( 0, 3 ) = pt ( 3 );
			cam ( 1, 3 ) = pt ( 3 );
			cam ( 2, 3 ) = pt ( 3 );
			return cam;
		}

		AVector CameraPointDuality::calculateReducedCameraDualFromReducedFMatrix ( const AMatrix &fmatrix )
		{
			fshuffle ( 0, 0 ) = fmatrix ( 0, 1 );
			fshuffle ( 0, 1 ) = fmatrix ( 1, 0 );
			fshuffle ( 0, 2 ) =  0;
			fshuffle ( 1, 0 ) = fmatrix ( 0, 2 );
			fshuffle ( 1, 1 ) =  0;
			fshuffle ( 1, 2 ) = fmatrix ( 2, 0 );
			fshuffle ( 2, 0 ) =  0;
			fshuffle ( 2, 1 ) = fmatrix ( 1, 2 );
			fshuffle ( 2, 2 ) = fmatrix ( 2, 1 );

			//fshuffle * ratio1 = 0; get ratio1, a column vector
			AMatrix &V_T3 = store3.VT;
			svdVT ( fshuffle, store3 );

			double A = V_T3 ( 2, 0 ) / V_T3 ( 2, 1 );
			double B = V_T3 ( 2, 1 ) / V_T3 ( 2, 2 );

			//ratio2 * fmatrix = 0; get ratio2, a row vector
			fmatrixT = trans ( fmatrix );
			svdVT ( fmatrixT, store3 );

			double C = V_T3 ( 2, 0 ) / V_T3 ( 2, 1 );
			double D = V_T3 ( 2, 1 ) / V_T3 ( 2, 2 );

			//A, B, C, and D lead via ratiomat to abcd
			ratioMat ( 0, 0 ) =  1;
			ratioMat ( 0, 1 ) =  -A;
			ratioMat ( 0, 2 ) =  0;
			ratioMat ( 0, 3 ) =  0;
			ratioMat ( 1, 0 ) =  0;
			ratioMat ( 1, 1 ) =  1;
			ratioMat ( 1, 2 ) =  B;
			ratioMat ( 1, 3 ) =  0;
			ratioMat ( 2, 0 ) =  -1;
			ratioMat ( 2, 1 ) =  C;
			ratioMat ( 2, 2 ) =  0;
			ratioMat ( 2, 3 ) =  1.0-C;
			ratioMat ( 3, 0 ) =  0;
			ratioMat ( 3, 1 ) =  -1;
			ratioMat ( 3, 2 ) =  D;
			ratioMat ( 3, 3 ) =  1.0-D;

			//solve for abcd (that is the last row in V_T4)
			AMatrix &V_T4 = store4.VT;
			svdVT ( ratioMat, store4 );

			AVector dual ( 4 );
			dual ( 0 ) = V_T4 ( 3, 0 );
			dual ( 1 ) = V_T4 ( 3, 1 );
			dual ( 2 ) = V_T4 ( 3, 2 );
			dual ( 3 ) = V_T4 ( 3, 3 );

			return dual;
		}

//		6 Point Algo   /////////////////////////////////////////////

		void CameraPointDuality::initFirst8ColumnsOfM()
		{
			M = 0;
			M ( 0, 0 ) =  -1;
			M ( 2, 1 ) =  -1;
			M ( 4, 2 ) =  -1;
			M ( 6, 3 ) =  -1;

			M ( 1, 4 ) =  -1;
			M ( 3, 5 ) =  -1;
			M ( 5, 6 ) =  -1;
			M ( 7, 7 ) =  -1;

			M ( 8, 0 ) =  -1;
			M ( 8, 1 ) =  -1;
			M ( 8, 2 ) =  -1;
			M ( 8, 3 ) =  -1;

			M ( 9, 4 ) =  -1;
			M ( 9, 5 ) =  -1;
			M ( 9, 6 ) =  -1;
			M ( 9, 7 ) =  -1;
		}

		void CameraPointDuality::initLast4ColumnsOfM ( double x1, double y1,
		                double x2, double y2,
		                double x3, double y3,
		                double x4, double y4,
		                double x5, double y5
		                                             )
		{
			M ( 0, 8 ) =  x1;
			M ( 1, 8 ) =  y1;
			M ( 2, 9 ) =  x2;
			M ( 3, 9 ) =  y2;
			M ( 4, 10 ) =  x3;
			M ( 5, 10 ) =  y3;
			M ( 6, 11 ) =  x4;
			M ( 7, 11 ) =  y4;

			M ( 8, 8 ) =  x5;
			M ( 8, 9 ) =  x5;
			M ( 8, 10 ) =  x5;
			M ( 8, 11 ) =  x5;

			M ( 9, 8 ) =  y5;
			M ( 9, 9 ) =  y5;
			M ( 9, 10 ) =  y5;
			M ( 9, 11 ) =  y5;
		}

		void CameraPointDuality::calculateQ ( const AMatrix &cam1, const AMatrix &cam2, const double x, const double y )
		{
			tmpPt = x, y, 1.0;
			matMul2Skew( tmpPt, cam2, tmpCam );
			matMul2_AT ( cam1, tmpCam, tmpA4x4 );

			matMul2Skew( tmpPt, cam1, tmpCam );
			matMul2_AT ( cam2, tmpCam, tmpB4x4 );

			sub2( tmpA4x4, tmpB4x4, Q );
		}

		void CameraPointDuality::wFrom6Points ( double x1, double y1,
		                                        double x2, double y2,
		                                        double x3, double y3,
		                                        double x4, double y4,
		                                        double x5, double y5,
		                                        double x6, double y6,
		                                        AMatrix &w, int rowOfW
		                                      )
		{
			initFirst8ColumnsOfM();
			initLast4ColumnsOfM ( x1, y1, x2, y2, x3, y3, x4, y4, x5, y5 );

//			printf( "calculating w\n" );
//			printf( "coors: (%f, %f) (%f, %f) (%f, %f) (%f, %f) (%f, %f)\n", x1, y1, x2, y2, x3, y3, x4, y4, x5, y5 );

			AMatrix &V_T12 = store10_12.VT;
			int ret = svdVT ( M, store10_12 );
			assert_msg ( ret == 0, "lapack ret = %d\n", ret );
			ADORE_UNUSED ( ret );

			convertLineCam ( V_T12 ( 10,  Range::all() ), cam1 );
			convertLineCam ( V_T12 ( 11,  Range::all() ), cam2 );

//			DBGOUT( DBGNORMAL, "Cam1 of pencil is:\n" );
//			printMat( DBGNORMAL, cam1 );
//			DBGOUT( DBGNORMAL, "Cam2 of pencil is:\n" );
//			printMat( DBGNORMAL, cam2 );

			calculateQ ( cam1, cam2, x6, y6 );

//			DBGOUT( DBGNORMAL, "Q is:\n" );
//			printMat( DBGNORMAL, Q );

			w ( rowOfW, 0 ) = Q ( 0, 1 );
			w ( rowOfW, 1 ) = Q ( 0, 2 );
			w ( rowOfW, 2 ) = Q ( 1, 2 );
			w ( rowOfW, 3 ) = Q ( 1, 3 );
			w ( rowOfW, 4 ) = Q ( 2, 3 );
		}

		void CameraPointDuality::convertLineCam ( const AVector &lineCam, AMatrix &cam )
		{
			int lineCol = 0;
			for ( int i = 0; i < 4; ++i ) cam ( 0, i ) = lineCam ( lineCol++ );
			for ( int i = 0; i < 4; ++i ) cam ( 1, i ) = lineCam ( lineCol++ );
			for ( int i = 0; i < 4; ++i ) cam ( 2, i ) = lineCam ( lineCol++ );
		}

		int CameraPointDuality::reconstructFrom6PointCorrespondences ( const std::vector<std::vector<AVector2D> > &points, AVector &X6_1, AVector &X6_2, AVector &X6_3 )
		{
			int numViews = points.size();
			assert_msg ( numViews == 3, "more views not yet implemented\n" );

			for ( int i = 0; i < numViews; ++i )
			{
				wFrom6Points ( points[i][0] ( 0 ), points[i][0] ( 1 ),
				               points[i][1] ( 0 ), points[i][1] ( 1 ),
				               points[i][2] ( 0 ), points[i][2] ( 1 ),
				               points[i][3] ( 0 ), points[i][3] ( 1 ),
				               points[i][4] ( 0 ), points[i][4] ( 1 ),
				               points[i][5] ( 0 ), points[i][5] ( 1 ), w, i );
			}

			DBGOUT( DBGCOND, "Condition of w from pts6Stable:\n" );
			printCondition( DBGCOND, w );
			pressKey( DBGCOND );

//			DBGOUT( DBGNORMAL, "Dual: w is:\n" );
//			printMat( DBGNORMAL, w );

			AMatrix &V_T5 = store3_5.VT;
			svdVT ( w, store3_5 );

			double alpha1, alpha2, alpha3;
			int numRoots = solveForAlpha ( V_T5, alpha1, alpha2, alpha3 );

			double a, b, c, d, e;
			if ( numRoots > 0 )
			{
//				printf( "root1: %f\n", alpha1 );
				a = alpha1 * ( V_T5 ( 3, 0 ) - V_T5 ( 4, 0 ) ) + V_T5 ( 4, 0 );
				b = alpha1 * ( V_T5 ( 3, 1 ) - V_T5 ( 4, 1 ) ) + V_T5 ( 4, 1 );
				c = alpha1 * ( V_T5 ( 3, 2 ) - V_T5 ( 4, 2 ) ) + V_T5 ( 4, 2 );
				d = alpha1 * ( V_T5 ( 3, 3 ) - V_T5 ( 4, 3 ) ) + V_T5 ( 4, 3 );
				e = alpha1 * ( V_T5 ( 3, 4 ) - V_T5 ( 4, 4 ) ) + V_T5 ( 4, 4 );
				computeX6 ( a, b, c, d, e, X6_1 );
//				printf( "X6_1:\n" );
//				printCvMat( X6_1 );
			}
			if ( numRoots == 3 )
			{
//				printf( "root2: %f\n", alpha2 );
				a = alpha2 * ( V_T5 ( 3, 0 ) - V_T5 ( 4, 0 ) ) + V_T5 ( 4, 0 );
				b = alpha2 * ( V_T5 ( 3, 1 ) - V_T5 ( 4, 1 ) ) + V_T5 ( 4, 1 );
				c = alpha2 * ( V_T5 ( 3, 2 ) - V_T5 ( 4, 2 ) ) + V_T5 ( 4, 2 );
				d = alpha2 * ( V_T5 ( 3, 3 ) - V_T5 ( 4, 3 ) ) + V_T5 ( 4, 3 );
				e = alpha2 * ( V_T5 ( 3, 4 ) - V_T5 ( 4, 4 ) ) + V_T5 ( 4, 4 );
				computeX6 ( a, b, c, d, e, X6_2 );
//				printf( "X6_2:\n" );
//				printCvMat( X6_2 );

//				printf( "root3: %f\n", alpha3 );
				a = alpha3 * ( V_T5 ( 3, 0 ) - V_T5 ( 4, 0 ) ) + V_T5 ( 4, 0 );
				b = alpha3 * ( V_T5 ( 3, 1 ) - V_T5 ( 4, 1 ) ) + V_T5 ( 4, 1 );
				c = alpha3 * ( V_T5 ( 3, 2 ) - V_T5 ( 4, 2 ) ) + V_T5 ( 4, 2 );
				d = alpha3 * ( V_T5 ( 3, 3 ) - V_T5 ( 4, 3 ) ) + V_T5 ( 4, 3 );
				e = alpha3 * ( V_T5 ( 3, 4 ) - V_T5 ( 4, 4 ) ) + V_T5 ( 4, 4 );
				computeX6 ( a, b, c, d, e, X6_3 );
//				printf( "X6_3:\n" );
//				printCvMat( X6_3 );
			}

			return numRoots;
		}

		void CameraPointDuality::computeX6 ( double a, double b, double c, double d, double e, AVector &X6 )
		{
			designMat = 0;
			designMat ( 0, 0 ) =  e-d;
			designMat ( 0, 3 ) =  a-b;
			designMat ( 1, 0 ) =  e-c;
			designMat ( 1, 2 ) =  a;
			designMat ( 2, 0 ) =  d-c;
			designMat ( 2, 1 ) =  b;
			designMat ( 3, 1 ) =  e-b;
			designMat ( 3, 2 ) =  a-d;
			designMat ( 4, 1 ) =  e;
			designMat ( 4, 3 ) =  a-c;
			designMat ( 5, 2 ) =  d;
			designMat ( 5, 3 ) =  b-c;

			svdVT ( designMat, store4 );
			X6 = store4.VT ( 3, Range::all() );
		}

		int CameraPointDuality::solveForAlpha ( const AMatrix &V_T5, double &alpha1, double &alpha2, double &alpha3 )
		{
			double a[2];
			double b[2];
			double c[2];
			double d[2];
			double e[2];

			a[0] = V_T5 ( 3, 0 );
			b[0] = V_T5 ( 3, 1 );
			c[0] = V_T5 ( 3, 2 );
			d[0] = V_T5 ( 3, 3 );
			e[0] = V_T5 ( 3, 4 );

			a[1] = V_T5 ( 4, 0 );
			b[1] = V_T5 ( 4, 1 );
			c[1] = V_T5 ( 4, 2 );
			d[1] = V_T5 ( 4, 3 );
			e[1] = V_T5 ( 4, 4 );

			double coeff0 = 0.0;
			double coeff1 = 0.0;
			double coeff2 = 0.0;
			double coeff3 = 0.0;

			coeff3 += getSummand3 ( a, c, e );
			coeff3 -= getSummand3 ( a, b, e );
			coeff3 += getSummand3 ( b, d, e );
			coeff3 -= getSummand3 ( a, d, e );
			coeff3 += getSummand3 ( a, b, d );
			coeff3 -= getSummand3 ( b, c, d );

			coeff2 += getSummand2 ( a, c, e );
			coeff2 -= getSummand2 ( a, b, e );
			coeff2 += getSummand2 ( b, d, e );
			coeff2 -= getSummand2 ( a, d, e );
			coeff2 += getSummand2 ( a, b, d );
			coeff2 -= getSummand2 ( b, c, d );

			coeff1 += getSummand1 ( a, c, e );
			coeff1 -= getSummand1 ( a, b, e );
			coeff1 += getSummand1 ( b, d, e );
			coeff1 -= getSummand1 ( a, d, e );
			coeff1 += getSummand1 ( a, b, d );
			coeff1 -= getSummand1 ( b, c, d );

			coeff0 += e[1] * c[1] * a[1];
			coeff0 -= a[1] * b[1] * e[1];
			coeff0 += b[1] * d[1] * e[1];
			coeff0 -= a[1] * d[1] * e[1];
			coeff0 += a[1] * b[1] * d[1];
			coeff0 -= b[1] * c[1] * d[1];

			coeff2 /= coeff3;
			coeff1 /= coeff3;
			coeff0 /= coeff3;

			return getRealRootsOfCubic ( coeff2, coeff1, coeff0, &alpha1, &alpha2, &alpha3 );
		}

	}
}
