

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include "correspondencecorrector.h"
#include <adrcore/util/storageorder.h>
#include <adrcore/util/printmat.h>
#include <adrcore/util/floatshift.h>
#include <adrcore/util/matmul2.h>

namespace adore
{
	namespace core
	{

		CorrespondenceCorrector::CorrespondenceCorrector() :
				T1 ( 3, 3 ),
				T2 ( 3, 3 ),
				T1inv ( 3, 3 ),
				T2inv ( 3, 3 ),
				R1 ( 3, 3 ),
				R2 ( 3, 3 ),
				x1 ( 3 ), x2 ( 3 ), e1 ( 3 ), e2 ( 3 ),
				Ftemp1 ( 3, 3 ), Ftemp1T ( 3, 3 ), Ftemp2 ( 3, 3 ),
				abcd ( 4 )
		{
			setIdentity ( T1 );
			setIdentity ( T2 );
			setIdentity ( T1inv );
			setIdentity ( T2inv );
			R1 = 0;
			R2 = 0;
		}


		CorrespondenceCorrector::~CorrespondenceCorrector()
		{
		}


		bool CorrespondenceCorrector::operator() ( AVector2D &p1, AVector2D &p2, AMatrix &fmat )
		{
			assert_msg ( p1.size() == 2 || p1 ( 2 ) == 1.0, "(%f, %f, %f)\n", p1 ( 0 ), p1 ( 1 ), p1 ( 2 ) );
			assert_msg ( p2.size() == 2 || p2 ( 2 ) == 1.0, "(%f, %f, %f)\n", p2 ( 0 ), p2 ( 1 ), p2 ( 2 ) );
			return operator() ( p1 ( 0 ), p1 ( 1 ), p2 ( 0 ), p2 ( 1 ), fmat );
		}

		bool CorrespondenceCorrector::operator() ( double &p1x, double &p1y, double &p2x, double &p2y, AMatrix &fmat )
		{
			//DBGOUT(DBGALL, "entered corrector functor\n");

			normalizeExponents ( fmat.begin(), fmat.end() );

//			printf( "fmat:\n" );
//			printMat( fmat, 20 );
//			printf( "points: (%f, %f)  (%f, %f)\n", p1x, p1y, p2x, p2y );

			//check, if points lie on epipoles
			trans ( fmat, Ftemp1T );
			fcalc.calculateEpipolarPoint ( fmat, e1 );
			fcalc.calculateEpipolarPoint ( Ftemp1T, e2 );
			double e1coor[2];
			double e2coor[2];
			e1coor[0] = e1 ( 0 ) / e1 ( 2 );
			e1coor[1] = e1 ( 1 ) / e1 ( 2 );
			e2coor[0] = e2 ( 0 ) / e2 ( 2 );
			e2coor[1] = e2 ( 1 ) / e2 ( 2 );

//			printf( "epipoles: (%f, %f) (%f, %f)\n", e1coor[0], e1coor[1], e2coor[0], e2coor[1] );

			//handle special cases with points on epipoles
			double p1IsEpi = false;
			double p2IsEpi = false;

			if ( fabs ( e1coor[0] - p1x ) < FLT_EPSILON && fabs ( e1coor[1] - p1y ) < FLT_EPSILON ) p1IsEpi = true;
			if ( fabs ( e2coor[0] - p2x ) < FLT_EPSILON && fabs ( e2coor[1] - p2y ) < FLT_EPSILON ) p2IsEpi = true;

			if ( p1IsEpi && p2IsEpi ) return true;
			else if ( p1IsEpi )
			{
				//is this really true? TODO
				p2x = e2coor[0];
				p2y = e2coor[1];
				return false;//not corrigible
			}
			else if ( p2IsEpi )
			{
				//is this really true? TODO
				p1x = e1coor[0];
				p1y = e1coor[1];
				return false;//not corrigible
			}

			//matrices, that translate the points to the origin
			T1 ( 0, 2 ) =  -p1x;
			T1 ( 1, 2 ) =  -p1y;
			T2 ( 0, 2 ) =  -p2x;
			T2 ( 1, 2 ) =  -p2y;

			//inverse for later use
			T1inv ( 0, 2 ) =  p1x;
			T1inv ( 1, 2 ) =  p1y;
			T2inv ( 0, 2 ) =  p2x;
			T2inv ( 1, 2 ) =  p2y;

//			printf( "T1:\n" );
//			printMat( T1, 20 );
//			printf( "T2:\n" );
//			printMat( T2, 20 );

//			printf( "T1inv:\n" );
//			printMat( T1inv, 20 );
//			printf( "T2inv:\n" );
//			printMat( T2inv, 20 );

			//calculate the fundamental matrix of the translated coordinate system
			matMul2 ( fmat, T1inv, Ftemp2 );
//			printf( "mul1:\n" );
//			printMat( Ftemp2 );

			matMul2_AT ( T2inv, Ftemp2, Ftemp1 );
			checkStorageOrder ( Ftemp1 );
//			printf( "tranformed FMat:\n" );
//			printMat( Ftemp1, 20 );

			trans ( Ftemp1, Ftemp1T );
//			printf( "trans'd fmat:\n" );
//			printMat( Ftemp1T, 20 );


			//DBGOUT(DBGALL, "calculating epipoles\n");
			//calculate epipoles
			fcalc.calculateEpipolarPoint ( Ftemp1, e1 );
			fcalc.calculateEpipolarPoint ( Ftemp1T, e2 );

//			printf( "Epipole1:\n" );
//			printMat( e1, 20 );
//			printf( "Epipole2:\n" );
//			printMat( e2, 20 );

			//DBGOUT(DBGALL, "normalizing epipoles\n");
			//normalize epipoles and build rotation matrices
			double e1div = 1.0, e2div = 1.0;
			e1div = SQR ( e1 ( 0 ) ) + SQR ( e1 ( 1 ) );
			e2div = SQR ( e2 ( 0 ) ) + SQR ( e2 ( 1 ) );

			//DBGOUT(DBGALL, "normalizing divisors: %.10f, %.10f\n", e1div, e2div);

			double e11 = e1 ( 0 ) / e1div;
			double e12 = e1 ( 1 ) / e1div;
			double e13 = e1 ( 2 ) / e1div;
			double e21 = e2 ( 0 ) / e2div;
			double e22 = e2 ( 1 ) / e2div;
			double e23 = e2 ( 2 ) / e2div;

			//DBGOUT(DBGALL, "building rotation matrix\n");
			R1 ( 0, 0 ) =  e11;
			R1 ( 0, 1 ) =  e12;
			R1 ( 1, 0 ) =  -e12;
			R1 ( 1, 1 ) =  e11;
			R1 ( 2, 2 ) =  1.0;

			R2 ( 0, 0 ) =  e21;
			R2 ( 0, 1 ) =  e22;
			R2 ( 1, 0 ) =  -e22;
			R2 ( 1, 1 ) =  e21;
			R2 ( 2, 2 ) =  1.0;

//			//print("R1:\n");
//			//printMat(R1, 20);
//			//print("R2:\n");
//			//printMat(R2, 20);
//
			//calculate the final fundamental fmatrix (12.3)
			matMul2_BT ( Ftemp1, R1, Ftemp2 );
			matMul2 ( R2, Ftemp2, Ftemp1 );

//			printf( "final fmat:\n" );
//			printMat( Ftemp1, 20 );

			//initializing the class members representing the main variables
			f1 = e13;
			f2 = e23;
			abcd[0] = Ftemp1 ( 1, 1 );
			abcd[1] = Ftemp1 ( 1, 2 );
			abcd[2] = Ftemp1 ( 2, 1 );
			abcd[3] = Ftemp1 ( 2, 2 );

			normalizeExponents ( abcd.begin(), abcd.end() );

			a = abcd[0];
			b = abcd[1];
			c = abcd[2];
			d = abcd[3];

			//DBGOUT(DBGALL, "preparing g(t)\n");
			prepareGt();

			//sometimes the floating point limits are reached
			if ( reachedInfty() )
			{
				DBGOUT ( DBGNORMAL, "unusual numbers in correspondencecorrector:\n" );
				printMat ( DBGNORMAL, Ftemp1 );
				DBGOUT ( DBGNORMAL, "points were: (%f, %f) and (%f, %f)\n", p1x, p1y, p2x, p2y );
				return false;
			}

			//DBGOUT(DBGALL, "calculating roots of g(t)\n");
			if ( !getRootsOfGt ( rootsR, rootsI ) )
			{
				return false;
			}

			//find the minimum of St by comparing its values at the roots of Gt
			//DBGOUT(DBGALL, "finding minimum of s(t)\n");
			double mint = DBL_MAX;
			double minVal = DBL_MAX;
			for ( int i = 0; i < 6; ++i )
			{
				//Evaluate only the real roots
				if ( rootsI[i] == 0.0 )
				{
					double val = getStAt ( rootsR[i] );
					//printf("R = %.10f   I = %.10f   s(R) = %.10f  g(t) = %.10f\n", rootsR[i], rootsI[i], val, getGtAt(rootsR[i]));
					if ( val < minVal )
					{
						minVal = val;
						mint = rootsR[i];
					}
				}
			}
			//printf("gT-test: %.10f   %.10f   %.10f   %.10f\n", getGtAt(rootsR[0] + 1.5), getGtAt(rootsR[0] + 15), getGtAt(rootsR[0] + 150), getGtAt(rootsR[0] + 1500));

			//print("mint before checking infty = %.10f  s(mint) = %.10f\n", mint, getStAt(mint));

			//check for t = infinity (asymtotic St)
			double infVal = 1.0/SQR ( f1 ) + SQR ( c ) / ( SQR ( a ) + SQR ( f2*c ) );
			//print("s(infinity) = %.10f\n", infVal);
			if ( infVal < minVal )
			{
				//this is what I think should happen. Not computet, because the result is infty anyway

				//mint = DBL_MAX;
				//printf("mint = infty, f = %f\n", f1);
				assert_msg ( f1 != 0.0, "f1 in illegal state in CorrepondenceCorrector." );
				assert_msg ( f2 != 0.0, "f2 in illegal state in CorrepondenceCorrector." );

				x1 ( 0 ) =  1;
				x1 ( 1 ) =  0;
				x1 ( 2 ) =  f1;

				x2 ( 0 ) =  1;
				x2 ( 1 ) =  0;
				x2 ( 2 ) =  f2;
			}
			else
			{
				//printf("mint = %.10f  s(mint) = %.10f\n", mint, getStAt(mint));

				//calculate optimal points (l and l' not explicitly computed)
				x1 ( 0 ) =  f1 * mint * mint;
				x1 ( 1 ) =  mint;
				x1 ( 2 ) =  f1 * f1 * mint + 1;

				//now compute the same for point 2:
				double ctd = c * mint + d;
				double atb = a * mint + b;

				x2 ( 0 ) =  f2*ctd*ctd;
				x2 ( 1 ) =  atb*ctd;
				x2 ( 2 ) =  f2*f2*ctd*ctd + atb*atb;

			}
			//transfer the points back, using e1 and e2 as vars
			matMul2_AT ( R1, x1, e1 );
			matMul2 ( T1inv, e1, x1 );

			//now, x1 and x2 are the points that best fist the given fundamental matrix
			p1x = x1 ( 0 ) / x1 ( 2 );
			p1y = x1 ( 1 ) / x1 ( 2 );

			matMul2_AT ( R2, x2, e2 );
			matMul2 ( T2inv, e2, x2 );

			p2x = x2 ( 0 ) / x2 ( 2 );
			p2y = x2 ( 1 ) / x2 ( 2 );

			//DBGOUT(DBGALL, "leaving corrector functor\n");
			return true;
		}


		void CorrespondenceCorrector::prepareGt()
		{
//			DBGOUT( DBGALL, "initials: %.10f, %.10f, %.10f, %.10f, %.10f, %.10f\n", f1, f2, a, b, c, d );

			double abd2 = a * b * d * d;
			double abc2 = a * b * c * c;
			double b2cd = b * b * c * d;
			double a2cd = a * a * c * d;
			double a2 = a * a;
			double b2 = b * b;
			double c2 = c * c;
			double d2 = d * d;
			double f22 = f2 * f2;
			double f24 = f22 * f22;
			double f12 = f1 * f1;
			double f14 = f12 * f12;

			t[6] = abd2 - b2cd;
			t[5] = b2*b2 + f24*d2*d2 + 2*b2*d2*f22 + b2*c2 - a2*d2;
			t[4] = 4*a*b2*b + 4*abd2*f22 + 4*b2cd*f22 + 4*c*d2*d*f24 + abc2 + 2*b2cd*f12 - a2cd - 2*abd2*f12;
			t[3] = 2*a2*b2 + 4*a2*b2 + 4*c2*d2*f22 + 2*a2*d2*f22 + 8*a*b*c*d*f22 + 2*b2*c2*f22 + 2*c2*d2*f24 + 2*b2*c2*f12 - 2*a2*d2*f12;
			t[2] = 4*a2*a*b + 4*a2cd*f22 + 4*abc2*f22 + 4*c2*c*d*f24 + b2cd*f14 + 2*abc2*f12 - abd2*f14 - 2*a2cd*f12;
			t[1] = a2*a2 + c2*c2*f24 + 2*a2*c2*f22 + b2*c2*f14 - a2*d2*f14;
			t[0] = a2cd*f14 - abc2*f14;
		}

		/*
		evaluates the function g(t) of algo 12.1 (HZ_Book)
		The variables are named along this algo, so don't blame me
		*/
		bool CorrespondenceCorrector::getRootsOfGt ( double rootsR[6], double rootsI[6] )
		{
			int degree = 6;
//			DBGOUT( DBGNORMAL, "roots of: %.10f, %.10f, %.10f, %.10f, %.10f, %.10f, %.10f\n", t[0], t[1], t[2], t[3], t[4], t[5], t[6] );
			bool success = rpoly ( t, degree, rootsR, rootsI );
			//DBGOUT(DBGALL, "success of rootfinding: %s, degree: %d\n", success ? "true" : "false", degree);
			if ( !success )
			{
				DBGOUT ( DBGALL, "Rootfinding failed. Degree is: %d\n", degree );
//				DBGOUT( DBGALL, "roots of: %.10f, %.10f, %.10f, %.10f, %.10f, %.10f, %.10f\n", t[0], t[1], t[2], t[3], t[4], t[5], t[6] );
//				pressKey();
			}
			return success;
		}

		double CorrespondenceCorrector::getGtAt ( double r )
		{
			double x = 1.0;
			double result = t[6] * x;
			x *= r;
			result += t[5] *x;
			x *= r;
			result += t[4] *x;
			x *= r;
			result += t[3] *x;
			x *= r;
			result += t[2] *x;
			x *= r;
			result += t[1] *x;
			x *= r;
			result += t[0] *x;

			return result;
		}

		double CorrespondenceCorrector::getStAt ( double t )
		{
			double ctd = SQR ( c * t + d );
			double atb = SQR ( a * t + b );
			double sum1 = t*t / ( 1 + f1 * f1 * t * t );
			double sum2 = ctd/ ( atb + f2 * f2 * ctd );
			//printf("ctd: %.10f, atb: %.10f, sum1: %.10f, sum2: %.10f\n", ctd, atb, sum1, sum2);
			return sum1 + sum2;
		}


	}
}
