

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include "fmatrixcalculator.h"
#include <adrcore/util/printmat.h>
#include <adrcore/util/floatshift.h>
#include <adrcore/util/storageorder.h>

namespace adore
{

	namespace core
	{


		/*
		    Enforces the 3x3-matrix mat to be of rank 2 by setting the lowest singular value of the SVD to zero.

		*/
		double FMatrixCalculator::enforceRank2 ( AMatrix &mat )
		{
			svdUVT ( mat, store3 );

			//setting the lowest singular value to zero
//			cout << rank2singulars << endl;
			double oldSingular = store3.D ( 2 );
			store3.D ( 2 ) = 0;

			//interpreting the vector rank2singulars as a diagonal matrix
			//this is a matmul
			mat ( 0, Range::all() ) = store3.D ( 0 ) * store3.VT ( 0, Range::all() );
			mat ( 1, Range::all() ) = store3.D ( 1 ) * store3.VT ( 1, Range::all() );
			mat ( 2, Range::all() ) = store3.D ( 2 ) * store3.VT ( 2, Range::all() );

			//abusing store3.VT as a temp-variable
			matMul2 ( store3.U, mat, store3.VT );

			//copying the result back to mat
			mat = store3.VT;

			return oldSingular;
		}

		int FMatrixCalculator::calculateFundamentalMatrixBase()
		{
			if ( numPoints < 7 )
			{
				DBGOUT ( DBGALL, "FMatrixCalculator::Too few correspondences to calculate a fundamental matrix\n" );
				return 0;
			}

			int ret = calculateFundamentalMatrixBase8Point();

			if ( ret > 0 )
			{
				//normalizeExponents(result1);//OPT
				enforceRank2 ( result1 );
				//normalizeExponents(result1);//OPT
			}
			if ( ret == 3 )
			{
				//normalizeExponents(result2);//OPT
				//normalizeExponents(result3);//OPT
				enforceRank2 ( result2 );
				enforceRank2 ( result3 );
				//normalizeExponents(result2);//OPT
				//normalizeExponents(result3);//OPT
			}
			return ret;
		}

		/**
		    ptsMat is altered
		    @return number of solutions
		*/
		int FMatrixCalculator::calculateFundamentalMatrixBase8Point()
		{
			return calculateFundamentalMatrix8Point ( ptsMat );
		}

		void extractAndPrintPoints ( const AMatrix &mat )
		{
			for ( unsigned row = 0; row < mat.rows(); ++row )
			{
				double x1 = mat ( row,6 );
				double y1 = mat ( row,7 );
				double x2 = mat ( row,2 );
				double y2 = mat ( row,5 );
				DBGOUT ( DBGNORMAL, "(%.4f, %.4f) - (%.4f, %.4f)\n", x1, y1, x2, y2 );
				ADORE_UNUSED( x1 );
				ADORE_UNUSED( y1 );
				ADORE_UNUSED( x2 );
				ADORE_UNUSED( y2 );
			}
		}


		int FMatrixCalculator::calculateFundamentalMatrix8Point ( const AMatrix& ptsMat )
		{
			/*			cout << "ptsMat: " << ptsMat << endl;
						store9.init(ptsMat.rows(), ptsMat.cols(), SVD_U);*/
			AVector &singulars = store9.D;
			singulars.resize ( std::min ( ptsMat.rows(), ptsMat.cols() ) );
			AMatrix &V_T = store9.VT;
			store9.changeRowsWorkspaceOnly ( ptsMat.rows() );
			store9.repairD();

			int numRoots = 1;

			singulars = 0;//only min(rows, 9) are filled by svd. To prevent uninitialized values

//			DBGOUT( DBGNORMAL, "8-point algo with %d points/rows (%d columns)\n", ptsMat.rows(), ptsMat.cols() );
//			printMat( ptsMat );printf( "\n" );
//			int retval = svdUVT ( ptsMat, store9 );
			int retval = svdVT ( ptsMat, store9 );
			if(retval > 0)
			{
				printMat( singulars );
				printMat( V_T );
				printMat( ptsMat );
//				pressKey();
			}
			assert_msg ( retval == 0, "dgesvd returns %d in fmatrixcalculator\n", retval ); ADORE_UNUSED ( retval );
//			cout << "singulars: " << singulars << endl;
//			cout << "V_T: " << V_T << endl;
//			AMatrix &U = store9.U;
//			cout << "U: " << U << endl;
//			AMatrix S( ptsMat.rows(), ptsMat.cols() );
//			S = 0;
//			for ( int i = 0; i < ptsMat.rows(); ++i )
//				S( i, i ) = singulars( i );
//			cout << "S: " << S << endl;
//			AMatrix temp( ptsMat.rows(), ptsMat.cols() );
//			AMatrix temp1( ptsMat.rows(), ptsMat.cols() );
//			matMul2( U, S, temp );
//			matMul2( temp, V_T, temp1 );
//			cout << "temp1: " << temp1 << endl; assert( false );

			//test, if something went wrong
			if ( singulars ( 6 ) <= eps )
			{
				//less than 7 points
				DBGOUT ( DBGNORMAL, "FMatrix: Linear Solution underdetermined\n" );
				DBGOUT ( DBGNORMAL, "FMatrix: The sixth singular value is: %.10f (calc'd from %zd points)\n", singulars ( 6 ), singulars.rows() );
#ifndef NDEBUG
				extractAndPrintPoints ( ptsMat );
#endif
				return 0;
			}
			if ( singulars.rows() == 7 || singulars ( 7 ) <= eps )
			{
				//fall back to 7 point algo
//				DBGOUT( DBGALL, "FMatrix: 7-Point-Algo\n" );
				numRoots = calculateFundamentalMatrixBase7Point ( V_T );
//				DBGOUT( DBGALL, "1:\n" );
//				printMat( DBGALL, result1 );
//				if ( numRoots == 3 )
//				{
//					DBGOUT( DBGALL, "2:\n" );
//					printMat( DBGALL, result2 );
//					DBGOUT( DBGALL, "3:\n" );
//					printMat( DBGALL, result3 );
//				}
			}
			else
			{
				result1 ( 0, 0 ) = V_T ( 8, 0 );
				result1 ( 0, 1 ) = V_T ( 8, 1 );
				result1 ( 0, 2 ) = V_T ( 8, 2 );
				result1 ( 1, 0 ) = V_T ( 8, 3 );
				result1 ( 1, 1 ) = V_T ( 8, 4 );
				result1 ( 1, 2 ) = V_T ( 8, 5 );
				result1 ( 2, 0 ) = V_T ( 8, 6 );
				result1 ( 2, 1 ) = V_T ( 8, 7 );
				result1 ( 2, 2 ) = V_T ( 8, 8 );
			}

			//core::core::printMat(singulars);printf("\n");

			return numRoots;
		}


		/*
		    V_T comes from the SVD of calculateFundamentalMatrix8Point.
		    This one is called as a fallback when exactly two singular values are zero.
		*/
		int FMatrixCalculator::calculateFundamentalMatrixBase7Point ( const AMatrix &V_T )
		{
			//using result1T and result2T as temporary variables.
			result1T ( 0, 0 ) = V_T ( 7, 0 );
			result1T ( 0, 1 ) = V_T ( 7, 1 );
			result1T ( 0, 2 ) = V_T ( 7, 2 );
			result1T ( 1, 0 ) = V_T ( 7, 3 );
			result1T ( 1, 1 ) = V_T ( 7, 4 );
			result1T ( 1, 2 ) = V_T ( 7, 5 );
			result1T ( 2, 0 ) = V_T ( 7, 6 );
			result1T ( 2, 1 ) = V_T ( 7, 7 );
			result1T ( 2, 2 ) = V_T ( 7, 8 );

			result2T ( 0, 0 ) = V_T ( 8, 0 );
			result2T ( 0, 1 ) = V_T ( 8, 1 );
			result2T ( 0, 2 ) = V_T ( 8, 2 );
			result2T ( 1, 0 ) = V_T ( 8, 3 );
			result2T ( 1, 1 ) = V_T ( 8, 4 );
			result2T ( 1, 2 ) = V_T ( 8, 5 );
			result2T ( 2, 0 ) = V_T ( 8, 6 );
			result2T ( 2, 1 ) = V_T ( 8, 7 );
			result2T ( 2, 2 ) = V_T ( 8, 8 );

			return calculateFundamentalMatricesFromNullSpaceSolutions ( result1T, result2T, result1, result2, result3 );
		}

		int FMatrixCalculator::calculateFundamentalMatricesFromNullSpaceSolutions (
		        const AMatrix &solution1, const AMatrix &solution2,
		        AMatrix &result1, AMatrix &result2, AMatrix &result3 )
		{
			//indices of primary diagonal of F in V_T (dl == down right)
			int drIdx[3][3] =
			{
				{0, 1, 2},
				{1, 2, 0},
				{2, 0, 1},
			};
			//indices of secondary diagonal of F in V_T (dr == down left)
			int dlIdx[3][3] =
			{
				{2, 1, 0},
				{0, 2, 1},
				{1, 0, 2},
			};

			//coefficients  of ax^3 + bx^2 + cx +d
			double a = 0;
			double b = 0;
			double c = 0;
			double d = 0;

			//diagonal elements of F1
			double f10 = 0;
			double f11 = 0;
			double f12 = 0;

			//diagonal elements of F2
			double f20 = 0;
			double f21 = 0;
			double f22 = 0;

			//difference between diagonal elements
			double d0 = 0.0;
			double d1 = 0.0;
			double d2 = 0.0;

			for ( int i = 0; i < 3; ++i )
			{
				//add primary diagonal
				f10 = solution1 ( 0, drIdx[i][0] );
				f11 = solution1 ( 1, drIdx[i][1] );
				f12 = solution1 ( 2, drIdx[i][2] );

				f20 = solution2 ( 0, drIdx[i][0] );
				f21 = solution2 ( 1, drIdx[i][1] );
				f22 = solution2 ( 2, drIdx[i][2] );

				d0 = f10 -f20;
				d1 = f11 -f21;
				d2 = f12 -f22;

				a +=  d0 *  d1 *  d2;
				b +=  d0 *  d1 * f22 +  d0 * f21 *  d2 + f20 *  d1 * d2;
				c +=  d0 * f21 * f22 + f20 *  d1 * f22 + f20 * f21 * d2;
				d += f20 * f21 * f22;

				//sub seconday digonal
				f10 = solution1 ( 0, dlIdx[i][0] );
				f11 = solution1 ( 1, dlIdx[i][1] );
				f12 = solution1 ( 2, dlIdx[i][2] );

				f20 = solution2 ( 0, dlIdx[i][0] );
				f21 = solution2 ( 1, dlIdx[i][1] );
				f22 = solution2 ( 2, dlIdx[i][2] );

				d0 = f10 -f20;
				d1 = f11 -f21;
				d2 = f12 -f22;

				a -=  d0 *  d1 *  d2;
				b -=  d0 *  d1 * f22 +  d0 * f21 *  d2 + f20 *  d1 * d2;
				c -=  d0 * f21 * f22 + f20 *  d1 * f22 + f20 * f21 * d2;
				d -= f20 * f21 * f22;
			}

			//now one has to solve ax^3 + bx^2 + cx +d = 0;
			//the x determines the linear combination xF1+(1-x)F2
			double r1, r2, r3;
			int numRoots = core::getRealRootsOfCubic ( b/a, c/a, d/a, &r1, &r2, &r3 );

			if ( numRoots > 0 )
			{
				//one one root, calculate the fundamental matrix
				result1 ( 0, 0 ) = r1 * solution1 ( 0, 0 ) + ( 1.0 - r1 ) * solution2 ( 0, 0 );
				result1 ( 0, 1 ) = r1 * solution1 ( 0, 1 ) + ( 1.0 - r1 ) * solution2 ( 0, 1 );
				result1 ( 0, 2 ) = r1 * solution1 ( 0, 2 ) + ( 1.0 - r1 ) * solution2 ( 0, 2 );
				result1 ( 1, 0 ) = r1 * solution1 ( 1, 0 ) + ( 1.0 - r1 ) * solution2 ( 1, 0 );
				result1 ( 1, 1 ) = r1 * solution1 ( 1, 1 ) + ( 1.0 - r1 ) * solution2 ( 1, 1 );
				result1 ( 1, 2 ) = r1 * solution1 ( 1, 2 ) + ( 1.0 - r1 ) * solution2 ( 1, 2 );
				result1 ( 2, 0 ) = r1 * solution1 ( 2, 0 ) + ( 1.0 - r1 ) * solution2 ( 2, 0 );
				result1 ( 2, 1 ) = r1 * solution1 ( 2, 1 ) + ( 1.0 - r1 ) * solution2 ( 2, 1 );
				result1 ( 2, 2 ) = r1 * solution1 ( 2, 2 ) + ( 1.0 - r1 ) * solution2 ( 2, 2 );
			}

			if ( numRoots == 3 )
			{
				//tree roots, compute the remaining 2 fmatrices
				result2 ( 0, 0 ) = r2 * solution1 ( 0, 0 ) + ( 1.0 - r2 ) * solution2 ( 0, 0 );
				result2 ( 0, 1 ) = r2 * solution1 ( 0, 1 ) + ( 1.0 - r2 ) * solution2 ( 0, 1 );
				result2 ( 0, 2 ) = r2 * solution1 ( 0, 2 ) + ( 1.0 - r2 ) * solution2 ( 0, 2 );
				result2 ( 1, 0 ) = r2 * solution1 ( 1, 0 ) + ( 1.0 - r2 ) * solution2 ( 1, 0 );
				result2 ( 1, 1 ) = r2 * solution1 ( 1, 1 ) + ( 1.0 - r2 ) * solution2 ( 1, 1 );
				result2 ( 1, 2 ) = r2 * solution1 ( 1, 2 ) + ( 1.0 - r2 ) * solution2 ( 1, 2 );
				result2 ( 2, 0 ) = r2 * solution1 ( 2, 0 ) + ( 1.0 - r2 ) * solution2 ( 2, 0 );
				result2 ( 2, 1 ) = r2 * solution1 ( 2, 1 ) + ( 1.0 - r2 ) * solution2 ( 2, 1 );
				result2 ( 2, 2 ) = r2 * solution1 ( 2, 2 ) + ( 1.0 - r2 ) * solution2 ( 2, 2 );

				result3 ( 0, 0 ) = r3 * solution1 ( 0, 0 ) + ( 1.0 - r3 ) * solution2 ( 0, 0 );
				result3 ( 0, 1 ) = r3 * solution1 ( 0, 1 ) + ( 1.0 - r3 ) * solution2 ( 0, 1 );
				result3 ( 0, 2 ) = r3 * solution1 ( 0, 2 ) + ( 1.0 - r3 ) * solution2 ( 0, 2 );
				result3 ( 1, 0 ) = r3 * solution1 ( 1, 0 ) + ( 1.0 - r3 ) * solution2 ( 1, 0 );
				result3 ( 1, 1 ) = r3 * solution1 ( 1, 1 ) + ( 1.0 - r3 ) * solution2 ( 1, 1 );
				result3 ( 1, 2 ) = r3 * solution1 ( 1, 2 ) + ( 1.0 - r3 ) * solution2 ( 1, 2 );
				result3 ( 2, 0 ) = r3 * solution1 ( 2, 0 ) + ( 1.0 - r3 ) * solution2 ( 2, 0 );
				result3 ( 2, 1 ) = r3 * solution1 ( 2, 1 ) + ( 1.0 - r3 ) * solution2 ( 2, 1 );
				result3 ( 2, 2 ) = r3 * solution1 ( 2, 2 ) + ( 1.0 - r3 ) * solution2 ( 2, 2 );

			}
			return numRoots;
		}


		void FMatrixCalculator::calculateEpipolarPoint ( const AMatrix &fmatrix, AVector &epi )
		{
			svdVT ( fmatrix, store3 );
			epi ( 0 ) = store3.VT ( 2, 0 );
			epi ( 1 ) = store3.VT ( 2, 1 );
			epi ( 2 ) = store3.VT ( 2, 2 );
		}

		void FMatrixCalculator::calculateEpipolarPoint ( const AMatrix &fmatrix, AVector &epi, SVDStoraged &store3 )
		{
			svdVT ( fmatrix, store3 );
			epi ( 0 ) = store3.VT ( 2, 0 );
			epi ( 1 ) = store3.VT ( 2, 1 );
			epi ( 2 ) = store3.VT ( 2, 2 );
		}

		AVector FMatrixCalculator::calculateEpipolarPoint ( const AMatrix &fmatrix )
		{
			AVector epi ( 3 );
			SVDStoraged store3 ( 3, 3, SVD_VT );
			calculateEpipolarPoint ( fmatrix, epi, store3 );
			return epi;
		}


		void FMatrixCalculator::calculateCamMatrix ( const AMatrix &fmat, const AVector &epipole, AMatrix &result )
		{
			assert ( result.rows() == 3 );
			assert ( result.cols() == 4 );

			AMatrix tmpMat = result ( Range::all(), Range ( 0, 2 ) );
			matMul2Skew ( epipole, fmat, tmpMat );
			result ( Range::all(), Range ( 0, 2 ) ) = tmpMat;

			result ( 0, 3 ) = epipole ( 0 );
			result ( 1, 3 ) = epipole ( 1 );
			result ( 2, 3 ) = epipole ( 2 );

			checkStorageOrder ( result );
			//debugging
//			assert( isRightHand( result ) );
#ifndef NDEBUG
			AMatrix fm = calculateFundamentalMatrix1Cam ( result );
			for ( int i = 0; i < 3; ++i ) for ( int j = 0; j < 3; ++j ) assert_msg ( fabs ( fm ( i, j ) ) - fabs ( fmat ( i, j ) ) < FLT_EPSILON,
									"FMatrixCalculator::calculateCamMatrix: result is wrong, maybe fmatrix was trans'd or the incorrect epipole was used (diff =  %.20f)\n", fabs ( fm ( i, j ) ) - fabs ( fmat ( i, j ) ) );
#endif
		}
		
		void FMatrixCalculator::calculateCamMatrixFullRankComponent ( const AMatrix &fmat, const AVector &epipole, AVector &result )
		{
			AMatrix mm(3, 3);
			matMul2Skew ( epipole, fmat, mm );
			
			SVDStoraged store3( 3, 3, SVD_VT );
			int retval = svdVT ( mm, store3 );
			assert_msg ( retval == 0, "dgesvd returns %d in fmatrixcalculator\n", retval ); ADORE_UNUSED ( retval );
			AVector a = store3.VT(0, Range::all());
			AVector b = store3.VT(1, Range::all());
			AVector c = cross(a, b);

			//store3.D(1) as a factor sets the third singular value to the same values as the second
			result(0) = -c(0) * store3.D(1);
			result(1) = -c(1) * store3.D(1);
			result(2) = -c(2) * store3.D(1);
			result(3) = store3.D(1);//TODO: find a good lambda
		}
		
		void FMatrixCalculator::calculateCamMatrixFullRank ( const AMatrix &fmat, const AVector &epipole, const AVector &vl, AMatrix &result )
		{
			assert ( result.rows() == 3 );
			assert ( result.cols() == 4 );

			AMatrix tmpMat = result ( Range::all(), Range ( 0, 2 ) );
			matMul2Skew ( epipole, fmat, tmpMat );
			
			result ( 0, 0 ) = tmpMat ( 0, 0 ) + vl(0) * epipole ( 0 );
			result ( 0, 1 ) = tmpMat ( 0, 1 ) + vl(1) * epipole ( 0 );
			result ( 0, 2 ) = tmpMat ( 0, 2 ) + vl(2) * epipole ( 0 );
			result ( 0, 3 ) = vl(3) * epipole ( 0 );

			result ( 1, 0 ) = tmpMat ( 1, 0 ) + vl(0) * epipole ( 1 );
			result ( 1, 1 ) = tmpMat ( 1, 1 ) + vl(1) * epipole ( 1 );
			result ( 1, 2 ) = tmpMat ( 1, 2 ) + vl(2) * epipole ( 1 );
			result ( 1, 3 ) = vl(3) * epipole ( 1 );

			result ( 2, 0 ) = tmpMat ( 2, 0 ) + vl(0) * epipole ( 2 );
			result ( 2, 1 ) = tmpMat ( 2, 1 ) + vl(1) * epipole ( 2 );
			result ( 2, 2 ) = tmpMat ( 2, 2 ) + vl(2) * epipole ( 2 );
			result ( 2, 3 ) = vl(3) * epipole ( 2 );

			if(!isRightHand( result )) result *= -1;
			
			checkStorageOrder ( result );
			//debugging
			assert( isRightHand( result ) );
		}
		
		void FMatrixCalculator::calculateCamMatrixFullRank ( const AMatrix &fmat, const AVector &epipole, AMatrix &result )
		{
			AVector vl(4);
			calculateCamMatrixFullRankComponent ( fmat, epipole, vl );
			calculateCamMatrixFullRank ( fmat, epipole, vl, result );
		}

		void FMatrixCalculator::calculateFundamentalMatrix1Cam ( const AMatrix &camera, AMatrix &fmat )
		{
			AMatrix M = camera ( Range::all(), Range ( 0, 2 ) );
			AVector t = camera ( Range::all(), 3 );
			matMul2Skew ( t, M, fmat );
		}

		AMatrix FMatrixCalculator::calculateFundamentalMatrix1Cam ( const AMatrix &camera )
		{
			AMatrix fmat ( 3, 3 );
			calculateFundamentalMatrix1Cam ( camera, fmat );
			return fmat;
		}

		AMatrix FMatrixCalculator::calculateFundamentalMatrix2Cams ( AMatrix &camera1,
		                const AMatrix &camera2 )
		{
			AMatrix fmat ( 3, 3 );
			calculateFundamentalMatrix2Cams ( camera1,	camera2, fmat );
			return fmat;
		}

		void FMatrixCalculator::calculateFundamentalMatrix2Cams ( AMatrix &camera1,
		                const AMatrix &camera2, AMatrix &fmat )
		{
			//calculate oriented camera center
			normalizeExponents ( camera1.begin(), camera1.end() );
// 			printf ( "cFM: normalized cam mat1:\n" );
// 			printMat ( camera1 );
			AVector C = getVectorOfMinors3x4 ( camera1 );
// 			printf ( "cFM: oriented camera centre:\n" );
// 			printMat ( C );

//			AVector epi2 = matMul ( camera2, C );
			AVector epi2(3);
			matMul2 ( camera2, C, epi2 );
// 			printf ( "cFM: epi2:\n" );
// 			printMat ( epi2 );

			AMatrix epi2Skew = getCrossMat ( epi2 );
// 			printf ( "epi2skew:\n" );
// 			printMat ( epi2Skew );

			AMatrix pseudo1 = pseudoInvertDeficient ( camera1 );
// 			printf ( "camera1:\n" );
// 			printMat ( camera1 );
// 			printf ( "pseudo1:\n" );
// 			printMat ( pseudo1 );

//			AMatrix camsVal = matMul ( camera2, pseudo1 );
			AMatrix camsVal(3, 3);
			matMul2 ( camera2, pseudo1, camsVal );
// 			printf ( "cam2 * pseudo:\n" );
// 			printMat ( camsVal );

			matMul2 ( epi2Skew, camsVal, fmat );
		}

		/*bool FMatrixCalculator::isOriented ( AMatrix &fmatrix,
		                                     const AVector &point1, const AVector &point2 )
		{
			AVector epi2(3);
			transInPlace ( fmatrix );
			calculateEpipolarPoint ( fmatrix, epi2 );
			transInPlace ( fmatrix );
			return isOriented ( fmatrix, epi2, point1, point2 );
		}*/

		bool FMatrixCalculator::isOriented ( const AMatrix &fmatrix, const AVector &epi2,
		                                     const AVector &point1, const AVector &point2 )
		{
			//calculate the third element of the equation
			//e2 x p2 +~ F p1
			//it is sufficient to test the positive scale factor

			double left = epi2 ( 0 ) * point2 ( 1 ) - epi2 ( 1 ) * point2 ( 0 );
			double right = fmatrix ( 2, 0 ) * point1 ( 0 ) + fmatrix ( 2, 1 ) * point1 ( 1 ) + fmatrix ( 2, 2 ) * point1 ( 2 );
			if ( ( left * right ) >= 0 ) return true;
			return false;
		}

		AMatrix FMatrixCalculator::calculateLineHomography ( const AMatrix &fmatrix,
		                const AVector &epipole )
		{
			AMatrix cross = getCrossMat ( epipole );
//			AMatrix result = matMul ( fmatrix, cross );
			AMatrix result(3, 3);
			matMul2 ( fmatrix, cross, result );
//			printf( "homo:\n" );
//			printMat( result );
			result ( Range::all(), 2 ) = result ( Range::all(), 2 ) + epipole;
//			printMat( result );
			assert ( getRank ( result ) == 3 );

			return result;
		}

		void FMatrixCalculator::calculateOrientedEpipolarPoint ( const AMatrix &fmatrix,
		                const AVector &epipole, AVector &result )
		{
			assert ( result.rows() == 3 );

//			double matData[9];
//			AMatrix mat33 ( matData, 3, 3 );
			AMatrix mat33 ( 3, 3 );

			mat33 ( Range::all(), 2 ) = epipole;

			mat33 ( Range::all(), 0 ) = fmatrix ( Range::all(), 1 );
			mat33 ( Range::all(), 1 ) = fmatrix ( Range::all(), 2 );
//			printf( "mat33\n" );
//			printMat( mat33 );
			result ( 0 ) = -det3 ( mat33 );

			mat33 ( Range::all(), 0 ) = fmatrix ( Range::all(), 0 );
//			printf( "mat33\n" );
//			printMat( mat33 );
			result ( 1 ) = det3 ( mat33 );

			mat33 ( Range::all(), 1 ) = fmatrix ( Range::all(), 1 );
//			printf( "mat33\n" );
//			printMat( mat33 );
			result ( 2 ) = -det3 ( mat33 );
		}

		AVector FMatrixCalculator::calculateOrientedEpipolarPoint ( const AMatrix &fmatrix, const AVector &epipole )
		{
			AVector result ( 3 );
			calculateOrientedEpipolarPoint ( fmatrix, epipole, result );
			return result;
		}
	}
}
