

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_NVIEWFMATRIXCALCULATOR_H
#define ADORE_NVIEWFMATRIXCALCULATOR_H

#include <vector>
#include <cfloat>
#include <adrcore/util/debug.h>
#include <adrcore/util/mathutil.h>
#include <adrcore/util/solvesvd.h>
#include <adrcore/types/aarray.h>

#ifndef SQRT2
#define SQRT2 1.414213562373095048763788073031832936977
#endif

namespace adore
{
	namespace core
	{

		/**
		 * This is an abstract base class for the calculation of the fundamental matrix an related values.
		 * The functions provided are using the OpenCv CvMat struct. To provide a nicer interface and
		 * additional funtionalities such as normalization/numerical preconditioning, derive your own class.
		 *
		 * This class allocates the neccessary memory for storing the diverse matrices used while calculating.
		 *  
		 */
		class FMatrixCalculator
		{
				FMatrixCalculator ( const FMatrixCalculator& ) { }
			public:

				/**
				 * Constructor, allocated the matrices according to numPoints via a call of init().
				 * @param numPoints the number of point correspondences.
				 */
				FMatrixCalculator ( int numPoints = 0 ) :
						eps ( FLT_EPSILON ),
						ptsMat(),
						result1 ( 3, 3 ),
						result2 ( 3, 3 ),
						result3 ( 3, 3 ),
						result1T ( 3, 3 ),
						result2T ( 3, 3 ),
						result3T ( 3, 3 ),
						store3( 3, 3, SVD_U | SVD_VT ),
						store9( 0, 9, SVD_VT )
				{
					if ( numPoints > 0 ) init ( numPoints );
				}

				/**
				 * Initialize FMatrixCalculator
				 * Essetially adjusts the size of ptsMat. Preserves the content as far as possible
				 * @param numPoints the number of point correspondences.
				 */
				void init ( int numPoints )
				{
					this->numPoints = numPoints;
					ptsMat.resizeAndPreserve ( numPoints, 9 );
				}

				/**
				 * Destructor. Deletes the matrices.
				 */
				virtual ~FMatrixCalculator()
				{
				}

				/**
				 * virtual funtion for the computation of the fundamental matrix.
				 * returns the number of solutions
				 */
				virtual int calculateFundamentalMatrix()
				{
					return 0;
				}

				/**
				 * Calculates the fundamental matrix from the camera matrix of the second view
				 * corresponding to the identity cam for the first view.
				 * F = [t]_x M, for cam = |M|t|
				 */
				static AMatrix calculateFundamentalMatrix1Cam ( const AMatrix &cam );
				static void calculateFundamentalMatrix1Cam ( const AMatrix &camera, AMatrix &fmat );

				/**
				 * Calculates the fundamental matrix from two general camera matrices
				 * F = [e2]_x P2 P1+, with P1+ the pseudo-inverse of P1
				 */
				static AMatrix calculateFundamentalMatrix2Cams ( AMatrix &cam1, const AMatrix &cam2 );
				static void calculateFundamentalMatrix2Cams ( AMatrix &cam1, const AMatrix &cam2, AMatrix &fmat );

				/**
					Creates a FMatrix of the type
					0	0	0
					0	0	1
					0	-1	0
					which maps a point to line crossing it horizontally
				*/
				static inline AMatrix createHorizontalFMatrix()
				{
					AMatrix fmat ( 3, 3 );
					createHorizontalFMatrix( fmat );
					return fmat;
				}
				static inline void createHorizontalFMatrix( AMatrix &fmat )
				{
					fmat = 0, 0, 0, 0, 0, 1, 0, -1, 0;
				}

				/**
				Creates a FMatrix of the type
				0	0	1
				0	0	0
				-1	0	0
				which maps a point to line crossing it vertically
				 */
				static inline AMatrix createVerticalFMatrix()
				{
					AMatrix fmat ( 3, 3 );
					createVerticalFMatrix( fmat );
					return fmat;
				}
				static inline void createVerticalFMatrix( AMatrix &fmat )
				{
					fmat = 0, 0, 1, 0, 0, 0, -1, 0, 0;
				}

				/**
				 * calculates an epipolar point given a fundamental matrix.
				 * That is, the right null space of F is computed.
				 * Fe = 0;
				 * F^Te' = 0;
				 * @param fmatrix a 3x3 fundamental matrix.
				 * @param result a 3x1 vector in which the result is stored.
				 */
				void calculateEpipolarPoint ( const AMatrix &fmat, AVector &epi );
				static void calculateEpipolarPoint ( const AMatrix &fmat, AVector &epi, SVDStoraged &store3 );
				static AVector calculateEpipolarPoint ( const AMatrix &fmatrix );

				/**
				 * calculates a correct oriented epipolarpoint
				 * given a Fundamental Matrix and the other epipolar point.
				 * (e^T, 0)^T +=+ (F|e')^^
				 * or
				 * (e'^T, 0)^T +=+ (F^T|e)^^
				 * */
				static void calculateOrientedEpipolarPoint ( const AMatrix &fmat, const AVector &epipole, AVector &result );
				static AVector calculateOrientedEpipolarPoint ( const AMatrix &fmat, const AVector &epipole );

				const AMatrix & getSolution ( int i )
				{
					extern AMatrix NullMatrix;
					switch ( i )
					{
						case 0:
							return result1;
							break;
						case 1:
							return result2;
							break;
						case 2:
							return result3;
							break;
						default:
							return NullMatrix;
							break;
					}
					return NullMatrix;
				}

				/**
				 * @return a pointer to the first fundamental matrix. There may be up to three, if
				 * the 7-point-algorithm was used.
				 */
				const AMatrix &getFMatrix1()
				{
					return result1;
				}

				/**
				 * @return a pointer to the second fundamental matrix. Only valid, if the
				 * 7-point-algorithm was used.
				 */
				const AMatrix &getFMatrix2()
				{
					return result2;
				}

				/**
				 * @return a pointer to the third fundamental matrix. Only valid, if the
				 * 7-point-algorithm was used.
				 */
				const AMatrix &getFMatrix3()
				{
					return result3;
				}

				/**
				 * @return a tranposed version of the first fundamental matrix.
				 */
				const AMatrix &getFMatrix1T()
				{
					result1T = trans ( result1 );
					return result1T;
				}

				/**
				 * @return a tranposed version of the second fundamental matrix.
				 */
				const AMatrix &getFMatrix2T()
				{
					result2T = trans ( result2 );
					return result2T;
				}

				/**
				 * @return a tranposed version of the third fundamental matrix.
				 */
				const AMatrix &getFMatrix3T()
				{
					result3T = trans ( result3 );
					return result3T;
				}

				/**
				 * This function calculates the second of the canonical AMatrix Matrices (Result 9.14 of HZ-Book).
				 * (For the fist invoke setIdentity(matrix))
				 * cam = |[e']_x F|e'|
				 */
				static void calculateCamMatrix ( const AMatrix &fmat, const AVector &epipole, AMatrix &result );

				/**
				 * like calculateCamMatrix, but applies a Heuristik that shall ensure the rank-3-ness of the result.
				 */
				static void calculateCamMatrixFullRank ( const AMatrix &fmat, const AVector &epipole, AMatrix &result );
				
				/**
				 * same as above, but uses a provided v and lambda
				 */
				static void calculateCamMatrixFullRank ( const AMatrix &fmat, const AVector &epipole, const AVector &vl, AMatrix &result );
				
				/**
				 * computes the v and lamba component to make a good camera matrix
				 * the result vector must have 4 components to hold v1, v2, v3, and lamba.
				 */
				static void calculateCamMatrixFullRankComponent ( const AMatrix &fmat, const AVector &epipole, AVector &result );
				
				/**
				 * If you have to fundamental matrices F1 and F2 and you know your correct fundamental matrix is
				 * a linear combination of them, you may want to use this function.
				 * The result may consists of one or three fundamtal matrices.
				 * @param solution1 F1, the first null space generator
				 * @param solution2 F2, the second null space generator
				 * @param result1 a CvMat of the appropriate size to hold a fundamental matrix.
				 * @param result2 a CvMat of the appropriate size to hold a fundamental matrix.
				 * @param result3 a CvMat of the appropriate size to hold a fundamental matrix.
				 */
				static int calculateFundamentalMatricesFromNullSpaceSolutions (
				        const AMatrix &solution1, const AMatrix &solution2,
				        AMatrix &result1, AMatrix &result2, AMatrix &result3 );

				/**
				 * test, whether the fundamental matrix is oriented or not.
				 * This test requires a known correspondence and the epipole of the second image.
				 */
				static bool isOriented ( const AMatrix &fmatrix, const AVector &epi2,
				                         const AVector &point1, const AVector &point2 );

				/**
				 * test, whether the fundamental matrix is oriented or not.
				 * This test requires a known correspondence of the second image.
				 * In this version, the epipole is computed for you.
				 */
				//static bool isOriented ( AMatrix &fmatrix, const AVector &point1, const AVector &point2 );

				/**
				 * Calculates a homography relating epipolar lines
				 * l' = H l  =>  H =  F [e]_x
				 * l  = H l' =>  H = FT [e']_x
				 * Make sure to pass the correct fmatrix/epipole combination.
				 */
				AMatrix calculateLineHomography ( const AMatrix &fmatrix, const AVector &epipole );

				/**
					Set the lowest singular value to zero. This enforces a rank of 2.
					@param mat The 3x3-Matrix.
					@return the old value of the lowest singular value
				*/
				double enforceRank2 ( AMatrix &mat );

				int calculateFundamentalMatrix8Point ( const AMatrix& ptsMat );

			protected:
				double eps;
				bool createdFixedMat;
				bool createdVarMat;

				int numPoints;
				AMatrix ptsMat;

				AMatrix result1;
				AMatrix result2;
				AMatrix result3;
				AMatrix result1T;
				AMatrix result2T;
				AMatrix result3T;

				SVDStoraged store3;
				SVDStoraged store9;

				/**
				    calculates the fundamental Matrix.
				    @return the number of fundamental matrices (0/1->1 or 3->3), or -1 if something went wrong
				*/
				int calculateFundamentalMatrixBase8Point();

				int calculateFundamentalMatrixBase();

				/**
				 *  calculates the fundamental Matrix.
				 *  In the case of just 7 point correspondeces, calculateFundamentalMatrix8Point calls this method as a fallback.
				 *  This static funtions is used for the minimal solution of the trifocal tensor as well, therefor it is static.
				 *  @return the number of fundamental matrices (1 or 3), or -1 if something went wrong
				*/
				int calculateFundamentalMatrixBase7Point ( const AMatrix &V_T );

		};

	}

}

#endif
