

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include "merger.h"
#include <adrcore/types/pointcorr.h>
#include <adrcore/util/printmat.h>
#include <adrcore/util/floatshift.h>
#include <list>
#include <adrcore/random/ransac.h>
#include <adrcore/nview/ransacclientmergercp.h>
#include <adrcore/minimization/lm2.h>
#include <adrcore/minimization/lmfunctionjacobi2.h>
#include <adrcore/nview/mergerhminfunc.h>

#include <adrcore/mestimate/filterransaccp.h>
#include <adrcore/random/prosac.h>

#include <adrcore/nview/homographycalculator.h>

namespace adore
{

	namespace core
	{

		Merger::Merger():
                    useCameras( true ),
                    usePoints3D( false ),
                    minimizeReproErr( false ),
                    assumeMetric( false )
		{
			p3dCalc.setMinimizeError( true );
		}


		Merger::~Merger()
		{
		}

		//TODO: put this in automerger
		bool Merger::merge( NViewMap &map, const ViewKnitter &knit, const bool delSourceNViews )
		{
			set<NView *> sourceNViews;
			for ( ViewKnitter::const_iterator it = knit.begin(); it != knit.end(); ++it )
			{
				const Tuple &tup = *it;

				VTuple *inA = tup.get<0>();
				VTuple *inB = tup.get<1>();
				VTuple *inAorB = tup.get<2>();

				NView *nv1 = map.getNView( inA->begin(), inA->end() );
				assert( nv1 );
				NView *nv2 = map.getNView( inB->begin(), inB->end() );
				assert( nv2 );
				NView *result = map.createNView( inAorB->begin(), inAorB->end() );
				assert( result );

				if ( !merge( *nv1, *nv2, *result, tup ) )
				{
					DBGOUT( DBGALL, "merging failed\n" );
					error( "blub\n" );
					return false;
				}

				if ( nv1->isNormalized() && nv2->isNormalized() )
				{
					result->setNormalized( true );
				}
				else if ( !nv1->isNormalized() && !nv2->isNormalized() )
				{
					result->setNormalized( false );
				}
				else
				{
					error( "Merger found mixed normalizations\n" );
				}

				//if one of those has 3d points already, we assume a recalculation
				if ( nv1->getNumPoints3D() || nv2->getNumPoints3D() )
				{
					p3dCalc.modify( *result );
				}

//				if ( bundleAdjust )
//				{
//					ba.modify( *result );
//					DBGOUT( DBGNORMAL, "Merger: BA after merging %d views needs %d iterations\nresidual(before - after): %.30f - %.30f\n", result->getNumViews(), ba.getLastIterations(), ba.getInitialError(), ba.getLastError() );
//				}

				DBGOUT( DBGNORMAL, "Merger: (%d + %d)->%d\n", nv1->getNumViews(), nv2->getNumViews(), result->getNumViews() );
				if ( delSourceNViews )
				{
					sourceNViews.insert( nv1 );
					sourceNViews.insert( nv2 );
				}
			}
			if ( delSourceNViews )
			{
				set<NView *>::iterator it = sourceNViews.begin();
				while ( it != sourceNViews.end() )
				{
					//save the iterator first to prevent invalidation
					set<NView *>::iterator oldIt = it; ++it;
					NView *nv = *oldIt;
					sourceNViews.erase( oldIt );
					assert_msg( nv->getViewResetMode() != DELETE_VIEW, "Dangerous VIEW-RESET-Mode in merger. The merging-result will point to nonexisting views." );
					assert_msg( nv->getCorrResetMode() == DELETE_CORR, "Leaky CORR-RESET-Mode. Old corrs wont get deleted." );
					map.setNView( nv->getViewsBegin(), nv->getViewsEnd(), NULL );
					delete nv;
				}
			}
			return true;
		}

		bool Merger::merge( NView &set1, NView &set2, NView &result )
		{
			DBGOUT( DBGALL, "merging\n" );

			//copy views and correspondences
			result.addViews( set1 );
			DBGOUT( DBGALL, "%d correspondences in set1\n", set1.getNumPointCorrs() );
			result.addViews( set2 );
			DBGOUT( DBGALL, "%d correspondences in set2\n", set2.getNumPointCorrs() );
			result.addPointCorrs( set1 );
			result.addPointCorrs( set2 );
			DBGOUT( DBGALL, "%d correspondences in result-nview\n", result.getNumPointCorrs() );

			return true;
		}

		bool Merger::merge( NView &set1, NView &set2, NView &result, Tuple tup )
		{
			bool ret = merge( set1, set2, result );
			if ( ( useCameras && set1.getNumCameras() && set2.getNumCameras() ) ||
			                ( usePoints3D && set1.getNumPoints3D() && set2.getNumPoints3D() ) )
			{
				ret = mergeByHomography( set1, set2, result, tup );
			}
			return ret;
		}

		void Merger::getCommonPoints( NView &set1, NView &set2, vector<ViewBase *> &v,
		                              vector<AVector3D *> &points1, vector<AVector3D *> &points2,
		                              bool useOr )
		{
			vector<AVector3D *> pt3d1( set1.getNumPoints3D() );
			set1.getPoints3D( pt3d1.begin() );
			for ( uint i = 0; i < pt3d1.size(); ++i )
			{
				AVector3D *p2 = NULL;

				bool success = !useOr;

				for ( uint j = 0; j < v.size(); ++j )
				{
					FeaturePoint *nvp1 = set1.getFeaturePoint( pt3d1[i], v[j] );
					if ( !useOr )
					{
						if ( !set2.hasFeaturePoint( nvp1 ) )
						{
							success = false;
						}
						else
						{
							p2 = set2.getPoint3D( nvp1 );
						}
					}
					else if ( useOr && set2.hasFeaturePoint( nvp1 ) )
					{
						success = true;
						p2 = set2.getPoint3D( nvp1 );
						break;
					}
				}

				if ( success )
				{
					points1.push_back( pt3d1[i] );
					points2.push_back( p2 );
				}
			}
		}

		bool Merger::mergeByHomography( NView &set1, NView &set2, NView &result, Tuple tup )
		{
			VTuple &inAandB = *tup.get<3>();
			assert( inAandB.size() > 0 );

			if(inAandB.size() == 1)
			{
				return mergeByHomography1ViewOverlap( set1, set2, result, tup );
			}
			else
			{
				return mergeByHomography2ViewOverlap( set1, set2, result, tup );
			}
		}

		bool Merger::mergeByHomography1ViewOverlap( NView &set1, NView &set2, NView &result, Tuple tup )
		{
			const bool miniRepro = minimizeReproErr;
			vector<vector<AVector2D*> > feats1View(1);
			AMatrix H( 4, 4 ), transformedCam( 3, 4 );

//			VTuple &inA = *tup.get<0>();
			VTuple &inB = *tup.get<1>();
			VTuple &inAandB = *tup.get<3>();
			assert( inAandB.size() == 1 );

			vector<AMatrix *> cams1, cams2;
			if ( useCameras || miniRepro )
			{
				cams1.push_back( &set1.getCamera( inAandB[0] )->getProjMat() );
				cams2.push_back( &set2.getCamera( inAandB[0] )->getProjMat() );
			}

			vector< AVector3D* > commonPts1, commonPts2;//initial candidates
			getCommonPoints( set1, set2, inAandB, commonPts1, commonPts2 );
			DBGOUT( DBGNORMAL, "Merger: 3D points in common: %zd\n", commonPts1.size() );

			if ( commonPts1.size() == 0 )
			{
				DBGOUT( DBGALL, "no common points in %zd overlapping views\n", inAandB.size() );
				return false;
			}

			assert( commonPts1.size() == commonPts2.size() );

			vector< AVector3D* > pts3D1, pts3D2;
			if ( assumeMetric )
			{
				//RANSAC filtering for one-view-case
				performRANSAC( cams1, cams2, commonPts1, commonPts2, pts3D1, pts3D2, &set1, &set2 );
				feats1View[0].resize( pts3D1.size() );
				for ( uint p = 0; p < pts3D1.size(); ++p )
				{
					feats1View[0][p] = set1.getFeaturePoint( pts3D1[p], inAandB[0] );
				}
				HomographyCalculator hcalc;
				bool ok = hcalc.calculateAssumingMetricUseAll( *cams1[0], *cams2[0], pts3D1, pts3D2, H );
				if(!ok) return false;
			}
			else
			{
				//ransac filtering for general case
				performRANSAC( cams1, cams2, commonPts1, commonPts2, pts3D1, pts3D2, &set1, &set2 );
				H = calculateHomographyFromCamerasAndPoints( cams1, cams2, pts3D1, pts3D2 );
			}

			double error = getErrorOfHomography( cams1, cams2, pts3D1, pts3D2, feats1View, H );
			ADORE_UNUSED( error );
			DBGOUT( DBGNORMAL, "Merger: Error of homography: %f\n", error );

#ifndef NDEBUG
			DBGOUT( DBGNORMAL, "Merger: before mini" );
			testHomography( cams1, cams2, pts3D1, pts3D2, H );
//			pressKey();
#endif

			if ( miniRepro )
			{
				if ( result.getNumPointCorrs() > 0 )
				{
					DBGOUT( DBGNORMAL, "Merger: result has %d corrs\n", result.getNumPointCorrs() );
					minimizeReprojectionError( pts3D1, pts3D2, cams1, cams2, feats1View, H );
				}
				else
				{
					DBGOUT( DBGCRITICAL, "Merger: minimizeReprojectionError impossible: NView lacks Corrs\n" );
				}

#ifndef NDEBUG
				DBGOUT( DBGNORMAL, "Merger: after mini" );
				testHomography( cams1, cams2, pts3D1, pts3D2, H );
//				pressKey();
#endif
			}

//			removeOutliers( set1, *pts3D1best, inAandB );
//			removeOutliers( set2, *pts3D2best, inAandB );

			//copy the cameras of set1 to the result
			DBGOUT( DBGNORMAL, "Merger: copying %d cameras\n", set1.getNumCameras() );
			result.addCameras( set1 );

			//transform cameras of set2 and then copy them, too
			vector<ViewBase *>::iterator it1 = inB.begin();
			DBGOUT( DBGNORMAL, "Merger: adding transformed camera\n" );
			AMatrix Hinv(4, 4);
			Hinv = H;
			invert(Hinv);
			for ( ; it1 != inB.end(); ++it1 )
			{
				Camera *camB = set2.getCamera( *it1 );
//				cout << "cam B: " << camB->getProjMat();

				if( result.getCamera( *it1, false) != NULL )
				{
					Camera *camA = result.getCamera( *it1, true );//create, if not available
//					cout << "cam A: " << camA->getProjMat();
					matMul2( camB->getProjMat(), Hinv, transformedCam );
//					cout << "transformed: " << transformedCam;
					matMul2( camA->getProjMat(), H, transformedCam );
//					cout << "transformed: " << transformedCam;
					continue;
				}

				Camera * camA = result.getCamera( *it1, true );//create, if not available
				matMul2( camB->getProjMat(), Hinv, transformedCam );
//				cout << "transformed: " << transformedCam;
				camA->getProjMat() = transformedCam;
			}

			//transform cameras of set1 and then copy them, too
//			vector<ViewBase *>::iterator it1 = inA.begin();
//			DBGOUT( DBGNORMAL, "Merger: adding transformed camera\n" );
//			for ( ; it1 != inA.end(); ++it1 )
//			{
//				Camera *camA = set1.getCamera( *it1 );
//				Camera *camB = result.getCamera( *it1, true );//create, if not available
//				matMul2( camA->getProjMat(), H, transformedCam );
//				cout << camA->getProjMat();
//				if(set2.getCamera( *it1, false) != NULL)
//					cout << set2.getCamera(*it1)->getProjMat();
//				cout << transformedCam;
//				camB->getProjMat() = transformedCam;
//			}

			//copy the cameras of set2 to the result
//			DBGOUT( DBGNORMAL, "Merger: copying %d cameras\n", set2.getNumCameras() );
//			result.addCameras( set2 );

			feats1View.clear();

			return true;
		}

		bool Merger::mergeByHomography2ViewOverlap( NView &set1, NView &set2, NView &result, Tuple tup )
		{
			VTuple &inAandB = *tup.get<3>();
			assert( inAandB.size() > 1 );
			VTuple &inA = *tup.get<0>();

			bool miniRepro = minimizeReproErr;

			AMatrix H( 4, 4 );// a homography
			AMatrix transformedCam( 3, 4 );

			vector<AMatrix *> cams1;
			vector<AMatrix *> cams2;

			if ( useCameras || miniRepro )
			{
				cams1.reserve( inAandB.size() );
				cams2.reserve( inAandB.size() );
				for ( uint i = 0; i < inAandB.size(); ++i )
				{
					cams1.push_back( &set1.getCamera( inAandB[i] )->getProjMat() );
					cams2.push_back( &set2.getCamera( inAandB[i] )->getProjMat() );
					DBGOUT( DBGNORMAL, "Merger: normalizing camera exponents\n" );
					normalizeExponents( cams1.back()->begin(), cams1.back()->end() );
					normalizeExponents( cams2.back()->begin(), cams2.back()->end() );
				}
			}

			vector< AVector3D* > candPts3D1;//initial candidates
			vector< AVector3D* > candPts3D2;
			vector< AVector3D* > pts3D1;//results of n-view-case-ransac
			vector< AVector3D* > pts3D2;
			vector< AVector3D* > *pts3D1best = &candPts3D1;//points to one- or n-view-case
			vector< AVector3D* > *pts3D2best = &candPts3D2;

			vector<vector<AVector2D *> > feats1View( inAandB.size() );
			vector<vector<AVector2D *> > featsNView( inAandB.size() );
			vector<vector<AVector2D *> > *bestFeats = &featsNView;

			if ( miniRepro || usePoints3D )
			{
				getCommonPoints( set1, set2, inAandB, candPts3D1, candPts3D2 );
				DBGOUT( DBGNORMAL, "Merger: 3D points in common: %zd\n", candPts3D1.size() );
				if ( candPts3D1.size() < 1 ) miniRepro = false;

				if ( usePoints3D )
				{
					if ( candPts3D1.size() == 0 )
					{
						DBGOUT( DBGALL, "no common points in %zd overlapping views\n", inAandB.size() );
						return false;
					}

					assert( candPts3D1.size() == candPts3D2.size() );

					//ransac filtering for general case
					performRANSAC( cams1, cams2, candPts3D1, candPts3D2, pts3D1, pts3D2, &set1, &set2 );
					pts3D1best = &pts3D1;
					pts3D2best = &pts3D2;
				}
			}

			H = calculateHomographyFromCamerasAndPoints( cams1, cams2, pts3D1, pts3D2 );
			double error = getErrorOfHomography( cams1, cams2, pts3D1, pts3D2, feats1View, H );
			ADORE_UNUSED( error );
			DBGOUT( DBGNORMAL, "Merger: Error of homography: %f\n", error );

			if ( miniRepro )
			{
				//remeber: bestFeats defaults to featsNView
				for ( uint i = 0; i < inAandB.size(); ++i )
				{
					DBGOUT( DBGNORMAL, "resizing featsNView[%d] to %zd\n", i, pts3D1best->size() );
					featsNView[i].resize( pts3D1best->size() );
				}
				for ( uint p = 0; p < pts3D1best->size(); ++p )
				{
					for ( uint i = 0; i < inAandB.size(); ++i )
					{
						ViewBase *v = inAandB[i];
						AVector *pt3D = ( *pts3D1best )[p];
						FeaturePoint *fp = set1.getFeaturePoint( pt3D, v );
						featsNView[i][p] = fp;
						if ( !fp )
						{
							DBGOUT( DBGNORMAL, "Merger: No Feat for 3D Point %p in View %p\n", pt3D, v );
						}
					}
				}
			}

#ifndef NDEBUG
			DBGOUT( DBGNORMAL, "Merger: before mini" );
			testHomography( cams1, cams2, *pts3D1best, *pts3D2best, H );
//			pressKey();
#endif
			//remember: bestFeats defaults to featsNView
			if ( miniRepro )
			{
				if ( result.getNumPointCorrs() > 0 )
				{
					DBGOUT( DBGNORMAL, "Merger: result has %d corrs\n", result.getNumPointCorrs() );
					minimizeReprojectionError( *pts3D1best, *pts3D2best, cams1, cams2, *bestFeats, H );
				}
				else
				{
					DBGOUT( DBGCRITICAL, "Merger: minimizeReprojectionError impossible: NView lacks Corrs\n" );
				}

#ifndef NDEBUG
				DBGOUT( DBGNORMAL, "Merger: after mini" );
				testHomography( cams1, cams2, *pts3D1best, *pts3D2best, H );
//				pressKey();
#endif
			}

//			removeOutliers( set1, *pts3D1best, inAandB );
//			removeOutliers( set2, *pts3D2best, inAandB );

			//transform cameras of set1 and then copy them, too
			vector<ViewBase *>::iterator it1 = inA.begin();
			DBGOUT( DBGNORMAL, "Merger: adding transformed camera\n" );
			for ( ; it1 != inA.end(); ++it1 )
			{
				Camera *camA = set1.getCamera( *it1 );
				Camera *camB = result.getCamera( *it1, true );//create, if not available
				matMul2( camA->getProjMat(), H, transformedCam );
				camB->getProjMat() = transformedCam;
			}

			//copy the cameras of set2 to the result
			DBGOUT( DBGNORMAL, "Merger: copying %d cameras\n", set2.getNumCameras() );
			result.addCameras( set2 );

			featsNView.clear();

			return true;
		}

		/** _________________ MERGESTEP BEGIN ____________________ */

		bool Merger::merge( NView & set1, NView & set2, NView & result, vector< ViewBase * > & inA, vector< ViewBase * > & inAandB )
		{
			bool ret = merge( set1, set2, result );
			if ( ( useCameras && set1.getNumCameras() && set2.getNumCameras() ) ||
			                ( usePoints3D && set1.getNumPoints3D() && set2.getNumPoints3D() ) )
			{
				ret = mergeByHomography( set1, set2, result, inA, inAandB );
			}
			if ( !ret )
			{
				DBGOUT( DBGALL, "merging failed\n" );
				error( "blub\n" );
				return false;
			}
			if ( set1.isNormalized() && set2.isNormalized() )
				result.setNormalized( true );
			else if ( !set1.isNormalized() && !set2.isNormalized() )
				result.setNormalized( false );
			else
				error( "Merger found mixed normalizations\n" );

			//if one of those has 3d points already, we assume a recalculation
			if ( set1.getNumPoints3D() || set2.getNumPoints3D() )
				p3dCalc.modify( result );

			DBGOUT( DBGNORMAL, "Merger: (%d + %d)->%d\n", set1.getNumViews(), set2.getNumViews(), result.getNumViews() );
			return true;
		}

		bool Merger::mergeByHomography( NView &set1, NView &set2, NView &result, vector<ViewBase*>& inA, vector<ViewBase*>& inAandB )
		{
			assert ( inAandB.size() > 0 );

			bool miniRepro = minimizeReproErr;

			AMatrix H ( 4, 4 );// a homography
			AMatrix transformedCam ( 3, 4 );

			vector<AMatrix *> cams1;
			vector<AMatrix *> cams2;

			if ( useCameras || miniRepro )
			{
				cams1.reserve ( inAandB.size() );
				cams2.reserve ( inAandB.size() );

				for ( uint i = 0; i < inAandB.size(); ++i )
				{
					cams1.push_back ( &set1.getCamera ( inAandB[i] )->getProjMat() );
					cams2.push_back ( &set2.getCamera ( inAandB[i] )->getProjMat() );
					DBGOUT ( DBGNORMAL, "Merger: normalizing camera exponents\n" );
					normalizeExponents ( cams1.back()->begin(), cams1.back()->end() );
					normalizeExponents ( cams2.back()->begin(), cams2.back()->end() );
				}
			}

			vector< AVector3D* > candPts3D1;//initial candidates

			vector< AVector3D* > candPts3D2;
			vector< AVector3D* > pts3D1one;//results of one-view-case-ransac
			vector< AVector3D* > pts3D2one;
			vector< AVector3D* > pts3D1;//results of n-view-case-ransac
			vector< AVector3D* > pts3D2;
			vector< AVector3D* > *pts3D1best = &candPts3D1;//points to one- or n-view-case
			vector< AVector3D* > *pts3D2best = &candPts3D2;

			vector<vector<AVector2D *> > feats1View ( inAandB.size() );
			vector<vector<AVector2D *> > featsNView ( inAandB.size() );
			vector<vector<AVector2D *> > *bestFeats = &featsNView;

			if ( miniRepro || usePoints3D )
			{
				getCommonPoints ( set1, set2, inAandB, candPts3D1, candPts3D2 );
				DBGOUT ( DBGNORMAL, "Merger: 3D points in common: %zd\n", candPts3D1.size() );

				if ( candPts3D1.size() < 1 ) miniRepro = false;

				if ( usePoints3D )
				{
					if ( candPts3D1.size() == 0 )
					{
						DBGOUT ( DBGALL, "no common points in %zd overlapping views\n", inAandB.size() );
						return false;
					}

					assert ( candPts3D1.size() == candPts3D2.size() );

					//ransac filtering for general case
					performRANSAC ( cams1, cams2, candPts3D1, candPts3D2, pts3D1, pts3D2, &set1, &set2 );

					pts3D1best = &pts3D1;

					pts3D2best = &pts3D2;
				}
			}

			H = calculateHomographyFromCamerasAndPoints ( cams1, cams2, pts3D1, pts3D2 );

			if ( miniRepro )
			{
				//remeber: bestFeats defaults to featsNView
				for ( uint i = 0; i < inAandB.size(); ++i )
				{
					DBGOUT ( DBGNORMAL, "resizing featsNView[%d] to %zd\n", i, pts3D1best->size() );
					featsNView[i].resize ( pts3D1best->size() );
				}

				for ( uint p = 0; p < pts3D1best->size(); ++p )
				{
					for ( uint i = 0; i < inAandB.size(); ++i )
					{
						ViewBase *v = inAandB[i];
						AVector *pt3D = ( *pts3D1best ) [p];
						FeaturePoint *fp = set1.getFeaturePoint ( pt3D, v );
						featsNView[i][p] = fp;

						if ( !fp )
						{
							DBGOUT ( DBGNORMAL, "Merger: No Feat for 3D Point %p in View %p\n", pt3D, v );
						}
					}
				}
			}

#ifndef NDEBUG
			//DEBUG output
			DBGOUT ( DBGNORMAL, "Merger: before mini" );

			testHomography ( cams1, cams2, *pts3D1best, *pts3D2best, H );

			// 	pressKey();
#endif

			//remember: bestFeats defaults to featsNView
			if ( miniRepro )
			{
				if ( result.getNumPointCorrs() > 0 )
				{
					DBGOUT ( DBGNORMAL, "Merger: result has %d corrs\n", result.getNumPointCorrs() );
					minimizeReprojectionError ( *pts3D1best, *pts3D2best, cams1, cams2, *bestFeats, H );
				}
				else
				{
					DBGOUT ( DBGCRITICAL, "Merger: minimizeReprojectionError impossible: NView lacks Corrs\n" );
				}

#ifndef NDEBUG
				//DEBUG output
				DBGOUT ( DBGNORMAL, "Merger: after mini" );

				testHomography ( cams1, cams2, *pts3D1best, *pts3D2best, H );

				// 	pressKey();
#endif
			}

			//     removeOutliers(set1, *pts3D1best, inAandB);
			//     removeOutliers(set2, *pts3D2best, inAandB);

			//transform cameras of set1 and then copy them, too
			vector<ViewBase *>::iterator it1 = inA.begin();

			DBGOUT ( DBGNORMAL, "Merger: adding transformed camera\n" );

			for ( ; it1 != inA.end(); ++it1 )
			{
				Camera *camA = set1.getCamera ( *it1 );
				Camera *camB = result.getCamera ( *it1, true );//create, if not available
				matMul2 ( camA->getProjMat(), H, transformedCam );
				camB->getProjMat() = transformedCam;
			}

			//copy the cameras of set2 to the result
			DBGOUT ( DBGNORMAL, "Merger: copying %d cameras\n", set2.getNumCameras() );

			result.addCameras ( set2 );

			feats1View.clear();

			featsNView.clear();

			return true;
		}

		/** _________________ MERGESTEP END ____________________ */

		void Merger::removeOutliers( NView &nview, vector< AVector3D* > &inlier3D, vector<ViewBase *> &inAandB )
		{
			NView::Point3DIterator pit = nview.getPoints3DBegin();
			for ( ; pit != nview.getPoints3DEnd(); ++pit )
			{
				bool inlier = false;
				vector< AVector3D* >::iterator it = inlier3D.begin();
				for ( ; it != inlier3D.end(); ++it )
				{
					if ( *it == *pit )
					{
						inlier = true;
						break;
					}
				}
				if ( !inlier )
				{
					for ( uint i = 0; i < inAandB.size(); ++i )
					{
						FeaturePoint *p = nview.getFeaturePoint( *pit, inAandB[i] );
						nview.removeFeaturePoint( p );
					}
				}
			}
		}

		void Merger::performRANSAC( vector<AMatrix *> &cams1, vector<AMatrix *> &cams2,
		                            vector< AVector3D* > &candPts3D1, vector< AVector3D* > &candPts3D2,
		                            vector< AVector3D* > &pts3D1, vector< AVector3D* > &pts3D2,
                                            NView * nv1, NView * nv2
		                          )
		{
			const double fraction = 0.5;

			if( assumeMetric )
			{
				if( useCameras )
				{
					FilterRANSACCP_PC ransac_client;
					ransac_client.init(candPts3D1, candPts3D2, *cams1[0], *cams2[0], nv1, nv2);
					ransac_client.setFractionThresh( fraction );

					RansacParams params(1000000);
					PROSAC prosac;
					prosac( &ransac_client, params );

					pts3D1.reserve(candPts3D1.size());
					pts3D2.reserve(candPts3D1.size());
					for ( FilterRANSACCP_PC::iterator it = ransac_client.begin(); it != ransac_client.end(); ++it )
					{
						uint where = *it;
						pts3D1.push_back(ransac_client.getInlier1(where));
						pts3D2.push_back(ransac_client.getInlier2(where));
					}
				}
			}
			else
			{
				int numSamples = useCameras ? 3 : 6;

				RANSACClientMergerCP rClient;
				rClient.setPoints3D( &candPts3D1, &candPts3D2 );
				if ( useCameras ) rClient.setCameras( &cams1, &cams2 );
				rClient.setFractionThresh( fraction );
//				rClient.setOneViewCase( specialOne );

				vector<int> inliers( candPts3D1.size() );
				fill( inliers.begin(), inliers.end(), 1 );

				RANSAC ransac( 1, numSamples, 50 );
				ransac.setClient( &rClient );
				ransac( inliers );

				//copy inliers
				pts3D1.reserve( candPts3D1.size() );
				pts3D2.reserve( candPts3D2.size() );
				for ( uint p = 0; p < candPts3D1.size(); ++p )
				{
					if ( inliers[p] )
					{
						pts3D1.push_back( candPts3D1[p] );
						pts3D2.push_back( candPts3D2[p] );
					}
				}
			}
		}

		AMatrix Merger::calculateHomographyFromCamerasAndPoints( vector<AMatrix *> &cams1, vector<AMatrix *> &cams2,
		                vector< AVector3D* > &pts3D1, vector< AVector3D* > &pts3D2 )
		{
			assert( cams1.size() == cams2.size() );
			assert( pts3D1.size() == pts3D2.size() );
			int ncams = cams1.size();
			int npoints = pts3D1.size();
			assert( ncams * 11 + npoints * 3 >= 16 );//make sure there is enough information

			const int columns = 16;
			AMatrix eq( 11 * ncams + 3 * npoints, columns );
			SVDStoraged store16( eq.rows(), 16, SVD_VT );
			AMatrix &V_T = store16.VT;


			eq = 0;
			int eqRow = 0;

//			for ( uint i = 0; i < cams1.size(); ++i )
//			{
//				printf( "P1[%d]    |   P2[%d]:\n", i, i );
//				printMats( *cams1[i], *cams2[i] );
//			}
//			pressKey();

			for ( int c = 0; c < ncams; ++c )
			{
//				normalizeExponents( cams1[c]->begin(), cams1[c]->end() );
//				normalizeExponents( cams2[c]->begin(), cams2[c]->end() );

				int pivotRow = 1;
				int pivotCol = 1;
				int minExp = INT_MAX;//minimum exponent found so far
				//find the optimal element to eliminate the homogeneous factor
				//the pivot is the element in the second cam that is closest to 1.0
				for ( int row = 0; row < 3; ++row )
				{
					for ( int col = 0; col < 4; ++col )
					{
						int exp = abs( getBinaryExponent( ( *cams2[c] )( row, col ) ) );
						if ( exp < minExp )
						{
							pivotRow = row;
							pivotCol = col;
							minExp = exp;
						}
					}
				}

				double pivotVal = ( *cams2[c] )( pivotRow, pivotCol );
				DBGOUT( DBGNORMAL, "Merger: pivot value is %f\n", pivotVal );

				//iterate over cam two elements
				for ( int col = 0; col < 4; ++col )
				{
					for ( int row = 0; row < 3; ++row )
					{
						if ( row == pivotRow && col == pivotCol ) continue;

						//col is also the column of H
						//row is the row of Cam1

						double cam2Val = ( *cams2[c] )( row, col );

						for ( int i = 0; i < 4; ++i )
						{
							eq( eqRow, 4 * col + i ) = pivotVal * ( *cams1[c] )( row, i );
							eq( eqRow, 4 * pivotCol + i ) -= cam2Val * ( *cams1[c] )( pivotRow, i );
						}

						++eqRow;
					}
				}


//				int cRow;
//				for ( cRow = 0; cRow < 3; ++cRow )
//				{
//					for ( int k = 0; k < 4; ++k )
//					{
//						if ( cRow == pivotRow && k == pivotCol ) continue;
//
//						for ( int j = 0; j < 4; ++j )
//						{
////                     printf("pos(%2d): +(%2d, %2d)(%2d, %2d)\n", 4 * k + j, pivotRow, pivotCol, cRow, j);
////                     printf("pos(%2d): -(%2d, %2d)(%2d, %2d)\n", k + 4 * pivotCol, pivotRow, pivotCol, cRow, j);
//							eq( eqRow, 4 * k + j ) = pivotVal * ( *cams1[c] )( cRow, j );
//							eq( eqRow, 4 * pivotCol + j ) -= ( *cams1[c] )( pivotRow, j ) * ( *cams2[c] )( cRow, k );
//						}
//
//						++eqRow;
//					}
//				}
			}

//			for ( uint i = 0; i < cams1.size(); ++i )
//			{
//				printf( "P1[%d]    |   P2[%d]:\n", i, i );
//				printMats( *cams1[i], *cams2[i] );
//			}
//			pressKey();

//			printf( "CameraEquation\n" );
//			printMat( eq );

			//bild the system of equations modelling the points
			for ( uint i = 0; i < pts3D1.size(); ++i )
			{
				AVector3D &p1 = ( *pts3D1[i] );
				AVector3D &p2 = ( *pts3D2[i] );
				normalizeExponents( p1.begin(), p1.end() );
				normalizeExponents( p2.begin(), p2.end() );

				int pivotRow = 3;
				double minDiffTo1 = DBL_MAX;
				//find the optimal element to eliminate the homogeneous factor
				for ( int row = 0; row < 4; ++row )
				{
					double diff = fabs( 1.0 - p1( row ) );
					if ( diff < minDiffTo1 )
					{
						pivotRow = row;
						minDiffTo1 = diff;
					}
				}
				double pivotVal = p1( pivotRow );

				for ( int j = 0; j < pivotRow; ++j )
				{
					for ( int k = 0; k < 4; ++k )
					{
						eq( eqRow, k * 4 + j ) =  pivotVal * p2( k );
						eq( eqRow, k * 4 + pivotRow ) =  -p1( j ) * p2( k );
					}
					++eqRow;
				}
				for ( int j = pivotRow + 1; j < 4; ++j )
				{
					for ( int k = 0; k < 4; ++k )
					{
						eq( eqRow, k * 4 + j ) =  pivotVal * p2( k );
						eq( eqRow, k * 4 + pivotRow ) =  -p1( j ) * p2( k );
					}
					++eqRow;
				}
			}

			svdVT( eq, store16 );

//			printf( "sing:\n" );
//			printMat( sing );

			//copy last row of V_T into H
			AMatrix H( 4, 4 );
			int idx = 0;
			for ( unsigned col = 0; col < H.cols(); ++col )
				for ( unsigned row = 0; row < H.rows(); ++row )
				{
					H( row, col ) = V_T( columns - 1, idx++ );
				}

			//to prevent left-handed cams; it's a guess
			AMatrix sub = H( Range( 0, 2 ), Range( 0, 2 ) );
			if ( det3( sub ) < 0 )
			{
				H *= -1;
			}

			return H;
		}

		void Merger::minimizeReprojectionError( vector< AVector3D* > &p3D1, vector< AVector3D* > &p3D2,
		                                        vector< AMatrix * > &cams1, vector< AMatrix * > &cams2,
		                                        vector< vector< AVector2D *> > &feats, AMatrix &H )
		{
			assert( feats.size() > 0 );
			assert( feats[0].size() > 0 );
			assert( p3D1.size() > 0 );
			assert( p3D2.size() > 0 );
			assert( cams1.size() > 0 );
			assert( cams2.size() > 0 );

			AVector Hvec( 16 );
			uint goalSize = 2 * feats.size() * feats[0].size();
			assert( goalSize == 2 * p3D1.size() * cams1.size() );

			int idx = 0;

			//copy H into vectorized form
			for ( int row = 0; row < 4; ++row )
				for ( int col = 0; col < 4; ++col )
				{
					Hvec( idx++ ) =  H( row, col );
				}

			MergerHMinFunc minFunc( &feats );
			minFunc.set3DPoints( &p3D1, &p3D2 );
			minFunc.setCameras( &cams1, &cams2 );
			LMFunctionJacobi2<MergerHMinFunc> jacobi( minFunc, 16, goalSize );
			LM2<MergerHMinFunc, LMFunctionJacobi2<MergerHMinFunc>, ZeroErrorFunc2<double>, InvCovarUnused<double>::type > lm;

			minFunc.useBoth( false );
			lm.solve( Hvec, minFunc, goalSize, jacobi );

			DBGOUT( DBGNORMAL, "Merger: Nonlinear merging takes %d iterations (residual: %.20f - %.20f)\n", lm.getLastIterations(), lm.getInitialError(), lm.getLastError() );

			idx = 0;
			for ( int row = 0; row < 4; ++row )
				for ( int col = 0; col < 4; ++col )
				{
					H( row, col ) = Hvec( idx++ );
				}
		}

		double Merger::getErrorOfHomography( vector<AMatrix *> &cams1, vector<AMatrix *> &cams2,
		                                     vector< AVector3D* > &pts3D1, vector< AVector3D* > &pts3D2,
		                                     vector< vector< AVector2D *> > &feats,
		                                     AMatrix &H )
		{
			double error = 0;
			AMatrix Hinv( 4, 4 );
			AMatrix PH( 3, 4 );
			AVector2D vector3( 3 );

			Hinv = H;
			invert( Hinv );

			for ( uint p = 0; p < pts3D2.size(); ++p )
			{
				for ( uint c = 0; c < cams1.size(); ++c )
				{
					double fz = ( *feats[c][p] )( 2 );
					assert( fz != 0 );
					double fx = ( *feats[c][p] )( 0 ) / fz;
					double fy = ( *feats[c][p] )( 1 ) / fz;

					//error of PHX' <-> xij

					matMul2( *cams1[c], H, PH );
					matMul2( PH, *pts3D2[p],vector3 );

					double z = vector3( 2 );
					assert( z != 0 );

					double dx1 = vector3( 0 ) / z - fx;
					dx1 *= dx1;
					double dy1 = vector3( 1 ) / z - fy;
					dy1 *= dy1;

					//error of P'H1X <-> xij

					matMul2( *cams2[c], Hinv, PH );
					matMul2( PH, *pts3D1[p], vector3 );

					z = vector3( 2 );
					assert( z != 0 );

					double dx2 = vector3( 0 ) / z - fx;
					dx2 *= dx2;
					double dy2 = vector3( 1 ) / z - fy;
					dy2 *= dy2;

					error += dx1 + dy1 + dx2 + dy2;
				}
			}

			return error;
		}

		void Merger::testHomography( vector<AMatrix *> &cams1, vector<AMatrix *> &cams2,
		                             vector< AVector3D* > &pts3D1, vector< AVector3D* > &pts3D2,
		                             AMatrix &H )
		{
			printf( "Merger: printing result of Homography:\n" );
			printMat( DBGNORMAL, H );

			double scale, variance;
			AMatrix testCam( 3, 4 );
			for ( uint i = 0; i < cams1.size(); ++i )
			{
				DBGOUT( DBGNORMAL, "orig Cam was oriented: %s\n", isRightHand( *cams1[i] ) ? "true" : "false" );
				matMul2( *cams1[i], H, testCam );
				DBGOUT( DBGNORMAL, "Merger: P1[%d]H    |   P2[%d]:\n", i, i );
				printMats( testCam, *cams2[i] );
				scaleEqualityTest( testCam, *cams2[i], scale, variance );
				DBGOUT( DBGNORMAL, "Merger: scale: %f, variance: %f\n", scale, variance );
				if(!isRightHand( testCam ))
				{
					DBGOUT( DBGCRITICAL, "Merger::testHomography: Camera is left-handed! Its values are:\n" );
					printMat( DBGCRITICAL, testCam );
				}
			}

//			for ( uint p = 0; p < pts3D1.size(); ++p )
//			{
//				AVector3D &pt1 = *pts3D1[p];
//				AVector3D pt2 = matMul( H, *pts3D2[p] );
//				printMats( pt1, pt2, 8, stdout, 2 );
//				printf( "\n" );
//			}
		}



	}
}//namespace
