

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_NVIEWMERGER_H
#define ADORE_NVIEWMERGER_H

#include <adrcore/types/nview.h>
#include <adrcore/types/nviewmap.h>
#include <adrcore/viewmatch/viewknitter.h>
#include <adrcore/nview/point3dcalculator.h>

namespace adore
{
	namespace core
	{


		/**
		 * This class merges NViews hierarchically.
		 * It begins with the NViewSets and builds up a tree from the leafs up to the root.
		 * After each merging, an optional bundle adjustment is carried out.
		 *
		 *  
		*/
		class Merger
		{
				typedef ViewKnitter::VTuple VTuple;
				typedef ViewKnitter::Tuple Tuple;

			public:

				enum MODE
				{
					LINEAR_CLOSED,
					LINEAR_OPEN
				};

				Merger();

				~Merger();

				bool merge ( NViewMap &nviewmap, const ViewKnitter &knit, const bool delSourceNViews = false );
				bool merge( NView &set1, NView &set2, NView &result, vector<ViewBase*>& inA, vector<ViewBase*>& inAandB );

				bool merge ( NView &set1, NView &set2, NView &result, Tuple tup );

				/**
				 * corrs and views only, very basic
				 */
				bool merge ( NView &set1, NView &set2, NView &result );

				/**
				 * arbitrary overlap, linear solution using cameras and points.
				 * the condition
				 * ncams * 11 + npoints * 3 >= 16
				 * should be met.
				 */
				static AMatrix calculateHomographyFromCamerasAndPoints ( vector<AMatrix *> &cams1, vector<AMatrix *> &cams2,
				                vector< AVector3D* > &pts3D1, vector< AVector3D* > &pts3D2 );

				static AMatrix calculateHomographyFromCameras( vector<AMatrix *> &cams1, vector<AMatrix *> &cams2 )
				{
					vector< AVector3D* > pts3D1, pts3D2;
					return calculateHomographyFromCamerasAndPoints ( cams1, cams2, pts3D1, pts3D2 );
				}

				void setUseCameras ( bool theValue )
				{
					useCameras = theValue;
				}


				bool getUseCameras() const
				{
					return useCameras;
				}

				void setUsePoints3D ( bool theValue )
				{
					usePoints3D = theValue;
				}

				bool getUsePoints3D() const
				{
					return usePoints3D;
				}

				void setMinimizeReproErr ( bool theValue )
				{
					minimizeReproErr = theValue;
				}

				bool getMinimizeReproErr() const
				{
					return minimizeReproErr;
				}

				void setAssumeMetric ( bool theValue )
				{
						assumeMetric = theValue;
				}

				bool getAssumeMetric() const
				{
						return assumeMetric;
				}

				/**
				 * a debug function, use it to test the outcome of calculateHomographyFromCamerasAndPoints
				 */
				static void testHomography ( vector<AMatrix *> &cams1, vector<AMatrix *> &cams2,
				                             vector< AVector3D* > &pts3D1, vector< AVector3D* > &pts3D2, AMatrix &H );

				static void testHomography ( vector<AMatrix *> &cams1, vector<AMatrix *> &cams2, AMatrix &H )
				{
					vector< AVector3D* > pts3D1, pts3D2;
					testHomography ( cams1, cams2, pts3D1, pts3D2, H );
				}

			private:

				bool useCameras;
				bool usePoints3D;
				bool minimizeReproErr;
                                bool assumeMetric;

				Point3DCalculator p3dCalc;

				bool mergeByHomography( NView &set1, NView &set2, NView &result, vector<ViewBase*>& inA, vector<ViewBase*>& inAandB );
				bool mergeByHomography ( NView &set1, NView &set2, NView &result, Tuple tup );

				bool mergeByHomography1ViewOverlap( NView &set1, NView &set2, NView &result, Tuple tup );
				bool mergeByHomography2ViewOverlap( NView &set1, NView &set2, NView &result, Tuple tup );

				/**
				 * useOr = true means that the common point must be found in just one of the views and not in all
				 * of them
				 */
				void getCommonPoints ( NView &set1, NView &set2, vector<ViewBase *> &v,
				                       vector<AVector3D *> &points1, vector<AVector3D *> &points2,
				                       bool useOr = true );

				void minimizeReprojectionError ( vector< AVector3D* > &p3D1, vector< AVector3D* > &p3D2,
				                                 vector< AMatrix * > &cams1, vector< AMatrix * > &cams2,
				                                 vector< vector< AVector2D *> > &feats, AMatrix &H );

				double getErrorOfHomography ( vector<AMatrix *> &cams1, vector<AMatrix *> &cams2,
				                              vector< AVector3D* > &pts3D1, vector< AVector3D* > &pts3D2,
				                              vector< vector< AVector2D *> > &feats,
				                              AMatrix &H );

				void performRANSAC ( vector<AMatrix *> &cams1, vector<AMatrix *> &cams2,
				                     vector< AVector3D* > &candPts3D1, vector< AVector3D* > &candPts3D2,
				                     vector< AVector3D* > &pts3D1, vector< AVector3D* > &pts3D2,
                                                     NView * nv1, NView * nv2
				                   );

				void removeOutliers ( NView &nview, vector< AVector3D* > &inlier3D, vector<ViewBase *> &inBoth );

		};

		/**
			convenience function. merges and increments the knitt step
		*/
		inline bool mergeStep( Merger &m, NViewMap &nviewmap, ViewKnitter &knit, const bool delSourceNViews = false )
		{
			bool result = m.merge( nviewmap, knit, delSourceNViews );
			knit.next();
			return result;
		}

	}
}

#endif
