

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_PARALLELMODIFIER_H
#define ADORE_PARALLELMODIFIER_H


#include <adrcore/nview/nviewoperator.h>
#include <adrcore/util/thread.h>

#define PMODDBG DBGNORMAL

namespace adore
{
	namespace core
	{
		class ParallelModifier
		{
			private:
				class MJob : public Job
				{
					private:
						NViewOperator *op;
						NView *nv;

					public:
						MJob():Job(), op( NULL ),nv( NULL ) {}

						void setNViewOperator( NViewOperator *no )
						{
							op = no;
						}

						NViewOperator *getNViewOperator()
						{
							return op;
						}

						void setNView( NView *n )
						{
							nv = n;
						}

					protected:
						//overloaded
						void execute()
						{
							assert( op );
							assert( nv );
							op->modify( *nv );
						}
				};

				ThreadPool pool;
				vector<MJob> jobs;

			public:

				/**
					pass NULL as op to initialize later
				*/
				ParallelModifier( NViewOperator *op, uint nThreads ):pool( nThreads ), jobs( nThreads )
				{
					assert( pool.getNumThreads() == nThreads );

					DBGOUT( PMODDBG, "ParallelModifier creating %d threads\n", nThreads );

					if ( op ) setOperator( op );
				}

				~ParallelModifier()
				{
					deleteClones();
				}

				void deleteClones()
				{
					for ( uint i = 1; i < pool.getNumThreads(); ++i )
					{
						delete jobs[i].getNViewOperator();
						jobs[i].setNViewOperator( NULL );
					}
				}

				void init( NViewOperator *op )
				{
					assert( op );
					deleteClones();
					setOperator( op );
				}

				template<class Iterator>
				void modify ( Iterator begin, Iterator end )
				{
					DBGOUT( PMODDBG, "parallel modify - begin (%d)\n", getNumThreads() );
					for ( ; begin != end; ++begin )
					{
						DBGOUT( PMODDBG, "parallel modify - nview %p\n", *begin );
						//blocks, if none is idle
						JobThread *t = pool.getIdle();
						NView *nv = *begin;
						MJob *j = static_cast<MJob *> ( t->getJob() );
						j->setNView( nv );
						printf( "assigning job to thread %p\n", t );
						t->doJob( j );
					}
					DBGOUT( PMODDBG, "parallel modify - joining\n" );
					pool.joinJobs();
					DBGOUT( PMODDBG, "parallel modify - end\n" );
				}

				void modifyAll ( NViewMap &map )
				{
					modify( map.getEntriesBegin(), map.getEntriesEnd() );
				}

				void modifyAll ( NViewMap &map, NViewOperator &op )
				{
					NViewOperator *pop = &op;
					init( pop );
					modifyAll ( map );
				}

				void modifyNumViews( NViewMap &map, uint numViews )
				{
					vector<NView *> nv;
					nv.reserve( map.getNumEntries() );
					NViewMap::NViewIterator it = map.getEntriesBegin();
					for ( ; it != map.getEntriesEnd(); ++it )
					{
						if ( ( *it )->getNumViews() == numViews ) nv.push_back( *it );
					}
					modify( nv.begin(), nv.end() );
				}

				void modifyNumViews( NViewMap &map, uint numViews, NViewOperator &op )
				{
					NViewOperator *pop = &op;
					init( pop );
					modifyNumViews ( map, numViews );
				}

			private:

				void setOperator( NViewOperator *op )
				{
					assert( op );
					jobs[0].setNViewOperator( op );
					pool.getThread( 0 )->setJob( &jobs[0] );
					for ( uint i = 1; i < pool.getNumThreads(); ++i )
					{
						jobs[i].setNViewOperator( op->clone() );
						pool.getThread( i )->setJob( &jobs[i] );
					}
				}
		};

	}
}//ns

#undef PMODDBG

#endif
