

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_NVIEWPROJECTIVITYCALCULATOR_H
#define ADORE_NVIEWPROJECTIVITYCALCULATOR_H

#include <vector>
#include <adrcore/util/debug.h>
#include <adrcore/types/camera.h>
#include <adrcore/types/rect.h>
#include <adrcore/types/view.h>
#include <adrcore/types/pointcorr.h>
#include <adrcore/util/solvesvd.h>

namespace adore
{
	namespace core
	{

		using namespace std;

		typedef AVector PointT;

		/**
		 * calculates a projectivity from point correspondences, such that
		 * x' = Hx
		 *  
		*/
		class ProjectivityCalculator
		{
			private:
				SVDStoraged store;
				AMatrix leftSide;
				AMatrix H;
			public:

				/**
				 * This Constructor intializes everything you will need to calculate a projectivity.
				 */
				ProjectivityCalculator():store( 0, 9, SVD_VT ), H( 3, 3 )
				{
				}

				/**
				 * Calculates the projectivity
				 */
				template <class CorrIterator, class ViewIterator >
				int calculateProjectivity ( CorrIterator it, int size, ViewIterator vit )
				{
					ViewBase *v1 = *vit; ++vit;
					ViewBase *v2 = *vit; --vit;//vit is used later
					CorrIterator cit = it;

					int row = 0;
					reserveSpace( size );
					for ( int i = 0; i < size; ++i, ++cit )
					{
						PointCorr *corr = *cit;
						if ( !corr ) continue;
						FeaturePoint *fp1 = corr->getFeature ( v1 );
						if ( !fp1 ) continue;
						FeaturePoint *fp2 = corr->getFeature ( v2 );
						if ( !fp2 ) continue;

						assert( ( *fp1 ) ( 2 ) == 1 );
						assert( ( *fp2 ) ( 2 ) == 1 );

						row = append( ( *fp1 ) ( 0 ), ( *fp1 ) ( 1 ), ( *fp2 ) ( 0 ), ( *fp2 ) ( 1 ), row );
					}
					assert( row > 7 );
					truncateSpace( row );

					return solve( H );
				}

				const AMatrix& getH() const
				{
					return H;
				}

				void reserveSpace( unsigned numCorrs )
				{
					numCorrs <<= 1;
					if ( leftSide.rows() != numCorrs ) leftSide.resize( numCorrs, 9 );
				}

				void truncateSpace( int nextRowToUse )
				{
					leftSide.resizeAndPreserve( nextRowToUse, 9 );
				}

				/**
				 * @return the row that may be used next
				 */
				int append( double x1, double y1, double x2, double y2, int nextRowToUse )
				{
					leftSide( nextRowToUse, 0 ) = -x1;
					leftSide( nextRowToUse, 1 ) = -y1;
					leftSide( nextRowToUse, 2 ) = -1;
					leftSide( nextRowToUse, 3 ) = 0;
					leftSide( nextRowToUse, 4 ) = 0;
					leftSide( nextRowToUse, 5 ) = 0;
					leftSide( nextRowToUse, 6 ) = x2 * x1;
					leftSide( nextRowToUse, 7 ) = x2 * y1;
					leftSide( nextRowToUse, 8 ) = x2;
					++nextRowToUse;
					leftSide( nextRowToUse, 0 ) = 0;
					leftSide( nextRowToUse, 1 ) = 0;
					leftSide( nextRowToUse, 2 ) = 0;
					leftSide( nextRowToUse, 3 ) = -x1;
					leftSide( nextRowToUse, 4 ) = -y1;
					leftSide( nextRowToUse, 5 ) = -1;
					leftSide( nextRowToUse, 6 ) = y2 * x1;
					leftSide( nextRowToUse, 7 ) = y2 * y1;
					leftSide( nextRowToUse, 8 ) = y2;
					++nextRowToUse;
					return nextRowToUse;
				}

				/**
				 * @return number of solutions
				 */
				int solve( AMatrix &H )
				{
					AVector &singulars = store.D;
					singulars.resize( min ( leftSide.rows(), leftSide.cols() ) );
					AMatrix &V_T = store.VT;
					store.changeRowsWorkspaceOnly( leftSide.rows() );
					store.repairD();
					int retval = svdVT ( leftSide, store );
					assert_msg ( retval == 0, "dgesvd returns %d in projectivitycalculator\n", retval ); ADORE_UNUSED ( retval );

					H ( 0, 0 ) = V_T ( 8, 0 );
					H ( 0, 1 ) = V_T ( 8, 1 );
					H ( 0, 2 ) = V_T ( 8, 2 );
					H ( 1, 0 ) = V_T ( 8, 3 );
					H ( 1, 1 ) = V_T ( 8, 4 );
					H ( 1, 2 ) = V_T ( 8, 5 );
					H ( 2, 0 ) = V_T ( 8, 6 );
					H ( 2, 1 ) = V_T ( 8, 7 );
					H ( 2, 2 ) = V_T ( 8, 8 );

					return 1;
				}
		};

	}
}

#endif
