

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_NVIEWPROJECTIVITYCALCULATORFMAT_H
#define ADORE_NVIEWPROJECTIVITYCALCULATORFMAT_H

#include <vector>
#include <adrcore/util/debug.h>
#include <adrcore/types/camera.h>
#include <adrcore/types/rect.h>
#include <adrcore/types/view.h>
#include <adrcore/types/pointcorr.h>
#include <adrcore/util/solvesvd.h>

namespace adore
{
	namespace core
	{

		using namespace std;

		typedef AVector PointT;

		/**
		 * calculates a projectivity from point correspondences, such that
		 * x' = Hx
		 * while the result has to comply with given fundamental matrix
		 *  
		*/
		class ProjectivityCalculatorFMat
		{
			private:
				SVDStoraged store;
				AMatrix leftSide;
				AMatrix H;

				AMatrix constraintMat;//the mat to get the span mat
				AMatrix spanMat;//the homographies that span the solution space
				SVDStoraged spanStore;
				AMatrix solveThis;

//				AMatrix debugFMat;
			public:

				/**
				 * This Constructor intializes everything you will need to calculate a projectivity.
				 */
				ProjectivityCalculatorFMat():store( 0, 4, SVD_VT ), H( 3, 3 ), constraintMat( 5, 9 ), spanMat( 9, 4 ), spanStore( constraintMat, SVD_VT ), solveThis( 0, 4 )/*, debugFMat(3, 3)*/
				{
				}

				/**
				 * Calculates the projectivity
				 */
				template <class CorrIterator, class ViewIterator >
				int calculateProjectivity ( CorrIterator it, int size, ViewIterator vit )
				{
					ViewBase *v1 = *vit; ++vit;
					ViewBase *v2 = *vit; --vit;//vit is used later
					CorrIterator cit = it;

					int row = 0;
					reserveSpace( size );
					for ( int i = 0; i < size; ++i, ++cit )
					{
						PointCorr *corr = *cit;
						if ( !corr ) continue;
						FeaturePoint *fp1 = corr->getFeature ( v1 );
						if ( !fp1 ) continue;
						FeaturePoint *fp2 = corr->getFeature ( v2 );
						if ( !fp2 ) continue;

						assert( ( *fp1 ) ( 2 ) == 1 );
						assert( ( *fp2 ) ( 2 ) == 1 );

						row = append( ( *fp1 ) ( 0 ), ( *fp1 ) ( 1 ), ( *fp2 ) ( 0 ), ( *fp2 ) ( 1 ), row );
					}
					assert( row > 3 );
					truncateSpace( row );

					return solve( H );
				}

				AMatrix & getH() { return H; }
				const AMatrix & getH() const { return H; }

				void reserveSpace( int numCorrs )
				{
					numCorrs <<= 1;
					if ( (int)leftSide.rows() != numCorrs )
					{
						leftSide.resize( numCorrs, 9 );
						solveThis.resize( numCorrs, 4 );
					}
				}

				void truncateSpace( int nextRowToUse )
				{
					leftSide.resizeAndPreserve( nextRowToUse, 9 );
					solveThis.resize( nextRowToUse, 4 );
				}

				/**
				 * @return the row that may be used next
				 */
				int append( double x1, double y1, double x2, double y2, int nextRowToUse )
				{
					leftSide( nextRowToUse, 0 ) = -x1;
					leftSide( nextRowToUse, 1 ) = -y1;
					leftSide( nextRowToUse, 2 ) = -1;
					leftSide( nextRowToUse, 3 ) = 0;
					leftSide( nextRowToUse, 4 ) = 0;
					leftSide( nextRowToUse, 5 ) = 0;
					leftSide( nextRowToUse, 6 ) = x2 * x1;
					leftSide( nextRowToUse, 7 ) = x2 * y1;
					leftSide( nextRowToUse, 8 ) = x2;
					++nextRowToUse;
					leftSide( nextRowToUse, 0 ) = 0;
					leftSide( nextRowToUse, 1 ) = 0;
					leftSide( nextRowToUse, 2 ) = 0;
					leftSide( nextRowToUse, 3 ) = -x1;
					leftSide( nextRowToUse, 4 ) = -y1;
					leftSide( nextRowToUse, 5 ) = -1;
					leftSide( nextRowToUse, 6 ) = y2 * x1;
					leftSide( nextRowToUse, 7 ) = y2 * y1;
					leftSide( nextRowToUse, 8 ) = y2;
					++nextRowToUse;
					return nextRowToUse;
				}

				/**
				 * @return number of solutions
				 */
				int solve( AMatrix &H )
				{
					matMul2( leftSide, spanMat, solveThis );
					assert( solveThis.cols() == 4 );

//					DBGOUT( DBGNORMAL, "rank of solveThis: %d\n", getRank( solveThis ) );

					AVector &singulars = store.D;
					singulars.resize( min ( solveThis.rows(), solveThis.cols() ) );
					AMatrix &V_T = store.VT;
					store.changeRowsWorkspaceOnly( solveThis.rows() );
					store.repairD();
					int retval = svdVT ( solveThis, store );
					assert_msg ( retval == 0, "dgesvd returns %d in projectivitycalculator\n", retval ); ADORE_UNUSED ( retval );

//					printMat( singulars );
//					printMat( V_T );

					//now construct H as a linear combination of spanMat's columns
					H = 0;
					for ( int i = 0; i < 4; ++i )
					{
						double factor = V_T ( 3, i );
						H ( 0, 0 ) += factor * spanMat( 0, i );
						H ( 0, 1 ) += factor * spanMat( 1, i );
						H ( 0, 2 ) += factor * spanMat( 2, i );
						H ( 1, 0 ) += factor * spanMat( 3, i );
						H ( 1, 1 ) += factor * spanMat( 4, i );
						H ( 1, 2 ) += factor * spanMat( 5, i );
						H ( 2, 0 ) += factor * spanMat( 6, i );
						H ( 2, 1 ) += factor * spanMat( 7, i );
						H ( 2, 2 ) += factor * spanMat( 8, i );
					}
//					printf( "skewTest (%f, %f, %f, %f):\n", V_T ( 3, 0 ), V_T ( 3, 1 ), V_T ( 3, 2 ), V_T ( 3, 3 ) );
//					printMat( debugFMat );
//					skewTest( H, debugFMat );
//					pressKey();

					return 1;
				}

				void setFMatrix( const AMatrix &f )
				{
//					printf( "Singular Values of FMat:\n" );
//					printSingularValues( f );

					AMatrix &C = constraintMat;
					C =   f( 0, 0 ),       0,       0, f( 1, 0 ),       0,       0, f( 2, 0 ),       0,       0,
					      0, f( 0, 1 ),       0,       0, f( 1, 1 ),       0,       0, f( 2, 1 ),       0,
					      0,       0, f( 0, 2 ),       0,       0, f( 1, 2 ),       0,       0, f( 2, 2 ),
					      f( 0, 1 ), f( 0, 0 ),       0, f( 1, 1 ), f( 1, 0 ),       0, f( 2, 1 ), f( 2, 0 ),       0,
					      0, f( 0, 2 ), f( 0, 1 ),       0, f( 1, 2 ), f( 1, 1 ),       0, f( 2, 2 ), f( 2, 1 );

					AMatrix &V_T = spanStore.VT;
					int retval = svdVT ( constraintMat, spanStore );
					assert_msg ( retval == 0, "dgesvd returns %d in projectivitycalculatorfmat\n", retval ); ADORE_UNUSED ( retval );

//					printf( "Singular Values of constraintMat:\n" );
//					printMat( spanStore.D );
//					printMat( DBGNORMAL, V_T );

					transInPlace( V_T );
					spanMat = V_T( Range::all(), Range( 5, 8 ) );

//					printMat( DBGNORMAL, spanMat );

//					printMat( f );
//					spanMatTest( f );
//					debugFMat = f;
				}

			private:

				void spanMatTest( const AMatrix & fmat )
				{
					AMatrix H( 3, 3 );
					for ( int i = 0; i < 4; ++i )
					{
						H ( 0, 0 ) = spanMat( 0, i );
						H ( 0, 1 ) = spanMat( 1, i );
						H ( 0, 2 ) = spanMat( 2, i );
						H ( 1, 0 ) = spanMat( 3, i );
						H ( 1, 1 ) = spanMat( 4, i );
						H ( 1, 2 ) = spanMat( 5, i );
						H ( 2, 0 ) = spanMat( 6, i );
						H ( 2, 1 ) = spanMat( 7, i );
						H ( 2, 2 ) = spanMat( 8, i );
						printf( "skewTest %d:\n", i );
						skewTest( H, fmat );
					}
				}

				void skewTest( const AMatrix &H, const AMatrix &fmat )
				{
					AMatrix result( 3, 3 );
					matMul2_AT( H, fmat, result );
					printMats( H, result );
				}



		};

	}
}

#endif
