

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Sergey Cheremukhin


#include "relativeorientationcalculator.h"

#include <cassert>

namespace adore {
namespace core {

bool RelativeOrientationCalculator::calculate(const PointColl& pts1, const PointColl& pts2)
{
	assert(pts1.size() == pts2.size());

	if(pts1.size() < 3 || pts2.size() < 3) return false;

	get_centroid(pts1, m_c1);
	get_centroid(pts2, m_c2);

	get_translated_points(pts1, m_c1, m_pts1);
	get_translated_points(pts2, m_c2, m_pts2);

	if( !get_rotation(m_pts1, m_pts2, m_R) )
	{
		DBGOUT( DBGNORMAL, "RelativeOrientationCalculator::calculate: failed to find relative rotation.\n" );
		return false;
	}

	if( !get_scale(m_pts1, m_pts2, m_scale) )
	{
		DBGOUT( DBGNORMAL, "RelativeOrientationCalculator::calculate: failed to find relative scale.\n" );
		return false;
	}

	get_translation(m_c1, m_c2, m_R, m_scale, m_t);

	return true;
}

void RelativeOrientationCalculator::get_centroid(const PointColl& pts, AVector& c)
{
	double sumx(0), sumy(0), sumz(0);

	for(PointColl::const_iterator it = pts.begin(); it != pts.end(); ++it)
	{
		const AVector& pt = **it;
		sumx += pt(0);
		sumy += pt(1);
		sumz += pt(2);
	}

	const double n(pts.size());
	sumx /= n;
	sumy /= n;
	sumz /= n;

	c = sumx, sumy, sumz;
}

void RelativeOrientationCalculator::get_translated_points(const PointColl& a, const AVector& t, PointColl& b)
{
	cleanup(b);

	for(PointColl::const_iterator it = a.begin(); it != a.end(); ++it)
	{
		const AVector& pt = **it;
		AVector * ptt = new AVector(3);
		*ptt = pt - t;
		b.push_back(ptt);
	}
}

void RelativeOrientationCalculator::cleanup(PointColl& pts)
{
	for(PointColl::const_iterator it = pts.begin(); it != pts.end(); ++it)
		delete *it;

	pts.clear();
}

bool RelativeOrientationCalculator::get_rotation(const PointColl& a, const PointColl& b, AMatrix& R)
{
	get_M(a, b, m_temp33);

	svdUVT(m_temp33, m_store.store);

	const AMatrix& U = m_store.store.U;
	const AMatrix& VT = m_store.store.VT;
	const AVector& D = m_store.store.D;
	int rank = 3;
	if(D(2) <= FLT_EPSILON)
	{
		--rank;
		if(D(1) <= FLT_EPSILON)
		{
			DBGOUT( DBGNORMAL, "RelativeOrientationCalculator::get_rotation: RANK <= 1 !!!!!\n" );
			return false;
		}
	}

	matMul2_AT_BT(VT, U, R);

	if(det3(R) < 0)
	{
		if(rank == 3)
		{
			DBGOUT( DBGNORMAL, "RelativeOrientationCalculator::get_rotation: rank is 3 and det of R is negative!\n" );
			return false;
		}

		// rank == 2
		m_temp33 =
				 VT(0, 0),  VT(0, 1),  VT(0, 2),
				 VT(1, 0),  VT(1, 1),  VT(1, 2),
				-VT(2, 0), -VT(2, 1), -VT(2, 2);

		matMul2_AT_BT(m_temp33, U, R);

		if(det3(R) < 0) {
			DBGOUT( DBGNORMAL, "RelativeOrientationCalculator::get_rotation: rank is 2 and dets of all R's' are negative!\n" );
			return false;
		}
	}

	return true;
}

void RelativeOrientationCalculator::get_M(const PointColl& a, const PointColl& b, AMatrix& M)
{
	M = 0;
	for(uint i = 0; i < a.size(); ++i)
	{
		const AVector & pt1 = *a[i];
		const AVector & pt2 = *b[i];
		matMulAdd2_BT(pt2, pt1, M);
	}
}

bool RelativeOrientationCalculator::get_scale(const PointColl& a, const PointColl& b, double& scale)
{
	double asum = 0, bsum = 0;
	for(uint i = 0; i < a.size(); ++i)
	{
		const AVector& pt1 = *a[i];
		const AVector& pt2 = *b[i];
		asum += pt1(0) * pt1(0) + pt1(1) * pt1(1) + pt1(2) * pt1(2);
		bsum += pt2(0) * pt2(0) + pt2(1) * pt2(1) + pt2(2) * pt2(2);
	}

	if(asum <= FLT_EPSILON || bsum <= FLT_EPSILON)
		return false;

	scale = sqrt(asum / bsum);
	return true;
}

void RelativeOrientationCalculator::get_translation(const AVector& c1, const AVector& c2, const AMatrix& R, const double s, AVector& t)
{
	matMul2(R, c2, t, s);
	t = c1 - t;
}

} // core
} // adore
