

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_NVIEWTTENSORCALCULATOR_H
#define ADORE_NVIEWTTENSORCALCULATOR_H

#include <adrcore/nview/cameracalculator.h>
#include <adrcore/nview/fmatrixcalculator.h>
#include <adrcore/types/camera.h>
#include <adrcore/nview/camerapointduality.h>
#include <adrcore/types/aarray.h>
#include <adrcore/util/mathutil.h>
#include <adrcore/minimization/lm2.h>
#include <adrcore/minimization/lmfunctionjacobi2.h>
#include <adrcore/nview/ttalgebraicfunc.h>
//#include <adrcore/test/progressobserver.h>

namespace adore
{
	namespace core
	{

		class CameraCalculator;

		/**
			A tripplet of cameras that is the solution
			of the 6-point algorithm
		*/
		class TTCamSolution
		{
			public:
				Camera &getCam ( int idx )
				{
					return cam[idx];
				}
				AMatrix &getMat ( int idx )
				{
					return cam[idx].getProjMat();
				}
			private:
				Camera cam[3];
		};

		/**
		 * This class hold all the tools you need to compute a trifocal tensor.
		 *  
		*/
		class TTensorCalculator
		{

				friend class TTAlgebraicFunc; //to use the same matrices
				friend class TTAlgebraicFunc2; //to use the same matrices

			public:

				enum
				{
					CAMERA_P1,
					CAMERA_P2,
					FMATRIX_F21,
					FMATRIX_F31,
					FMATRIX_F32
				};

				/**
				 * Constructor.
				 */
				TTensorCalculator();

				/**
				 * Destructor.
				 */
				virtual ~TTensorCalculator();

				/**
				 * Init the internal variables by telling the TTensorCalculator how many correspondences of which type will be used.
				 * the sizes of the matrices will be adjusted and their content preserved as far as possilbe
				 * @param numPPP the number of Point-Point-Point-Correspondences
				 * @param numLLL the number of Line-Line-Line-Correspondences
				 * @param numPLL the number of Point-Line-Line-Correspondences
				 * @param numPPL the number of Point-Point-Line-Correspondences
				 */
				void init ( int numPPP, int numLLL = 0, int numPLL = 0, int numPPL = 0 );

				/**
				 * A simple DLT calculation. Make sure to build the equation matrices first (and normalize the input!).
				 */
				int calculateTrifocalTensorBaseLinear();

				/**
				 * The minimal solution. Calculate one or three Trifocal Tensors from 6 Point-Point-Point correspondences.
				 * Make sure, to set up the equation matrix first.
				 */
				int calculateTrifocalTensorBase6Point();
				int calculateTrifocalTensorBase6Point ( TTCamSolution sol[3] );

				/**
				 * Calculate a trifocal tensor from three Cameras.
				 * @param tensor the Tensor that will hold the result
				 * @param cams the three cams
				 */
				void calculateTrifocalTensorBase3CamsT ( const AMatrix &cam1, const AMatrix &cam2, const AMatrix &cam3, ATensor3 &tensor );

				/**
				 * Calculate a trifocal tensor from three Cameras.
				 * @param lineTensor the Tensor that will hold the result (as a 27-vector)
				 * @param cams the three cams
				 */

				AVector calculateTrifocalTensorBase3CamsV ( const AMatrix &cam1, const AMatrix &cam2, const AMatrix &cam3 );

				/**
				 * Calculate a trifocal tensor from two Cameras. The first camera is assumed to be the identity cam.
				 * @param tensor the Tensor that will hold the result
				 * @param cams the two cams
				 */
				AVector calculateTrifocalTensorBase2CamsV ( const AMatrix &cam1, const AMatrix &cam2 );

				ATensor3 calculateTrifocalTensorBase2CamsT ( const AMatrix &cam1, const AMatrix &cam2 );
				void calculateTrifocalTensorBase2CamsT ( const AMatrix &cam1, const AMatrix &cam2, ATensor3 &tensor );

				/**
				 * Calculate an algebraic valid Tensor. As this methos uses the linear method first, make sure to fill the
				 * equation matrix as adviced.
				 * @param doIterate Perform a LM-minimization to optimize the result (strongly recommended)
				 * @param cam1 if not NULL, the camera of the second View can be gained as a side-effect of this algo
				 * @param cam2 if not NULL, the camera of the third View can be gained as a side-effect of this algo
				 */
				int calculateTrifocalTensorBaseMinAlgebraicError ( bool doIterate = true, AMatrix *cam1 = NULL, AMatrix *cam2 = NULL );

				/**
				 * Retrieve the epipoles from a Tensor.
				 * @param tensor the Tensor
				 * @param epipole will hold the epipole.
				 * @param transpose if true, the epipole of the second view is calculated, otherwise the epipole of the third View is.
				 */
				AVector2D calculateEpipole ( const ATensor3 &tensor, bool transpose );
				void calculateEpipole ( const ATensor3 &tensor, bool transpose, AVector2D &epi );

				/**
				 * Retrieve the epipoles from a Tensor given as a 27-vector.
				 * @param lineTensor the Tensor
				 * @param epipole a CvMat that will hold the epipole.
				 * @param transpose if true, the epipole of the second view is calculated, otherwise the epipole of the third View is.
				 */
				AVector2D calculateEpipole ( const AVector &lineTensor, bool transpose );

				/**
				 * Retrieve the Cameras from a Tensor using it's epipoles.
				 * @param tensor the Tensor.
				 * @param epi1 the epipole in the second View.
				 * @param epi2 the epipole in the third View.
				 * @param which CAMERA_P1: the Camera of the second View, CAMERA_P2 the Camera of the third View.
				 * @param camera the Camera
				 */
				AMatrix calculateCameraMatrix ( const ATensor3 &tensor, const AVector2D &epi1, const AVector2D &epi2, int which );
				void calculateCameraMatrix ( const ATensor3 &tensor, const AVector2D &epi1, const AVector2D &epi2, int which, AMatrix &cam );

				/**
				 * Retrieve the Cameras from a Tensor using it's epipoles.
				 * @param lineTensor the Tensor.
				 * @param epi1 the epipole in the second View.
				 * @param epi2 the epipole in the third View.
				 * @param which CAMERA_P1: the Camera of the second View, CAMERA_P2 the Camera of the third View.
				 * @return camera the Camera
				 */
				AMatrix calculateCameraMatrix ( const AVector &colVector, const AVector2D &epi1, const AVector2D &epi2, int which );

				/**
				 * Retrieve the fundamental matrices from a Tensor using it's epipoles.
				 * @param tensor the Tensor.
				 * @param epi1 the epipole in the second View.
				 * @param epi2 the epipole in the third View.
				 * @param which FMATRIX_F21: The fundamental matrix between the first and the second View.
				 * FMATRIX_F31: The fundamental matrix between the first and the third View.
				 * FMATRIX_F32: The fundamental matrix between the second and the third View (implementation may be a little inefficient for this one).
				 * @return fmat the result is stored in this CvMat
				 */
				AMatrix calculateFundamentalMatrix ( const ATensor3 &tensor, const AVector2D &epi1, const AVector2D &epi2, int which );

				const ATensor3& getSolution ( int i )
				{
					extern ATensor3 NullTensor3;
					switch ( i )
					{
						case 0:
							return tensor1;
							break;
						case 1:
							return tensor2;
							break;
						case 2:
							return tensor3;
							break;
						default:
							return NullTensor3;
							break;
					}
					return NullTensor3;
				}

				/**
				 * Retrieve the first Tensor.
				 * @return a Tensor
				 */
				ATensor3 &getTensor1()
				{
					return tensor1;
				}

				/**
				 * Retrieve the second Tensor.
				 * This only makes sense in the minimum case (6 Point-Point-Point-Correspondences).
				 * @return a Tensor
				 */
				ATensor3 &getTensor2()
				{
					return tensor2;
				}

				/**
				 * Retrieve the third Tensor.
				 * This only makes sense in the minimum case (6 Point-Point-Point-Correspondences).
				 * @return a Tensor
				 */
				ATensor3 &getTensor3()
				{
					return tensor3;
				}

				/**
				 * Converts a 27-vector to a Tensor.
				 * @param tensor the tensor that should hold the result.
				 * @param colVector the 27-vector
				 */
				ATensor3 buildTensorFromLineTensor ( const AVector &colVector );
				void buildTensorFromLineTensor ( const AVector &colVector, ATensor3 &tensor );
				AVector buildLineTensorFromTensor ( const ATensor3 &tensor );
				void buildLineTensorFromTensor ( const ATensor3 &tensor, AVector &colVector );

				/**
				 * Adds 4 equations corresponding to the three points (p23 and p33 omitted intentionally).
				 * This build the equation for all cases but the minimum case.
				 * @param p11 coordinate 1 of point 1
				 * @param p12 coordinate 2 of point 1
				 * @param p13 coordinate 3 of point 1
				 * @param p21 coordinate 1 of point 2
				 * @param p22 coordinate 2 of point 2
				 * @param p31 coordinate 1 of point 3
				 * @param p32 coordinate 2 of point 3
				 */
				void setPPP ( int row, double p11, double p12, double p13, double p21, double p22, double p31, double p32 );

				/**
				 * Adds one correspondence for the numerically stable 6-Point algo.
				 * @param p11 coordinate 1 of point 1
				 * @param p12 coordinate 2 of point 1
				 * @param p21 coordinate 1 of point 2
				 * @param p22 coordinate 2 of point 2
				 * @param p31 coordinate 1 of point 3
				 * @param p32 coordinate 2 of point 3
				 */
				void setPPP ( int row, double p11, double p12, double p21, double p22, double p31, double p32 );

				static AVector2D getPointInThirdView ( const AVector2D &pt1, const AVector2D &pt2, const AMatrix &fmat21, const ATensor3 &ttensor );

			protected:

				vector<vector<AVector2D> > pts6Stable; //3 views, 6 point a 2 coors
				int pts6StableEqRow;
				AMatrix T[3];//these are used twice: as tensor-sub-matrices and Transformations
				AMatrix solution1;
				AMatrix solution2;
				AMatrix result1;
				AMatrix result2;
				AMatrix result3;
				AVector3D X1;
				AVector3D X2;
				AVector3D X3;
				AMatrix camSelections;
				ATensor3 tensor1;
				ATensor3 tensor2;
				ATensor3 tensor3;

				AMatrix linearEq; //the n*27-equation-matrix
				AVector lineTensor;//all 27 elements in one column
				SVDStoraged store27;
				AMatrix E;//matrix for algebraic minimization

				//algebraic constraint enforcement
				SVDStoraged store27_18;
				SVDStoraged store15;
				AVector candT15;
				AMatrix matsize_15;
				AVector2D epipole1;
				AVector2D epipole2;
				AVector epipoles; //6-vector for LM
				AVector lmGoal;
				AVector camRemains;
				SVDStoraged store3;

				//temporary space
				AVector vec3;
				AMatrix mat33;

				CameraPointDuality dual;
				CameraCalculator cam;

				int calculateTrifocalTensorBaseSixthPoint ( const AVector &X6, TTCamSolution &sol );
				void fillMatE ( AMatrix &E, const AVector2D &epi1, const AVector2D &epi2 );
				void enforceEpipolesConstraint ( AMatrix &linearEq, AMatrix &E, AVector &lineTensor, AVector *camRemains = NULL );
				AMatrix calculateCameraMatrix ( AMatrix T[3], const AVector2D &epi1, const AVector2D &epi2, int which );
				void calculateCameraMatrix ( AMatrix T[3], const AVector2D &epi1, const AVector2D &epi2, int which, AMatrix &cam );

				//debug
				void tauCheck ( ATensor3 &tensor );
				void tauCheck ( AVector &lineTensor );

			private:
				//prevent copy constructor (thread-threat)
				TTensorCalculator ( const TTensorCalculator& )
				{
					assert( false );
				}
		};

	}

}

#endif
