

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include "calibrationparamversatile.h"
#include <adrcore/util/matmul2.h>
#include <adrcore/util/mathutil.h>
#include <adrcore/util/printmat.h>

namespace adore
{

	namespace core
	{


		CalibrationParamVersatile::CalibrationParamVersatile()
				: CalibrationParamMixIn(),
				choice(),
				softConstraints(),
				constSoftConstraints(),
				biasID ( false ),
				numQParams ( 8 )
				/*K(0)*/
		{
			K.reserve ( 10 );
			ensureKSize ( 1 );
		}

		CalibrationParamVersatile::CalibrationParamVersatile ( const CalibrationParamVersatile &cv )
				: CalibrationParamMixIn(),
				choice ( cv.choice ),
				softConstraints ( cv.softConstraints ),
				constSoftConstraints ( cv.constSoftConstraints ),
				biasID ( cv.biasID ),
				numQParams ( 8 )
		{
			K.reserve ( 10 );
			ensureKSize ( 1 );
		}

		CalibrationParamVersatile::~CalibrationParamVersatile()
		{
			//for(uint i = 0; i < K.size(); ++i) delete K[i];
		}

		int CalibrationParamVersatile::getNumParams() const
		{
//			assert ( choice );
//			printf( "versatile: %d, %d, %d, %d\n", numQParams, choice.getNumViews(), choice.getNumCalibrationParams(), choice.getNumCONSTParams() );
			return numQParams +
			       choice.getNumViews() * choice.getNumCalibrationParams() +
			       choice.getNumCONSTParams();
		}

		void CalibrationParamVersatile::calculateQuadricAndConics ( const AVector &params, int paramOff, AMatrix &quad, vector<AMatrix> &conics )
		{
			calculateConics ( params, paramOff, conics );
			if ( biasID )
			{
				calculateQuadric ( params, paramOff, conics[0], quad );
			}
			else
			{
				calculateQuadric ( params, paramOff, quad );
			}
		}

		void CalibrationParamVersatile::calculateConics ( const AVector &params, int paramOff, vector<AMatrix> &conics )
		{
			ensureKSize ( conics.size() );
			paramOff += numQParams;
			for ( uint i = 0; i < conics.size(); ++i )
			{
				choice.getCalibrationMatrixFromAllParams ( params, K[i], ( int ) i, ( int ) paramOff );
				matMul2_BT ( K[i], K[i], conics[i] );
			}
		}

		void CalibrationParamVersatile::getInitializedParams ( AVector &param, int paramOff, AMatrix *initQ, vector<AMatrix> *w )
		{
			if ( biasID )
			{
				double a1 = 0.0;
				double a2 = 0.0;
				double a3 = 0.0;
				if ( initQ )
				{
					a1 = ( *initQ ) ( 0, 3 );
					a2 = ( *initQ ) ( 1, 3 );
					a3 = ( *initQ ) ( 2, 3 );
				}
                param ( paramOff++ ) =  a1;
                param ( paramOff++ ) =  a2;
                param ( paramOff++ ) =  a3;
			}
			else
			{
				double a = 1.0;
				double b = 0.0;
				double c = 0.0;
				double d = 0.0;
				double e = 1.0;
				double f = 0.0;
				double g = 0.0;
				double h = 0.0;
				if ( initQ )
				{
					a = ( *initQ ) ( 0, 0 );
					b = ( *initQ ) ( 0, 1 );
					c = ( *initQ ) ( 0, 2 );
					d = ( *initQ ) ( 0, 3 );
					e = ( *initQ ) ( 1, 1 );
					f = ( *initQ ) ( 1, 2 );
					g = ( *initQ ) ( 1, 3 );
					h = ( *initQ ) ( 2, 3 );
				}
                param ( paramOff++ ) =  a;
                param ( paramOff++ ) =  b;
                param ( paramOff++ ) =  c;
                param ( paramOff++ ) =  d;
                param ( paramOff++ ) =  e;
                param ( paramOff++ ) =  f;
                param ( paramOff++ ) =  g;
                param ( paramOff++ ) =  h;
			}

			if ( w )
			{
				ensureKSize ( w->size() );
				for ( uint i = 0; i < K.size(); ++i )
				{
					int retVal = cholesky ( ( *w ) [i], K[i] );
					if ( retVal != 0 )
					{
						DBGOUT( DBGCRITICAL, "CalibrationParamVersatile: cholesky failed for w[%d], return value was %d, w was:\n", ( int )i, retVal );
						printMat( DBGCRITICAL, ( *w ) [i] );
						DBGOUT( DBGCRITICAL, "CalibrationParamVersatile: replacing w with identity" );
						setIdentity( ( *w ) [i] );
					}
				}
			}
			else
			{
				for ( uint i = 0; i < K.size(); ++i )
				{
					setIdentity ( K[i] );
				}
			}

			int increment = choice.sameCameras() ? 0 : choice.getNumCalibrationParams();
			for ( uint i = 0; i < K.size(); ++i )
			{
				choice.getParamFromCalibrationMatrix ( K[i], param, paramOff );
				paramOff += increment;
			}
		}

		void CalibrationParamVersatile::ensureKSize ( uint size )
		{
			for ( uint i = K.size(); i < size; ++i ) K.push_back ( AMatrix ( 3, 3 ) );
		}

		int CalibrationParamVersatile::getNumSoftConstraints()
		{
			int sum = 0;
			/*if ( constSoftConstraints ) */
			sum += constSoftConstraints.getNumConstraints();
			/*if ( softConstraints ) */
			sum += softConstraints.getNumConstraints();
			return sum;
		}

		void CalibrationParamVersatile::getSoftConstraintsGoal ( AVector &goal, int offset )
		{
			if ( /*constSoftConstraints &&*/ constSoftConstraints.getNumViews() )
			{
				constSoftConstraints.getGoal ( goal, offset, 0 );
				offset += constSoftConstraints.getNumConstraints();
			}
//			if ( softConstraints )
//			{
				uint limit = MIN ( softConstraints.getNumViews(), ( uint ) K.size() );
				for ( uint i = 0; i < limit; ++i )
				{
					softConstraints.getGoal ( goal, offset, i );
					offset += constSoftConstraints.getNumConstraints ( i );
				}
//			}
		}

		void CalibrationParamVersatile::evaluateSoftConstraints ( AVector &err, int errOff )
		{
			if ( /*constSoftConstraints &&*/ constSoftConstraints.getNumViews() )
			{
				constSoftConstraints.getError ( K[0], err, errOff, 0 );
				errOff += constSoftConstraints.getNumConstraints();
			}
//			if ( softConstraints )
//			{
				uint limit = MIN ( softConstraints.getNumViews(), ( uint ) K.size() );
				for ( uint i = 0; i < limit; ++i )
				{
					softConstraints.getError ( K[i], err, errOff, i );
					errOff += constSoftConstraints.getNumConstraints ( i );
				}
//			}
		}

	}
}
