

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Sergey Cheremukhin


#include "quaternion.h"

namespace adore
{
	namespace core
	{
		static const int Q_X = 0;
		static const int Q_Y = 1;
		static const int Q_Z = 2;
		static const int Q_W = 3;

		Quaternion::Quaternion()
				:AVector( 4 )
		{
		}

		void Quaternion::fromAngleAxis( double rad, double x, double y, double z )
		{
			// Here we calculate the sin( theta / 2) once for optimization
			double result = std::sin( 0.5 * rad );
			// Calcualte the w value by cos( theta / 2 )
			operator()( Q_W ) = std::cos( 0.5 * rad );

			// Calculate the x, y and z of the quaternion
			operator()( Q_X ) = x * result;
			operator()( Q_Y ) = y * result;
			operator()( Q_Z ) = z * result;
		}

		void Quaternion::fromRotationMatrix( const AMatrix& rotmat )
		{
			double trace = rotmat( 0,0 ) + rotmat( 1,1 ) + rotmat( 2,2 );
			double mem_inv_x4;
			if ( trace>0.0 )
			{
				operator()( Q_W ) = std::sqrt( trace + 1.0 ) * 0.5;

				mem_inv_x4 = 0.25 / operator()( Q_W );
				operator()( Q_X ) = ( rotmat( 2,1 ) - rotmat( 1,2 ) ) * mem_inv_x4;
				operator()( Q_Y ) = ( rotmat( 0,2 ) - rotmat( 2,0 ) ) * mem_inv_x4;
				operator()( Q_Z ) = ( rotmat( 1,0 ) - rotmat( 0,1 ) ) * mem_inv_x4;
			}
			else
			{
				if ( rotmat( 0,0 ) > rotmat( 1,1 ) && rotmat( 0,0 ) > rotmat( 2,2 ) )
				{
					// rotmat(0,0) is the maximum diagonal term
					operator()( Q_X ) = std::sqrt( rotmat( 0,0 ) - rotmat( 1,1 ) - rotmat( 2,2 ) + 1.0 ) * 0.5;
					mem_inv_x4 = 0.25 / operator()( Q_X );
					operator()( Q_W ) = ( rotmat( 2,1 ) - rotmat( 1,2 ) ) * mem_inv_x4;
					operator()( Q_Y ) = ( rotmat( 0,1 ) + rotmat( 1,0 ) ) * mem_inv_x4;
					operator()( Q_Z ) = ( rotmat( 0,2 ) + rotmat( 2,0 ) ) * mem_inv_x4;
				}
				else if ( rotmat( 1,1 ) > rotmat( 2,2 ) )
				{
					// rotmat(1,1) is the maximum diagonal term
					operator()( Q_Y ) = std::sqrt( rotmat( 1,1 ) - rotmat( 0,0 ) - rotmat( 2,2 ) + 1.0 ) * 0.5;
					mem_inv_x4 = 0.25 / operator()( Q_Y );
					operator()( Q_W ) = ( rotmat( 0,2 ) - rotmat( 2,0 ) ) * mem_inv_x4;
					operator()( Q_X ) = ( rotmat( 0,1 ) + rotmat( 1,0 ) ) * mem_inv_x4;
					operator()( Q_Z ) = ( rotmat( 1,2 ) + rotmat( 2,1 ) ) * mem_inv_x4;
				}
				else
				{
					// rotmat(2,2) is the maximum diagonal term
					operator()( Q_Z ) = std::sqrt( rotmat( 2,2 ) - rotmat( 1,1 ) - rotmat( 0,0 ) + 1.0 ) * 0.5;
					mem_inv_x4 = 0.25 / operator()( Q_Z );
					operator()( Q_W ) = ( rotmat( 1,0 ) - rotmat( 0,1 ) ) * mem_inv_x4;
					operator()( Q_X ) = ( rotmat( 0,2 ) + rotmat( 2,0 ) ) * mem_inv_x4;
					operator()( Q_Y ) = ( rotmat( 1,2 ) + rotmat( 2,1 ) ) * mem_inv_x4;
				}
			}
		}

		AMatrix Quaternion::toRotationMatrix()
		{
			AMatrix mat( 3,3 );
			toRotationMatrix( mat );
			return mat;
		}

		void Quaternion::toRotationMatrix(AMatrix& mat)
		{
			double x = operator()( Q_X );
			double y = operator()( Q_Y );
			double z = operator()( Q_Z );
			double w = operator()( Q_W );

			// First row
			mat( 0,0 ) = 1.0 - 2.0 * ( y * y + z * z );
			mat( 0,1 ) = 2.0 * ( x * y - w * z );
			mat( 0,2 ) = 2.0 * ( x * z + w * y );

			// Second row
			mat( 1,0 ) = 2.0 * ( w * z + x * y );
			mat( 1,1 ) = 1.0 - 2.0 * ( x * x + z * z );
			mat( 1,2 ) = 2.0 * ( y * z - w * x );

			// Third row
			mat( 2,0 ) = 2.0 * ( x * z - w * y );
			mat( 2,1 ) = 2.0 * ( y * z + w * x );
			mat( 2,2 ) = 1.0 - 2.0 * ( x * x + y * y );
		}

		const Quaternion& Quaternion::operator*( const Quaternion& q )
		{
			double& x = operator()( Q_X );
			double& y = operator()( Q_Y );
			double& z = operator()( Q_Z );
			double& w = operator()( Q_W );

			double nw = w*q( Q_W ) - x*q( Q_X ) - y*q( Q_Y ) - z*q( Q_Z );
			double nx = w*q( Q_X ) + x*q( Q_W ) + y*q( Q_Z ) - z*q( Q_Y );
			double ny = w*q( Q_Y ) + y*q( Q_W ) + z*q( Q_X ) - x*q( Q_Z );
			double nz = w*q( Q_Z ) + z*q( Q_W ) + x*q( Q_Y ) - y*q( Q_X );

			x = nx;
			y = ny;
			z = nz;
			w = nw;

			return *this;
		}
	}
}
