

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include "ransac.h"
#include <adrcore/util/avgvar.h>

namespace adore
{
	namespace core
	{


		RANSAC::RANSAC( int numResults, int samples, int maxIts ):
				maxIterations( maxIts ),
				maxNumResults( numResults ),
				sampleSize( samples ),
				sample( sampleSize ),
				bestSample( sampleSize ),
				bestResult( -1 )
		{
			fill( bestSample.begin(), bestSample.end(), -1 );
		}


		RANSAC::~RANSAC()
		{
		}

		void RANSAC::operator()( vector< int > &indices )
		{
//			printf( "entering ransac\n" );
			if ( maxIterations < 1 ) return;

			uint numIndices = indices.size();
			assert( numIndices >= sampleSize );

			//initialize inliermask for randomset to prevent the choice of outliers
			int mSize = 0;
			for ( uint i = 0; i < numIndices; ++i ) if ( indices[i] == 1 ) ++mSize;
			inMask.resize( mSize );
			mSize = numIndices - mSize;
			outMask.resize( mSize );
			int oidx = 0, iidx = 0;
			for ( uint i = 0; i < numIndices; ++i )
			{
				if ( indices[i] == 1 ) inMask[iidx++] = i;
				else outMask[oidx++] = i;
			}

			//is everything okay?
			bool okay = user->checkPrerequisitsAndInit( inMask.size(), indices );
			if ( !okay )
			{
				DBGOUT( DBGNORMAL, "Prerequisits for this filter are not met. Don't filter anything\n" );
				return;
			}

			//initialize the parameters controlling the ransac interation
			int N = INT_MAX;
			int scount = 0;
			double curErr = 0.0; //current number of inliers;
			double minErr = DBL_MAX; //the minimum current error
			int mostIn = 0;
			double errRatio;
			const double prob = 0.95; //the probability for at least one sample consisting exclusivly of inliers
			const double logdivident = log( 1.0-prob );

			//everything to remember inliers and their number
			int in[maxNumResults];//the number of inliers for every result based on the sample.
			double err[maxNumResults];
			vector<int> iFlags( numIndices ); //the inlier flags after ransac
			fill( iFlags.begin(), iFlags.end(), 0 );
			vector<int> inlierFlags[maxNumResults];

			//everything to remember the distances of the samples to resolve ties via comparison of variance
			vector<double> dVals( numIndices );
			vector<double> distVals[maxNumResults];

			for ( int i = 0; i < maxNumResults; ++i )
			{
				inlierFlags[i].resize( numIndices );
				distVals[i].resize( numIndices );
			}

			int bestInlierIdx = -1; // best of maxNumResults

			//variables to manage results with the same number of inliers
			int ties[maxNumResults];
			int numTies;

			//the ransac algorithm
//			printf( "RANSAC: N: %d, scount: %d\n", N, scount );
			while ( N > scount )
			{
				createChoice( numIndices );

//				printf( "choice created\n" );

				curErr = DBL_MAX;
				int numIn = -1;
				numTies = 0;
				bestInlierIdx = -1;

				int numResults = user->applySample( sample );

				if ( numResults > 0 )
				{
					//compute the number of inliers for every result and determin the maximum number of inliers.
//					printf( "getting inliers of %d results\n", numResults );

					for ( int r = 0; r < numResults; ++r )
					{
						user->computeError( r, inlierFlags[r], distVals[r] );
						err[r] = user->getError();
						in[r] = user->getNumIn();

//						printf( "RESULT %d: err: %f, numIn: %d\n", r, user->getError(), user->getNumIn() );

						if ( err[r] < 0 )
						{
							DBGOUT( DBGALL, "tensor rejected in getNumberOfInliers\n" );
							continue;
						}

						if ( in[r] > numIn || ( in[r] == numIn && err[r] < curErr ) )
						{
							curErr = err[r];
							bestInlierIdx = r;
							numTies = 0;
							numIn = in[r];
						}
						else if ( in[r] == numIn && err[r] == curErr )
						{
							++numTies;
						}
					}

					//printf("resolving %d ties\n", numTies);
					if ( numIn >= mostIn && curErr <= minErr && numTies > 0 )
					{
						int idx = 0;
						for ( int r = 0; r < numResults; ++r )
						{
							if ( err[r] == curErr ) ties[idx++] = r;
						}
						int res = resolveTies( ties, numTies, inlierFlags, distVals );
						if ( res < 0 )
						{
							DBGOUT( DBGALL, "Tie between results could not be resolved (numIn = %f)\n", curErr );
						}
						else
						{
							curErr = err[res];
							bestInlierIdx = res;
							numIn = in[res];
						}
					}

//					DBGOUT( DBGNORMAL, "numIn: %d, mostIn: %d, iterations-max: %d\n", numIn, mostIn, N );

					if ( numIn > mostIn || ( numIn == mostIn && curErr <= minErr ) )
					{
						bool optimized = user->optimize( bestInlierIdx );
						if ( optimized )//LO-step, modifies numIn and curErr
						{
							//optimization took place
							//invalidates numIn, curErr, bestInlierIdx, inlierFlags, distVals
							bestInlierIdx = 0;
							user->computeError( 3, inlierFlags[0], distVals[0] );
//							printf( "%d <> %d\n", numIn, user->getNumIn() );
//							printf( "%f <> %f\n", curErr, user->getError() );
//							assert_msg( numIn <= user->getNumIn(), "%d > %d\n", numIn, user->getNumIn() );
//							assert_msg( curErr >= user->getError(), "%f < %f\n", curErr, user->getError() );
							curErr = user->getError();
							//after optimization, there may be fewer inliers.
							//if this happens, then we fake the number of
							//inliers to be the greater one
							numIn = MAX( numIn, user->getNumIn() );
						}

						mostIn = numIn;

						if ( !optimized && curErr == minErr ) //a tie
						{
							//tie between samples
							double avg, varNew, varOld;
							//                 printf("tie between sample, bestInlierIdx: %d, numIn: %d\n", bestInlierIdx, numIn);
							//                 printf("calculating new variance, sizes:(%d, %d)\n", distVals[bestInlierIdx].size(), inlierFlags[bestInlierIdx].size());
							getAvgVar( distVals[bestInlierIdx], avg, varNew, inlierFlags[bestInlierIdx] );
							//                 printf("calculating old variance, sizes:(%d, %d)\n", dVals.size(), iFlags.size());
							getAvgVar( dVals, avg, varOld, iFlags );

							if ( varNew < varOld )
							{
//								printf( "b: bestInlierIndex: %d\n", bestInlierIdx );
//								printf( "copy iFlags\n" );
								copy( inlierFlags[bestInlierIdx].begin(), inlierFlags[bestInlierIdx].end(), iFlags.begin() );
//								printf( "copy dVals\n" );
								copy( distVals[bestInlierIdx].begin(), distVals[bestInlierIdx].end(), dVals.begin() );
//								printf( "copy sample\n" );
								copy( sample.begin(), sample.end(), bestSample.begin() );
								bestResult = bestInlierIdx;
							}
							else if ( varNew == varOld )
							{
								DBGOUT( DBGNORMAL, "Tie between runs could not be resolved (curErr = %f, variance: %f)\n", curErr, varNew );
							}

						}
						else
						{
							minErr = curErr;
//							printf( "a: bestInlierIndex: %d\n", bestInlierIdx );
//							printf( "copy iFlags\n" );
							copy( inlierFlags[bestInlierIdx].begin(), inlierFlags[bestInlierIdx].end(), iFlags.begin() );
//							printf( "copy dVals\n" );
							copy( distVals[bestInlierIdx].begin(), distVals[bestInlierIdx].end(), dVals.begin() );
//							printf( "copy sample\n" );
							copy( sample.begin(), sample.end(), bestSample.begin() );
							bestResult = bestInlierIdx;

							//recalculate N
							DBGOUT( DBGNORMAL, "numIn: %d, curErr: %f, minErr: %f, iterations-max: %d\n", numIn, curErr, minErr, N );
							DBGOUT( DBGNORMAL, "Old N: %d (%d done) inMaskSize: %zd\n", N, scount, inMask.size() );
							errRatio = numIn / double( inMask.size() );
							if ( errRatio == 1.0 )
							{
								N = 0; //stops immediately.
							}
							else
							{
								double logdivisor = log( 1.0-intPow( errRatio, ( int )inMask.size() ) );
								if ( -logdivisor > FLT_EPSILON )
								{
									N = ( int ) ( logdivident/logdivisor );
									if ( N < 0 ) N = INT_MAX;
								}
								DBGOUT( DBGNORMAL, "New N: %d (%d done),  errRatio: %f\n", N, scount, errRatio );
							}
						}
					}

					if ( scount >= maxIterations ) //if no improvement occurrs, check for iteration limit
					{
						N = 0; //stops immediately.
					}
					++scount;
				}
				else if ( numResults == 0 )
				{
					DBGOUT( DBGNORMAL, "RANSAC: no result of %d\n", maxNumResults );
				}
				else
				{
					//something wrong, maybe points are collinear or something
					DBGOUT( DBGCRITICAL, "something wrong while computing a RANSAC sample (%d out of %zd), sample is:\n", sampleSize, inMask.size() );
#ifndef NDEBUG
					for ( uint i = 0; i < sample.size(); ++i ) printf( " %d ", sample[i] );
					printf( "\n" );
#endif
//					pressKey();
				}
			} // end of ransac algorithm.

			//now mark the inliing correspondences
			for ( uint i = 0; i < numIndices; ++i )
			{
				indices[i] = iFlags[i];
			}
			DBGOUT( DBGNORMAL, "RANSAC Filter: found error %f after %d of %d iterations\n", minErr, scount, N );

			user->ransacDone();
		}

		void RANSAC::createChoice( int size )
		{
			randSet.createChoice( sample, 0, size - 1, &outMask );

#ifndef NDEBUG
			for ( uint i = 0 ; i < sampleSize; ++i )
			{
				if ( testChoice( sample[i] ) )
				{
					printf( "Inlier \n" );
					for ( uint j = 0; j < inMask.size(); ++j )
					{
						printf( "%d  ", inMask[j] );
					}
					printf( "\n" );
					printf( "Outlier \n" );
					for ( uint j = 0; j < outMask.size(); ++j )
					{
						printf( "%d  ", outMask[j] );
					}
					printf( "\n" );
					getchar();
				}
			}
#endif
		}

		bool RANSAC::testChoice( int val )
		{
			//printf("testing Choice\n");
			int inCount = 0;
			int outCount = 0;

			for ( uint i = 0; i < inMask.size(); ++i )
			{
				if ( val == inMask[i] ) ++inCount;
			}
			for ( uint i = 0; i < outMask.size(); ++i )
			{
				if ( val == outMask[i] ) ++outCount;
			}
			if ( outCount > 0 )
			{
				printf( "Choice is an outlier!\n" );
				return true;
			}
			if ( outCount + inCount != 1 )
			{
				printf( "Choice not in or out!\n" );
				return true;
			}
			//printf("Choice tested\n");
			return false;
		}

		int RANSAC::resolveTies( int *ties, int numTies, vector<int> *inlierFlags, vector<double> *distVals )
		{
			bool notResolved = false;
			double minVar, var, avg;
			int bestIdx = ties[0];
			getAvgVar( distVals[ties[0]], avg, minVar, inlierFlags[ties[0]] );
			for ( int i = 1; i < numTies; ++i )
			{
				getAvgVar( distVals[ties[i]], avg, var, inlierFlags[ties[i]] );
				if ( var < minVar )
				{
					minVar = var;
					bestIdx = ties[i];
					notResolved = false;
				}
				else if ( var == minVar )
				{
					notResolved = true;
				}
			}
			return notResolved ? -1 : ties[bestIdx];
		}

	}
}
