

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_CORECOLLECTORS_H
#define ADORE_CORECOLLECTORS_H

#include <algorithm>
#include <vector>
#include <deque>
#include <adrcore/util/c++0x.h>
#include <adrcore/util/templatelimits.h>

namespace adore
{

	namespace core
	{

		using namespace std;

		/**
			A priority queue with a fixed amount of space
		*/
		template<class T>
		class FixedPQ
		{
			public:
				typedef T value_type;
				typedef vector<value_type> CollT;
				typedef	typename CollT::iterator iterator;
				typedef	typename CollT::const_iterator const_iterator;
			private:
				int maxN;
				CollT elems;
				std::less<T> comp;
			public:
				FixedPQ ( int n ) :maxN ( n ), elems() {elems.reserve ( maxN );}

				inline void push ( const T &e )
				{
					elems.push_back ( e );
					std::push_heap ( elems.begin(), elems.end(), comp );
				}

				inline const T& top()
				{
					assert ( !empty() );
					return elems.front();
				}

				inline void pop()
				{
					assert ( !empty() );
					std::pop_heap ( elems.begin(), elems.end(), comp );
					elems.pop_back();
				}

				inline bool empty() {return elems.empty();}

				inline void clear() {elems.clear();}
				inline iterator begin() {return elems.begin();}
				inline const_iterator begin() const {return elems.begin();}
				inline iterator end() {return elems.end();}
				inline const_iterator end() const {return elems.end();}
		};

		/**
			A comparator that may be used to sort the tuples inside the collectors
			according to their distance
		*/
		template<class T, class DistT>
		class KDTupleComparator
		{
			private:
				typedef	tuple<T*, DistT> value_type;
			public:
				inline bool operator() ( const value_type &t1, const value_type &t2 ) const
				{
					return t1.get<1>() < t2.get<1>();
				}
		};

		/**
		 * this collector does nothing
		 * this one has the essential methods a collector must define
		 */
		class FakeCollector
		{
			public:
				/**
					inserts an element into the collector
				*/
				inline void insert ( const void*, const float ) const {;}

				/**
					the pruning distance
					the branches whose hyperectanlge is further away
					than this are not visited
				*/
				inline float getDist() {return maxLimit<float>();}
		};

		extern FakeCollector DummyCollector;

		/**
			this collector collects one single best matching sample
		*/
		template<class T, class DistT>
		class SingleBest
		{
			public:
				typedef	tuple<T*, DistT> value_type;
				value_type best;
			public:
				SingleBest () : best( NULL, maxLimit<DistT>() ) {;}

				inline void insert ( T* sample, const DistT dist )
				{
					if ( dist < best.get<1>() )
					{
						best.get<0>() = sample;
						best.get<1>() = dist;
					}
				}

				inline DistT getDist() {return best.get<1>();}

				inline value_type getBest() {return best;}
		};

		/**
		 * collects anything in range
		 */
		template<class T, class DistT>
		class RangeCollector
		{
			public:
				typedef	tuple<T*, DistT> value_type;
				typedef deque<value_type> CollT;
			private:
				DistT range;
				CollT elems;
			public:
				typedef	typename CollT::iterator iterator;
				typedef	typename CollT::const_iterator const_iterator;
			public:
				RangeCollector ( const DistT r ) :range ( r ) {;}
				inline void insert ( T* sample, const DistT dist )
				{
					//printf("adding %f / %f\n", dist, range);
					if ( dist <= range )
					{
						elems.push_back ( value_type ( sample, dist ) );
					}
				}

				inline DistT getDist() {return range;}

				inline void clear() {elems.clear();}
				inline iterator begin() {return elems.begin();}
				inline const_iterator begin() const {return elems.begin();}
				inline iterator end() {return elems.end();}
				inline const_iterator end() const {return elems.end();}

				inline int size() {return elems.size();}
		};

		/**
		 * collects the n nearest neighbors
		 */
		template<class T, class DistT>
		class NearestCollector
		{
			public:
				typedef	tuple<T*, DistT> value_type;
				typedef	KDTupleComparator<T, DistT> Comparator;
			public:
				typedef vector<value_type> CollT;
				typedef	typename CollT::iterator iterator;
				typedef	typename CollT::const_iterator const_iterator;
			private:
				int maxN;
				const Comparator comp;
				CollT elems;
			public:
				NearestCollector ( int n ) :maxN ( n ), comp(), elems() {elems.reserve ( maxN );}
				inline void insert ( T* sample, const DistT dist )
				{
					if ( ( int ) elems.size() < maxN )
					{
						elems.push_back ( value_type ( sample, dist ) );
						std::push_heap ( elems.begin(), elems.end(), comp );
					}
					else if ( dist < elems.front().get<1>() )
					{
						std::pop_heap ( elems.begin(), elems.end(), comp );
						elems.pop_back();
						elems.push_back ( value_type ( sample, dist ) );
						std::push_heap ( elems.begin(), elems.end(), comp );
					}
				}

				inline void sort() {sort_heap<iterator, Comparator> ( begin(), end(), comp );}

				inline DistT getDist()
				{
					if ( ( int ) elems.size() < maxN ) return maxLimit<DistT>();
					return elems.front().get<1>();
				}

				inline void clear() {elems.clear();}
				inline iterator begin() {return elems.begin();}
				inline const_iterator begin() const {return elems.begin();}
				inline iterator end() {return elems.end();}
				inline const_iterator end() const {return elems.end();}

				inline int size() {return elems.size();}
		};
	}

}

#endif
