

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef CPICKBUFFER_H
#define CPICKBUFFER_H

#include <cfloat>
#include <cmath>
#include <cstring>
#include <set>
#include <vector>
#include <iostream>
#include <adrcore/util/debug.h>
#include <adrcore/util/mathutil.h>
#include <adrcore/util/unptr.h>

namespace adore
{
	namespace core
	{
		using namespace std;

#define DBGGRID DBGALL
#define MAX_IND_IDX (4 * bufX + 4 * bufY)

		enum GRID_ERR
		{
			GRID_SUCCESS = 0,
			GRID_NULL_PTR,
			GRID_IDX_OUT_OF_BOUNDS, //a serious error that should not happen
			GRID_COORS_OUT_OF_BOUNDS,
			GRID_POINT_EXISTS
		};

		inline void printGridError( GRID_ERR e )
		{
			switch ( e )
			{
				case GRID_NULL_PTR:
					printf( "Grid2D: NULL pointer\n" );
					break;
				case GRID_IDX_OUT_OF_BOUNDS:
					printf( "Grid2D: index out of bounds (very bad!)\n" );
					break;
				case GRID_COORS_OUT_OF_BOUNDS:
					printf( "Grid2D: coordinates out of bounds\n" );
					break;
				case GRID_POINT_EXISTS:
					printf( "Grid2D: point exists\n" );
					break;
				default:
					break;
			}
		}

		/**
		  This class represents a two dimensional grid.
		  This grid can be used to store some data in it that is addressed via a (x, y)-coordinate pair.
		  For example, it can be used to find the nearest point p stored in this grid to some point q.
		  The nearest point is found via a spiral search on the bins near the point p.
		 *
		 * As the features are managed internally by a set, choose your comparator wisely. If two distinct
		 * point are qualified by the comparator as the same point, only one will be added.
		   
		*/
		template <class PointT, class SubClassT, class SetComparator = less<PointT > >
		class Grid2D
		{
			public:
				typedef typename set< PointT, SetComparator>::iterator Iterator ;
				typedef typename set< PointT, SetComparator>::const_iterator Const_Iterator ;
				typedef typename set< PointT, SetComparator>::iterator iterator ;
				typedef typename set< PointT, SetComparator>::const_iterator const_iterator ;
				typedef PointT Point;
				typedef typename Ptr<PointT>::type PtPtr;
				typedef typename UnPtr<PointT>::type PtVal;

			private:
				ToPointer<PointT> toPointer;
				ToValue<PointT> toValue;

				vector< PointT > *pickBuffer;
				bool deletePoints;

				int *indices;
				int indIdx;
				int *visited;

				double sameDist; //dist <= sameDist bedeute den gleichenPunkt

				int bufX;//number of bins in x direction
				int bufY;//number of bins in y direction
				double x;
				double y;
				double width;
				double height;

				SetComparator defaultComparator;
			public://TODO remove public
				set<PointT , SetComparator> pointSet;

			public:
				/**
				*  Constructor
				*/
				Grid2D ( SetComparator &comparator, bool delPoints = true ) :
						pickBuffer ( NULL ),
						deletePoints ( delPoints ),
						sameDist ( 0 ),
						bufX ( -1 ),
						bufY ( -1 ),
						x ( 0 ),
						y ( 0 ),
						width ( -1 ),
						height ( -1 ),
						pointSet ( comparator )
				{
				}

				/**
				 *  Constructor
				 */
				Grid2D ( bool delPoints = true ) :
						pickBuffer ( NULL ),
						deletePoints ( delPoints ),
						sameDist ( 0 ),
						bufX ( -1 ),
						bufY ( -1 ),
						x ( 0 ),
						y ( 0 ),
						width ( -1 ),
						height ( -1 ),
						defaultComparator(),
						pointSet ( defaultComparator )
				{
				}

				/**
					copy constructor, makes a deep copy
				*/
				Grid2D ( const Grid2D &grid ):
						pickBuffer ( NULL ),
						deletePoints ( grid.deletePoints ),
						sameDist ( grid.sameDist ),
						bufX ( grid.bufX ),
						bufY ( grid.bufY ),
						x ( grid.x ),
						y ( grid.y ),
						width ( grid.width ),
						height ( grid.height ),
						defaultComparator( grid.defaultComparator ),
						pointSet ( )//dont copy at this point, because init will delete pointSet
				{
					init ( x, y, width, height, bufX, bufY );
					if ( grid.pickBuffer )
					{
						for ( int i = 0; i < bufX * bufY; ++i )
						{
							pickBuffer[i] = grid.pickBuffer[i];
						}
					}
					pointSet = grid.pointSet;//instead, copy now
				}


				/**
				*  Destructor
				*/
				virtual ~Grid2D()
				{
					DBGOUT ( DBGGRID, "deleting grid\n" );
					clean();
					destroy();
				}

				/**
				 * Initializes the Grid2D.
				 * @param x the width of the grid.
				 * @param y the height of the grid.
				 * @param bufStepX the number of x-subdivisions.
				 * @param bufStepY the number of y-subdivisions.
				 */
				void init ( double x0, double y0, double w, double h, int bufStepX = 20, int bufStepY = 20 );


				int getIndex ( double coorX, double coorY )
				{
					int xIdx = int ( ( coorX - x ) * bufX / width );
					int yIdx = int ( ( coorY - y ) * bufY / height );
					return yIdx * bufX + xIdx;
				}

				/**
				 * Adds a point and it's associated data to the grid.
				 * @param pt2D the point.
				 * @param checkExistence if true and a point exists within sameDist, an error is returned
				 * @return true, if point was added succeccfully or point was already in the grid and the additional data was added successfully
				 */
				GRID_ERR add ( PointT const &pt2D, bool checkExistence = true );

				bool exists ( PointT const &pt2D );

				/**
				 * returns the pointer to the element tha has been removed
				 * don't use it afterwards, if deletePoints has been set to TRUE,
				 * because it then points to freed memory
				 */
				PtPtr remove ( double coorX, double coorY, double maxDist );

				PtPtr remove ( double coorX, double coorY )
				{
					return remove( coorX, coorY, sameDist );
				}

				/**
				* Resets the Grid2D. Any references to points and there data are lost.
				*/
				void clean();

				/**
				 * The nearest neighbour of the point in the argument is returned
				 * @param pt2D of this point you want the nearest neighbor.
				 * @param maxDist the maximum distance a point may have to be considered as the same point.
				 * @return the nearest neighbour and it's associated data.
				 */
				PtPtr pick ( double coorX, double coorY, double maxDist = 1.0 );

				/**
				* A vector of all Points in the Grid2D is returned.
				 * @return the vector
				*/
				Iterator getBegin() {return pointSet.begin();}
				Const_Iterator getBegin() const { return pointSet.begin(); }

				Iterator getEnd() {return pointSet.end();}
				Const_Iterator getEnd() const { return pointSet.end(); }

				iterator begin() {return pointSet.begin();}

				iterator end() {return pointSet.end();}

				int getSize() {return pointSet.size();}

				/**
				 * the vector is filled with all points nearer to pt2D than (the euklidian) dist.
				 * @param pt2D the point who's neighbors are wanted.
				 * @param dist the distance a neighbor must not exceed.
				 * @param list a vector. The result is stored in it.
				 */
				void pickNear ( double coorX, double coorY, double dist, vector<PointT > &list );


				/**
				 * Sets the distance, that determins whether two distinct points are treated as the same point.
				 * @param d the same-distance
				 */
				void setSameDist ( double d ) {sameDist = d * d;}

				/**
				 * Returns the distance, that determins whether two distinct points are treated as the same point.
				 * @return the current same-distance
				 */
				double getSameDist() {return sqrt ( sameDist );}

				void transform ( double addX, double addY, double scaleX, double scaleY );

				/**
				 * Barton and Nackman for getters
				 */
				SubClassT& asLeaf() { return static_cast<SubClassT &> ( *this ); }

				void setDeletePoints ( bool theValue )
				{
					deletePoints = theValue;
				}

				bool getDeletePoints() const
				{
					return deletePoints;
				}

			private:
				double xGetter ( PtVal const &p ) {return asLeaf().xGetter ( p );}
				double yGetter ( PtVal const &p ) {return asLeaf().yGetter ( p );}
				void setter ( PtVal &p, const double x, const double y ) {asLeaf().setter ( p, x, y );}

				/**
				 * Destroys the Grid2D.
				 */
				void destroy();

				/**
				 * Helper for pickNear().
				 */
				inline void findPointsNear ( int idx, double coorX, double coorY,
				                             double dist, vector<PointT> &list );

				PtPtr findPoint ( int idx, double coorX, double coorY );

				void getIndicesAtDistance ( double distX, double distY, double coorX, double coorY );

			public:
				//debug function
				bool sanityCheck()
				{
					//check for each entry in pointSet its existence in pickBuffer
					for ( iterator it = begin(); it != end(); ++it )
					{
						bool found = false;
						for ( int idx = 0; idx < bufX * bufY; ++idx )
						{
							for ( uint i = 0; i < pickBuffer[idx].size(); i++ )
							{
								if ( *it == pickBuffer[idx][i] )
								{
									found = true;
									break;
								}
							}
						}
						if ( !found ) return false;
					}
					return true;
				}

				//debug function
				int getBin( PointT p )
				{
					for ( int idx = 0; idx < bufX * bufY; ++idx )
					{
						for ( uint i = 0; i < pickBuffer[idx].size(); i++ )
						{
							if ( p == pickBuffer[idx][i] )
							{
								return idx;
							}
						}
					}
					return -1;
				}

				//debug function
				string binDiffToString( int idx, double coorX, double coorY )
				{
					ostringstream oss;
					for ( uint i = 0; i < pickBuffer[idx].size(); ++i )
					{
						PointT cur2D = pickBuffer[idx][i];
						double inX = xGetter ( toValue( cur2D ) );
						double inY = yGetter ( toValue( cur2D ) );

						double dist = getSqrDistance2D ( inX, inY, coorX, coorY );
						oss << "(" << cur2D << ", [" << inX << ", " << inY << "], dist: " << dist << ")" << endl;
					}
					return oss.str();
				}
		};

		template <class PointT, class SubClassT, class SetComparator>
		void Grid2D<PointT, SubClassT, SetComparator>::init ( double x0, double y0, double w, double h, int bufStepX, int bufStepY )
		{
			assert( bufStepX > 0 );
			assert( bufStepY > 0 );
			pointSet.clear();

			//alles Neu, wenn sich die Aufloeung geaendert hat
			if ( !pickBuffer || x != x0 || y != y0 || width != w || height != h || bufStepX != bufX || bufStepY != bufY )
			{
				if ( pickBuffer )
				{
					destroy();
				}

				width = w;
				height = h;
				x = x0;
				y = y0;

				bufX = MAX( 1, int( w / float( bufStepX ) ) );
				bufY = MAX( 1, int( h / float( bufStepY ) ) );

				DBGOUT ( DBGGRID, "creating Grid2D\n" );
				pickBuffer = new vector<PointT>[bufX * bufY];
				visited = new int[bufX * bufY];
				indices = new int[MAX_IND_IDX];
				bzero ( visited, bufX * bufY * sizeof ( int ) );
				indIdx = 0;
			}
			else
			{
				clean();
			}
			//printf("x: %f, y: %f, w: %f, h: %f, binSize: %d, bufX: %d, bufY: %d\n", x, y, width, height, binSize, bufX, bufY);
		}

		/**
		 * x -> x * scaleX + addX
		 * y -> y * scaleY + addY
		 */
		template <class PointT, class SubClassT, class SetComparator>
		void Grid2D<PointT, SubClassT, SetComparator>::transform ( double addX, double addY, double scaleX, double scaleY )
		{
			width *= scaleX;
			height *= scaleY;

			x = scaleX * x + addX;
			y = scaleY * y + addY;

//			printf( "Grid2D - normalizing %d entries\n", getSize() );
			for ( Iterator it = getBegin(); it != getEnd(); ++it )
			{
				double ptX = xGetter ( toValue( *it ) );
				double ptY = yGetter ( toValue( *it ) );
				setter ( toValue( *it ), ptX * scaleX + addX, ptY * scaleY + addY );

//				ptX = xGetter( **it );
//				ptY = yGetter( **it );
//				printf( "after trafo: (%f, %f)\n", ptX, ptY );
			}
		}

		template <class PointT, class SubClassT, class SetComparator>
		bool Grid2D<PointT, SubClassT, SetComparator>::exists ( PointT const &pt2DRef )
		{
			PtPtr pt2D = toPointer( pt2DRef );
			if ( !pt2D ) { return false; }

			double ptX = xGetter ( toValue( pt2D ) );
			double ptY = yGetter ( toValue( pt2D ) );

			if ( ptX < x || ptY < y || ptX > x + width || ptY > y + height ) { return false; }

			int idx = getIndex ( ptX, ptY );
			if ( idx < 0 || idx >= bufX * bufY ) { return false; }

			//schauen, ob der Punkt bereits enthalten ist.
			PtPtr in = findPoint ( idx, ptX, ptY );
			if ( in )
			{
				double inX = xGetter ( toValue( in ) );
				double inY = yGetter ( toValue( in ) );
				if ( getSqrDistance2D ( inX, inY, ptX, ptY ) <= sameDist )
				{
					return true;
				}
			}

			return false;
		}

		template <class PointT, class SubClassT, class SetComparator>
		GRID_ERR Grid2D<PointT, SubClassT, SetComparator>::add ( PointT const &pt2DRef, bool checkExistence )
		{
			typedef typename Grid2D<PointT, SubClassT, SetComparator>::PtPtr PtPtr;
			PtPtr pt2D = toPointer( pt2DRef );
			if ( !pt2D )
			{
				DBGOUT ( DBGGRID, "Grid2D: point is NULL\n" );
				return GRID_NULL_PTR;
			}

			double ptX = xGetter ( toValue( pt2D ) );
			double ptY = yGetter ( toValue( pt2D ) );

			if ( ptX < x || ptY < y || ptX > x + width || ptY > y + height )
			{
				DBGOUT ( DBGGRID, "Grid2D: coors out of range\n" );
				DBGOUT ( DBGGRID, "(%f, %f) where grid is (%f, %f)-(%f, %f)\n", ptX, ptY, x, y, x+width, y+height );
				return GRID_COORS_OUT_OF_BOUNDS;
			}

			int idx = getIndex ( ptX, ptY );

			if ( idx < 0 || idx >= bufX * bufY )
			{
				DBGOUT ( DBGGRID, "idx %d out of range\n", idx );
				return GRID_IDX_OUT_OF_BOUNDS;
			}

			//schauen, ob der Punkt bereits enthalten ist.
			if ( checkExistence )
			{
				PtPtr in = findPoint ( idx, ptX, ptY );
				if ( in )
				{
					double inX = xGetter ( toValue( in ) );
					double inY = yGetter ( toValue( in ) );
					if ( getSqrDistance2D ( inX, inY, ptX, ptY ) <= sameDist )
					{
						DBGOUT ( DBGGRID, "Grid2D: point already known\n" );
						return GRID_POINT_EXISTS;
					}
				}
			}

//			cout << "adding\n" << ( *pt2D )( 0 ) << ( *pt2D )( 1 ) << endl;

			//neuen Punkt einfgen
#ifndef NDEBUG
			uint sizeBefore = pointSet.size();
#endif
			pointSet.insert ( pt2D );
#ifndef NDEBUG
			assert_msg ( sizeBefore == pointSet.size() - 1, "%d and %zd\n", sizeBefore, pointSet.size() );
#endif

//			cout << "pointSet size: " << pointSet.size() << endl;
			pickBuffer[idx].push_back ( pt2D );

			return GRID_SUCCESS;
		}

		template <class PointT, class SubClassT, class SetComparator>
		typename Grid2D<PointT, SubClassT, SetComparator>::PtPtr Grid2D<PointT, SubClassT, SetComparator>::remove ( double coorX, double coorY, double maxDist )
		{
			if ( coorX < x || coorY < y ) {return NULL;}

			maxDist *= maxDist;
			int idx = getIndex ( coorX, coorY );
//			printf( "index in remove: %d\n", idx );

			PtPtr in = findPoint ( idx, coorX, coorY );
			if ( in )
			{
				double inX = xGetter ( toValue( in ) );
				double inY = yGetter ( toValue( in ) );

				float dist = getSqrDistance2D ( inX, inY, coorX, coorY );
				if ( dist <= maxDist )
				{
#ifndef NDEBUG
					DBGOUT( DBGALL, "(%f, %f)-(%f, %f)[%f]{%p}, size(%zd)\n", inX, inY, coorX, coorY, dist, in, pickBuffer[idx].size() );
//					cout << binDiffToString( idx, coorX, coorY );
					uint sizeBefore = pointSet.size();
#endif
					pointSet.erase ( in );
#ifndef NDEBUG
					assert_msg ( sizeBefore == pointSet.size() + 1, "%d and %zd\n", sizeBefore, pointSet.size() );
#endif
					int i = 0;
					while ( pickBuffer[idx][i] != in ) ++i;
					pickBuffer[idx].erase ( pickBuffer[idx].begin() + i );

					if ( deletePoints ) delete in;

					return in;
				}
			}
			return NULL;
		}

		template <class PointT, class SubClassT, class SetComparator>
		typename Grid2D<PointT, SubClassT, SetComparator>::PtPtr Grid2D<PointT, SubClassT, SetComparator>::pick ( double coorX, double coorY, double maxDist )
		{
			typedef typename Grid2D<PointT, SubClassT, SetComparator>::PtPtr PtPtr;
			int idx;
			PtPtr candPt = NULL;
			double candDist = 0;
			PtPtr bestPt = NULL;
			double bestDist = DBL_MAX;

			maxDist *= maxDist;

			//DBGOUT(DBGGRID, "Punkt ist (%f, %f)\n", pt2D[0], pt2D[1]);

			bzero ( visited, bufX * bufY * sizeof ( int ) );

			double bufStepX = width / double ( bufX );
			double bufStepY = height / double ( bufY );
			double ar = bufStepY/bufStepX;

			for ( double dist = 0.5 * bufStepX; dist < width; dist += bufStepX )
			{
				getIndicesAtDistance ( dist, ar * dist, coorX, coorY ); //fllt das Indizes-Feld
				//DBGOUT(DBGGRID, "indIdx = %d\n", indIdx);
				for ( int i = 0; i < indIdx; ++i )
				{
					idx = indices[i];
					//printf("testing idx %d\n", idx);
					if ( idx < 0 || idx >= bufX * bufY ) continue;
					if ( visited[idx] ) continue;

					visited[idx] = true;
					candPt = findPoint ( idx, coorX, coorY );

					if ( candPt ) //wenn einer gefunden wurde, schauen, ob's der beste ist.
					{
						double inX = xGetter ( toValue( candPt ) );
						double inY = yGetter ( toValue( candPt ) );
						candDist = getSqrDistance2D ( inX, inY, coorX, coorY );

						//DBGOUT(DBGGRID, "Cand-Punkt ist (%d, %d) at distance %f\n", (*candPt->pt)[0], (*candPt->pt)[1], candDist);
						if ( candDist <= maxDist && candDist < bestDist )
						{
							bestDist = candDist;
							bestPt = candPt;
//							printf( "cand index in pick: %d (%p), size(%d)\n", idx, bestPt, pickBuffer[idx].size() );
//							cout << binDiffToString( idx, coorX, coorY );
						}
					}
				}

				//nun kann es sein, dass die Entfernung des gefundenen Punktes doch so gross ist,
				//dass im naechsten Ring geschaut werden muss
				//also pruefen, ob evtl. noch ein Durchlauf der aeusseren for-Schleife
				//notwendig ist
				if ( ( bestDist >= maxDist ) || ( bestPt && bestDist < dist ) ) break; //fertig
			}

			/*			if(bestPt != NULL){
							DBGOUT(DBGGRID, "bester Punkt ist (%f, %f)\n",
										(*bestPt)(0), (*bestPt)(1));
						}
						else{
									DBGOUT(DBGGRID, "bester Punkt ist NULL\n");
						}*/

//			printf( "point in pick: %p\n", bestPt );
			return bestPt;
		}

		template <class PointT, class SubClassT, class SetComparator>
		void Grid2D<PointT, SubClassT, SetComparator>::pickNear ( double coorX, double coorY, double dist, vector<PointT > &list )
		{
			int idx;

			DBGOUT( DBGGRID, "Searchdist is %f, Punkt ist (%f, %f), bucket is %d\n", dist, coorX, coorY, getIndex( coorX, coorY ) );
			DBGOUT( DBGGRID, "Grid: (%f, %f)(%f, %f) [%d, %d]\n", x, y, width, height, bufX, bufY );

			bzero ( visited, bufX * bufY * sizeof ( int ) );

			double bufStepX = width / double ( bufX );
			double bufStepY = height / double ( bufY );
			double ar = bufStepY/bufStepX;
			for ( double d = 0.5 * bufStepX; d <= dist + bufStepX; d += bufStepX )
			{
				getIndicesAtDistance ( d, ar * d, coorX, coorY ); //fllt das Indizes-Feld und indIdx (der Ring im Abstand d)
				DBGOUT( DBGGRID, "--- Dist: %f, indIdx = %d\n", d, indIdx );
				for ( int i = 0; i < indIdx; i++ )
				{
					idx = indices[i];
					DBGOUT( DBGGRID, "idx = %d/%d, vis = %d\n", idx, bufX * bufY, visited[idx] );
					if ( idx < 0 || idx >= bufX * bufY ) continue;
					if ( visited[idx] ) continue;
//					DBGOUT( DBGNORMAL, "idx %d taken\n", idx );

					visited[idx] = true;
					findPointsNear ( idx, coorX, coorY, dist, list );
				}
			}
		}

		template <class PointT, class SubClassT, class SetComparator>
		void Grid2D<PointT, SubClassT, SetComparator>::findPointsNear ( int idx,
		                double coorX, double coorY, double dist, vector<PointT> &list )
		{
			PtPtr cur2D = NULL;
			double d;
			dist *=dist;
			DBGOUT( DBGGRID, "searching %zd points in bucket idx:%d\n", pickBuffer[idx].size(), idx );
			for ( uint i = 0; i < pickBuffer[idx].size(); i++ )
			{
				cur2D = pickBuffer[idx][i];
				double inX = xGetter ( toValue( cur2D ) );
				double inY = yGetter ( toValue( cur2D ) );

				d = getSqrDistance2D ( inX, inY, coorX, coorY );
				DBGOUT( DBGGRID, "distance of (%f, %f) : %f/%f\n", inX, inY, d, dist );

				if ( d < dist )
				{
					list.push_back ( cur2D );
				}
			}
		}

		template <class PointT, class SubClassT, class SetComparator>
		typename Grid2D<PointT, SubClassT, SetComparator>::PtPtr Grid2D<PointT, SubClassT, SetComparator>::findPoint ( int idx, double coorX, double coorY )
		{
			typedef typename Grid2D<PointT, SubClassT, SetComparator>::PtPtr PtPtr;
//			DBGOUT( DBGGRID, "idx: %d\n", idx );

			PtPtr cur2D = NULL;
			double dist;
			double lastDist = DBL_MAX;
			int bestIdx = -1;

			for ( uint i = 0; i < pickBuffer[idx].size(); ++i )
			{
				cur2D = pickBuffer[idx][i];
				double inX = xGetter ( toValue( cur2D ) );
				double inY = yGetter ( toValue( cur2D ) );

//				printf( "in findPoint:\n" );
//				cout << "(" << ( *cur2D->getPoint() )[0] << ", " << ( *cur2D->getPoint() )[1] << ") (" << pt2D[0] << ", " << pt2D[1] << ")"<< endl;

				dist = getSqrDistance2D ( inX, inY, coorX, coorY );
				if ( dist < lastDist )
				{
					lastDist = dist;
					bestIdx = i;
					if ( dist == 0.0 ) break;
				}
			}

			if ( bestIdx == -1 ) return NULL;
			return pickBuffer[idx][bestIdx];
		}

		template <class PointT, class SubClassT, class SetComparator>
		void Grid2D<PointT, SubClassT, SetComparator>::getIndicesAtDistance ( double distX, double distY, double coorX, double coorY )
		{
			//die vier Seiten abwandern
			double xDiff, yDiff;
			int xIdx, yIdx;
			indIdx = 0;

			double ptX = ( coorX - x );
			double ptY = ( coorY - y );
			double bufStepX = width / double ( bufX );
			double bufStepY = height / double ( bufY );

//			DBGOUT( DBGGRID, "distX: %f, distY: %f\n", distX, distY );
//			DBGOUT( DBGGRID, "ptX: %f, ptY: %f\n", ptX, ptY );
//			DBGOUT( DBGGRID, "bufStepX: %f, bufStepY: %f\n", bufStepX, bufStepY );

			if ( distX == 0 && distY == 0 )
			{
				xIdx = int ( ptX * bufX / width );
				yIdx = int ( ptY * bufY / height );
				assert_msg ( indIdx < MAX_IND_IDX, "idx %d >= %d\n", indIdx, MAX_IND_IDX );
				indices[indIdx++] = yIdx * bufX + xIdx;
				return;
			}

			yDiff = -distY;
			yIdx = int ( ( ptY + yDiff ) * bufY / height );
			for ( xDiff = -distX; xDiff < distX; xDiff+=bufStepX )
			{
//				printf( "xDiff: %f, distX: %f, bufStepX: %f\n", xDiff, distX, bufStepX );
//				int sig = signum( xDiff );
				xIdx = int ( ( ptX + xDiff/* - sig */ ) * bufX / width );
				assert_msg ( indIdx < MAX_IND_IDX, "idx %d >= %d\n", indIdx, MAX_IND_IDX );
				indices[indIdx++] = yIdx * bufX + xIdx;
				DBGOUT( DBGGRID, "a: index %d added x/y:%d/%d\n", indices[indIdx-1], xIdx, yIdx );
			}
			yDiff = distY;
			yIdx = int ( ( ptY + yDiff ) * bufY / height );
			for ( xDiff = distX; xDiff > -distX; xDiff-=bufStepX )
			{
//				int sig = signum( xDiff );
				xIdx = int ( ( ptX + xDiff/* - sig*/ ) * bufX / width );
				assert_msg ( indIdx < MAX_IND_IDX, "idx %d >= %d\n", indIdx, MAX_IND_IDX );
				indices[indIdx++] = yIdx * bufX + xIdx;
				DBGOUT( DBGGRID, "b: index %d added x/y:%d/%d\n", indices[indIdx-1], xIdx, yIdx );
			}

			xDiff = distX;
			xIdx = int ( ( ptX + xDiff ) * bufX / width );
			for ( yDiff = -distY; yDiff < distY; yDiff+=bufStepY )
			{
//				printf( "yDiff: %f, distY: %f, bufStepY: %f\n", yDiff, distY, bufStepY );
//				int sig = signum( yDiff );
				int yIdx = int ( ( ptY + yDiff/* - sig*/ ) * bufY / height );
				assert_msg ( indIdx < MAX_IND_IDX, "idx %d >= %d\n", indIdx, MAX_IND_IDX );
				indices[indIdx++] = yIdx * bufX + xIdx;
				DBGOUT( DBGGRID, "c: index %d added x/y:%d/%d\n", indices[indIdx-1], xIdx, yIdx );
			}
			xDiff = -distX;
			xIdx = int ( ( ptX + xDiff ) * bufX / width );
			for ( yDiff = distY; yDiff > -distY; yDiff-=bufStepY )
			{
//				int sig = signum( yDiff );
				int yIdx = int ( ( ptY + yDiff/* - sig*/ ) * bufY / height );
				assert_msg ( indIdx < MAX_IND_IDX, "idx %d >= %d\n", indIdx, MAX_IND_IDX );
				indices[indIdx++] = yIdx * bufX + xIdx;
				DBGOUT( DBGGRID, "d: index %d added x/y:%d/%d\n", indices[indIdx-1], xIdx, yIdx );
			}
		}

		template <class PointT, class SubClassT, class SetComparator>
		void Grid2D<PointT, SubClassT, SetComparator>::destroy()
		{
			pointSet.clear();

			if ( pickBuffer )
			{
				DBGOUT ( DBGGRID, "deleting Grid2D\n" );
				delete [] pickBuffer;
				delete [] indices;
				delete [] visited;

				pickBuffer = NULL;
				indices = NULL;
				visited = NULL;
			}
		}

		template <class PointT, class SubClassT, class SetComparator>
		void Grid2D<PointT, SubClassT, SetComparator>::clean()
		{
			if ( deletePoints )
			{
				DBGOUT ( DBGGRID, "Grind2D: deleting %zd points\n", pointSet.size() );
				for ( Iterator it = getBegin(); it != getEnd(); ++it )
				{
					delete *it;
				}
			}

			pointSet.clear();
			DBGOUT ( DBGGRID, "Grind2D: pointSet size: %zd\n", pointSet.size() );
			if ( pickBuffer )
			{
				for ( int i = 0; i < bufX *  bufY; i++ )
				{
					pickBuffer[i].clear();
				}
			}
			//if(visited) bzero(visited, bufX * bufY * sizeof(int));  braucht nicht, weil bei jedem pick zurckgesetzt
			indIdx = 0;
		}

		////////////////// SubClasses  //////////////////////////


		template <class PointT, class SetComparator = less<PointT > >
		class Grid2DHomo:public Grid2D<PointT, Grid2DHomo<PointT, SetComparator>, SetComparator>
		{
			public:
				typedef Grid2D<PointT, Grid2DHomo<PointT, SetComparator>, SetComparator> BaseT;
				typedef typename BaseT::Iterator Iterator;
				typedef typename BaseT::Const_Iterator Const_Iterator;
				typedef typename BaseT::iterator iterator;
				typedef typename BaseT::const_iterator const_iterator;
				typedef typename BaseT::PtVal PtVal;

			public:
				Grid2DHomo ( SetComparator &comparator, bool delPoints = true ) :Grid2D<PointT, Grid2DHomo<PointT, SetComparator>, SetComparator> ( comparator, delPoints ) {;}
				Grid2DHomo ( bool delPoints = true ) :Grid2D<PointT, Grid2DHomo<PointT, SetComparator>, SetComparator> ( delPoints ) {;}
				Grid2DHomo ( const Grid2DHomo &grid ):Grid2D<PointT, Grid2DHomo<PointT, SetComparator>, SetComparator> ( grid ) {;}

				inline double xGetter ( PtVal const &p ) const
				{
					assert ( p.size() == 3 );
					return p ( 0 ) / p ( 2 );
				}

				inline double yGetter ( PtVal const &p ) const
				{
					assert ( p.size() == 3 );
					return p ( 1 ) / p ( 2 );
				}

				inline void setter ( PtVal &p, const double x, const double y )
				{
					assert ( p.size() == 3 );
					p ( 0 ) = x;
					p ( 1 ) = y;
					p ( 2 ) = 1.0;
				}
		};

		template <class PointT, class SetComparator = less<PointT > >
		class Grid2DInhomo:public Grid2D<PointT, Grid2DInhomo<PointT, SetComparator>, SetComparator>
		{
			public:
				typedef Grid2D<PointT, Grid2DInhomo<PointT, SetComparator>, SetComparator> BaseT;
				typedef typename BaseT::Iterator Iterator;
				typedef typename BaseT::Const_Iterator Const_Iterator;
				typedef typename BaseT::iterator iterator;
				typedef typename BaseT::const_iterator const_iterator;
				typedef typename BaseT::PtVal PtVal;

			public:
				Grid2DInhomo ( SetComparator &comparator, bool delPoints = true ) :Grid2D<PointT, Grid2DInhomo<PointT, SetComparator>, SetComparator> ( comparator, delPoints ) {;}
				Grid2DInhomo ( bool delPoints = true ) :Grid2D<PointT, Grid2DInhomo<PointT, SetComparator>, SetComparator> ( delPoints ) {;}
				Grid2DInhomo ( const Grid2DInhomo &grid ):Grid2D<PointT, Grid2DInhomo<PointT, SetComparator>, SetComparator> ( grid ) {;}

				inline double xGetter ( PtVal const &p ) const
				{
					return p ( 0 );
				}

				inline double yGetter ( PtVal const &p ) const
				{
					return p ( 1 );
				}

				inline void setter ( PtVal &p, const double x, const double y )
				{
					p ( 0 ) = x;
					p ( 1 ) = y;
				}
		};

	}
};


#endif
