

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Sergey Cheremukhin


#ifndef QUADTREE_H
#define QUADTREE_H

#include <vector>

#include <boost/shared_ptr.hpp>

#include <adrcore/types/aarray.h>
#include <adrcore/types/featurepoint.h>
#include <adrcore/util/mathutil.h>
#include <adrcore/features/linedistfilter.h>
#include <adrcore/util/printmat.h>

/** y assumed to grow from top to bottom
 *  no checks!
 *  no points assumed to be at same position.
 *  one should not copy this tree.
*/ 

namespace adore {
	namespace core {

		class QuadTree {

			class Node {
				AVectord center;
				double width, height;

				FeaturePoint* feature;

			public:
				typedef boost::shared_ptr<Node> NodePtr;

			private:
				NodePtr nw, ne, sw, se;

			public:
				Node(const AVectord& c, double w, double h) :
                                        center(c), width(w), height(h), feature(NULL) {
				}

				void insert(FeaturePoint* fp) {
					if(!nw && !feature) {
						feature = fp;
						return;
					}
					else if(!nw) {
						double newW = 0.5 * width;
						double newH = 0.5 * height;
						double newHalfW = 0.5 * newW;
						double newHalfH = 0.5 * newH;
						AVectord c(2);
						c = center(0) - newHalfW, center(1) - newHalfH;
						nw.reset(new Node(c, newW, newH));
						c(0) = center(0) + newHalfW;
						ne.reset(new Node(c, newW, newH));
						c(1) = center(1) + newHalfH;
						se.reset(new Node(c, newW, newH));
						c(0) = center(0) - newHalfW;
						sw.reset(new Node(c, newW, newH));

						insertPriv(feature);
						feature = NULL;
					}
					insertPriv(fp);
				}

			private:
				void insertPriv(FeaturePoint* fp) {
					if( (*fp)(0) <= center(0) ) {
						if( (*fp)(1) <= center(1) ) {
							nw->insert(fp);
						} else {
							sw->insert(fp);
						}
					} else {
						if( (*fp)(1) <= center(1) ) {
							ne->insert(fp);
						} else {
							se->insert(fp);
						}
					}
				}

			public:
				double getWidth() const { return width; }
				double getHeight() const { return height; }
				const AVectord& getCenter() const { return center; }
				bool isLeaf() const { return NULL == nw; }
				bool empty() const { return NULL == nw && NULL == feature; }
				NodePtr getNW() const { return nw; }
				NodePtr getNE() const { return ne; }
				NodePtr getSE() const { return se; }
				NodePtr getSW() const { return sw; }
				FeaturePoint* getValue() const { return feature; }
			};

		public:
			typedef Node::NodePtr NodePtr;

		private:
			NodePtr root;

			QuadTree(const QuadTree&) {}
			QuadTree& operator=(const QuadTree&) { return *this; }

		public:
			QuadTree(double w, double h) {
				AVectord c(2);
				c = w * 0.5, h * 0.5;
				root.reset(new Node(c, w, h));
			}
			QuadTree() {}

			void insert(FeaturePoint* fp) {
				assert(root);
				if(!fp) return;
				root->insert(fp);
			}

			NodePtr getRoot() const { return root; }

			void build(const std::vector<FeaturePoint*>& pts) {
				double minX = std::numeric_limits<int>::max() >> 1;
				double maxX = -minX;
				double minY = minX, maxY = maxX;
				for(size_t i = 0; i < pts.size(); ++i) {
					FeaturePoint* fp = pts[i];
					if(!fp) continue;
					minX = std::min(minX, (*fp)(0));
					minY = std::min(minY, (*fp)(1));
					maxX = std::max(maxX, (*fp)(0));
					maxY = std::max(maxY, (*fp)(1));
				}
				double width = maxX - minX, height = maxY - minY;
				AVectord c(2);
				c = width * 0.5, height * 0.5;
				root.reset(new Node(c, width, height));
				for(size_t i = 0; i < pts.size(); ++i) {
					root->insert(pts[i]);
				}
			}
		};


		class QuadTreeLineAssessor {
			const QuadTree& tree;
			std::vector<FeaturePoint*> result;
			typedef QuadTree::NodePtr NodePtr;
			std::vector<NodePtr> toDo;
			std::vector<NodePtr> toAdd;
			std::vector<FeaturePoint*> toMeasure;

		public:
			QuadTreeLineAssessor(const QuadTree& t, size_t num) : tree(t)
			{
				result.reserve(num);
				toDo.reserve(num);
				toAdd.reserve(num);
				toMeasure.reserve(num);
			}

			const std::vector<FeaturePoint*>& assess(const AVectord& p1, const AVectord& p2, double thresh)
			{
				const double factor = sqrt(2.0) * 0.5;
				const double halfThresh = thresh * 0.5;

				result.clear();
				toDo.clear();
				toAdd.clear();
				toMeasure.clear();

				AVectord line(3);
				cross(p1, p2, line);
				LineDistance ldist( line( 0 ), line( 1 ), line( 2 ) );
				if ( !ldist.isOkay() )
				{
					DBGOUT( DBGNORMAL, "QuadTreeLineAssessor: insane line occurred" );
					printMat( DBGNORMAL, line );
					return result;
				}

				toDo.push_back(tree.getRoot());
				while(!toDo.empty()) {
					NodePtr node = toDo.back();
					toDo.pop_back();
					if(node->empty())
						continue;
					if(node->isLeaf()) {
						toMeasure.push_back(node->getValue());
						continue;
					}

					NodePtr child = node->getNE();
					double diag = factor * std::max(child->getWidth(), child->getHeight());
					double maxDist = diag + thresh;
					testNode(node->getNE(), diag, maxDist, thresh, halfThresh, ldist);
					testNode(node->getSE(), diag, maxDist, thresh, halfThresh, ldist);
					testNode(node->getSW(), diag, maxDist, thresh, halfThresh, ldist);
					testNode(node->getNW(), diag, maxDist, thresh, halfThresh, ldist);
				}

				for(size_t i = 0; i < toMeasure.size(); ++i) {
					FeaturePoint* cand = toMeasure[i];
					double dist = fabs( ldist.getDistance( ( * cand ) ( 0 ), ( * cand ) ( 1 ) ) );
					if(dist <= thresh)
						result.push_back(cand);
				}

				while(!toAdd.empty()) {
					NodePtr node = toAdd.back();
					toAdd.pop_back();
					if(node->empty())
						continue;
					if(node->isLeaf()) {
						result.push_back(node->getValue());
						continue;
					}
					toAdd.push_back(node->getNE());
					toAdd.push_back(node->getSE());
					toAdd.push_back(node->getSW());
					toAdd.push_back(node->getNW());
				}
				return result;
			}

		private:
			inline void testNode(NodePtr node, double diag, double maxDist, double thresh, double halfThresh, const LineDistance& ldist)
			{
				const AVectord& p = node->getCenter();
				double dist = fabs( ldist.getDistance( p ( 0 ), p ( 1 ) ) );

				if(dist <= maxDist) {
					if(diag <= halfThresh && dist <= diag)
						toAdd.push_back(node);
					else
						toDo.push_back(node);
				}
			}
		};
	}
}

#endif // QUADTREE_H
