

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Sergey Cheremukhin


#include "surfdescriptor.h"
#include "integralimage.h"

#include <cassert>
#include <memory>

#include <adrcore/imgprocessing/rgbtogray.h>
#include <adrcore/imgprocessing/channelfunctions.h>

#include <adrcore/util/mathutil.h>
#include <adrcore/util/gaussians.h>

namespace adore
{
	namespace core
	{

		using namespace std;

		const double pi = M_PI;
		const double pi_2 = pi * 2;
		const double pi_over_3 = pi / 3;
		const double pi_over_9 = pi / 9;

		SurfDescriptor::SurfDescriptor( uint featID, uint descrID, bool upright, int subregs, bool split ) :
				DescriptorCreator( featID, descrID ),
				state( Invalid ),
				image_gray( NULL ),
				iimage( NULL ),
				simple( upright ),
				subregions( subregs ),
				splitSums( split ),
				tmpVec2(2)
		{
			if ( SubRegsMin > subregions ) subregions = SubRegsMin;
			else if ( SubRegsMax < subregions ) subregions = SubRegsMax;
			windowSize = subregions * SubRegSize;
			orients.reserve(30);
		}

		SurfDescriptor::SurfDescriptor( const SurfDescriptor & sd ) :
				DescriptorCreator( sd.getFeatureID(), sd.getDescriptorID() ),
				state( Invalid ),
				image_gray( NULL ),
				iimage( NULL ),
				simple( sd.simple ),
				subregions( sd.subregions ),
				splitSums( sd.splitSums ),
				tmpVec2(2)
		{
			if ( SubRegsMin > subregions ) subregions = SubRegsMin;
			else if ( SubRegsMax < subregions ) subregions = SubRegsMax;
			windowSize = subregions * SubRegSize;
			orients.reserve(30);
		}

		DescriptorCreator * SurfDescriptor::clone()
		{
			return new SurfDescriptor( *this );
		}

		SurfDescriptor::~SurfDescriptor()
		{
			reset();
		}

		void SurfDescriptor::reset()
		{
			state = Invalid;
			image_gray = NULL;
			local_image_gray.reset();
			iimage = NULL;
			local_iimage.reset();
		}

		void SurfDescriptor::prepare( ViewT * view )
		{
			reset();

			ImageT * src = view->getImage();
			if ( NULL == src ) return;

			if ( src->getNumChannels() == 3 )
			{
				DBGOUT( DBGNORMAL, "Converting RGB to grayscale\n" );
				local_image_gray.reset(new ImageT( ( int ) src->getWidth(), ( int ) src->getHeight(), 1 ));

				RGBtoGray<imgT> converter;
				Channel<imgT>::Set in( true );
				src->createChannels( in );

				converter.process( in, local_image_gray.get() );
				image_gray = local_image_gray.get();
			}
			else if ( 1 == src->getNumChannels() )
			{
				image_gray = src;
			}
			else return;

			local_iimage.init(image_gray);
			iimage = &local_iimage;

			state = Valid;
		}

		void SurfDescriptor::prepare( ImageT* image, SurfIntegralImageT * intImage )
		{
			reset();

			if ( NULL == image || 1 != image->getNumChannels() ) return;

			image_gray = image;

			if ( NULL != intImage ) iimage = intImage;
			else
			{
				local_iimage.init( image_gray );
				iimage = &local_iimage;
			}

			state = Valid;
		}

		void SurfDescriptor::performActionDerived( ViewT * view )
		{
			prepare( view );
			if(state != Valid) {
				DBGOUT( DBGNORMAL, "state is invalid. not initialized.\n" );
				return;
			}
			makeDescriptors( view );
		}

		void SurfDescriptor::makeDescriptors( ViewT * view )
		{
			ViewT::PointsIterator it = view->getPointsBegin( fid );
			ViewT::PointsIterator end = view->getPointsEnd( fid );
			while ( it != end )
			{
				if ( NULL != *it ) makeDescriptorPriv( **it );
				++it;
			}
		}

		void SurfDescriptor::makeDescriptor( FeaturePoint & fp ) const
		{
			if(state != Valid) {
				DBGOUT( DBGNORMAL, "state is invalid. not initialized.\n" );
				return;
			}
			makeDescriptorPriv(fp);
		}

		void SurfDescriptor::makeDescriptorPriv( FeaturePoint & fp ) const
		{
			assert( fp( 2 ) == 1.0 );

			if ( !simple )
			{
				DBGOUT(DBGALL, "SurfDescriptor: calculating rotinv descriptor.\n");
				getMainOrientations( fp );
				if ( orients.empty() ) {
					DBGOUT(DBGALL, "SurfDescriptor: no orientations found.\n");
					return;
				}
				rotationMatrix2d( -orients[0], fp.getInvAffineMat() );
				calculateDescriptors( fp, orients );
			}
			else
			{
				fp.getInvAffineMat() = 1, 0, 0, 1;
				calculateDescriptorSimple( fp );
			}
		}

		bool SurfDescriptor::checkBorders( FeaturePoint& fp, int radius ) const
		{
			int ptx = ( int )round( fp( 0 ) ), pty = ( int )round( fp( 1 ) );

			if (ptx - radius < 0 )
			{
				return false;
			}
			if (pty - radius < 0 )
			{
				return false;
			}
			if (ptx + radius >= iimage->getWidth() )
			{
				return false;
			}
			if (pty + radius >= iimage->getHeight() )
			{
				return false;
			}

			return true;
		}

		void SurfDescriptor::getMainOrientations( FeaturePoint & fp ) const
		{
			orients.clear();
			const int ptx = ( int )round( fp( 0 ) ), pty = ( int )round( fp( 1 ) );
			const real_type s = fp.getScale();
			const int step = ( int )round( s ), sampleSz = step * OrientSampleSize;
			const int scaledRadius = OrientRadius * step;

			const int checkRadius = scaledRadius + (sampleSz / 2);
			if( !checkBorders(fp, checkRadius ) )
			{
				DBGOUT(DBGALL, "SurfDescriptor: checkborders failed; feat (%f, %f), radius %d, integralimage(%d, %d).\n", fp(0), fp(1), checkRadius, iimage->getWidth(), iimage->getHeight() );
				return;
			}

			int step2 = step * 2;
			if ( !hasTable( scaledRadius, step2 ) )
				prepareGaussTable( scaledRadius, step2 );

			/// circular neighbourhood of radius 6s
			histogram.clear();
			int minX = -scaledRadius, minY = -scaledRadius, maxX = scaledRadius, maxY = scaledRadius;

			const int scaledRadius_sq = scaledRadius * scaledRadius;
			for ( int y = minY; y <= maxY; y += step )
			{
				bool hadX = false;
				for ( int x = minX; x <= maxX; x += step )
				{
					if ( ( x * x + y * y ) > scaledRadius_sq ) {
						if(hadX) break;
						else continue;
					} else hadX = true;

					responseXandY( ptx + x, pty + y, sampleSz, tmpVec2 );
					tmpVec2(1) = -tmpVec2(1);
					real_type g = gauss( x, y, step2 );
					histogram.add( tmpVec2, g );
				}
			}

			histogram.getMainOrientations(orients);
		}

		bool SurfDescriptor::isWithin( real_type x, real_type y ) const
		{
			return x >= 0 && x < image_gray->getWidth() && y >= 0 && y < image_gray->getHeight();
		}

		bool SurfDescriptor::checkBorders( FeaturePoint& fp, const RotMatT & A, real_type halfSz ) const
		{
			vec2 temp1(2), temp2(2);
			temp1 = halfSz, halfSz;
			matMul2(A, temp1, temp2);
			if(!isWithin(fp(0) + temp2(0), fp(1) + temp2(1)))
				return false;
			temp1 = halfSz, -halfSz;
			matMul2(A, temp1, temp2);
			if(!isWithin(fp(0) + temp2(0), fp(1) + temp2(1)))
				return false;
			temp1 = -halfSz, halfSz;
			matMul2(A, temp1, temp2);
			if(!isWithin(fp(0) + temp2(0), fp(1) + temp2(1)))
				return false;
			temp1 = -halfSz, -halfSz;
			matMul2(A, temp1, temp2);
			if(!isWithin(fp(0) + temp2(0), fp(1) + temp2(1)))
				return false;
//			if(!isWithin(fp(0) - temp2(0), fp(1) - temp2(1)))
//				return false;
//			if(!isWithin(fp(1) - temp2(1), fp(0) + temp2(0)))
//				return false;
//			if(!isWithin(fp(1) + temp2(1), fp(0) - temp2(0)))
//				return false;
			return true;
		}

		void SurfDescriptor::calculateDescriptors( FeaturePoint & fp, const vector<SurfDescriptor::real_type>& orients ) const
		{
			float posx = fp( 0 ), posy = fp( 1 );
			real_type s = fp.getScale();
			int step = ( int )round( s ), step2 = step * 2, step3 = step * 3, stepH = ( step * subregions * SubRegSize ) / 2;
			int width = subregions * SubRegSize * step + step2;

			RotMatT A( 2, 2 );

			ImageT patch( width, width, 1 );
			int numSums = splitSums ? 8 : 4;
			vec2 sums( numSums );
			auto_ptr<FeatureDescriptor> fd(new FeatureDescriptor( 0/*size*/, did ));

			if ( !hasTable( stepH, step3 ) )
				prepareGaussTable( stepH, step3 );

			bool haveSomeDescriptor = false;
			for ( size_t i = 0; i < orients.size(); ++i )
			{
				const real_type& o = orients[i];
				rotationMatrix2d( -o, A );
				bool ok = checkBorders(fp, A, width * 0.5);
				if(!ok) continue;
				else haveSomeDescriptor = true;

				warpAffine2( *image_gray, posx, posy, A, patch.begin(), width, width );
				SurfIntegralImageT intImage( &patch );

				DescVecT & vec = fd->addVector( 1 + subregions * subregions * numSums ); // 1 - sign
				int pos = 1;

				for ( int y = 0; y < 5 * subregions; y += 5 )
				{
					for ( int x = 0; x < 5 * subregions; x += 5 )
					{
						sums = 0;
						/// each subregion is 5x5 sample points
						for ( int yy = y; yy < y + 5; ++yy )
						{
							for ( int xx = x; xx < x + 5; ++xx )
							{
								responseTopXandTopY( xx, yy, step2, intImage, tmpVec2 );
								real_type g = gauss( xx * step - stepH, yy * step - stepH, step3 );
								tmpVec2 = tmpVec2 * g;
								if ( !splitSums )
								{
									sums( 0 ) += tmpVec2( 0 ); sums( 1 ) += abs( tmpVec2( 0 ) );
									sums( 2 ) += tmpVec2( 1 ); sums( 3 ) += abs( tmpVec2( 1 ) );
								}
								else
								{
									if ( 0 > tmpVec2( 0 ) )
									{
										sums( 0 ) += tmpVec2( 0 ); sums( 1 ) += abs( tmpVec2( 0 ) );
									}
									else
									{
										sums( 2 ) += tmpVec2( 0 ); sums( 3 ) += abs( tmpVec2( 0 ) );
									}
									if ( 0 > tmpVec2( 1 ) )
									{
										sums( 4 ) += tmpVec2( 1 ); sums( 5 ) += abs( tmpVec2( 1 ) );
									}
									else
									{
										sums( 6 ) += tmpVec2( 1 ); sums( 7 ) += abs( tmpVec2( 1 ) );
									}
								}
							}
						}
						vec( Range( pos, pos + numSums - 1 ) ) = sums;
						pos += numSums;
					}
				}
				vec( 0 ) = fp.getSignedStrength() < 0 ? -1 : 1;
				DescVecT::iterator it = vec.begin();
				++it;
				normalize( it, vec.end(), 1.0 );
			}

			if(!haveSomeDescriptor) {
				COUT( DBGALL, fp << endl);
				COUT( DBGALL, "SurfDescriptor::calculateDescriptor: no descriptor calculated." << endl);
				return;
			}

			fp.setDescriptor( fd.get() );
			fd.release();
		}

		void SurfDescriptor::calculateDescriptorSimple( FeaturePoint & fp ) const
		{
			typedef AArray< real_type, 1 > vec2;

			int posx = ( int )round( fp( 0 ) ), posy = ( int )round( fp( 1 ) );
// 			float posx = fp( 0 ), posy = fp( 1 );//not, because of responceXandY below
			real_type s = fp.getScale();
			int step = ( int )round( s ), step2 = step * 2, step3 = step * 3, stepH = ( step * subregions * 5 ) / 2;

			int numSums = splitSums ? 8 : 4;
			int tempX, tempY;
			vec2 sums( numSums );

			FeatureDescriptor * fd = new FeatureDescriptor( 0/*size*/, did );
			DescVecT & vec = fd->addVector( 1 + subregions * subregions * numSums ); // 1 - sign
			int pos = 1;

			if ( !hasTable( stepH, step3 ) )
				prepareGaussTable( stepH, step3 );

			/// 4x4 square subregions
			for ( int y = 0; y < 5 * subregions; y += 5 )
			{
				for ( int x = 0; x < 5 * subregions; x += 5 )
				{
					sums = 0;
					for ( int yy = y; yy < y + 5; ++yy )
					{
						for ( int xx = x; xx < x + 5; ++xx )
						{
							bool ok = checkBorders(fp, stepH + 1);
							if(!ok) continue;

							tempX = xx * step - stepH, tempY = yy * step - stepH;
							real_type g = gauss( tempX, tempY, step3 );
							responseXandY( tempX + posx, tempY + posy, step2, tmpVec2 );
							tmpVec2 = tmpVec2 * g;
							if ( !splitSums )
							{
								sums( 0 ) += tmpVec2( 0 ); sums( 1 ) += abs( tmpVec2( 0 ) );
								sums( 2 ) += tmpVec2( 1 ); sums( 3 ) += abs( tmpVec2( 1 ) );
							}
							else
							{
								if ( 0 > tmpVec2( 0 ) )
								{
									sums( 0 ) += tmpVec2( 0 ); sums( 1 ) += abs( tmpVec2( 0 ) );
								}
								else
								{
									sums( 2 ) += tmpVec2( 0 ); sums( 3 ) += abs( tmpVec2( 0 ) );
								}
								if ( 0 > tmpVec2( 1 ) )
								{
									sums( 4 ) += tmpVec2( 1 ); sums( 5 ) += abs( tmpVec2( 1 ) );
								}
								else
								{
									sums( 6 ) += tmpVec2( 1 ); sums( 7 ) += abs( tmpVec2( 1 ) );
								}
							}
						}
					}
					vec( Range( pos, pos + numSums - 1 ) ) = sums;
					pos += numSums;
				}
			}

			vec( 0 ) = fp.getSignedStrength() < 0 ? -1 : 1;
			normalize( vec.begin(), vec.end(), 1.0 );

			fp.setDescriptor( fd );
		}

		void SurfDescriptor::responseXandY( int x, int y, int sz, vec2 &resp ) const
		{
			assert(x < iimage->getWidth());
			assert(y < iimage->getHeight());
			int step = sz / 2;
			int topX = x - step;
			int topY = y - step;
			int wx = sz;
			int wy = sz;
			if(topX < 0 ) {topX = 0; wx += topX;}
			if(topY < 0 ) {topY = 0; wy += topY;}
			if(topX + wx >= iimage->getWidth()) wx = iimage->getWidth() - topX - 1;
			if(topY + wy >= iimage->getHeight()) wy = iimage->getHeight() - topY - 1;

			if ( 0 == sz % 2 )
			{
				real_type white = iimage->getSum( topX, topY, wx, wy );
				resp( 0 ) = white - ( real_type )2.0 * iimage->getSum( topX, topY, step, wy );
				resp( 1 ) = white - ( real_type )2.0 * iimage->getSum( topX, topY, wx, step );
			}
			else
			{
				resp( 0 ) = iimage->getSum( x + 1, topY, step, wy ) - iimage->getSum( topX, topY, step, wy );
				resp( 1 ) = iimage->getSum( topX, y + 1, wx, step ) - iimage->getSum( topX, topY, wx, step );
			}
		}

		void SurfDescriptor::responseTopXandTopY( int x, int y, int sz, SurfIntegralImageT& intImage, vec2 &resp ) const
		{
			assert(x < iimage->getWidth());
			assert(y < iimage->getHeight());
			int step = sz / 2;
			int topX = x * step, topY = y * step;
			int wx = sz;
			int wy = sz;
			if(topX < 0 ) {topX = 0; wx += topX;}
			if(topY < 0 ) {topY = 0; wy += topY;}
			if(topX + wx >= iimage->getWidth()) wx = iimage->getWidth() - topX - 1;
			if(topY + wy >= iimage->getHeight()) wy = iimage->getHeight() - topY - 1;

			if ( 0 == sz % 2 )
			{
				real_type white = intImage.getSum( topX, topY, wx, wy );
				resp( 0 ) = white - ( real_type )2.0 * intImage.getSum( topX, topY, step, wy );
				resp( 1 ) = white - ( real_type )2.0 * intImage.getSum( topX, topY, wx, step );
			}
			else
			{
				resp( 0 ) = intImage.getSum( x + 1, topY, step, wy ) - iimage->getSum( topX, topY, step, wy );
				resp( 1 ) = intImage.getSum( topX, y + 1, wx, step ) - iimage->getSum( topX, topY, wx, step );
			}
		}

		void SurfDescriptor::rotationMatrix2d( real_type angle, RotMatT &A ) const
		{
			real_type cos_ = cos( angle ), sin_ = sin( angle );
			A = ( real_type )cos_, -( real_type )sin_,
			    ( real_type )sin_, ( real_type )cos_;
		}

		bool SurfDescriptor::hasTable( int max, int s ) const
		{
			if ( s >= ( int )gaussTables.size() ) return false;
			AMatrix* table = gaussTables[s].get();
			if ( NULL == table || (int) table->cols() <= max || (int) table->rows() <= max )
				return false;
			return true;
		}

		void SurfDescriptor::prepareGaussTable( uint max, int s ) const
		{
			if ( s >= ( int )gaussTables.size() )
				gaussTables.resize( s + 1 );
			gaussTables[s] = boost::shared_ptr<AMatrix>( new AMatrix( max + 1, max + 1 ) );
			AMatrix& table = *gaussTables[s];
			for ( int y = 0; y <= ( int )max; ++y )
			{
				for ( int x = y; x <= ( int )max; ++x )
				{
					table( x, y ) = table( y, x ) = 1.0 / exp( ( x * x + y * y ) / ( 2.0 * s * s ) );
				}
			}
		}

		SurfDescriptor::real_type SurfDescriptor::gauss( int x, int y, int s ) const
		{
			assert( 0 < s );
			return ( *gaussTables[s] )( abs( x ), abs( y ) );
		}

	} // core
} // adore
