

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Andrei Tchouev


#ifndef ADORE_SURFACE_GEOMETRY2_H
#define ADORE_SURFACE_GEOMETRY2_H

#include <list>
#include <set>
#include <cmath>
#include <cassert>

namespace adore
{
	namespace core
	{
		/**	2D point
		*
		*	A standard mathematical 2 dimensional point with
		*	double precision coordinates.
		*
		*	 
		*/
		class Point2D
		{
				double v[2];			///< point coordinates

			public:
				/**	Default constructor
				*/
				Point2D()
				{
				};

				/**	Constructor with initialization
				*/
				Point2D( double x, double y )
				{
					v[0] = x;
					v[1] = y;
				};

				/**	Access point components as an array
				*/
				double& operator[]( int i )
				{
					return v[i];
				};

				/**	Access point components as an array
				*/
				const double operator[]( int i ) const
				{
					return v[i];
				};

				/**	Point metric (squared distance to the origin)
				*/
				double operator()() const
				{
					return v[0]*v[0] + v[1]*v[1];
				};

				/** Get mirror point about the origin
				*/
				Point2D operator-() const
				{
					return Point2D( -v[0],-v[1] );
				};

				/**	Vector-like dot product
				*/
				double operator*( const Point2D& a ) const
				{
					return v[0]*a.v[0]+v[1]*a.v[1];
				};

				/**	Vector-like point-scalar product
				*/
				Point2D operator*( double a )
				{
					return Point2D( v[0]*a,v[1]*a );
				};

				/**	Vector-like point-inverse scalar product
				*/
				inline Point2D operator/( const double a ) const
				{
					double inv_a = 1.0 / a;
					return Point2D( v[0]*inv_a,v[1]*inv_a );
				};

				/**	Vector-like point addition
				*/
				Point2D operator+( const Point2D& a ) const
				{
					return Point2D( v[0]+a.v[0],v[1]+a.v[1] );
				};

				/**	Vector-like point subtraction
				*/
				Point2D operator-( const Point2D& a ) const
				{
					return Point2D( v[0]-a.v[0],v[1]-a.v[1] );
				};

				/**	Point equality operator
				*/
				bool operator==( const Point2D& a ) const
				{
					return v[0]==a.v[0] && v[1]==a.v[1];
				};

				/**	Point inequality operator
				*/
				bool operator!=( const Point2D& a ) const
				{
					return v[0]!=a.v[0] || v[1]!=a.v[1];
				};
		};

		/**	Vector-like point-scalar product
		*/
		inline Point2D operator*( const double a, const Point2D b )
		{
			return Point2D( b[0]*a,b[1]*a );
		}

		/**	2D face
		*
		*	A standard mathematical 2 dimensional face, which is an
		*	edge on the surface.
		*
		*	@warning Vertex indices must not excceed the capacity of 16 bits
		*
		*	 
		*/
		class Face2D
		{
				unsigned v[2];				///< face vertices

				/**	Pack vertex indices in one 32 bit word
				*/
				unsigned pack() const
				{
					if ( v[0] < v[1] )
						return ( v[0] & 0x0000FFFF ) | ( v[1]<<16 );
					else
						return ( v[1] & 0x0000FFFF ) | ( v[0]<<16 );
				};

			public:
				/**	Default constructor
				*/
				Face2D()
				{
				};

				/**	Constructor with initialization
				*/
				Face2D( unsigned a, unsigned b )
				{
					v[0] = a;
					v[1] = b;
				};

				/**	Access face vertices as an array
				*/
				unsigned& operator[]( int i )
				{
					return v[i];
				};

				/**	Access face vertices as an array
				*/
				const unsigned operator[]( int i ) const
				{
					return v[i];
				};

				/**	Face equality operator
				*/
				bool operator==( const Face2D& rv ) const
				{
					return pack()==rv.pack();
				};

				/**	Face equality operator
				*/
				bool operator!=( const Face2D& rv ) const
				{
					return pack()!=rv.pack();
				};

				/**	Face comparison operator
				*/
				bool operator<( const Face2D& rv ) const
				{
					return pack() < rv.pack();
				};

				/**	Face packing for hash operation
				*/
				operator size_t() const
				{
					return static_cast<size_t>( pack() );
				};
		};

		/**	Determine the signed triangle area, depending on the triangle winding.
		*	@param src Source array of points
		*	@param a Index of the first triangle vertex in the source index
		*	@param b Index of the second triangle vertex in the source index
		*	@param c Index of the third triangle vertex in the source index
		*	@return Signed triangle area
		*/
		inline double area( Point2D* src, unsigned a, unsigned b, unsigned c )
		{
			return 0.5 * ( ( src[b][0]-src[a][0] )*( src[c][1]-src[a][1] ) - ( src[c][0]-src[a][0] )*( src[b][1]-src[a][1] ) );
		}

		/**	Determine the halfspace of a point relative to the face
		*	@param src Source array of points
		*	@param f Face with indices indicating points in the source array
		*	@param a Point for which to determine the halfspace
		*	@return -1 for negative, +1 for positiove halfspace, 0 if the point lies on line of the edge
		*/
		inline int halfspace( Point2D* src, Face2D f, Point2D a, double epsilon )
		{
			//alpha = (x2 - x1) * (y - y1) - (y2 - y1) * (x - x1)
			double dx = src[f[1]][0] - src[f[0]][0];
			double dy = src[f[1]][1] - src[f[0]][1];
			double x = a[0] - src[f[0]][0];
			double y = a[1] - src[f[0]][1];
			double alpha = dx * y - dy * x;

			if ( alpha > epsilon )
				return -1;
			else if ( alpha<-epsilon )
				return 1;
			else
				return 0;
		}

		/**	2D simplex
		*
		*	A standard mathematical 2 dimensional simplex, which is a
		*	triangle on the surface.
		*
		*	 
		*/
		class Simplex2D
		{
				unsigned v[3];

			public:
				/**	Default constructor
				*/
				Simplex2D()
				{
				};

				/**	Constructor with initialization
				*/
				Simplex2D( unsigned a, unsigned b, unsigned c )
				{
					v[0] = a;
					v[1] = b;
					v[2] = c;
				};

				/**	Access simplex vertices as an array
				*/
				unsigned& operator[]( int i )
				{
					return v[i];
				};

				/**	Access simplex vertices as an array
				*/
				const unsigned operator[]( int i ) const
				{
					return v[i];
				};

				/**	Test if the simplex is null.
				*	The simplex is null if any of it's two vertices are identical.
				*/
				bool nullSimplex()
				{
					return v[0]==v[1] || v[0]==v[2] || v[1]==v[2];
				};
		};

		/**	2D circle
		*
		*	 
		*/
		struct Circle2D
		{
			Point2D center;					///< circle center
			double radius;					///< circle radius
		};

		/**	Determine the radius of the circumcirlce of triangle
		*	@param src Source array of points
		*	@param a Index of the first triangle vertex in the source index
		*	@param b Index of the second triangle vertex in the source index
		*	@param c Index of the third triangle vertex in the source index
		*	@return circumradius
		*/
		inline double circumradius( Point2D* src, unsigned a, unsigned b, unsigned c )
		{
			assert( a!=b && b!=c && c!=a );

			double dca = ( src[c]-src[a] ) * ( src[b]-src[a] );
			double dba = ( src[c]-src[b] ) * ( src[a]-src[b] );
			double dcb = ( src[a]-src[c] ) * ( src[b]-src[c] );

			double radius = 0.5 * std::sqrt( ( dca+dba )*( dba+dcb )*( dcb+dca ) / ( dba*dcb+dcb*dca+dca*dba ) );

			return radius;
		}

		/**	Determine the center of the circumcirlce of triangle
		*	@param src Source array of points
		*	@param a Index of the first triangle vertex in the source index
		*	@param b Index of the second triangle vertex in the source index
		*	@param c Index of the third triangle vertex in the source index
		*	@return circumcenter
		*/
		inline Point2D circumcenter( Point2D* src, unsigned a, unsigned b, unsigned c )
		{
			assert( a!=b && b!=c && c!=a );

			double dca = ( src[c]-src[a] )*( src[b]-src[a] );
			double dba = ( src[c]-src[b] )*( src[a]-src[b] );
			double dcb = ( src[a]-src[c] )*( src[b]-src[c] );

			double n1 = dba * dcb;
			double n2 = dcb * dca;
			double n3 = dca * dba;

			Point2D center = 0.5 * ( ( n2+n3 )*src[a]+( n3+n1 )*src[b]+( n1+n2 )*src[c] )/( n1+n2+n3 );

			return center;
		}

		/**	Determine the circumcirlce of triangle
		*	@param src Source array of points
		*	@param a Index of the first triangle vertex in the source index
		*	@param b Index of the second triangle vertex in the source index
		*	@param c Index of the third triangle vertex in the source index
		*	@return circumcirlce
		*/
		inline Circle2D circumcircle( Point2D* src, unsigned a, unsigned b, unsigned c )
		{
			assert( a!=b && b!=c && c!=a );

			Circle2D result;

			double dca = ( src[c]-src[a] ) * ( src[b]-src[a] );
			double dba = ( src[c]-src[b] ) * ( src[a]-src[b] );
			double dcb = ( src[a]-src[c] ) * ( src[b]-src[c] );

			double n1 = dba * dcb;
			double n2 = dcb * dca;
			double n3 = dca * dba;

			result.center = 0.5 * ( ( n2+n3 )*src[a]+( n3+n1 )*src[b]+( n1+n2 )*src[c] )/( n1+n2+n3 );
			result.radius = 0.5 * std::sqrt( ( dca+dba )*( dba+dcb )*( dcb+dca ) / ( dba*dcb+dcb*dca+dca*dba ) );

			return result;
		}

		/**	Active face list that stores its members in STL-type set
		*
		*	 
		*/
		template <class T> class ActiveFaceList2D
		{
				T storage;					///< face storing container
				typename T::iterator it;	///< container iterator

			public:
				/**	Set container iterator to the beginning of the container
				*/
				void start()
				{
					it = storage.begin();
				};

				/**	Increment container iterator and return if the end of the container is not reached yet
				*	@param f current iterated element, if the end is not reached
				*/
				bool next( Face2D& f )
				{
					if ( it==storage.end() )
						return false;
					else
					{
						f = *it;
						++it;
						return true;
					}
				};

				/**	Insert face into the set
				*	@param f Face to insert
				*/
				void insert( Face2D f )
				{
					storage.insert( f );
				};

				/**	Remove and return the first element
				*/
				Face2D extract()
				{
					typename T::iterator iter = storage.begin();
					Face2D result = *iter;
					storage.erase( iter );
					return result;
				};

				/**	Remove face from the set
				*	@param f Face to remove
				*/
				void erase( Face2D f )
				{
					storage.erase( f );
				};

				/**	Test if the face is already in the set
				*	@param f Face to test
				*/
				bool member( Face2D f )
				{
					return storage.find( f )!=storage.end();
				};

				/**	Update face in the set.
				*	The face is inserted into the set if it is not in the set
				*	The face is removed from the set if it is already in the set
				*	@param f Face to update
				*/
				void update( Face2D f )
				{
					typename T::iterator iter = storage.find( f );
					if ( iter==storage.end() )
						storage.insert( f );
					else
						storage.erase( iter );
				};

				/**	Get number of elements
				*/
				unsigned size()
				{
					return static_cast<unsigned>( storage.size() );
				};

				/**	Test if the set is empty
				*/
				bool empty()
				{
					return storage.empty();
				};
		};

		/**	Simplex list that stores its members in STL-type sequence container
		*
		*	 
		*/
		template <class T = std::set<Face2D> > class SimplexStore2D
		{
				T storage;					///< face storing container
				typename T::iterator it;	///< container iterator

			public:
				/**	Set container iterator to the beginning of the container
				*/
				void start()
				{
					it = storage.begin();
				};

				/**	Increment container iterator and return if the end of the container is not reached yet
				*	@param f current iterated element, if the end is not reached
				*/
				bool next( Simplex2D& t )
				{
					if ( it==storage.end() )
						return false;
					else
					{
						t = *it;
						++it;
						return true;
					}
				};

				/**	Insert the simplex into the list
				*	@param t Simplex to add
				*/
				void insert( Simplex2D& t )
				{
					storage.push_back( t );
				};
		};

		/**	Simplex list that outputs its members directly to the receiver
		*
		*	 
		*/
		class SimplexCallback2D
		{
				typedef void ( *callback )( unsigned* vcallback, void* extra );	///< Triangle callback function prototype
				callback receiver;						///< Triangle callback function
				void* eparam;							///< Extra parameter for callback function

			public:
				/**	Constructor
				*	@param cb Triangle callback function
				*	@param extra Extra parameter for callback function
				*/
				SimplexCallback2D( callback cb, void* extra )
						:receiver( cb ),eparam( extra ) {};

				/**	Insert the simplex into the list
				*	@param t Simplex to add
				*/
				void insert( Simplex2D& t )
				{
					receiver( ( unsigned* )&t,eparam );
				};
		};

		/**	Type definitions
		*/
		typedef ActiveFaceList2D< std::set<Face2D> > FaceList2D;
		typedef SimplexStore2D< std::list<Simplex2D> > SimplexStoreList2D;
		typedef SimplexCallback2D SimplexList2D;
	}
}

#endif	// ADORE_SURFACE_GEOMETRY2_H
