

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_AARRAY_H
#define ADORE_AARRAY_H

#include "aarrayrange.h"
#include "aarrayinitializer.h"
#include "aarrayview.h"

#include <boost/multi_array.hpp>

#include <boost/static_assert.hpp>

#include <iostream>

namespace adore {
namespace core {

template<typename T, int rank> class AArray;

/** ***************** IteratorTypeGen ******************* **/
template <typename T, int rank> struct IteratorTypeGen
{
    typedef T * iterator;
    typedef const T * const_iterator;

    static iterator begin(boost::multi_array<T, rank>& a) { return a.origin(); }
    static const_iterator begin(const boost::multi_array<T, rank>& a) { return a.origin(); }
    static iterator end(boost::multi_array<T, rank>& a) { return a.origin() + a.num_elements(); }
    static const_iterator end(const boost::multi_array<T, rank>& a) { return a.origin() + a.num_elements(); }
};
/** ***************** IteratorTypeGen ******************* **/




/** ***************** AArray ******************* **/

template<typename T, int rank> class AArray
{
public:
    typedef T scalar_type;
    enum { get_rank = rank };
    typedef scalar_type& reference;
	typedef const scalar_type& const_reference;
	typedef boost::multi_array<T, rank> array_type;
    typedef AArray<T, rank> this_type;
    typedef typename array_type::size_type size_type;

    typedef typename IteratorTypeGen<T, rank>::iterator iterator;
    typedef typename IteratorTypeGen<T, rank>::const_iterator const_iterator;

private:
    array_type m_data;

public:

    AArray() : m_data() {}

    explicit AArray(const std::vector<size_type>& extents) :
        m_data( extents ) { }

    explicit AArray( size_type sz1 ) :
        m_data( boost::extents[sz1] )
    {
        BOOST_STATIC_ASSERT(rank == 1);
    }

    explicit AArray( size_type sz1, size_type sz2 ) :
        m_data( boost::extents[sz1][sz2] )
    {
        BOOST_STATIC_ASSERT(rank == 2);
    }

    explicit AArray( size_type sz1, size_type sz2, size_type sz3 ) :
        m_data( boost::extents[sz1][sz2][sz3] )
    {
        BOOST_STATIC_ASSERT(rank == 3);
    }

	AArray( const AArray& rhs ) :
		m_data( rhs.m_data )
	{
//		BOOST_STATIC_ASSERT(rank == 33);
	}

	AArray& operator=( const AArray& rhs )
	{
		//		BOOST_STATIC_ASSERT(rank == 33);
		m_data = rhs.internal_array();
		return *this;
	}

	template <typename A> AArray( const AArrayView<A, rank>& view) :
		m_data( view.internal_array() )
    {
		// TODO
//		BOOST_STATIC_ASSERT(rank == 33);
    }

	template <typename A> this_type& operator=( const AArrayView<A, rank>& view )
	{
		internal_array() = view.internal_array();
		return *this;
	}

    const size_type* shape() const { return m_data.shape(); }

    void swap( this_type & rhs )
    {
		std::swap(m_data, rhs.data());
    }

    array_type& internal_array() { return m_data; }
    const array_type& internal_array() const { return m_data; }

    scalar_type* data() { return m_data.data(); }
    const scalar_type* data() const { return m_data.data(); }

    size_type size() const { return m_data.num_elements(); }

    size_type size1() const { return m_data.shape()[0]; }

    size_type rows() const
    {
        BOOST_STATIC_ASSERT(rank == 1 || rank == 2);
        return m_data.shape()[0];
    }

    size_type size2() const
    {
        BOOST_STATIC_ASSERT(rank > 1);
        return m_data.shape()[1];
    }

    size_type cols() const
    {
        BOOST_STATIC_ASSERT(rank == 1 || rank == 2);
        return m_data.shape()[1];
    }

	size_type size3() const
	{
		BOOST_STATIC_ASSERT(rank > 2);
		return m_data.shape()[2];
	}

	void resize(size_type sz)
    {
        BOOST_STATIC_ASSERT(rank == 1);
		resizeAndPreserve(sz); // :( copy on resize
    }

	void resizeAndPreserve(size_type sz)
	{
		BOOST_STATIC_ASSERT(rank == 1);
		m_data.resize(boost::extents[sz]);
	}

	void resize(size_type sz1, size_type sz2)
    {
        BOOST_STATIC_ASSERT(rank == 2);
		resizeAndPreserve(sz1, sz2); // :( copy on resize
    }

	void resizeAndPreserve(size_type sz1, size_type sz2)
	{
		BOOST_STATIC_ASSERT(rank == 2);
		m_data.resize(boost::extents[sz1][sz2]);
	}

	void resize(size_type sz1, size_type sz2, size_type sz3)
	{
		BOOST_STATIC_ASSERT(rank == 3);
		resizeAndPreserve(sz1, sz2, sz3); // :( copy on resize
	}

	void resizeAndPreserve(size_type sz1, size_type sz2, size_type sz3)
	{
		BOOST_STATIC_ASSERT(rank == 3);
		m_data.resize(boost::extents[sz1][sz2][sz3]);
	}

    AArrayView<this_type, 1> operator() (const Range& r)
    {
        BOOST_STATIC_ASSERT(rank == 1);
        return AArrayView<this_type, 1>(*this, r);
    }

    AArrayView<const this_type, 1> operator() (const Range& r) const
    {
        BOOST_STATIC_ASSERT(rank == 1);
        return AArrayView<const this_type, 1>(*this, r);
    }

    AArrayView<this_type, 2> operator() (const Range& r0, const Range& r1)
    {
        BOOST_STATIC_ASSERT(rank == 2);
        return AArrayView<this_type, 2>(*this, r0, r1);
    }

    AArrayView<const this_type, 2> operator() (const Range& r0, const Range& r1) const
    {
        BOOST_STATIC_ASSERT(rank == 2);
        return AArrayView<const this_type, 2>(*this, r0, r1);
    }

    AArrayView<this_type, 1> operator() (const Range& r0, int i1)
    {
        BOOST_STATIC_ASSERT(rank == 2);
        return AArrayView<this_type, 1>(*this, r0, i1);
    }

    AArrayView<const this_type, 1> operator() (const Range& r0, int i1) const
    {
        BOOST_STATIC_ASSERT(rank == 2);
        return AArrayView<const this_type, 1>(*this, r0, i1);
    }

    AArrayView<this_type, 1> operator() (int i0, const Range& r1)
    {
        BOOST_STATIC_ASSERT(rank == 2);
        return AArrayView<this_type, 1>(*this, i0, r1);
    }

    AArrayView<const this_type, 1> operator() (int i0, const Range& r1) const
    {
        BOOST_STATIC_ASSERT(rank == 2);
        return AArrayView<const this_type, 1>(*this, i0, r1);
    }

    AArrayView<const this_type, 2> operator() (int i0, const Range& r1, const Range& r2) const
    {
        BOOST_STATIC_ASSERT(rank == 3);
        return AArrayView<const this_type, 2>(*this, i0, r1, r2);
    }

    Initializer<this_type> operator=(const scalar_type val)
    {
        return Initializer<this_type>((*this), val);
    }

	reference operator() (size_type i) { BOOST_STATIC_ASSERT(rank == 1); return m_data[i]; }
	const_reference operator() (size_type i) const { BOOST_STATIC_ASSERT(rank == 1); return m_data[i]; }
    reference operator() (size_type i0, size_type i1) { BOOST_STATIC_ASSERT(rank == 2); return m_data[i0][i1]; }
	const_reference operator() (size_type i0, size_type i1) const { BOOST_STATIC_ASSERT(rank == 2); return m_data[i0][i1]; }
    reference operator() (size_type i0, size_type i1, size_type i2) { BOOST_STATIC_ASSERT(rank == 3); return m_data[i0][i1][i2]; }
	const_reference operator() (size_type i0, size_type i1, size_type i2) const { BOOST_STATIC_ASSERT(rank == 3); return m_data[i0][i1][i2]; }

    void write_at(std::ostream &os) const
    {
        std::copy((*this).begin(), (*this).end(), std::ostream_iterator<T>(os, " "));
        std::cout << std::endl;
    }

    iterator begin() { return IteratorTypeGen<T, rank>::begin( m_data ); }
    const_iterator begin() const { return IteratorTypeGen<T, rank>::begin( m_data ); }
    iterator end() { return IteratorTypeGen<T, rank>::end( m_data ); }
    const_iterator end() const { return IteratorTypeGen<T, rank>::end( m_data ); }

    this_type operator-() const
    {
        std::vector<size_t> ex; ex.assign( shape(), shape() + rank );
        this_type result( ex );
        const_iterator ita = begin(), enda = end();
        iterator itb = result.begin();
        for( ; ita != enda; ++ita, ++itb)
            (*itb) = -(*ita);
        return result;
    }

    void setIdentity()
    {
        ( *this ) = 0;
        for ( unsigned d = 0; d < std::min( rows(), cols() ); ++d )
        {
            ( *this )( d, d ) = 1;
        }
    }

	/**
	* calculates the sqare distance to another array
	*/
	T sqrDist( const this_type & arr2 ) const
    {
        assert( size() == arr2.size());

        T result = 0;
        const_iterator arr1It = begin(), arr1End = end();
        const_iterator arr2It = arr2.begin();
        for ( ; arr1It != arr1End; ++arr1It, ++arr2It )
        {
            T dist = *arr1It - *arr2It;
            result += dist * dist;
        }
        return result;
    }

    this_type& operator /= (const scalar_type val)
    {
        typename this_type::iterator it = begin(), _end = end();
        for ( ; it != _end; ++it )
            *it /= val;
        return *this;
    }

    this_type& operator *= (const scalar_type val)
    {
        typename this_type::iterator it = begin(), _end = end();
        for ( ; it != _end; ++it )
            *it *= val;
        return *this;
    }

    this_type& operator += (const this_type& val)
    {
            typename this_type::iterator it1 = begin(), _end = end();
            typename this_type::const_iterator it2 = val.begin();
            for ( ; it1 != _end; ++it1, ++it2 )
            {
                    *it1 += *it2;
            }
            return *this;
    }

	/**
	 * subtracts one array from another storing the result in an already allocated vector
	*/
	void sub2( const this_type &arr2, this_type &result ) const
    {
            const this_type & arr1 = *this;
            typename this_type::const_iterator arr1It = arr1.begin();
            typename this_type::const_iterator arr2It = arr2.begin();
            typename this_type::iterator outIt = result.begin(), outEnd = result.end();
            for ( ; outIt != outEnd; ++outIt, ++arr1It, ++arr2It )
            {
                    *outIt = *arr1It - *arr2It;
            }
    }

};

/** ***************** AArray ******************* **/

template<typename T, int rank>
std::ostream & operator<<(std::ostream & os, const AArray<T, rank> & arr)
{
    arr.write_at(os);
    return os;
}

template<typename T, int rank>
void sub(const AArray<T, rank>& a, const AArray<T, rank>& b, AArray<T, rank>& c)
{
    assert( (a.size() == b.size()) && (a.size() == c.size()) );
    typename AArray<T, rank>::const_iterator ita = a.begin(), enda = a.end(), itb = b.begin();
    typename AArray<T, rank>::iterator itc = c.begin();
    for( ; ita != enda; ++ita, ++itb, ++itc)
        *itc = (*ita) - (*itb);
}

template<typename T, int rank>
AArray<T, rank> operator-(const AArray<T, rank>& a, const AArray<T, rank>& b)
{
    assert(a.size() == b.size());
    std::vector<size_t> ex; ex.assign( a.shape(), a.shape() + rank );
    AArray<T, rank> result(ex);
    sub(a, b, result);
    return result;
}

template<typename A, typename B, typename C>
void sum(const A& a, const B& b, C& c)
{
    assert( (a.size() == b.size()) && (a.size() == c.size()) );
    typename A::const_iterator ita = a.begin(), enda = a.end();
    typename B::const_iterator itb = b.begin();
    typename C::iterator itc = c.begin();
    for( ; ita != enda; ++ita, ++itb, ++itc)
            *itc = (*ita) + (*itb);
}

template<typename T, int rank>
AArray<T, rank> operator+(const AArray<T, rank>& a, const AArray<T, rank>& b)
{
    assert(a.size() == b.size());
    std::vector<size_t> ex; ex.assign( a.shape(), a.shape() + rank );
    AArray<T, rank> result(ex);
    sum(a, b, result);
    return result;
}

template<typename A, typename T, int rank>
AArray<T, rank> operator+(const AArray<T, rank>& a, const AArrayView<A, rank>& b)
{
    assert(a.size() == b.size());
    std::vector<size_t> ex; ex.assign( a.shape(), a.shape() + rank );
    AArray<T, rank> result(ex);
    sum(a, b, result);
    return result;
}

template<typename A, typename B, typename C>
void mul(const A& a, const B& b, C& c)
{
    assert( (a.size() == b.size()) && (a.size() == c.size()) );
    typename A::const_iterator ita = a.begin(), enda = a.end();
    typename B::const_iterator itb = b.begin();
    typename C::iterator itc = c.begin();
    for( ; ita != enda; ++ita, ++itb, ++itc)
            *itc = (*ita) * (*itb);
}

template<typename T, int rank>
AArray<T, rank> operator*(const AArray<T, rank>& a, const AArray<T, rank>& b)
{
    assert(a.size() == b.size());
    std::vector<size_t> ex; ex.assign( a.shape(), a.shape() + rank );
    AArray<T, rank> result(ex);
    mul(a, b, result);
    return result;
}

template<typename A, typename T, int rank>
AArray<T, rank> operator*(const AArray<T, rank>& a, const AArrayView<A, rank>& b)
{
    assert(a.size() == b.size());
    std::vector<size_t> ex; ex.assign( a.shape(), a.shape() + rank );
    AArray<T, rank> result(ex);
    mul(a, b, result);
    return result;
}

template<typename T, int rank>
AArray<T, rank> operator*(const AArray<T, rank>& a, const T factor)
{
    std::vector<size_t> ex; ex.assign( a.shape(), a.shape() + rank );
    AArray<T, rank> result(ex);
    typename AArray<T, rank>::const_iterator ita = a.begin(), enda = a.end();
    typename AArray<T, rank>::iterator itr = result.begin();
    for( ; ita != enda; ++ita, ++itr)
        *itr = *ita * factor;
    return result;
}

template<typename T, int rank>
AArray<T, rank> operator*(const int factor, const AArray<T, rank>& a)
{
    return a * (T)factor;
}

template<typename T, int rank>
AArray<T, rank> operator*(const T factor, const AArray<T, rank>& a)
{
    return a * factor;
}

template <typename T, int rank>
inline bool isNullArray( const AArray<T, rank>& a )
{
    return ( a.size() == 0 );
}

template <typename T, int rank>
inline void setNullArray( AArray<T, rank>& a )
{
	assert(false);
}

template <typename T>
inline void setNullArray( AArray<T, 1>& a )
{
	a.resize(0);
}

template <typename T>
inline void setNullArray( AArray<T, 2>& a )
{
	a.resize(0, 0);
}

template <typename T>
inline void setNullArray( AArray<T, 3>& a )
{
	a.resize(0, 0, 0);
}

template<typename T>
T dot(const AArray<T, 1>& a, const AArray<T, 1>& b)
{
    T result = 0;
    typename AArray<T, 1>::const_iterator arr1It = a.begin(), arr1End = a.end();
    typename AArray<T, 1>::const_iterator arr2It = b.begin();
    for ( ; arr1It != arr1End; ++arr1It, ++arr2It )
    {
        result += *arr1It * *arr2It;
    }
    return result;
}

/**
   * tronspose a matrix.
   * The transpose doesn't move the values, which is bad when using lapack
   */
template<typename T> inline void transInPlace( AArray<T, 2> &m )
{
    assert( m.rows() == m.cols() );

    for ( unsigned i = 0; i < m.rows() - 1; ++i )
        for ( unsigned j = i+1; j < m.cols(); ++j )
        {
            double temp = m( i, j );
            m( i, j ) = m( j, i );
            m( j, i ) = temp;
        }
}

/**
   * @return a transposed version of the matrix
   */
template<typename T> inline AArray<T, 2> trans( const AArray<T, 2> &m )
{
    AArray<T, 2> result( m.cols(), m.rows() );

    for ( unsigned i = 0; i < m.rows(); ++i )
        for ( unsigned j = 0; j < m.cols(); ++j )
        {
            result( j, i ) = m( i, j );
        }

    return result;
}

/**
   * @return transpose a matrix
   */
template<typename T> inline void trans( const AArray<T, 2> &m, AArray<T, 2> &result )
{
    for ( unsigned i = 0; i < m.rows(); ++i )
        for ( unsigned j = 0; j < m.cols(); ++j )
        {
            result( j, i ) = m( i, j );
        }
}

typedef AArray<double, 1> AVector;
typedef AArray<double, 1> AVector2D;
typedef AArray<double, 1> AVector3D;
typedef AArray<double, 2> AMatrix;
typedef AArray<double, 3> ATensor3;

typedef AArray<float, 1> AVectorf;
typedef AArray<double, 1> AVectord;

typedef AArray<float, 2> AMatrixf;
typedef AArray<double, 2> AMatrixd;

typedef AArrayView<AVector, 1> AVectorView;
typedef AArrayView<AMatrix, 2> AMatrixView;

extern AMatrixf Identity2x2f;
extern AMatrixd Identity2x2d;

} // core
} // adore

#endif // ADORE_AARRAY_H
