

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_TYPESBIJECTIVITY_H
#define ADORE_TYPESBIJECTIVITY_H

#include <adrcore/types/inorderset.h>
#include <map>
#include <cassert>

namespace adore
{
	namespace core
	{

		using namespace std;

		/**
		 * A Bijectivity is a one-to-one-relation between an entity class A and an entity class B.
		 *  
		*/
		template<class A, class B, class ComparatorA = less<A>, class ComparatorB = less<B> >
		class Bijectivity
		{
			public:

//				typedef InOrderSet< A, ComparatorA > InOrderSetAT;
//				typedef InOrderSet< B, ComparatorB > InOrderSetBT;
				typedef InOrderSet< A> InOrderSetAT;
				typedef InOrderSet< B> InOrderSetBT;

				typedef typename InOrderSetAT::ConstIterator AIterator;
				typedef typename InOrderSetBT::ConstIterator BIterator;

			public:
				/**
				 * Constructor.
				 * @param aNull the Value that should be returned if a entity of class A does not exist.
				 * @param bNull the Value that should be returned if a entity of class A does not exist.
				 */
				Bijectivity( A aNull, B bNull )
				{
					this->aNull = aNull;
					this->bNull = bNull;
				}

				/**
				 * Destructor.
				 */
				~Bijectivity() {;}

				/**
				 * Add a one-to-one-relation to this Bijectivity.
				 * @param a an entity of class A
				 * @param b an entity of class B
				 */
				void add( A a, B b )
				{
//					not symmetric, removed TODO check if this causes bugs
//					AIterator it = As.find( a );
//					if ( it != As.getEnd() ) return;

					aToB[a] = b;
					bToA[b] = a;
					As.insert( a );
					Bs.insert( b );
				}

				void remove( A a )
				{
					AIterator it = As.find( a );
					if ( it == As.getEnd() ) return;
					B b = getB( a );
					remove( a, b );
				}

				void remove( B b )
				{
					BIterator it = Bs.find( b );
					if ( it == Bs.getEnd() ) return;
					B a = getA( b );
					remove( a, b );
				}

				/**
				 * Get an iterator to the begin of all entities of class A.
				 * @return the iterator.
				 */
				AIterator getAsBegin() const
				{
					return As.getBegin();
				}

				/**
				 * Get an iterator to the end of all entities of class A.
				 * @return the iterator.
				 */
				AIterator getAsEnd() const
				{
					return As.getEnd();
				}

				/**
				 * Get an iterator to the begin of all entities of class B.
				 * @return the iterator.
				 */
				BIterator getBsBegin() const
				{
					return Bs.getBegin();
				}

				/**
				 * Get an iterator to the end of all entities of class B.
				 * @return the iterator.
				 */
				BIterator getBsEnd() const
				{
					return Bs.getEnd();
				}

				/**
				 * Get the entity a, that relates to b.
				 * @param b an entity of class B
				 * @return an entity of class A
				 */
				const A & getA( const B & b ) const
				{
					typename map<B, A>::const_iterator it = bToA.find( b );
					if ( it == bToA.end() )
						return aNull;
					return it->second;
				}

				/**
				 * Get the entity b, that relates to a.
				 * @param a an entity of class A
				 * @return an entity of class B
				 */
				const B & getB( const A & a ) const
				{
					typename map<A, B>::const_iterator it = aToB.find( a );
					if ( it == aToB.end() )
						return bNull;
					return it->second;
				}

				/**
				 * Get the entity b, that relates to a.
				 * @param a an entity of class A
				 * @return an entity of class B
				 */
				const B & operator[]( const A & a ) const
				{
					return getB( a );
				}

				/**
				 * Get the entity a, that relates to b.
				 * @param b an entity of class B
				 * @return an entity of class A
				 */
				const A & operator[]( const B & b ) const
				{
					return getA( b );
				}

				/**
				 * deletes all relations.
				 */
				void clear()
				{
					As.clear();
					Bs.clear();
					aToB.clear();
					bToA.clear();
				}

				/**
				 * returns the number of relations.
				 * @return the number of relations.
				 */
				int getSize() const
				{
					assert( As.getSize() == Bs.getSize() );
					return As.getSize();
				}

			protected:
				A aNull;
				B bNull;

				InOrderSetAT As;
				InOrderSetBT Bs;
				map<A, B> aToB;
				map<B, A> bToA;

				void remove( A a, B b )
				{
					typename map<A, B>::iterator it1 = aToB.find( a );
					aToB.erase( it1 );
					typename map<B, A>::iterator it2 = bToA.find( b );
					bToA.erase( it2 );

					As.remove( a );
					Bs.remove( b );
				}
		};

	}

}

#endif
