

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include "camera.h"
#include <adrcore/util/printmat.h>
#include <adrcore/util/floatshift.h>
#include <adrcore/util/storageorder.h>

namespace adore
{

	namespace core
	{

		Camera::Camera() :
				projMat ( 3, 4 ),
				//projMat (camData, 3, 4),
				K ( 3, 3 ),
				Kinv ( 3, 3 ),
				R ( 3, 3 ),
				loc ( 4 ),
				temp33 ( 3, 3 ), temp3 ( 3 )
		{
			loc ( 3 ) = 1.0;
			setIdentity ( projMat );
		}

		Camera::Camera ( const Camera &c ) :
                projMat ( c.getProjMat() ),
				K ( 3, 3 ),
				Kinv ( 3, 3 ),
				R ( 3, 3 ),
				loc ( 4 ),
				temp33 ( 3, 3 ), temp3 ( 3 )
		{
//			printf( "copying cam\n" );
			loc ( 3 ) = 1.0;
		}

		Camera::Camera ( const AMatrix &c ) :
                projMat ( c ),
				K ( 3, 3 ),
				Kinv ( 3, 3 ),
				R ( 3, 3 ),
				loc ( 4 ),
				temp33 ( 3, 3 ), temp3 ( 3 )
		{
			loc ( 3 ) = 1.0;
		}

		Camera::~Camera()
		{
		}

		void Camera::compose()
		{
            AMatrix sub(3, 3);
            AVector subcol(3);
			AVector2D subLoc = loc ( Range ( 0, 2 ) );

			matMul2 ( K, R, sub );
			getProjMat() ( Range::all(), Range ( 0, 2 ) ) = sub;

			matMul2 ( sub, subLoc, subcol, -1.0 );
			getProjMat() ( Range::all(), 3 ) = subcol;
		}

		void Camera::decompose() const
		{
			assert ( loc ( 3 ) == 1.0 );

			temp33 = getProjMat() ( Range::all(), Range ( 0, 2 ) );

			rq ( temp33, K, R );
			makeDiagOfROfRQPosisitve ( K, R );

			Kinv = K;
			invertUpperTriangular3x3Matrix ( Kinv );

			AVector subcol = getProjMat() ( Range::all(), 3 );
			matMul2 ( Kinv, subcol, temp3 );

			AVector2D subLoc(3);
			matMul2_AT ( R, temp3, subLoc, -1.0 );//trans of R is the inverse of R
			loc ( Range ( 0, 2 ) ) = subLoc;

			normalizeK ( K );
		}

		void Camera::normalizeK ( AMatrix &K ) /*const*/
		{
			double div = K ( 2, 2 );
			assert ( fabs( div ) > 0 );
			if ( fabs( div ) < FLT_EPSILON ) return;

			div = 1.0 / div;
			for ( int row = 0; row < 3; ++row )
			{
				for ( int col = row; col < 3; ++col )
				{
					K ( row, col ) = K ( row, col ) * div;
				}
			}

		}

		const AVector Camera::calculateCameraCenter()
		{
			normalizeExponents ( projMat.begin(), projMat.end() );
			return getVectorOfMinors3x4 ( projMat );
		}

		void Camera::calculateCameraCenter ( AVector &result )
		{
			normalizeExponents ( projMat.begin(), projMat.end() );
			getVectorOfMinors3x4 ( projMat, result );
		}


		void Camera::getValuesForDepthCalculation ( const AVector &pt, double &detM, double &T, double &w, double &m ) const
		{
			AMatrix sub = getProjMat() ( Range::all(), Range ( 0, 2 ) );

			detM = det3 ( sub );
			m = sqrt ( sub ( 2, 0 ) * sub ( 2, 0 ) + sub ( 2, 1 ) * sub ( 2, 1 ) + sub ( 2, 2 ) * sub ( 2, 2 ) );

            const AMatrix & PM = getProjMat();
			w = PM ( 2, 0 ) * pt ( 0 ) + PM ( 2, 1 ) * pt ( 1 ) + PM ( 2, 2 ) * pt ( 2 ) + PM ( 2, 3 ) * pt ( 3 );
			T = pt ( 3 );
		}

		double Camera::getDepth ( const AVector &pt ) const
		{
			double detM, T, w, m;
			getValuesForDepthCalculation ( pt, detM, T, w, m );
			return signum ( detM ) * w / T / m;
		}

		int Camera::getSign ( const AVector &pt ) const
		{
			double detM, T, w, m;
			getValuesForDepthCalculation ( pt, detM, T, w, m );
			return int ( signum ( detM * T * w ) );
		}


		bool isIdentityCam ( Camera *c )
		{
			assert ( c );
            AMatrix & cam = c->getProjMat();
			return isIdentityCam ( cam );
		}

        bool isIdentityCam ( const AMatrix &cam )
		{
			if ( cam ( 0, 0 ) != 1 ) return false;
			if ( cam ( 0, 1 ) != 0 ) return false;
			if ( cam ( 0, 2 ) != 0 ) return false;
			if ( cam ( 0, 3 ) != 0 ) return false;
			if ( cam ( 1, 0 ) != 0 ) return false;
			if ( cam ( 1, 1 ) != 1 ) return false;
			if ( cam ( 1, 2 ) != 0 ) return false;
			if ( cam ( 1, 3 ) != 0 ) return false;
			if ( cam ( 2, 0 ) != 0 ) return false;
			if ( cam ( 2, 1 ) != 0 ) return false;
			if ( cam ( 2, 2 ) != 1 ) return false;
			if ( cam ( 2, 3 ) != 0 ) return false;
			return true;
		}

	}
}

