

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_COREDESCRIPTORBACKREF_H
#define ADORE_COREDESCRIPTORBACKREF_H

#include "featuredescriptor.h"

namespace adore
{

	namespace core
	{

		template<class BackT> class BackRefRuler;

		/**
		 *	This implements a Descriptorvector-To-Feature back reference
		 * this is usefull in conjunction with Features and KD-Trees
			 
		*/
		template <class BackT>
		class DescriptorBackRef
		{
			private:

			public:
				typedef BackRefRuler<BackT> Ruler;
				typedef FeatureDescriptor::T_numtype T_numtype;
				typedef FeatureDescriptor::DescVecT* DescVecT;
				typedef T_numtype value_type;

				DescVecT vec;
				BackT backRef;

			public:

				DescriptorBackRef ( int vecsize, BackT bRef ) :vec ( vecsize ),backRef ( bRef ) {;}

				DescriptorBackRef ( DescVecT v, BackT bRef ) :vec ( v ), backRef ( bRef ) {;}

				DescriptorBackRef ( const DescriptorBackRef &br ) :vec ( br.vec ), backRef ( br.backRef ) {;}

				const DescriptorBackRef &operator= ( DescriptorBackRef &v )
				{
					vec = v.vec; // vec.reference ( v.vec );
					backRef = v.backRef;
					return *this;
				}

				T_numtype &operator[] ( int idx )
				{
					assert_msg ( idx < (int) vec->size(), "%d - %zd\n", idx, vec->size() );
					return (*vec) ( idx );
				}

				const T_numtype &operator[] ( int idx ) const
				{
					assert_msg ( idx < (int) vec->size(), "const %d - %zd\n", idx, vec->size() );
					return (*vec) ( idx );
				}

				uint size() const { return vec->size(); }

				void swap ( DescriptorBackRef &other )
				{
					std::swap ( backRef, other.backRef );
					std::swap ( vec, other.vec ); //vec.swap ( other.vec );
				}
		};

		template <class BackT>
		class BackRefRuler
		{
			public:
				typedef float distance_type;
				const int dims;

			public:
				BackRefRuler ( int d ) :dims ( d ) {;}

				float sqrDist ( const DescriptorBackRef<BackT> &v1, const DescriptorBackRef<BackT> &v2 ) const
				{
					float result = 0;
					for ( int i = dims - 1; i >= 0; --i )
					{
						float dist = v1[i] - v2[i];
						result += dist * dist;
					}
					return result;
				}
		};

		template <class BackT>
		inline void swap ( DescriptorBackRef<BackT> &v1, DescriptorBackRef<BackT> &v2 )
		{
			v1.swap ( v2 );
		}

		///// Adding backrefs to featurePoints
		///// A FeaturePoint* is added as the backref

		template<class Collection>
		inline void addBackRefs ( FeaturePoint *fp, const int descrId, Collection &coll )
		{
			assert ( fp );
			FeatureDescriptor *fd = fp->getDescriptor ( descrId );
			if ( fd )
			{
				for ( uint j = 0; j < fd->size(); ++j )
				{
					coll.push_back ( DescriptorBackRef<FeaturePoint *> ( ( *fd ) [j], fp ) );
				}
			}
		}

		template<class Collection>
		inline void addBackRefs ( ViewBase *v, const int featsId, const int descrId, Collection &coll )
		{
			assert ( v );
			for ( ViewBase::PointsIterator fit = v->getPointsBegin ( featsId ); fit != v->getPointsEnd ( featsId ); ++fit )
			{
				addBackRefs ( *fit, descrId, coll );
			}
		}

		template<class ViewIterator, class Collection>
		inline void addBackRefs ( ViewIterator vit, const int size, const int featsId, const int descrId, Collection &coll )
		{
			for ( int i = 0; i < size; ++i, ++vit )
			{
				ViewBase *v = *vit;
				addBackRefs ( v, featsId, descrId, coll );
			}
		}

		///// Adding backrefs to view index (this is used in image arranger)
		///// the integer value of the control varable of the for loop is added as a backref

		template<class Collection>
		inline void addBackRefIndices ( FeaturePoint *fp, const int idx, const int descrId, Collection &coll )
		{
			assert ( fp );
			FeatureDescriptor *fd = fp->getDescriptor ( descrId );
			if ( fd )
			{
				for ( uint j = 0; j < fd->size(); ++j )
				{
					coll.push_back ( DescriptorBackRef<int> ( &(*fd)[j], idx ) );
				}
			}
		}
		template<class Collection>
		inline void addBackRefIndices ( ViewBase *v, const int idx, const int featsId, const int descrId, Collection &coll )
		{
			assert ( v );
			for ( ViewBase::PointsIterator fit = v->getPointsBegin ( featsId ); fit != v->getPointsEnd ( featsId ); ++fit )
			{
				addBackRefIndices ( *fit, idx, descrId, coll );
			}
		}

		template<class ViewIterator, class Collection>
		inline void addBackRefIndices ( ViewIterator vit, const int size, const int featsId, const int descrId, Collection &coll )
		{
			for ( int i = 0; i < size; ++i, ++vit )
			{
				ViewBase *v = *vit;
				addBackRefIndices ( v, i, featsId, descrId, coll );
			}
		}


		/**
		 * helper function to determin the dimension of the descriptors of the features inside a view.
		 */
		inline int getDescriptorDim ( ViewBase *v, const int featsId, const int descrId )
		{
			int dim = 0;
			if ( v && v->getNumFeaturePoints ( featsId ) )
			{
				FeatureDescriptor *fd = ( *v->getPointsBegin ( featsId ) )->getDescriptor ( descrId );
				if ( fd )
				{
					dim = ( *fd ) [0].size();
				}
			}
			return dim;
		}

	}
}

#endif
