

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_FEATURESFEATURE_H
#define ADORE_FEATURESFEATURE_H

#include <adrcore/types/featuredescriptor.h>
#include <vector>


namespace adore
{

	namespace core
	{

		class ViewBase;
		template<class T> class View;

		using namespace std;

		uint nextFeatureId();
		void leastFeatureId( uint );

		/**
		 * The base class of all features
		 
		*/
		class FeatureBase
		{
			private:
				uint featureTypeId;

			public:
				FeatureBase( uint typeId, const float &_strength = 0, const float &_scale = 1.0F ):
						featureTypeId( typeId ),
						strength( _strength ),
						scale( _scale ),
						view( NULL ),
						id( nextFeatureId() )
				{;}

				virtual ~FeatureBase() {;}

				void setScale( float theValue )
				{
					scale = theValue;
				}

				float getScale() const
				{
					return scale;
				}

				void setStrength( float theValue )
				{
					strength = theValue;
				}

				float getStrength() const
				{
					return fabs( strength );
				}

				float getSignedStrength() const
				{
					return strength;
				}

				ViewBase *getView() const
				{
					return view;
				}

				void setView( ViewBase* theValue )
				{
					view = theValue;
				}

				template<class T>
				void setView( View<T>* theValue )
				{
					view = theValue;
				}

				template<class T>
				View<T>* getView() const
				{
					return static_cast<View<T> *>( view );
				}

				uint getFeatureTypeId() const
				{
					return featureTypeId;
				}

				uint getId() const
				{
					return id;
				}

				void setId( const uint i )
				{
					leastFeatureId( i );
					id = i;
				}

			protected:

				float strength;
				float scale;
				ViewBase *view;
				uint id;
		};

		/**
		 * A feature is described by descriptor. These Descriptors are deleted on deletion :)
		 */
		class Feature : public FeatureBase
		{
			public:
				Feature( uint typeID, const double &_strength = 0, const double &_scale = 1.0F ):
						FeatureBase( typeID, _strength, _scale )
				{
					for ( uint i = 0; i < maxDescriptorTypeNo; ++i ) descriptors[i] = NULL;
				}

				~Feature()
				{
					for ( uint i = 0; i < maxDescriptorTypeNo; ++i ) delete descriptors[i];
				}

				/**
				 * set the descriptor. The typeId of the descriptor is evaluated.
				 * the can be multiple descriptors as long they have different types
				 */
				void setDescriptor( FeatureDescriptor *d )
				{
					if ( d && descriptors[d->getId()] != d )
					{
						delete descriptors[d->getId()];
					}
					assert( !d || d->getId() < maxDescriptorTypeNo );
					descriptors[d->getId()] = d;
				}

				FeatureDescriptor *getDescriptor( int id = 0 )
				{
					return descriptors[id];
				}
				const FeatureDescriptor* getDescriptor( int id = 0 ) const
				{
					return descriptors[id];
				}

			protected:

				FeatureDescriptor *descriptors[maxDescriptorTypeNo];
		};

		template <class Comparator = greater<float> >
		class FeatureStrengthComparator
		{
			public:
				bool operator()( Feature *a, Feature *b ) const
				{
					assert( a );
					assert( b );
//					assert( a != b );//makes it less general, but discovers bigs

					bool ab = comparator( a->getStrength(), b->getStrength() );
					bool ba = comparator( b->getStrength(), a->getStrength() );
					if ( ab == ba )
					{
//						printf( "comparing pointers\n %f = %f", a->getStrength(), b->getStrength() );
						//if the strength are the same, distinguish
						//featurepoints through their pointer
						//this is crucial, because this comparator is used in a set
						return compPt( a, b );
					}
					else
					{
//						printf( "comparing strength\n %f = %f", a->getStrength(), b->getStrength() );
						return ab;
					}
				}

			private:
				Comparator comparator;
				less<Feature *> compPt;
		};

		template <class Comparator = greater<float> >
		class FeatureScaleComparator
		{
			public:
				bool operator()( Feature *a, Feature *b ) const
				{
					assert( a );
					assert( b );

					bool ab = comparator( a->getScale(), b->getScale() );
					bool ba = comparator( b->getScale(), a->getScale() );
					if ( ab == ba )
					{
						//if the strength are the same, distinguish
						//featurepoints through their pointer
						//this is crucial, because this comparator is used in a set
						return compPt( a, b );
					}
					else
					{
						return ab;
					}
				}

			private:
				Comparator comparator;
				less<Feature *> compPt;
		};

		template <class Comparator = less<ViewBase *> >
		class FeatureViewComparator
		{
			public:
				bool operator()( Feature *a, Feature *b ) const
				{
					assert( a );
					assert( b );

					bool ab = comparator( a->getView(), b->getView() );
					bool ba = comparator( b->getView(), a->getView() );
					if ( ab == ba )
					{
						//if the strength are the same, distinguish
						//featurepoints through their pointer
						//this is crucial, because this comparator is used in a set
						return compPt( a, b );
					}
					else
					{
						return ab;
					}
				}

			private:
				Comparator comparator;
				less<Feature *> compPt;
		};

		template <class Comparator = less<ViewBase *> >
		class FeatureViewAmbiComparator
		{
			public:
				bool operator()( Feature *a, Feature *b ) const
				{
					assert( a );
					assert( b );

					return comparator( a->getView(), b->getView() );
				}

			private:
				Comparator comparator;
				less<Feature *> compPt;
		};

	}

}

#endif
