

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include <adrcore/types/nview.h>
#include <adrcore/util/tensortrafo.h>

namespace adore
{

	namespace core
	{

		uint nextNViewId()
		{
			static int nextId = 0;
			return ++nextId;
		}

		NView::NView ( int vrm, int crm, int carm, int prm ) :
				viewToCam ( NULL, NULL ),
				p3DToCorr ( NULL, NULL ),
				info ( NULL ),
				normalized ( false ),
				id ( nextNViewId() ),
				viewResetMode ( vrm ),
				corrResetMode ( crm ),
				camResetMode ( carm ),
				point3dResetMode ( prm )
		{
		}

		NView::~NView()
		{
			reset();
		}

		void NView::reset()
		{
//			printf( "resetting NView" );
			if ( point3dResetMode & DELETE_CORR )
			{
				Point3DIterator it3 = getPoints3DBegin();
				for ( ; it3 != getPoints3DEnd(); ++it3 )
				{
					delete *it3;
				}
			}

			if ( corrResetMode & DELETE_CORR )
			{
				CorrIterator i1 = pointCorrs.getBegin();
				for ( ; i1 != pointCorrs.getEnd(); ++i1 )
				{
					delete *i1;
				}
				pointCorrs.clear();
			}
			
			else if ( viewResetMode & FORGET_VIEW ) pointCorrs.clear();

			if ( camResetMode & DELETE_CAM )
			{
				CameraIterator i3 = getCamerasBegin();
				for ( ; i3 != getCamerasEnd(); ++i3 )
				{
					delete *i3;
				}
			}

			if ( viewResetMode & DELETE_VIEW )
			{
				DBGOUT( DBGNORMAL, "NView deletes Views\n" );
				for ( ViewIterator i = getViewsBegin(); i != getViewsEnd(); ++i ) delete *i;
				views.clear();
			}
			else
			{
				if ( viewResetMode & RESET_VIEW )
				{
					for ( ViewIterator i = getViewsBegin(); i != getViewsEnd(); ++i )
					{
						( *i )->reset();
						DBGOUT( DBGNORMAL, "NView - reset: Num Feats after reset: %d\n", ( *i )->getNumAllFeaturePoints() );
					}
				}
				if ( viewResetMode & FORGET_VIEW )
				{
					views.clear();
				}
			}

			points3D.clear();
			ptToCorr.clear();
			viewToCam.clear();
			p3DToCorr.clear();

			delete info; info = NULL;
		}


		const NView& NView::operator=( const NView &nv )
		{
			points3D = nv.points3D;
			triangles = nv.triangles;
			views = nv.views;

			ptToCorr = nv.ptToCorr;
			pointCorrs = nv.pointCorrs;

			viewToCam = nv.viewToCam;
			p3DToCorr = nv.p3DToCorr;

			info = nv.info;

			normalized = nv.normalized;

			id = nv.id;
			viewResetMode = nv.viewResetMode;
			corrResetMode = nv.corrResetMode;

			return *this;
		}

		void NView::addView ( ViewBase *v )
		{
			assert ( v );
			views.insert ( v ); //views is a set, so every view is registered only once
		}


		void NView::addPointCorrs ( vector<vector<FeaturePoint *> > &pointsVec )
		{
			assert( false );
			vector<FeaturePoint *> pts ( pointsVec.size() );

			//printf("add corrs: %d views, %d corrs\n", pointsVec.size(), pointsVec[0].size());

			uint size = pointsVec[0].size();
			for ( uint i = 0; i < size; ++i )
			{
				int j = 0;
				vector<vector<FeaturePoint *> >::iterator it = pointsVec.begin();
				for ( ; it != pointsVec.end(); ++it, ++j )
				{
					pts[j] = ( *it ) [i];
//					if ( pts[j] ) printf( "%d ", pts[j]->getView()->getId() );
//					else printf( "-1 " );
				}
//				printf( "\n " );
				addPointCorr ( pts.begin(), pts.end() );
			}

#ifndef NDEBUG
			assert ( sanityCheck() );
#endif
		}

		void NView::addPointCorrs ( NView &nview )
		{
//			DBGOUT( DBGALL, "adding pointCorrs of another nview\n" );
			CorrIterator itC = nview.getPointCorrsBegin();
			for ( ; itC != nview.getPointCorrsEnd(); ++itC )
			{
				PointCorr *corr = *itC;
				addPointCorr ( corr->getBegin(), corr->getEnd() );
			}
		}

		bool NView::sanityCheck()
		{
			return true;//TODO remove this again

			CorrIterator it = getPointCorrsBegin();
			set<ViewBase *> viewSet;

			int corrFails = 0;
			int where = 0;

			for ( ; it != getPointCorrsEnd(); ++it )
			{
				int nulls = 0;
				PointCorr *corr = *it;
				PointCorr::Iterator it = corr->getBegin();
				viewSet.clear();
				for ( ; it != corr->getEnd(); ++it )
				{
					if ( *it )
					{
						viewSet.insert ( ( *it )->getView() );
					}
					else ++nulls;
				}
				if ( viewSet.size() + nulls < corr->getSize() )
				{
					++corrFails;
					DBGOUT ( DBGALL, "failed at %d of %d\n", where, pointCorrs.getSize() );
					DBGOUT ( DBGALL, "%zd + %d < %d\n", viewSet.size(), nulls, corr->getSize() );
					PointCorr::Iterator it = corr->getBegin();
					for ( ; it != corr->getEnd(); ++it )
					{
						if ( *it )
						{
							printf ( "(%f, %f, %f):%p ", ( **it ) ( 0 ), ( **it ) ( 1 ), ( **it ) ( 2 ), ( *it )->getView() );
						}
						printf ( "\n" );
					}
				}
				++where;
			}

			if ( corrFails != 0 ) printf ( "%d PointCorrs fail sanity check\n", corrFails );
			return corrFails == 0;
		}

		void NView::removeFeaturePoint ( FeaturePoint *p, PointCorr *c )
		{
			assert ( p );
			assert ( c );
			c->removeFeature ( p );
			map<FeaturePoint *, PointCorr *>::iterator it = ptToCorr.find ( p );
			assert ( it != ptToCorr.end() );
			ptToCorr.erase ( it );
		}

		void NView::removeFeaturePoint ( FeaturePoint *p )
		{
			if ( !p ) return;
			PointCorr *corr = getPointCorr ( p );

			//testing for some segfault
			if ( corr )
			{
				removeFeaturePoint ( p, corr );
			}
		}

		void NView::removePointCorrAnd3D ( FeaturePoint *p )
		{
			if ( !p ) return;
			PointCorr *corr = getPointCorr ( p );
			removePointCorrAnd3D ( corr );
		}

		void NView::removePointCorrAnd3D ( PointCorr *c )
		{
			if ( !c ) return;
//			DBGOUT( DBGNORMAL, "removing corr %p with %d pts\n", c, c->getSize() );
			PointCorr::Iterator it = c->getBegin();
			while ( it != c->getEnd() )
			{
				PointCorr::Iterator oldIt = it;
				++it;
				assert ( hasFeaturePoint ( *oldIt ) );
				removeFeaturePoint ( *oldIt, c );
			}
//			DBGOUT( DBGNORMAL, "points removed from corr %p\n", c );
			Point3D *p3d = p3DToCorr.getA ( c );
			if ( p3d )
			{
//				DBGOUT( DBGNORMAL, "deleting 3D point %p\n", p3d );
				points3D.remove ( p3d );
				p3DToCorr.remove ( p3d );
				delete p3d;
			}
//			DBGOUT( DBGNORMAL, "removing corr %p itself\n", c );
			pointCorrs.remove ( c );
			delete c;
		}

		Camera *NView::getCamera ( ViewBase *v, bool create )
		{
			if ( !v || !hasView ( v ) )
			{
				DBGOUT ( DBGNORMAL, "NView::getCamera - View is unknown to NView\n" );
				return NULL;
			}

			Camera *c = viewToCam[v];
			if ( !c && create )
			{
				c = new Camera();
				viewToCam.add ( v, c );
			}

			return c;
		}

		Camera* NView::getCamera( ViewBase* v) const
		{
			if ( !v || !hasView ( v ) )
			{
				DBGOUT ( DBGNORMAL, "NView::getCamera - View is unknown to NView\n" );
				return NULL;
			}
			return viewToCam[v];
		}

		void NView::addCameras ( NView &nview )
		{
			Camera *cam1 = NULL;
			Camera *cam2 = NULL;
			NView::ViewIterator it = nview.getViewsBegin();
			for ( ; it != nview.getViewsEnd(); ++it )
			{
				cam1 = nview.getCamera ( *it );
				assert ( cam1 );
				cam2 = getCamera ( *it, true ); //create, if not available
				assert ( cam2 );
//				printMats ( cam1->getProjMat(), cam2->getProjMat() );
				*cam2 = *cam1;
//				printMat ( cam2->getProjMat() );
			}

//			printf( "add camera debug:\n" );
//			for ( CameraIterator cit = getCamerasBegin(); cit != getCamerasEnd(); ++cit )
//			{
//				printMat( ( *cit )->getProjMat() );
//			}
		}

		void NView::addPointCorr ( PointCorr *corr )
		{
			pointCorrs.insert ( corr );
			for ( ViewIterator vit = getViewsBegin(); vit != getViewsEnd(); ++vit )
			{
				FeaturePoint *p = corr->getFeature ( *vit );
				if ( p ) ptToCorr[p] = corr;
			}
		}

		void NView::normalize()
		{
			if ( isNormalized() ) return;
			DBGOUT ( DBGALL, "NView normalizes views\n" );
			ViewIterator it = getViewsBegin();
            AMatrix temp34(3, 4);
			for ( ; it != getViewsEnd(); ++it )
			{
				( *it )->normalize();
				Camera *cam = getCamera ( *it );
				if ( cam )
				{
					AMatrix &proj = cam->getProjMat();
                    matMul2 ( ( *it )->getTrafo(), proj, temp34 );
                    proj = temp34;
				}
			}
			if ( isNViewSet() )
			{
				ViewIterator it = getViewsBegin();
				switch ( info->getType() )
				{
					case NViewInfo::PROJECTIVITY:
					{
                        const AMatrix & t1 = ( *it )->getInvTrafo(); ++it;
                        const AMatrix & t2 = ( *it )->getTrafo();
						transformH ( info->getH(), t1, t2 );
					}
					break;
					case NViewInfo::FMATRIX:
					{
                        const AMatrix & t1 = ( *it )->getInvTrafo(); ++it;
                        const AMatrix & t2 = ( *it )->getInvTrafo();
						transformFM ( info->getFMatrix(), t1, t2 );
					}
					break;
					case NViewInfo::TTENSOR:
					{
                        const AMatrix & t1 = ( *it )->getInvTrafo(); ++it;
                        const AMatrix & t2 = ( *it )->getTrafo(); ++it;
                        const AMatrix & t3 = ( *it )->getTrafo();
						transform ( info->getTTensor(), t1, t2, t3 );
					}
					break;
					default:
						break;
				}
			}
			normalized = true;
		}

		void NView::denormalize()
		{
			if ( !isNormalized() ) return;
			DBGOUT ( DBGNORMAL, "NView denormalizes views\n" );
			ViewIterator it = getViewsBegin();
            AMatrix temp34(3, 4);
            for ( ; it != getViewsEnd(); ++it )
			{
				( *it )->denormalize();
				Camera *cam = getCamera ( *it );
				if ( cam )
				{
					AMatrix &proj = cam->getProjMat();
                    matMul2 ( ( *it )->getInvTrafo(), proj, temp34 );
                    proj = temp34;
				}
			}
			if ( isNViewSet() )
			{
				DBGOUT ( DBGNORMAL, "NView is NViewSet of type %d\n", info->getType() );
				ViewIterator it = getViewsBegin();
				switch ( info->getType() )
				{
					case NViewInfo::PROJECTIVITY:
					{
						DBGOUT ( DBGNORMAL, "NView denormalizes Projectivity\n" );
                        const AMatrix & t1 = ( *it )->getTrafo(); ++it;
                        const AMatrix & t2 = ( *it )->getInvTrafo();
						transformH ( info->getH(), t1, t2 );
					}
					break;
					case NViewInfo::FMATRIX:
					{
						DBGOUT ( DBGNORMAL, "NView denormalizes FMatrix\n" );
                        const AMatrix & t1 = ( *it )->getTrafo(); ++it;
                        const AMatrix & t2 = ( *it )->getTrafo();
						transformFM ( info->getFMatrix(), t1, t2 );
					}
					break;
					case NViewInfo::TTENSOR:
					{
						DBGOUT ( DBGNORMAL, "NView denormalizes TTensor\n" );
                        const AMatrix & t1 = ( *it )->getTrafo(); ++it;
                        const AMatrix & t2 = ( *it )->getInvTrafo(); ++it;
                        const AMatrix & t3 = ( *it )->getInvTrafo();
						transform ( info->getTTensor(), t1, t2, t3 );
					}
					break;
					default:
						DBGOUT ( DBGCRITICAL, "NView can't denormalize unknown Tensor of type %d\n", info->getType() );
						break;
				}
			}
			normalized = false;
		}


	}
} //namespace
