

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_NVIEWNVIEW_H
#define ADORE_NVIEWNVIEW_H

#define DBGNVIEW DBGNORMAL

#include <adrcore/util/mathutil.h>
#include <adrcore/util/printmat.h>
#include <adrcore/types/aarray.h>
#include <adrcore/types/camera.h>
#include <adrcore/types/featurepoint.h>
#include <adrcore/types/featureline.h>
#include <adrcore/types/view.h>
#include <adrcore/types/bijectivity.h>
#include <adrcore/types/pointcorr.h>
#include <adrcore/types/corr1ton.h>
#include <adrcore/types/correspondence.h>
#include <adrcore/types/nviewinfo.h>

namespace adore
{
	namespace core
	{

		/**
		* How to affect ViewBases when the Map is deleted?
		* KEEP: leave views unaffected
		* RESET: deletes all Featurepoints
		* FORGET: as RESET, additionally cleans the internal references to the views.
		* DELETE: deletes the views
		 */
		enum
		{
			KEEP_VIEW = 0,   //leave views unaffected
			RESET_VIEW = 1,  //deletes all Featurepoints
			FORGET_VIEW = 2, //cleans the internal references.
			DELETE_VIEW = 4  //deletes the views
		};

		enum
		{
			KEEP_CORR = 0,   //leave views unaffected
			FORGET_CORR = 2, //cleans the internal references.
			DELETE_CORR = 4  //deletes the views
		};

		enum
		{
			KEEP_CAM = 0,   //leave views unaffected
			FORGET_CAM = 2, //cleans the internal references.
			DELETE_CAM = 4  //deletes the views
		};
		
		enum
		{
			KEEP_POINT3D = 0,
			FORGET_POINT3D = 2,
			DELETE_POINT3D = 4
		};

		/**
		 * The basic Entity describing a scene.
		 * It consists of 3D points, views, pointCorrs, and cameras
		 * Exept the views, everything is allocated and deleted by this class
		 *  
		*/
		class NView
		{

			public:


				/**
				 * Another abbreviation
				 */
				typedef AVector Point3D;

				//collections
				typedef InOrderSet< PointCorr *> CorrColl;
				typedef InOrderSet< Corr1ToN * > MultiCorrColl;
				typedef InOrderSet< LineCorr *> LineCorrColl;
				typedef InOrderSet<Point3D *> Point3DColl;
				typedef InOrderSet<ViewBase *> ViewColl;
				typedef Bijectivity<ViewBase *, Camera*> CamMap;
				typedef vector<unsigned> TriColl;
				typedef map<FeaturePoint *, PointCorr *> PtCorrMap;
				typedef map<Feature *, Corr1ToN *> LineMCorrMap;
				typedef map<FeatureLine *, LineCorr *> LineCorrMap;
				typedef Bijectivity<Point3D *, PointCorr*> Pt3DCorrMap;

				//iterator types
				typedef CorrColl::ConstIterator CorrIterator;
				typedef MultiCorrColl::ConstIterator MultiCorrIterator;
				typedef LineCorrColl::ConstIterator LineCorrIterator;
				typedef Point3DColl::Iterator Point3DIterator;
				typedef Point3DColl::ConstIterator ConstPoint3DIterator;
				typedef ViewColl::ConstIterator ViewIterator;
				typedef CamMap::BIterator CameraIterator;
				typedef TriColl::iterator TrianglesIterator;
				typedef TriColl::const_iterator ConstTrianglesIterator;
				typedef PtCorrMap::const_iterator PointToCorrIterator;

			private:

				Point3DColl points3D;
				TriColl triangles;
				ViewColl views;//here, inOrder is essential to relate the views with the tensor

				CorrColl pointCorrs;
				PtCorrMap ptToCorr;
				MultiCorrColl multiPoints, multiLines;
				LineMCorrMap lineMCorrMap;
				LineCorrColl lineCorrs;
				LineCorrMap lineCorrMap;

				CamMap viewToCam;
				Pt3DCorrMap p3DToCorr;

				NViewInfo *info;//further reconstruction info like tensors etc.

				bool normalized;

				int id;

				/**
				* the current reset mode affecting views.
				 */
				int viewResetMode;

				/**
				 * the current reset mode affecting views.
				 */
				int corrResetMode;

				/**
				 * the current reset mode affecting views.
				 */
				int camResetMode;
				
				/**
				 * the current reset mode affecting views.
				 */
				int point3dResetMode;

			public:

				/**
				 * Constructor.
				 */
				NView ( int vrm = FORGET_VIEW, int crm = DELETE_CORR, int carm = DELETE_CAM, int prm = DELETE_POINT3D );

				/**
				 * Destructor.
				 */
				~NView();

				/**
					sets the internal references to that of the assigned nview
					this is potentially dangerous and not well tested
				*/
				const NView& operator=( const NView &nv );

				/**
				 * Resets the NView, deleting all gained information.
				 */
				void reset();


				/**
				 * returns all entities.
				 * the features are in canonical order (see: getPointCorrs)
				 */
				template<class BaseClassT>
                                void getAll ( vector<Camera *> &cams, vector<AVector3D*> &vec3D, vector<BaseClassT *> &feats );

				//////////////////////////////
				////     managing views
				//////////////////////////////

				/**
				 * Get an iterator to the begin of all the ViewBases of this NView.
				 * @return the iterator
				 */
				ViewIterator getViewsBegin() const
				{
					return views.getBegin();
				}

				/**
				 * Get an iterator to the end of all the ViewBases of this NView.
				 * @return the iterator
				 */
				ViewIterator getViewsEnd() const
				{
					return views.getEnd();
				}

				/**
				 * fills a vector with pointers to all views.
				 * @param views the vector that should be filled.
				 */
				template<class ViewIteratorT>
				void getViews ( ViewIteratorT vit ) const
				{
					ViewIterator it = getViewsBegin();
					for ( ; it != getViewsEnd(); ++it, ++vit )
					{
						*vit = *it;
					}
				}

				/**
				 * adds a ViewBase.
				 * @param v the ViewBase.
				 */
				void addView ( ViewBase *v );

				/**
				 * Retrieve the number of ViewBases.
				 * @return the number of ViewBases.
				 */
				uint getNumViews() const { return views.getSize(); }


				/**
				 * Get the ViewBase that corresponds to a Camera
				 * @param c the Camera.
				 * @return the ViewBase.
				 */
				ViewBase *getView ( Camera *c )
				{
					return viewToCam[c];
				}

				/**
				 * Find out, whether a ViewBase is a member of this set.
				 * @param v the ViewBase
				 * @return true, if the ViewBase is known.
				 */
				bool hasView ( ViewBase *v ) const { return ( views.find ( v ) != views.getEnd() ); }

				void addViews ( NView &nview )
				{
					NView::ViewIterator it = nview.getViewsBegin();
					for ( ; it != nview.getViewsEnd(); ++it )
					{
						addView ( *it );
					}
				}


				//////////////////////////////
				////     managing cameras
				//////////////////////////////

				int getNumCameras() const
				{
					return viewToCam.getSize();
				}

				/**
				 * Get an iterator to the begin of all the Cameras of this NView.
				 * @return the iterator
				 */
				CameraIterator getCamerasBegin() const
				{
					return viewToCam.getBsBegin();
				}

				/**
				 * Get an iterator to the end of all the Cameras of this NViewSet.
				 * @return the iterator
				 */
				CameraIterator getCamerasEnd() const
				{
					return viewToCam.getBsEnd();
				}

				/**
				  * Get the camera that corresponds to a ViewBase.
				  * @param v the ViewBase.
				  * @return the Camera
				  */
				Camera *getCamera ( ViewBase *v, bool create = false );
				Camera *getCamera ( ViewBase * v) const;

				bool hasCamera ( ViewBase *v )
				{
					return ( viewToCam[v] != NULL );
				}

				void addCameras ( NView &nview );

				/**
				 * fills a vector with pointers to all cameras.
				 * @param cams the vector that should be filled.
				 */
				template<class CamIteratorT>
				void getCameras ( CamIteratorT cit ) const
				{
					CameraIterator it = getCamerasBegin();
					for ( ; it != getCamerasEnd(); ++it, ++cit )
					{
						*cit = *it;
					}
				}

				//////////////////////////////
				////     managing pointCorrs
				//////////////////////////////

				/**
				 * Get an iterator to all PointCorrs known by this NView.
				 * @return The pointCorrs.
				 */
				CorrIterator getPointCorrsBegin() const
				{
					return pointCorrs.getBegin();
				}

				CorrIterator getPointCorrsEnd() const
				{
					return pointCorrs.getEnd();
				}

				uint getNumPointCorrs() const
				{
					return pointCorrs.getSize();
				}

				/**
				 * Get the correspondence that resects to a given 3D-point.
				 * @param pt3D the 3D-point
				 * @return the PointCorr.
				 */
				PointCorr *getPointCorr ( Point3D *pt3D ) const
				{
					return p3DToCorr.getB ( pt3D );
				}

				/**
				 * Get the PointCorr of a 2D-point p.
				 * @param p the 2D-point
				 * @return the PointCorr
				 */
				PointCorr *getPointCorr ( FeaturePoint *p ) const
				{
					map<FeaturePoint *, PointCorr *>::const_iterator it = ptToCorr.find ( p );
					if ( it == ptToCorr.end() )
						return NULL;
					return ( *it ).second;
				}

				bool addFeaturePoint ( FeaturePoint *p, PointCorr *c )
				{
					if ( !p || !c ) return false;
					if ( c->getFeature ( p->getView() ) ) return false;
					if ( hasFeaturePoint ( p ) ) return false;
					addFeaturePointNoCheck ( p, c );
					return true;
				}

				void addFeaturePointNoCheck ( FeaturePoint *p, PointCorr *c )
				{
					c->addFeature ( p );
					ptToCorr[p] = c;
				}

				void removeFeaturePoint ( FeaturePoint *p, PointCorr *c );

				bool hasFeaturePoint ( FeaturePoint *pt )
				{
					map<FeaturePoint *, PointCorr *>::iterator it = ptToCorr.find ( pt );
					return ( it != ptToCorr.end() );
				}

				FeaturePoint *getFeaturePoint ( Point3D *p, ViewBase *v )
				{
					return getPointCorr ( p )->getFeature ( v );
				}

				void removeFeaturePoint ( FeaturePoint *p );

				void removePointCorrAnd3D ( FeaturePoint *p );

				void removePointCorrAnd3D ( PointCorr *c );

				/**
				 * Get all PointCorr as a vector of vectors. Every "sub-vector" represents the points of one ViewBase.
				 * E.g. feats[0][3] and feats[1][3] belong to the same PointCorr.
				 * @param feats the structure that will be filled with PointCorrs.
				 */
				template<class BaseClassT>
				void getPointCorrs ( vector< vector<BaseClassT *> > &feats );

				/**
				 * returns points in canonical ordering
				 * that is: pt[corr1][view1], pt[corr1][view2],.., pt[corr1][viewN]
				 *          pt[corr2][view1], pt[corr2][view2],.., pt[corr2][viewN]
				 *          ...
				 *          pt[corrN][view1], pt[corrN][view2],.., pt[corrN][viewN]
				*/
				template<class BaseClassT>
				void getPointCorrs ( vector<BaseClassT *> &feats );

				template<class BaseClassT>
				void getPointCorrs ( ViewBase *v, vector<BaseClassT *> &feats, bool fillNull = false );

				/**
				 * From a specific ViewBase, get the corresponding features of a given set of features.
				 * @param feats1 feature points from one ViewBase (ViewBase1 say)
				 * @param v2 another ViewBase (ViewBase2 say)
				 * @param feats2 this vector is filled with the feaature points of ViewBase2 that correspond to the feature points in feats1
				 * if a feature points doesn't have a correspondence, the entry in the vector is NULL.
				 */
				template<class BaseClassT>
				void getCorrespondingFeatures ( vector<BaseClassT *> &feats1, ViewBase &v2, vector<BaseClassT *> &feats2 );

				/**
				 * constructs a new pointcorr from a number of feature points
				 */
				template<class IteratorT>
				PointCorr *addPointCorr ( IteratorT begin, IteratorT end, float badness = 0.0F );

				/**
				 * same as above, but no checks. Preferably used during loading
				 */
				template<class IteratorT>
				PointCorr *addPointCorrNoCheck ( IteratorT begin, IteratorT end, float badness = 0.0F );

				/**
				 * adds a point coor without any checks
				 * only those feature points are registered for which a view is referenced
				 */
				void addPointCorr ( PointCorr *corr );

				void addPointCorrs ( vector<vector<FeaturePoint *> > &pointsVec );

				void addPointCorrs ( NView &nview );

				//////////////////////////////
				////     managing 3D-points
				//////////////////////////////

				/**
				 * Get the 3D-point that is the resection of all the points of a given PointCorr.
				 * @param corr the PointCorr
				 * @return the 3D-point
				 */
				Point3D *getPoint3D ( PointCorr *corr ) const
				{
					return p3DToCorr.getA ( corr );
				}

				/**
				 * Get the 3D-point corresponding to a 2D-point.
				 * @param pt the 2D-point.
				 * @return the 3D-point
				 */
				Point3D *getPoint3D ( FeaturePoint *pt ) const
				{
					map<FeaturePoint *, PointCorr *>::const_iterator it = ptToCorr.find ( pt );
					if ( it == ptToCorr.end() )
						return NULL;
					return p3DToCorr.getA ( ( *it ).second );
				}


				/**
				 * Set the 3D-point that correspond to a 2D-point pt. All other 2D-points in the PointCorr of pt will
				 * then correspond to the 3D-point, too.
				 * @param pt3D the 3D-point
				 * @param pt the 2D-point
				 * @return true, if successful (meaning a correspondence with pt exists)
				 */
				bool setPoint3D ( Point3D *pt3D, FeaturePoint *pt )
				{
					assert ( pt3D );
					assert ( pt );
					map<FeaturePoint *, PointCorr *>::iterator it = ptToCorr.find ( pt );
					if ( it == ptToCorr.end() ) return false;
//					assert_msg( it != ptToCorr.end(), "This pt3D<->pt map is illegal, because the pt is not yet know / has no Correpondence (ptToCorr.size(): %d, pointCorrs.size(): %d)\n", ptToCorr.size(), pointCorrs.getSize() );
					PointCorr *corr = ( *it ).second;
					assert ( corr );
					p3DToCorr.add ( pt3D, corr );
					points3D.add ( pt3D );
					return true;
				}

				/**
				 * Set the 3D-point of a PointCorr.
				 * @param pt3D the 3D-point
				 * @param corr the PointCorr
				 */
				void setPoint3D ( Point3D *pt3D, PointCorr *corr )
				{
					p3DToCorr.add ( pt3D, corr );
					points3D.add ( pt3D );
				}

				/**
				 * adds a 3D point without an associated correspondence
				 */
				void addPoint3D ( Point3D *pt3D )
				{
					p3DToCorr.add ( pt3D, NULL );
					points3D.add ( pt3D );
				}

				/**
				 * Get an iterator pointing to the begin of all 3D-points.
				 * @return the iterator.
				 */
				Point3DIterator getPoints3DBegin()
				{
					return points3D.getBegin();
				}

				ConstPoint3DIterator getPoints3DBegin() const
				{
					return points3D.getBegin();
				}

				/**
				 * Get an iterator pointing to the end of all registered 3D-points.
				 * @return the iterator.
				 */
				Point3DIterator getPoints3DEnd()
				{
					return points3D.getEnd();
				}

				ConstPoint3DIterator getPoints3DEnd() const
				{
					return points3D.getEnd();
				}

				/**
				 * fills a vector with pointers to all 3D points.
				 * @param pts3D the vector that should be filled.
				 */
				template<class Point3DIteratorT>
				void getPoints3D ( Point3DIteratorT pit ) const
				{
					ConstPoint3DIterator it = getPoints3DBegin();
					for ( ; it != getPoints3DEnd(); ++it, ++pit )
					{
						*pit = *it;
					}
				}

				/**
				 * fills a vector with pointers to all 3D points that correspond to given 2D points
				 * @param pts2D the vector with the indexing points.
				 * @param pts3D the vector that should be filled.
				 */
				void getPoints3D ( vector<FeaturePoint *> &pts2D, vector<Point3D *> &pts3D )
				{
					vector<FeaturePoint *>::iterator it = pts2D.begin();
					for ( ; it != pts2D.end(); ++it )
					{
						pts3D.push_back ( getPoint3D ( *it ) );
					}
				}

				uint getNumPoints3D() const { return points3D.getSize(); }

				PointToCorrIterator getPointToCorrBegin()
				{
					return ptToCorr.begin();
				}

				PointToCorrIterator getPointToCorrEnd()
				{
					return ptToCorr.end();
				}

				uint getNumPointToCorrs()
				{
					return ptToCorr.size();
				}

				template<class InputIter> void setTriangles ( InputIter begin, InputIter end )
				{
					triangles.assign ( begin,end );
				}

				uint getNumTriangles() const { return triangles.size(); }

				TrianglesIterator getTrianglesBegin()
				{
					return triangles.begin();
				}

				TrianglesIterator getTrianglesEnd()
				{
					return triangles.end();
				}

				ConstTrianglesIterator getTrianglesBegin() const
				{
					return triangles.begin();
				}

				ConstTrianglesIterator getTrianglesEnd() const
				{
					return triangles.end();
				}


				bool isNViewSet() const { return getInfo() != NULL; /*leaf were used for estimation*/ }

				/*
				 *  normalization
				 */
				void normalize();

				void denormalize();

				bool isNormalized() const {return normalized;}

				/**
				 * Use this function to tell the nview what status it has.
				 * This function doesn't change anything but that status.
				 * Once the NView is declared as (un)normalized, use normalize/denormalize
				 * to change the status und update 2D-points, cameras and the tensor.
				 */
				void setNormalized ( bool status )
				{
					normalized = status;
				}

				void setId ( int theValue )
				{
					id = theValue;
				}

				int getId() const
				{
					return id;
				}

				NViewInfo::TYPE getInfoType() const
				{
					if ( info )
					{
						return info->getType();
					}
					else return NViewInfo::NONE;
				}

				NViewInfo *getInfo() const {return info;}

				void setInfo ( NViewInfo *i ) {info = i;}

				/**
				 * Tell how ViewBases should be treated when a reset is invoked.
				 * @param theValue one of RESET_MODE
				 */
				void setViewResetMode ( const int& theValue )
				{
					viewResetMode = theValue;
				}

				/**
				 * Find out how ViewBases are treated when a reset is invoked.
				 * @return one of RESET_MODE
				 */
				int getViewResetMode() const
				{
					return viewResetMode;
				}

				void setCorrResetMode ( int theValue )
				{
					corrResetMode = theValue;
				}

				int getCorrResetMode() const
				{
					return corrResetMode;
				}
				
				void setPoint3DResetMode ( int theValue )
				{
					point3dResetMode = theValue;
				}

				int getPoint3DResetMode() const
				{
					return point3dResetMode;
				}

				void setCamResetMode ( int theValue )
				{
					camResetMode = theValue;
				}

				int getCamResetMode() const
				{
					return camResetMode;
				}

				/**
				 * Get an iterator to all MultiCorrs known by this NView.
				 * @return The multiCorrs.
				 */
				MultiCorrIterator getMultiLinesBegin() const
				{
					return multiLines.getBegin();
				}

				MultiCorrIterator getMultiLinesEnd() const
				{
					return multiLines.getEnd();
				}

				uint getNumMultiLines() const
				{
					return multiLines.getSize();
				}

				template<class IteratorT>
				Corr1ToN* addMultiLineNoCheck( Feature* f1, IteratorT begin, IteratorT end )
				{
					Corr1ToN* corr = new Corr1ToN;
					multiLines.insert( corr );
					corr->addFeature( f1 );
					lineMCorrMap[f1] = corr;
					for ( ; begin != end; ++begin )
					{
						Feature* f = begin->first;
						float badness = begin->second;
						corr->addFeature( f, badness );
						lineMCorrMap[f] = corr;
					}
					return corr;
				}

				Corr1ToN* getLineMultiCorr ( Feature* f )
				{
					LineMCorrMap::iterator it = lineMCorrMap.find ( f );
					if ( it == lineMCorrMap.end() )
						return NULL;
					return ( *it ).second;
				}

				template<class IteratorT>
				LineCorr * addLineCorrNoCheck ( IteratorT begin, IteratorT end, float badness )
				{
					LineCorr * c = new LineCorr();
					lineCorrs.insert ( c );
					for ( ; begin != end; ++begin )
					{
						FeatureLine * l = *begin;
						if ( !l ) continue;
						assert ( l->getView() );
						c->addFeature ( l );
						lineCorrMap[l] = c;
					}
					c->setBadness ( MAX ( c->getBadness(), badness ) );
					return c;
				}

				uint getNumLineCorrs() const
				{
					return lineCorrs.getSize();
				}

				LineCorrIterator getLineCorrsBegin() const
				{
					return lineCorrs.getBegin();
				}

				LineCorrIterator getLineCorrsEnd() const
				{
					return lineCorrs.getEnd();
				}

			private:

				NView( const NView & ):
						viewToCam ( NULL, NULL ),
						p3DToCorr ( NULL, NULL )
				{}

				/**
				 * Check for sanity. It's debug.
				 */
				bool sanityCheck();
		};


		template<class IteratorT>
		PointCorr *NView::addPointCorrNoCheck ( IteratorT begin, IteratorT end, float badness )
		{
			PointCorr *c = new PointCorr();
			pointCorrs.insert ( c );
			for ( ; begin != end; ++begin )
			{
				FeaturePoint *pt = *begin;
				if ( !pt ) continue;
				assert ( pt->getView() );
				addFeaturePointNoCheck ( pt, c );
			}
			c->setBadness ( MAX ( c->getBadness(), badness ) );
			return c;
		}

		template<class IteratorT>
		PointCorr *NView::addPointCorr ( IteratorT begin, IteratorT end, float badness )
		{
//			DBGOUT( DBGNVIEW, "adding correspondence %p\n", corr );
			int dist = distance ( begin, end );
			PointCorr *c = NULL;
			if ( dist < 2 )
			{
				DBGOUT ( DBGNVIEW, "PointCorr has to span 2 points at least (1)\n" );
				return c;
			}

			//check, if one of the points is already known
			bool corrAlreadySetUp = false;
			IteratorT it = begin;
			int nullCount = 0;
			for ( ; it != end; ++it )
			{
				FeaturePoint *pt = *it;
				if ( !pt )
				{
					++nullCount;
				}
				else
				{
					PointCorr *cc = getPointCorr ( pt );
					if ( cc )
					{
						corrAlreadySetUp = true;
//						assert( !c || c == cc );
						//make sure points don't come from different corrs
						if ( c && ( c != cc ) ) return NULL;
						else c = cc;
					}
				}
			}

			if ( dist - nullCount < 2 )
			{
				DBGOUT ( DBGNVIEW, "PointCorr has to span 2 points at least (2)\n" );
				return c;
			}

			if ( !corrAlreadySetUp )
			{
				c = new PointCorr();
				pointCorrs.insert ( c );
			}

			//add FeaturePoints into Corr
			for ( ; begin != end; ++begin )
			{
				FeaturePoint *pt = *begin;
				if ( !pt ) continue;
				assert ( pt->getView() );

				//currently, the matcher adds the points to their view
				FeaturePoint *known = c->getFeature ( pt->getView() );
				if ( !known )
				{
//					assert( !hasFeaturePoint( pt ) );
					addFeaturePointNoCheck ( pt, c );
				}
			}

#ifndef NDEBUG
			if ( !sanityCheck() )
			{
				printf ( "corr was already up: %s\n", corrAlreadySetUp ? "true" : "false" );
				abort();
			}
#endif

			c->setBadness ( MAX ( c->getBadness(), badness ) );
			return c;
		}

		template<class BaseClassT>
		void NView::getCorrespondingFeatures ( vector<BaseClassT *> &feats1, ViewBase &v2, vector<BaseClassT *> &feats2 )
		{
			typename vector<BaseClassT *>::iterator it = feats1.begin();
			int i = 0;
			for ( ; it != feats1.end(); ++it )
			{
				PointCorr *corr = getPointCorr ( *it );
				feats2.push_back ( corr->getFeature ( &v2 ) );
				++i;
			}
			//printf("NViewSet: found %d corrs\n", i);
		}

		template<class BaseClassT>
		void NView::getPointCorrs ( vector< vector<BaseClassT *> > &feats )
		{
			assert_msg ( feats.size() == getNumViews(), "Vector has wrong size (%zd != %d)", feats.size(), getNumViews() );

			CorrIterator it = pointCorrs.getBegin();
			for ( ; it != pointCorrs.getEnd(); ++it )
			{
				int viewI = 0;
				PointCorr *corr = *it;
				ViewIterator vi = views.getBegin();
				for ( ; vi != views.getEnd(); ++vi, ++viewI )
				{
					feats[viewI].push_back ( corr->getFeature ( *vi ) );
				}
			}

			//printf("views: %d, points: %d\n", feats.size(), feats[0].size());
		}

//		resturn points in canonical ordering
//that is: pt[corr1][view1], pt[corr1][view2],.., pt[corr1][viewN]
//		pt[corr2][view1], pt[corr2][view2],.., pt[corr2][viewN]
//		...
//		pt[corrN][view1], pt[corrN][view2],.., pt[corrN][viewN]
		template<class BaseClassT>
		void NView::getPointCorrs ( vector<BaseClassT *> &feats )
		{
			for ( Point3DIterator it = getPoints3DBegin(); it != getPoints3DEnd(); ++it )
			{
				AVector3D *p = *it;
				PointCorr *corr = getPointCorr ( p );

				for ( CameraIterator cit = getCamerasBegin(); cit != getCamerasEnd(); ++cit )
				{
					Camera *c = *cit;
					ViewBase *v = getView ( c );
					feats.push_back ( corr->getFeature ( v ) );
				}
			}
		}

		template<class BaseClassT>
                void NView::getAll ( vector<Camera *> &cams, vector<AVector3D*> &vec3D, vector<BaseClassT *> &feats )
		{
			cams.reserve ( getNumCameras() );
			vec3D.reserve ( getNumPoints3D() );
			feats.reserve ( cams.capacity() * vec3D.capacity() );

			for ( CameraIterator cit = getCamerasBegin(); cit != getCamerasEnd(); ++cit )
			{
				cams.push_back ( *cit );
			}

			for ( Point3DIterator it = getPoints3DBegin(); it != getPoints3DEnd(); ++it )
			{
				AVector3D *p = *it;
				PointCorr *corr = getPointCorr ( p );

				for ( vector<Camera *>::iterator cit = cams.begin(); cit != cams.end(); ++cit )
				{
					Camera *c = *cit;
					ViewBase *v = getView ( c );
					feats.push_back ( corr->getFeature ( v ) );
				}
                                vec3D.push_back ( p );
			}
		}



		template<class BaseClassT>
		void NView::getPointCorrs ( ViewBase *v, vector<BaseClassT *> &feats, bool fillNull )
		{
			assert( false );
			CorrIterator it = pointCorrs.getBegin();
			for ( ; it != pointCorrs.getEnd(); ++it )
			{
				PointCorr *corr = *it;
				FeaturePoint *pt = corr->getFeature ( v );
				if ( pt ) feats.push_back ( pt );
				else if ( fillNull ) feats.push_back ( NULL );
			}
		}

	}
}

#endif
