

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_NVIEWNVIEWMAP_H
#define ADORE_NVIEWNVIEWMAP_H

#include <stdarg.h>
#include <cstdlib>
#include <set>
#include <map>
#include <string>
#include <adrcore/util/debug.h>
#include <adrcore/util/stlutil.h>
#include <adrcore/types/bijectivity.h>
#include <adrcore/types/inorderset.h>
#include <adrcore/types/nview.h>
#include <adrcore/types/viewkeystring.h>
#include <adrcore/types/viewrelations.h>

namespace adore
{
	namespace core
	{

		enum
		{
			KEEP_NVIEW = 0,   //leave views unaffected
			DELETE_NVIEW = 1  //deletes the views
		};

		class NView;

		using namespace std;

		/**
		 * This classes purpose is to maintain multiple NViews and NViewSets. It provides
		 * a mapping from ViewBases to NViews:
		 * (ViewBase1 *, ViewBase2 *, ..., ViewBaseN *) -> (char1, char2, ..., char(sizeof(ViewBase*)*N)) -> string -> NViewSet
		 * So it can easily be checked if a NView to a set of views already exists.
		 *
		 * Additionally, iterators of frequently needed subsets of all NViews are available, such as all
		 * NViews, NViews that are not NViewSets or NViewSets.
		 *
		 *  
		*/
		class NViewMap
		{

			private:

				/**
				 * The key identifying an NView through its ViewBases.
				 */
				typedef basic_string<char> ViewsKeyString;


			public:

				/**
				 * An Iterator over all Entries (i.e. NViews)
				 */
				typedef Bijectivity<ViewsKeyString, NView *>::BIterator EntryIterator;

				/**
				 * An Iterator over all NViews (but not NViewSets)
				 */
				typedef InOrderSet<NView *>::ConstIterator NViewIterator;

				/**
				 * An Iterator over all ViewBase
				 */
				typedef NView::ViewIterator ViewIterator;

			private:

				/**
				 * the mapping from ViewBases to NViews
				 */
				Bijectivity<ViewsKeyString, NView *> viewsToSet;

				/**
				 * how to reset NViews
				 */
				int nviewResetMode;

				/**
				 * the master NView containing the unfiltered correspondences
				 * and all views
				 * this is used by matcher and by splitter.
				 * Spliter breaks it into 2-NViews to create the input for the RANSAC stage
				 */
				NView master;

				/**
				 * realtions and distance metric for view
				 */
				ViewRelations vrela;

			public:

				/**
				 * Constructor.
				 */
				NViewMap();

				/**
				 * Destructor.
				 */
				~NViewMap();

				/**
				 * Resets the map.
				 * This deletes all known NViews.
				 * The ViewBases are affected according to the RESET_MODE.
				 */
				void reset();

				/**
				 * Get the NView of two ViewBases. Optionally, create a new NView.
				 * @param v1 a ViewBase
				 * @param v2 a ViewBase
				 */
				NView *getNView ( ViewBase *v1, ViewBase *v2 );
				NView *createNView ( ViewBase *v1, ViewBase *v2 );


				/**
				 * Get the NView of three ViewBases. Optionally, create a new NView.
				 * @param v1 a ViewBase
				 * @param v2 a ViewBase
				 * @param v3 a ViewBase
				 */
				NView *getNView ( ViewBase *v1, ViewBase *v2, ViewBase *v3 );
				NView *createNView ( ViewBase *v1, ViewBase *v2, ViewBase *v3 );


				NView *getNView ( ViewBase *v1, ViewBase *v2, ViewBase *v3, ViewBase *v4 );
				NView *createNView ( ViewBase *v1, ViewBase *v2, ViewBase *v3, ViewBase *v4 );

				/**
				 * Get the NView of n ViewBases. Optionally, create a new NView.
				 * @param begin An iterator pointing to the begin of the views.
				 * @param end An iterator pointing to the end of the views.
				 */
				template<class IteratorT>
				NView *getNView ( IteratorT begin, IteratorT end )
				{
					ViewsKeyString key = getKey ( begin, end );
					return viewsToSet[key];
				}

				template<class IteratorT>
				NView *createNView ( IteratorT begin, IteratorT end );

				/**
				 * Set the NView of n ViewBases. If a NView already exists, nothing happens.
				 * @param begin An iterator pointing to the begin of the views.
				 * @param end An iterator pointing to the end of the views.
				 * @param nview the brand new NView
				 */
				template<class ViewBaseIteratorT>
				void setNView ( ViewBaseIteratorT begin, ViewBaseIteratorT end, NView *nview )
				{
					ViewsKeyString key = getKey ( begin, end );

					if ( nview )
					{
						for ( ; begin != end; ++begin )
						{
							ViewBase *v = *begin;
							assert ( v );
							addViewToNView ( v, nview );
						}
						viewsToSet.add ( key, nview );
					}
					else
					{
						viewsToSet.remove( key );
					}
				}

				/**
				 * Set the NView of two ViewBases. If a NView already exists, nothing happens.
				 * @param v1 a ViewBase
				 * @param v2 a ViewBase
				 * @param nview the brand new NView
				 */
				void setNView ( ViewBase *v1, ViewBase *v2, NView *nview );

				/**
				 * Set the NView of three ViewBases. If a NView already exists, nothing happens.
				 * @param v1 a ViewBase
				 * @param v2 a ViewBase
				 * @param v3 a ViewBase
				 * @param nview the brand new NView
				 */
				void setNView ( ViewBase *v1, ViewBase *v2, ViewBase *v3, NView *nview );

				void setNView ( ViewBase *v1, ViewBase *v2, ViewBase *v3, ViewBase *v4, NView *nview );

				/**
				 * Retrieve the number of known NViews.
				 * @return the number of NViews.
				 */
				int getNumEntries() const {return viewsToSet.getSize();} // const - for serialization

				/**
				 * Get an iterator pointing to the begin of all NViews.
				 * @return the iterator.
				 */
				EntryIterator getEntriesBegin() { return viewsToSet.getBsBegin(); }
				EntryIterator getEntriesBegin() const { return viewsToSet.getBsBegin(); } // const - for serialization

				/**
				 * Get an iterator pointing to the end of all NViews.
				 * @return the iterator.
				 */
				EntryIterator getEntriesEnd() { return viewsToSet.getBsEnd(); }
				EntryIterator getEntriesEnd() const { return viewsToSet.getBsEnd(); } // const - for serialization

				/**
				 * Fills a vector with pointers to all NViewSets.
				 * @param desc the vector
				 */
				void getNViewSets ( vector<NView *> &desc );

				/**
				 * Fills a vector with pointers to all NViews.
				 * @param desc the vector
				 */
				void getEntries ( vector<NView *> &desc );

				/**
				 * Starts at the NViewSets (and at the NViewSets ONLY) and follows their parent links
				 * until there is no more parent. All distinct roots are put in the vector.
				 * NViewSets without a parent do NOT count as a root.
				 * @param desc the vector
				 */
//				void getRoots ( vector<NView *> &desc );


				///////
				// all ViewBases
				///////

				void addView ( ViewBase *view )
				{
					master.addView ( view );
				}

				/**
				 * Get an iterator to the begin of all the ViewBases.
				 * @return the iterator
				 */
				ViewIterator getAllViewsBegin() const
				{
					return master.getViewsBegin();
				}

				/**
				 * Get an iterator to the end of all the ViewBases.
				 * @return the iterator
				 */
				ViewIterator getAllViewsEnd() const
				{
					return master.getViewsEnd();
				}

				/**
				 * fills a vector with pointers to all views.
				 * @param views the vector that should be filled.
				 */
				void getAllViews ( vector<ViewBase *> &views ) const
				{
					ViewIterator it = getAllViewsBegin();
					for ( ; it != getAllViewsEnd(); ++it )
					{
						views.push_back ( *it );
					}
				}

				/**
				 * Retrieve the number of all ViewBases.
				 * @return the number of all ViewBases.
				 */
				uint getNumAllViews() const { return master.getNumViews(); }

				/**
				 * Tell the Map how ViewBases should be treated when a reset is invoked.
				 * @param theValue one of RESET_MODE
				 */
				void setViewResetMode ( const int& theValue )
				{
					master.setViewResetMode ( theValue );
				}

				/**
				 * Find out how ViewBases are treated when a reset is invoked.
				 * @return one of RESET_MODE
				 */
				int getViewResetMode() const
				{
					return master.getViewResetMode();
				}

				void setCorrResetMode ( const int& theValue )
				{
					master.setCorrResetMode ( theValue );
				}

				int getCorrResetMode() const
				{
					return master.getCorrResetMode();
				}

				void setNViewResetMode ( const int& theValue )
				{
					nviewResetMode = theValue;
				}


				int getNViewResetMode() const
				{
					return nviewResetMode;
				}

				void setMasterNView ( const NView& theValue )
				{
					master = theValue;
				}

				NView &getMasterNView()
				{
					return master;
				}

				ViewRelations &getViewRelations() {return vrela;}

			private:

				//this is simply to work around the include restrictions of C/C++
				//used in the template functions
				inline void addViewToNView ( ViewBase *v, NView *nv )
				{
					addView ( v );
					nv->addView ( v );
				}
		};


		template<class IteratorT>
		NView *NViewMap::createNView ( IteratorT begin, IteratorT end )
		{
			ViewsKeyString key = getKey ( begin, end );

			NView *nv = viewsToSet[key];
			if ( nv == NULL )
			{
				nv = new NView();
				for ( ; begin != end; ++begin )
				{
					ViewBase *v = *begin;
					assert ( v );
					addViewToNView ( v, nv );
				}
				viewsToSet.add ( key, nv );
			}

			return nv;
		}


	}
}

#endif
