

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_CORERINGBUFFER_H
#define ADORE_CORERINGBUFFER_H

#include<vector>
#include<list>

namespace adore
{

	namespace core
	{

		/**
		 * a ringbuffer built out of a list and a vector
		 * the list holds the lements and the vector holds iterators
		 * that points to list entries. The vectors is uses to provide
		 * indexed access.
		 * If pointers are put into this, take care of deleting the
		 * objects the fall out of this buffer.
		 *
			 
		*/
		template<class TypeT>
		class RingBuffer
		{

			public:
				typedef std::list<TypeT> BufferT;
				typedef typename BufferT::iterator iterator;
				typedef std::vector<iterator> IndicesT;

			public:
				RingBuffer( int length = 1, TypeT initVal = TypeT( 0 ) ):
						ring( length ), start( 0 )
				{
					initBase( length, initVal );
				}

				~RingBuffer()
				{
					releaseRing();
				}

				/**
				 * if length differs from the current values, the buffer
				 * gets reallocated
				 */
				void init( int length, TypeT initVal = TypeT( 0 ) )
				{
					if ( length == getLength() ) return;
					initBase( length, initVal );
				}

				void resize( int length )
				{
					init( length );
				}

				/**
				 * copys values into the buffer beginning
				 */
				void insert( TypeT entity )
				{
					feed();
					ring.back() = entity;
				}

				TypeT &operator()( int idx ) const
				{
					return *indices[getIdxOfIdx( idx )];
				}

				TypeT &operator[]( int idx ) const
				{
					return *indices[getIdxOfIdx( idx )];
				}

				/**
				 * Iterate throug the iterators that point to the beginnings
				 * of each line.
				 */
				iterator begin()
				{
					return ring.begin();
				}

				iterator end()
				{
					return ring.end();
				}

				int getLength() const
				{
					return ring.size();
				}

				void feed()
				{
					start = ( start + 1 ) % ring.size();
					//put the begin at the end
					ring.splice( ring.end(), ring, ring.begin() );
				}

				int size() const
				{
					return getLength();
				}

			private:

				void releaseRing()
				{
				}

				int getIdxOfIdx( int row ) const
				{
					return ( start + row ) % ring.size();
				}

				void initBase( int length, TypeT initVal = TypeT( 0 ) )
				{
					releaseRing();
					ring.resize( length );
					indices.resize( length );

					typename  BufferT::iterator bit = ring.begin();
					typename IndicesT::iterator iit = indices.begin();
					for ( ; bit != ring.end(); ++iit, ++bit )
					{
						*bit = initVal;
						*iit = bit;
					}
				}

				BufferT ring;//this is a list to provide iterators
				IndicesT indices;//this is a vector to provide indexed access
				int start;
		};

	}

}

#endif
