

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORE_NVIEWVIEW_H
#define ADORE_NVIEWVIEW_H

#include <string>
#include <adrcore/types/featuretypes.h>
#include <adrcore/types/featurepoint.h>
#include <adrcore/types/featureline.h>
#include <adrcore/search/grid2d.h>
#include <adrcore/imgprocessing/imageplane.h>
#include <boost/shared_ptr.hpp>

namespace adore
{
	namespace core
	{

		using namespace std;
		typedef FeaturePoint::scalar_type coorT;

		uint nextViewId();
		void leastViewId ( uint );

		class ViewBase
		{
			public:
				enum POINT_MODE
				{
					POINT_FORGET,
                    POINT_DELETE
				};

				typedef Grid2DHomo< FeaturePoint * > FeatureContainer;
				typedef FeatureContainer::Iterator PointsIterator;
				typedef FeatureContainer::Const_Iterator PointsConstIterator;
				typedef vector<unsigned>::iterator TrianglesIterator;
				typedef vector<unsigned>::const_iterator TrianglesConstIterator;
				typedef boost::shared_ptr<FeatureLine> segmentT;
				typedef vector<segmentT>::iterator LinesIterator;
				typedef vector<segmentT>::const_iterator LinesConstIterator;

			public:

				ViewBase();

				/**
				 * Copy-Constructor. Does not copy images associated with the parameter-view.
				 * The images of the new view are set to NULL.
				  * Does not copy features
				 * @param view The view to be copied.
				  */
				ViewBase ( ViewBase &view );

				/**
					makes this view a copy of the argument
				*/
				void copy ( ViewBase *v );

				/**
				 * Destructor. Deletes everything but the image
				 */
				virtual ~ViewBase() = 0;

				/**
				 * resets the View, deleting all features.
				 * The normalizing transformation is _not_ resetted
				 */
				void reset();

				/**
				 * Get the features as a List of Points.
				 * @param list a list that shall be filled with features.
				 * @return the number of added feature points
				 */
				template<class Iterator>
				int getFeaturePoints ( Iterator fit, int featsID = 0 )
				{
					if ( !features[featsID] ) return 0;
					int count = 0;
					PointsIterator it = features[featsID]->getBegin();
					for ( ; it != features[featsID]->getEnd(); ++it, ++fit )
					{
						*fit = *it; ++count;
					}
					return count;
				}

				int getNumFeaturePoints ( int featsID = 0 ) const
				{
					return features[featsID] ? features[featsID]->getSize() : 0;
				}

				/** Get number of indices in triangle table where every 3 indices correspond to 1 triangle
				 *  @warning Returns number of indices, divide by 3 to get number of triangles
				 */
				int getNumTriangles ( int featsID = 0 ) const
				{
					return triangles[featsID] ? triangles[featsID]->size() : 0;
				}

				int getNumAllFeaturePoints()
				{
					int sum = 0;
					for ( uint i = 0; i < maxFeatureTypeNo; ++i )
					{
						sum += getNumFeaturePoints ( i );
					}
					return sum;
				}

				template<class Iterator>
				int getAllFeaturePoints ( Iterator it )
				{
					int count = 0;
					for ( uint i = 0; i < maxFeatureTypeNo; ++i )
					{
						for ( PointsIterator pit = getPointsBegin ( i ); pit != getPointsEnd ( i ); ++pit, ++it )
						{
							*it = *pit; ++count;
						}
					}
					return count;
				}

				/**
				         * Pick the NvPoint that is close to the given 2D-point.
				         * @param x the x-coordinate of the 2D-point.
				         * @param y the y-coordinate of the 2D-point.
				         * @return the NvPoint
				 */
				FeaturePoint *pick ( double x, double y, int featsID = 0 );

				FeaturePoint *pick ( double x, double y, int featsID, double maxDist );

				FeaturePoint *pick ( const FeaturePoint &vec )
				{
					return pick ( vec ( 0 ) / vec ( 2 ), vec ( 1 ) / vec ( 2 ), vec.getFeatureTypeId() );
				}

				void pickNear ( double x, double y, std::vector<FeaturePoint*>& points, int featsID, double maxDist ) const;

				void pickNear ( const FeaturePoint & vec, std::vector<FeaturePoint*>& points, double maxDist ) const
				{
					pickNear ( vec(0) / vec(2), vec(1) / vec(2), points, vec.getFeatureTypeId(), maxDist);
				}

				/**
				         * Add a new FeaturePoint
				         * @param vec the point
				         * @return GRID_SUCCESS (== 0), if everything went well, an error code otherwise (see grid2d.h)
				 */
				GRID_ERR add ( FeaturePoint *const vec, bool checkExistence = true );

				GRID_ERR movePoint ( FeaturePoint *nP, double newX, double newY, double maxDist );

				void remove ( const double x, const double y, const uint featsID = 0 );

				void remove ( const FeaturePoint *const nP );

				bool exists ( FeaturePoint *const vec );

				template<class InputIter> void setTriangles ( InputIter begin, InputIter end,  const uint featsID = 0 )
				{
					assert ( featsID < maxFeatureTypeNo );
					if ( !triangles[featsID] )
						triangles[featsID] = new vector<unsigned> ( begin,end );
					else
						triangles[featsID]->assign ( begin,end );
				}

				void setWidth ( double theValue )
				{
					width = theValue;
				}

				double getWidth() const
				{
					return width;
				}

				void setHeight ( double theValue )
				{
					height = theValue;
				}


				double getHeight() const
				{
					return height;
				}

				void setX ( double theValue )
				{
					x = theValue;
				}

				double getX() const
				{
					return x;
				}

				void setY ( double theValue )
				{
					y = theValue;
				}

				double getY() const
				{
					return y;
				}

				void getMid ( double &mx, double &my )
				{
					mx = x + 0.5 * width;
					my = y + 0.5 * height;
				}

				void setDimensions ( double x, double y, double width, double height )
				{
					DBGOUT ( DBGALL, "View sets its dimensions to (%f, %f, %f, %f)\n", x,y, width, height );
					this->x = x;
					this->y = y;
					this->width = width;
					this->height = height;
				}

				/**
				         * initializes the view's grid and recalculates all internal values according to the
				         * given normalizing transformation.
				         * After setting an image or, alternatively, setting a width and a height, and
				         * after setting an optional normalizing transformation, this method _must_
				         * be called.
				 */
				void initialize();

				void setSameDist ( double theValue )
				{
					sameDist = theValue;
				}

				double getSameDist() const
				{
					return sameDist;
				}

				void setImagePath ( const string& theValue )
				{
					imagePath = theValue;
				}

				const string & getImagePath() const { return imagePath; }

				const AMatrix & getTrafo() const
				{
					return trafo;
				}

				const AMatrix&  getInvTrafo() const
				{
					return invTrafo;
				}

				AMatrix & getTrafo()
				{
					return trafo;
				}

				AMatrix & getInvTrafo()
				{
					return invTrafo;
				}

				void normalize();

				void denormalize();

				bool isNormalized() const {return normalized;}

				bool getNormalized() const {return normalized;}

				/**
				 * tell the view its normalization status.
				 * It's for the Loader, primary.
				 */
				void setNormalized ( const bool val ) {normalized = val;}

				uint getId() const {return id;}

				void setId ( const uint i ) {leastViewId ( i ); id = i;}

				bool hasFeatures ( int featsID =0 ) const
				{
					return features[featsID] != NULL;
				}

				bool hasTriangles ( int featsID =0 )
				{
					return triangles[featsID] != NULL;
				}

				PointsIterator getPointsBegin ( int featsID = 0 )
				{
					if ( !features[featsID] )
					{
						DBGOUT ( DBGNORMAL, "View.h: featsID %d not assigned\n", featsID );
						return empty.getBegin();
					}
					return features[featsID]->getBegin();
				}
				PointsConstIterator getPointsBegin ( int featsID = 0 ) const
				{
					if ( !features[featsID] )
					{
						DBGOUT ( DBGNORMAL, "View.h: featsID %d not assigned\n", featsID );
						return empty.getBegin();
					}
					return features[featsID]->getBegin();
				}

				PointsIterator getPointsEnd ( int featsID = 0 )
				{
					if ( !features[featsID] )
					{
						DBGOUT ( DBGNORMAL, "View.h: featsID %d not assigned\n", featsID );
						return empty.getEnd();
					}
					return features[featsID]->getEnd();
				}
				PointsIterator getPointsEnd ( int featsID = 0 ) const
				{
					if ( !features[featsID] )
					{
						DBGOUT ( DBGNORMAL, "View.h: featsID %d not assigned\n", featsID );
						return empty.getEnd();
					}
					return features[featsID]->getEnd();
				}

				TrianglesIterator getTrianglesBegin ( int featsID = 0 )
				{
					if ( !triangles[featsID] )
					{
						DBGOUT ( DBGNORMAL, "View.h: featsID %d no triangles\n", featsID );
						return dummy.begin();
					}
					return triangles[featsID]->begin();
				}
				TrianglesConstIterator getTrianglesBegin ( int featsID = 0 ) const
				{
					if ( !triangles[featsID] )
					{
						DBGOUT ( DBGNORMAL, "View.h: featsID %d no triangles\n", featsID );
						return dummy.begin();
					}
					return triangles[featsID]->begin();
				}

				TrianglesIterator getTrianglesEnd ( int featsID = 0 )
				{
					if ( !triangles[featsID] )
					{
						DBGOUT ( DBGNORMAL, "View.h: featsID %d no triangles\n", featsID );
						return dummy.end();
					}
					return triangles[featsID]->end();
				}
				TrianglesConstIterator getTrianglesEnd ( int featsID = 0 ) const
				{
					if ( !triangles[featsID] )
					{
						DBGOUT ( DBGNORMAL, "View.h: featsID %d no triangles\n", featsID );
						return dummy.end();
					}
					return triangles[featsID]->end();
				}

				void setPointMode ( const POINT_MODE& theValue );

				POINT_MODE getPointMode() const
				{
					return pointMode;
				}


			private:
				//members associated with grid2d

				FeatureContainer *features[maxFeatureTypeNo];
				FeatureContainer empty;	//dummy container, in case one wants an iterator to non-existing Featurecontainer

				vector<unsigned>* triangles[maxFeatureTypeNo];
				vector<unsigned> dummy;	// one dummy container more

				double x;
				double y;
				double width;
				double height;
				double sameDist;//two points are the same point, if there distance is less than this
				bool normalized;
				string imagePath;
				AMatrix trafo;
				AMatrix invTrafo;

				uint id;

				POINT_MODE pointMode;

				vector<segmentT> segments;
			public:
				void addSegment( const segmentT& s )
				{
					s.get()->setView(this);
					segments.push_back( s );
					if ( normalized )
						segments.back()->transform( trafo( 0, 2 ), trafo( 1, 2 ), trafo( 0, 0 ), trafo( 1, 1 ) );
				}
				size_t getNumSegments() const { return segments.size(); }
				const segmentT& getSegment( size_t idx ) const { return segments.at( idx ); }
				segmentT& getSegment( size_t idx ) { return segments.at( idx ); }
				LinesIterator getLinesBegin() { return segments.begin(); }
				LinesIterator getLinesEnd() { return segments.end(); }
				LinesConstIterator getLinesBegin() const { return segments.begin(); }
				LinesConstIterator getLinesEnd() const { return segments.end(); }
		};


		/**
		 * if d is the number of pixels in a not normalized image, this function will return
		 * the number in a normalized one
		 */
		inline float getNormalizedPixel( const float d, const ViewBase *v )
		{
			assert( v->isNormalized() );
			return d * v->getTrafo()( 0, 0 );
		}

		/**
		 * A View consists of an image and features.
		 *  
		*/

		template<class T>
		class View : public ViewBase
		{
			public:
				typedef T imgT;

			public:
				/**
				 * Constructor
				 * @param img Set the image associated with this View.
				 */
				View ( ImagePlane<T> *const img = NULL, bool deleteImg = false ) :
						ViewBase(),
						image ( img ),
						deleteImage ( deleteImg )
				{
					if ( img )
					{
						setImage ( img );
						adoptDimensionsFromImage();
					}
				}


				/**
				 * Copy-Constructor. Does not copy images associated with the parameter-view.
				 * The images of the new view are set to NULL.
				 * @param view The view to be copied.
				 */
				View ( View &view )
						:ViewBase ( view )
				{
					printf ( "Copy-Ctor in View\n" );
				}


				/**
				 * Destructor. Deletes the image
				 */
				~View()
				{
					DBGOUT ( DBGALL, "deleting View.\n" );
					if ( deleteImage&&image )
					{
						DBGOUT ( DBGALL, "deleting View's image %p.\n", image );
						delete image;
					}
				}

				void copy ( ViewBase *v )
				{
					View<T> *view = dynamic_cast<View<T> *> ( v );
					assert ( view );
					copy ( view );
				}

				void copy ( View<T> *view )
				{
					ViewBase::copy ( view );
					if ( deleteImage && image ) delete image;
					deleteImage = view->deleteImage;
					image = new ImagePlane<T> ( *view->image, true );
				}

				/**
				 * Retrieve the image associated with this View.
				 * @return the image
				 */
				ImagePlane<T> *getImage()
				{
					return image;
				}
				const ImagePlane<T> * getImage() const { return image; }

				/**
				 * Set the image associated with this View.
				 * @param i the image
				 */
				void setImage ( ImagePlane<T> *const i )
				{
					assert ( i );
					image = i;
				}

				void adoptDimensionsFromImage()
				{
					assert ( image );
					this->setDimensions ( 0, 0, image->getWidth(), image->getHeight() );
				}

				void releaseImage() {delete image; image = NULL;}

				bool hasImage() {return image != NULL;}

				int getNumChannels() {return image.getNumChannels();}

				ViewBase *getBase() {return static_cast<ViewBase *> ( this );}

				void setDeleteImage ( const bool b )
				{
					deleteImage = b;
				}

				bool getDeleteImage() const
				{
					return deleteImage;
				}

			private:

				ImagePlane<T> *image;

				bool deleteImage;

		};

	}

}

#endif
