

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef ADORECOMPARATORS_H
#define ADORECOMPARATORS_H

#include <cstdlib>
#include <functional>
#include "elementaccessor.h"
#include <adrcore/util/c++0x.h>

namespace adore
{
	namespace core
	{
		using namespace std;

		/**
		    Functor for the comparison of to vectors according to the value of a given Index.
		    @param T something vectorlike with []-operator
		    @param Cmp how to compare the elements of T. Standard is less&lt;double&gt;
		     
		*/
		template<class T, class Cmp = less<double> >
		class IndexComparator
		{
			public:
				/**
				 * Constructor.
				 * @param i the index of the elements that will be compared.
				 */
				IndexComparator( int i ):idx( i ) {;}

				/**
				 * comparison of references.
				 * @param a reference a.
				 * @param b reference b.
				 */
				bool operator()( const T &a, const T &b ) const
				{
					return comparator( a[idx], b[idx] );
				}

				/**
				 * comparison of pointers.
				 * @param a pointer a.
				 * @param b pointer b.
				 */
				bool operator()( const T *a, const T *b ) const
				{
					return comparator( ( *a )[idx], ( *b )[idx] );
				}

			private:
				const int idx;
				Cmp comparator;
		};

		template<	class T,
		class Indexer = ElemAtSB<T>,
		class Cmp = less<typename Indexer::value_type>
		>
		class IndexComparator3
		{
			public:
				/**
				     * Constructor.
				     * @param i the index of the elements that will be compared.
				     */
				IndexComparator3( int i ):idx( i ), idxAt(), comparator() {;}

				/**
				    	 * comparison of references.
				    	 * @param a reference a.
				    	 * @param b reference b.
				     */
				bool operator()( const T &a, const T &b ) const
				{
					return comparator( idxAt( a, idx ), idxAt( b, idx ) );
				}

			private:
				const int idx;
				const Indexer idxAt;
				const Cmp comparator;
		};

		template<	class TupleT,
		int idx,
		class Cmp = less<typename boost::tuples::element<idx,TupleT>::type >
		>
		class TupleComparator
		{
			public:
				/**
				     * Constructor.
				     * @param i the index of the elements that will be compared.
				     */
				TupleComparator():comparator() {;}

				/**
				    	 * comparison of references.
				    	 * @param a reference a.
				    	 * @param b reference b.
				     */
				bool operator()( const TupleT &a, const TupleT &b ) const
				{
					return comparator( a.get<idx>(), b.get<idx>() );
				}

			private:
				const Cmp comparator;
		};

		/**
		    Functor for the comparison of to vectors according to their euklidian distance relative to another vector.
		     
		*/
		template<class T, class DistT = double, class ElementCmp = less<double> >
		class RelativeDistanceComparator
		{
			public:
				/**
				 * Constructor.
				 * @param c the vector whose distance is essential.
				 */
				RelativeDistanceComparator( T &c )
				{
					cmp = &c;
				}

				/**
				 * comparison of references.
				 * @param a reference a.
				 * @param b reference b.
				 */
				bool operator()( const T &a, const T &b ) const
				{
					T dist = *cmp - a;
					DistT d1 = dist * dist;
					dist = *cmp - b;
					DistT d2 = dist * dist;
					return elemCmp( d1, d2 );
				}

				/**
				 * comparison of pointers.
				 * @param a pointer a.
				 * @param b pointer b.
				 */
				bool operator()( const T *a, const T *b ) const
				{
					T da = *a;
					T db = *b;

					T dist = *cmp - da;
					DistT d1 = dist * dist;
					dist = *cmp - db;
					DistT d2 = dist * dist;
					return elemCmp( d1, d2 );
				}

			private:
				T *cmp;
				ElementCmp elemCmp;
		};

//
//		the versions below use the operators that don't construct new vectors
//

		/**
		Functor for the comparison of to vectors according to their euklidian distance relative to another vector.
		 
		*/
		template<class T, class DistT = double, class ElementCmp = less<double> >
		class RelativeDistanceComparator2
	{
		public:
		/**
		* Constructor.
		* @param c the vector whose distance is essential.
		*/
		RelativeDistanceComparator2( T &c ):cmp( &c ),dist( c.size() )
	{
	}

		/**
		* comparison of references.
		* @param a reference a.
		* @param b reference b.
		*/
		bool operator()( const T &a, const T &b ) const
	{
		( *cmp ).sub2( a, dist );
		DistT d1 = dist.dot2( dist );
		( *cmp ).sub2( b, dist );
		DistT d2 = dist.dot2( dist );
		return elemCmp( d1, d2 );
	}

		/**
		* comparison of pointers.
		* @param a pointer a.
		* @param b pointer b.
		*/
		bool operator()( const T *a, const T *b ) const
	{
		T da = *a;
		T db = *b;

		( *cmp ).sub2( da, dist );
		DistT d1 = dist.dot2( dist );
		( *cmp ).sub2( db, dist );
		DistT d2 = dist.dot2( dist );
		return elemCmp( d1, d2 );
	}

		private:
		T *cmp;
		mutable T dist; //temporary element that holds the distance
		ElementCmp elemCmp;
	};

//
//		the versions below use a ruler object
//

		template<	class T,
		class Ruler,
		class ElementCmp = less<typename Ruler::distance_type>
		>
		class RelativeDistanceComparator3
	{
		private:
		typedef typename Ruler::distance_type DistT;

		const T &cmp;
		const Ruler &ruler;
		const ElementCmp elemCmp;

		public:
		/**
		* Constructor.
		* @param c the vector whose distance is essential.
		*/
		RelativeDistanceComparator3( T &c, Ruler &r ):cmp( c ), ruler( r ) {;}

		/**
		* comparison of references.
		* @param a reference a.
		* @param b reference b.
		*/
		bool operator()( const T &a, const T &b ) const
	{
		DistT d1 = ruler.sqrDist( cmp, a );
		DistT d2 = ruler.sqrDist( cmp, b );
		return elemCmp( d1, d2 );
	}

	};

	}
	}
#endif
