

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#ifndef SCALESHIFTTRAFO_H
#define SCALESHIFTTRAFO_H

#include <adrcore/types/aarray.h>
#include <adrcore/types/rect.h>

namespace adore
{
	namespace core
	{

		/**
		* into = t1 * t2
		* into may be identical with t1 or t2
		*/
		template<class calcT>
                void multScaleShiftTrafo(const AArray<calcT, 2> &t1, const AArray<calcT, 2> &t2, AArray<calcT, 2> &into)
		{
			//last column first to prevent interference
			into(0, 2) = t1(0, 2) + t1(0, 0) * t2(0, 2);
			into(1, 2) = t1(1, 2) + t1(1, 1) * t2(1, 2);
			into(0, 0) = t1(0, 0) * t2(0, 0);
			into(1, 1) = t1(1, 1) * t2(1, 1);
		}

		template<class calcT>
		void transformPoint2D(AArray<calcT, 1> &point, const AArray<calcT, 2> &trafo)
		{
			point(0) = point(0) * trafo(0, 0) + trafo(0, 2);
			point(1) = point(1) * trafo(1, 1) + trafo(1, 2);
		}

		template<class rectT, class calcT>
                void transformRect(Rect<rectT> &rect, const AArray<calcT, 2> &trafo)
		{
			rect.setX(rect.getX() * trafo(0, 0) + trafo(0, 2));
			rect.setY(rect.getY() * trafo(1, 1) + trafo(1, 2));
			rect.setWidth(rect.getWidth() * trafo(0, 0));
			rect.setHeight(rect.getHeight() * trafo(1, 1));
		}

		template<class calcT>
		void getScaleShiftTrafo(double srcW, double srcH, double dstW, double dstH, AArray<calcT, 2> &trafo)
		{
			calcT xScale = srcW > FLT_EPSILON ? dstW / srcW : 1.0;
			calcT yScale = srcH > FLT_EPSILON ? dstH / srcH : 1.0;
			trafo = xScale, 0, 0,
			0, yScale, 0,
			0, 0, 1;
		}
		
		template<class calcT>
		AArray<calcT, 2> getScaleShiftTrafo(double srcW, double srcH, double dstW, double dstH)
		{
			AArray<calcT, 2> trafo(3, 3);
			getScaleShiftTrafo(srcW, srcH, dstW, dstH, trafo);
			return trafo;
		}
		
		template<class calcT>
		void getScaleShiftTrafo(double shiftX, double shiftY, double srcW, double srcH, double dstW, double dstH, AArray<calcT, 2> &trafo)
		{
			calcT xScale = dstW / srcW;
			calcT yScale = dstH / srcH;
			calcT offX = -shiftX * xScale;
			calcT offY = -shiftY * yScale;
			trafo = xScale, 0, offX,
			0, yScale, offY,
			0, 0, 1;
		}

		template<class calcT>
		AArray<calcT, 2> getScaleShiftTrafo(double shiftX, double shiftY, double srcW, double srcH, double dstW, double dstH)
		{
			AArray<calcT, 2> trafo(3, 3);
			getScaleShiftTrafo(shiftX, shiftY, srcW, srcH, dstW, dstH, trafo);
			return trafo;
		}
		
		/**
		 * inverts a matrix of the form
		 * r 0 t1
		 * 0 s t2
		 * 0 0 1
		 * @param normMat the matrix.
		 */
		template<class calcT>
		inline void invertScaleShiftTrafo( AArray<calcT, 2> &normMat )
		{
			calcT scale = 1.0/normMat( 0, 0 );
			normMat( 0, 0 ) = scale;
			normMat( 0, 2 ) = -scale * normMat( 0, 2 );
			scale = 1.0/normMat( 1, 1 );
			normMat( 1, 1 ) = scale;
			normMat( 1, 2 ) = -scale * normMat( 1, 2 );
		}
		
		template<class calcT>
		inline void invertNormalizationMatrix( AArray<calcT, 2> &normMat )
		{
			invertScaleShiftTrafo( normMat );
		}

		/**
		 * The recommended normalization for an image.
		 * It improves the numerical stability.
		 */
		template<class calcT>
		void getNormalizationTrafo( const calcT width,
		                            const calcT height, AArray<calcT, 2> &trafo )
		{
			calcT diag = 2.0 / ( width + height );
			trafo( 0, 1 ) = 0;
			trafo( 1, 0 ) = 0;
			trafo( 2, 0 ) = 0;
			trafo( 2, 1 ) = 0;

			trafo( 0, 0 ) =  diag;
			trafo( 1, 1 ) =  diag;
			trafo( 0, 2 ) =  - width / ( width + height );
			trafo( 1, 2 ) =  - height / ( width + height );
			trafo( 2, 2 ) =  1.0;
		}
		
		template<class calcT>
		inline const AArray<calcT, 2> getNormalizationTrafo( const calcT width, const calcT height )
		{
			AArray<calcT, 2> trafo( 3, 3 );
			getNormalizationTrafo( width, height, trafo );
			return trafo;
		}

		/**
		 * if d is the number of pixels in a not normalized image, this function will return
		 * the number in a normalized one
		 */
		inline float getNormalizedPixel( const float d, const float width, const float height )
		{
			return 2.0F * d  / ( width + height );
		}

		/**
		 * Applies a transformation Matrix.
		 * This is still restricted to a 3x3 Matrix representing the normalization of 2D-Points.
		 * Of course, the inverse may be used as well.
		 * Make sure, the third coordinate is 1.
		 * @param begin Pointer to the Begin of the data
		 * @param numPoints the number of points.
		 * @param trafo the transformation.
		 * @param homogeneousInput the input points have dim + 1 Coordinates and the last coordinate should be set to 1.0
		 */
		template <class IteratorT >
                void applyScaleShiftTrafo( IteratorT begin, const int numPoints, const AMatrix &trafo, bool homogeneousInput = true )
		{
                        typedef typename Ptr<typename std::iterator_traits<IteratorT>::value_type>::type valT;

			double scale = trafo( 0, 0 );
			double xOff  = trafo( 0, 2 );
			double yOff  = trafo( 1, 2 );

			ToPointer<valT> toPointer;
			//scale and translate
			for ( int i = 0; i < numPoints; ++i, ++begin )
			{
				valT pt = toPointer( *begin );
				if ( !pt ) continue;
				( *pt )( 0 ) = scale * ( *pt )( 0 ) + xOff;
				( *pt )( 1 ) = scale * ( *pt )( 1 ) + yOff;
				if ( homogeneousInput ) ( *pt )( 2 ) = 1.0;
			}
		}
		
		template <class IteratorT >
                void apply2DNormalizationTrafo( IteratorT begin, const int numPoints, const AMatrix &trafo, bool homogeneousInput = true )
		{
			applyScaleShiftTrafo( begin, numPoints, trafo, homogeneousInput );
		}


	}
}
#endif // SCALESHIFTTRAFO_H
