

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include "thread.h"
#include <unistd.h>
#include <errno.h>
#include <sys/resource.h>
#include <adrcore/random/random.h>
#define THDBG DBGALL

namespace adore
{

	namespace core
	{

		static uint numThreads = 1;

		uint figureNumThreads()
		{
			long n = sysconf( _SC_NPROCESSORS_ONLN );
			if ( n > 0 )
			{
				numThreads = n;
				return numThreads;
			}
			else
			{
				printf( "sysconf does not support querying the number of processors\n" );
			}
			return 0;
		}

		void setNumThreads( uint n )
		{
			if ( n > 0 )
			{
				numThreads = n;
			}
			else
			{
				printf( "invalid number of threads requested: %u\n", n );
			}
		}

		uint getNumThreads()
		{
			return numThreads;
		}


		static uint threadsCreated = 0;

		uint getThreadsCreated()
		{
			return threadsCreated;
		}

		void incThreadsCreated()
		{
			++threadsCreated;
		}

		///////////////////////////////////////

		Thread::Thread():running( false ), stopSoon( false ), listener( NULL )
		{
		}


		Thread::~Thread()
		{
			if ( running ) stop();
		}

		int Thread::start( bool detached, bool sscope )
		{
			if ( ! running )
			{
				pthread_attr_init( &attr );
				if ( detached )
					pthread_attr_setdetachstate( & attr, PTHREAD_CREATE_DETACHED );

				if ( sscope )
					pthread_attr_setscope( & attr, PTHREAD_SCOPE_SYSTEM );

				running = true;//need to be set before creation, in case thread finishes quickly
				DBGOUT( DBGALL, "Thread %p creates pthread\n", this );
				int code = pthread_create( &id, &attr, Thread::entryPoint, this );
				DBGOUT( DBGALL, "Thread %p finished creating pthread\n", this );
				if ( code != 0 )
				{
					char *errstr = std::strerror( code );
					printf( "Thread: error during creation; code = %d, message = %s\n", code, errstr );
					if ( code == EAGAIN )
					{
						struct rlimit rlim;
						int ret = getrlimit( RLIMIT_NPROC, &rlim );
						assert( ret == 0 );
						ADORE_UNUSED( ret );
						printf( "Thread: threads created: %u - thread limits are %lu/%lu\n", getThreadsCreated(), rlim.rlim_cur, rlim.rlim_max );
					}
					assert( false );
					exit( 1 );
				}
				else
				{
					incThreadsCreated();
				}
				assert( code == 0 );

				DBGOUT( DBGALL, "running -> true\n" );
				return code;
			}
			else
			{
				DBGOUT( DBGALL, "Thread::start : thread is already running\n" );
			}
			return -1;
		}

		void Thread::run()
		{
			execute();
			DBGOUT( DBGALL, "running -> false\n" );
			running = false;
			if ( listener )
			{
				DBGOUT( DBGNORMAL, "Thread notifies StopListener\n" );
				listener->notify( this );
			}
		}

		/*static */
		void *Thread::entryPoint( void * pthis )
		{
			assert( pthis );
			pthread_cleanup_push( cleaner,( void * )0 );
			Thread * pt = ( Thread* )pthis;
			pt->run();
			pthread_cleanup_pop( 1 );
			return NULL;
		}

		/////////////////////////////////


		void JobThread::stop()
		{
			Thread::stop();
			if ( isIdle() )
			{
				runSem.increment();
			}
		}

		void JobThread::doJob( Job *j )
		{
			assert( isIdle() );
			job = j;
			setIdle( false );//need's to be done before increment
			runSem.increment();
		}

		void JobThread::execute()
		{
			DBGOUT( DBGALL, "%p JobThread execute\n", this );
			do
			{
				DBGOUT( DBGALL, "%p exec: locking runSemaphore\n", this );
				runSem.decrement();

				if ( shallStop() && idle )
				{
					DBGOUT( DBGNORMAL, "%p obeys shallStop()\n", this );
					break;
				}

				DBGOUT( DBGALL, "%p exec: passed runSemaphore\n", this );
				if ( job )
				{
//					assert( !isIdle() );//can fail - TODO: is this bad?
					DBGOUT( DBGALL, "%p executing job\n", this );
					job->execute();
					setIdle( true );
				}
				else if ( !idle )
				{
					DBGOUT( DBGCRITICAL, "%p job was NULL\n", this );
					assert( false );
					exit( 1 );
				}
			}
			while ( !( shallStop() ) );
			DBGOUT( DBGALL, "%p JobThread done\n", this );
		}

		bool JobThread::isIdle()
		{
			return idle;
		}

		void JobThread::setIdle( bool val )
		{
//			Lock l( idleMutex );
			if ( val )
			{
				DBGOUT( DBGALL, "%p idle true\n", this );
				idle = true;
				joblistener->jobNotify( this );
				joinMutex.unlock();
			}
			else
			{
				DBGOUT( DBGALL, "%p idle false\n", this );
				joinMutex.lock();
				idle = false;
			}
		}

		void JobThread::joinJob()
		{
			Lock l( joinMutex );//blocks, if not idle
		}

		////////////////////////////////////

		ThreadPool::ThreadPool( uint nThreads ):threads( nThreads ), semaphore( nThreads )
		{
			DBGOUT( DBGNORMAL, "ThreadPool creates %d threads\n", nThreads );
			for ( uint i = 0; i < nThreads; ++i )
			{
				threads[i].setJobFinishedListener( this );
				assert( threads[i].isIdle() );
			}
		}

		JobThread *ThreadPool::getIdle()
		{
			DBGOUT( THDBG, "i1-semaphore: %d\n", semaphore.getValue() );
			semaphore.decrement();//blocks at 0
			DBGOUT( THDBG, "i2-semaphore: %d\n", semaphore.getValue() );
			JobThread *result = NULL;
			for ( uint i = 0; i < threads.size(); ++i )
			{
				if ( threads[i].isIdle() )
				{
					result = &threads[i];
				}
			}
			assert( result );
			return result;
		}

		void ThreadPool::jobNotify( JobThread *t )
		{
			DBGOUT( THDBG, "%p notifies\n", t );
			assert( t->isIdle() );
			DBGOUT( THDBG, "n1-semaphore: %d\n", semaphore.getValue() );
			semaphore.increment();
			DBGOUT( THDBG, "n2-semaphore: %d\n", semaphore.getValue() );
		}

		void ThreadPool::joinJobs()
		{
//			printf( " -- joining --\n" );
			for ( uint i = 0; i < threads.size(); ++i )
			{
				if ( !threads[i].isIdle() ) threads[i].joinJob();
				assert( threads[i].isIdle() );
			}
			assert_msg( semaphore.getValue() == ( int )threads.size(), "%d/%d\n", semaphore.getValue(), threads.size() );
		}

		void ThreadPool::join()
		{
			for ( uint i = 0; i < threads.size(); ++i )
			{
				assert( threads[i].isIdle() );
				threads[i].stop();
				if ( threads[i].isRunning() ) threads[i].join();
				assert( threads[i].isIdle() );
				assert( !threads[i].isRunning() );
			}
		}

	}

}
