

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include "epilinematcher.h"
#include <adrcore/features/linedistfilter.h>
#include "matchsorter.h"
#include <adrcore/util/printmat.h>
#include <adrcore/imgprocessing/drawing.h>

namespace adore
{

	namespace core
	{

        void getEpi2( AMatrix & fmat, AVector & epipole, SVDStoraged & store3 )
		{
			transInPlace( fmat );
			FMatrixCalculator::calculateEpipolarPoint( fmat, epipole, store3 );
			transInPlace( fmat );
			inhomogenize<2>( epipole );
		}

		void EpiLineMatcher::match ( ViewBase *v1, ViewBase *v2, NView *set )
		{
			if ( isNullArray( fmat ) )
			{
				if ( set->getNumViews() != 2 )
				{
					DBGOUT( DBGALL, "EpiLineMatcher: wrong number of views\n" );
					return;
				}
				if ( !set->isNViewSet() )
				{
					DBGOUT( DBGALL, "EpiLineMatcher: not an NViewSet\n" );
					return;
				}
				NViewInfo *info = set->getInfo();
				if ( isNullArray( info->getFMatrix() ) )
				{
					DBGOUT( DBGALL, "EpiLineMatcher: No fundamental matrix\n" );
					return;
				}
				fmat.resize( 3, 3 );
				fmat = info->getFMatrix();
			}
			if(set->isNViewSet())
			{
				NViewInfo *info = set->getInfo();
				if ( !isNullArray( info->getFMatrix() ) )
				{
					DBGOUT( DBGALL, "EpiLineMatcher: fundamental matrix of NView overrides local FMat\n" );
					fmat.resize( 3, 3 );
					fmat = info->getFMatrix();
				}
			}
			if ( isNullArray( fmat ) )
			{
				DBGOUT( DBGALL, "EpiLineMatcher: Still No fundamental matrix\n" );
				return;
			}
			DBGOUT( DBGALL, "EpiLineMatcher: Everything is fine\n" );
			match ( v1, v2, set, fmat );
		}


		/**
		 * Has currently a runtime of O(n²)
		 */
		EpiLineMatcher::PairT EpiLineMatcher::findSuitable( FeaturePoint &fp, ViewBase *v2, const AMatrix &fmat, const AVector &epi2 )
		{
			PairT best( NULL, FLT_MAX );
			PairT good( NULL, FLT_MAX );

			AVector line = matMul( fmat, fp );
			LineDistance ldist( line( 0 ), line( 1 ), line( 2 ) );
			if ( !ldist.isOkay() )
			{
				DBGOUT( DBGALL, "EpiLineMatcher: insane line occurred" );
				printMat( DBGALL, fmat );
				printMat( DBGALL, fp );
				return best;
			}

			int count = 0;
			FeatureDescriptor *fd1 = fp.getDescriptor( descrID );
			if ( fd1 )
			{
				for ( uint k = 0; k < fd1->size(); ++k )
				{
					for ( ViewT::PointsIterator it = v2->getPointsBegin( matchID );
					                it != v2->getPointsEnd( matchID ); ++it )
					{
						FeaturePoint &cand = **it;
						assert( cand( 2 ) == 1.0 );
						float dist = fabsf( ldist.getDistance( cand( 0 ), cand( 1 ) ) );
// 						printf( "dist: %f\n", dist );
						if ( dist < lineDistThresh )
						{
							++count;
//							printf( "dist < %f\n", thresh );

							bool ori = FMatrixCalculator::isOriented ( fmat, epi2, fp, cand );
							if ( ori )
							{
								FeatureDescriptor *fd2 = ( *it )->getDescriptor( descrID );
								if ( fd2 )
								{
									for ( uint j = 0; j < fd2->size(); ++j )
									{
										float dist = getDistance( ( *fd1 )[k], ( *fd2 )[j] );
										if ( dist < featDistThresh )
										{
											if ( dist < best.second )
											{
												good = best;
												best.first = *it;
												best.second = dist;
											}
											else if ( dist < good.second )
											{
												good.first = *it;
												good.second = dist;
											}
										}
									}
								}
							}
							else
							{
								DBGOUT( DBGALL, "Epilinematcher featpair was not oriented\n" );
							}
						}
					}
				}
			}
			else
			{
				DBGOUT( DBGALL, "Epilinematcher needs a DESCRIPTOR\n" );
			}
			DBGOUT( DBGALL, "EpiLineMatcher: ckecked %d candidates\n", count );

			if ( best.second / good.second < ratioThresh )	return best;
			return PairT( NULL, FLT_MAX );
		}

		void EpiLineMatcher::match( ViewBase *v1, ViewBase *v2, NView *set, AMatrix &fmat )
		{
// 			DBGOUT( DBGALL, "EpiLineMatcher: matching view %p and %p of NView %p\n", v1, v2, set);
			DBGOUT( DBGALL, "EpiLineMatcher: matching %d with %d\n", v1->getNumFeaturePoints( matchID ), v2->getNumFeaturePoints( matchID ) );
			
// 			printMat(DBGALL, fmat);
// 			ImagePlane<uint8_t> *img = visualizeEpipolarGeometry( fmat, dynamic_cast<View<uint8_t>*>(v1), dynamic_cast<View<uint8_t>*>(v2) );
// 			img->save("epilinematcher_image.png");
// 			pressKey();

			getEpi2 ( fmat, epi2, store3 );

			//re-orient fmt in case is has been optimized and lost its sign
			for ( NView::CorrIterator cit = set->getPointCorrsBegin(); cit != set->getPointCorrsEnd(); ++cit )
			{
				PointCorr *pc = *cit;
				FeaturePoint *fp1 = pc->getFeature( v1 );
				FeaturePoint *fp2 = pc->getFeature( v2 );
				bool io1 = FMatrixCalculator::isOriented ( fmat, epi2, *fp1, *fp2 );
				bool io2 = FMatrixCalculator::isOriented ( -fmat, epi2, *fp1, *fp2 );
				if ( io1 != io2 )
				{
					if ( io2 ) fmat *= -1;
					break;
				}
			}

			MatchSorter mSort( 100/*reserve*/ );
			for ( ViewT::PointsIterator it = v1->getPointsBegin( matchID ); it != v1->getPointsEnd( matchID ); ++it )
			{
				FeaturePoint *fp1 = *it;
				PairT fp2 = findSuitable( *fp1, v2, fmat, epi2 );

				if ( fp2.first )
				{
					mSort.add( fp1, fp2.first, fp2.second );
				}
				else
				{
					DBGOUT( DBGALL, "EpiLineMatcher wasn't able to find a match.\n" );
//					printMats( DBGALL, set->getCamera( v1 )->getProjMat(), set->getCamera( v2 )->getProjMat() );
//					pressKey();
				}
			}
			mSort.sort();

			int count = 0;
			for ( MatchSorter::iterator it = mSort.begin();
			                ( it != mSort.end() ) && ( count < countThresh ); ++it )
			{
//				DBGOUT( DBGALL, "try adding match with badness %f\n", ( *it ).badness );
//				if ( ( *it ).badness > thresh ) break;

				DBGOUT( DBGALL, "adding match with badness %f\n", ( *it ).badness );
				FeaturePoint *fp1 = ( *it ).fp1;
				FeaturePoint *fp2 = ( *it ).fp2;
//				DBGOUT( DBGALL, "%p --- %p\n", fp1, fp2 );

				count += this->addPairToNview( fp2 /*certain*/, fp1 /*doubtful*/, set, ( *it ).badness ) ? 1 : 0;
			}

			DBGOUT( DBGNORMAL, "EpiLineMatcher added %d corrs\n", count );
		}



	}
}
