

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include "kdtreematcher.h"
#include "matchsorter.h"
#include <adrcore/search/kdtree3.h>

namespace adore
{

	namespace core
	{

		typedef KDTree3<KDTreeMatcher::iterator, KDTreeMatcher::BackRefRuler> KDTreeT;

		/**
		     * puts the descriptors into the backref data structure
		     */
		void KDTreeMatcher::takeVecs( ViewBase *v, CollectionT &backrefs )
		{
			assert( v );
			backrefs.clear();
			int numFeats = v->getNumFeaturePoints( matchID );
			if ( numFeats > 0 )
			{
				backrefs.reserve( numFeats );
				for ( ViewT::PointsIterator it = v->getPointsBegin( matchID ); it != v->getPointsEnd( matchID ); ++it )
				{
					FeaturePoint *fp = *it;
					assert( fp );
					FeatureDescriptor *fd = fp->getDescriptor( descrID );
					if ( fd )
					{
						for ( uint j = 0; j < fd->size(); ++j )
						{
							backrefs.push_back( DBackRef( &(*fd)[j], fp ) );
						}
					}
				}
			}
		}

		void KDTreeMatcher::match( ViewBase *v1, ViewBase *v2, NView *set )
		{
//			cout<<"KDTreeMatcher"<<endl;
			set->addView( v1 );//just in case ...
			set->addView( v2 );

			takeVecs( v1, backrefs1 );
			if ( backrefs1.size() < 1 )
				{DBGOUT( DBGCRITICAL, "KDTreeMatcher: nothing to match in view 1\n" ); return;}

			takeVecs( v2, backrefs2 );
			if ( backrefs2.size() < 1 )
				{DBGOUT( DBGCRITICAL, "KDTreeMatcher: nothing to match in view 2\n" ); return;}

			match( set );
		}

		void addBest( Feature *const f1, Feature *const f2, const float dist, MatchSorter &mSort )
		{
			FeaturePoint *fp1 = dynamic_cast<FeaturePoint *>( f1 );
			FeaturePoint *fp2 = dynamic_cast<FeaturePoint *>( f2 );
			assert( fp1 );
			assert( fp2 );
			mSort.add( fp1, fp2, dist );
		}

		void addRatioBest( KDTreeMatcher::DBackRef &vec, KDTreeT::NearestCollectorT &nc, float goodRatio, MatchSorter &mSort )
		{
			KDTreeT::NearestCollectorT::iterator it0 = nc.begin();
			KDTreeT::NearestCollectorT::iterator it1 = it0 + 1;
			float dist1 = ( *it1 ).get<1>();
			float dist2 = ( *it0 ).get<1>();
			assert( dist2 >= dist1 );
			if ( dist1 / dist2 <= goodRatio )
			{
				const KDTreeMatcher::DBackRef *best = ( *it1 ).get<0>();
				addBest( best->backRef, vec.backRef, dist1, mSort );
			}
		}

		/**
		 * adds the best into mSort if it fulfills the goodRatio-criterion
		 * the others are added into waste if their ratio to the best is not worse than goodRatio
		 */
		void addBetterThanRatio( KDTreeMatcher::DBackRef &vec, KDTreeT::NearestCollectorT &nc, float goodRatio, MatchSorter &mSort, MatchSorter &waste )
		{
			nc.sort();
			KDTreeT::NearestCollectorT::iterator it0 = nc.begin();
			KDTreeT::NearestCollectorT::iterator it1 = it0 + 1;
			float dist1 = ( *it0 ).get<1>();
			float dist2 = ( *it1 ).get<1>();
			assert( dist2 >= dist1 );
			KDTreeMatcher::DBackRef *best = ( *it0 ).get<0>();
			if ( dist1 / dist2 <= goodRatio )
			{
				//unique match
				addBest( best->backRef, vec.backRef, dist1, mSort );
			}
			else
			{
				//disputable match
				addBest( best->backRef, vec.backRef, dist1, waste );
				for ( ; it1 != nc.end(); ++it1 )
				{
					float dist2 = ( *it1 ).get<1>();
					assert( dist2 >= dist1 );
					if ( dist1 / dist2 > 0.9 )
					{
						best = ( *it1 ).get<0>();
						addBest( best->backRef, vec.backRef, dist2, waste );
					}
					else
					{
						break;
					}
				}
			}
		}

		void KDTreeMatcher::match( NView *set )
		{
			iterator set1begin = backrefs1.begin();
			iterator set1end = backrefs1.end();
			iterator set2begin = backrefs2.begin();
			iterator set2end = backrefs2.end();

			if ( set1begin == set1end || set2begin == set2end ) return;

			DBGOUT( DBGNORMAL, "KDTreeMatcher: matching set of size %zd and %zd\n", backrefs1.size(), backrefs2.size() );

			MatchSorter mSort( backrefs2.size() /*reserve*/ );

			int dim = ( *set1begin ).size();

			BackRefRuler ruler( dim );
			KDTreeT kdtree(
			        set1begin,
			        set1end,
			        dim,
			        ruler );

			for ( ; set2begin != set2end; ++set2begin )
			{
				DBackRef &vec = *set2begin;

				switch ( mode )
				{
					case KDTREE_SINGLE_BEST:
					{
						KDTreeT::SingleBestT sb;
						kdtree.findNearestNeighbour( vec, sb );
						DBackRef *best = sb.getBest().get<0>();
						float dist = sb.getBest().get<1>();
						addBest( best->backRef, vec.backRef, dist, mSort );
					}
					break;
					case KDTREE_BBF_SINGLE_BEST:
					{
						KDTreeT::SingleBestT sb;
						kdtree.findNearestNeighbourBBF( vec, maxBBFNodes, sb );
						DBackRef *best = sb.getBest().get<0>();
						float dist = sb.getBest().get<1>();
						addBest( best->backRef, vec.backRef, dist, mSort );
					}
					break;
					case KDTREE_RATIO_BEST:
					{
						KDTreeT::NearestCollectorT nc( 2 );
						kdtree.findNearestNeighbour<KDTreeT::NearestCollectorT>( vec, nc );
						addRatioBest( vec, nc, goodRatio, mSort );
					}
					break;
					case KDTREE_BBF_RATIO_BEST:
					{
						KDTreeT::NearestCollectorT nc( 2 );
						kdtree.findNearestNeighbourBBF<KDTreeT::NearestCollectorT>( vec, maxBBFNodes, nc );
						addRatioBest( vec, nc, goodRatio, mSort );
					}
					break;
					case KDTREE_BETTER_THAN_RATIO:
					{
						assert( waste );
						KDTreeT::NearestCollectorT nc( 4 );
						kdtree.findNearestNeighbour<KDTreeT::NearestCollectorT>( vec, nc );
						addBetterThanRatio( vec, nc, goodRatio, mSort, *waste );
					}
					break;
					case KDTREE_BBF_BETTER_THAN_RATIO:
					{
						assert( waste );
						KDTreeT::NearestCollectorT nc( 4 );
						kdtree.findNearestNeighbourBBF<KDTreeT::NearestCollectorT>( vec, maxBBFNodes, nc );
						addBetterThanRatio( vec, nc, goodRatio, mSort, *waste );
					}
					break;
					default:
						break;
				}


//				if ( best )
//				{
////                 DBGOUT(DBGNORMAL, "match found\n");
//					FeaturePoint *fp1 = dynamic_cast<FeaturePoint *>( best->backRef );
//					FeaturePoint *fp2 = dynamic_cast<FeaturePoint *>( vec.backRef );
//					assert( fp1 );
//					assert( fp2 );
//					mSort.add( fp1, fp2, dist1 );
//				}
			}

			mSort.sort();

			int count = 0;
			for ( MatchSorter::iterator it = mSort.begin();
			                ( it != mSort.end() ) && ( count < countThresh ); ++it )
			{
				if ( ( *it ).badness > thresh ) break;

//				DBGOUT( DBGALL, "adding match with badness %f\n", ( *it ).badness );
				FeaturePoint *fp1 = ( *it ).fp1;
				FeaturePoint *fp2 = ( *it ).fp2;
//				DBGOUT( DBGALL, "%f, %f\n", ( *fp1 )( 0 ), ( *fp1 )( 1 ) );
//				DBGOUT( DBGALL, "%f, %f\n", ( *fp2 )( 0 ), ( *fp2 )( 1 ) );
//				pressKey();

				//since the feats of the kdtree can be match more than once, test
				//them. this is necessary because the last one added wins
				count += this->addPairToNview( fp1, fp2, set, ( *it ).badness ) ? 1 : 0;
			}

		}


	}
}//Namespace
