

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Sergey Cheremukhin


#include "linesfilteringmerger.h"

#include <adrcore/types/view.h>
#include <adrcore/types/nview.h>
#include <adrcore/types/nviewmap.h>
#include <adrcore/util/scaleshifttrafo.h>

#include <vector>
#include <map>

#include <boost/tuple/tuple.hpp>

using namespace std;

namespace adore
{
	namespace core
	{

		typedef boost::tuple<FeatureLine*, FeatureLine*, FeatureLine*> LineCorr3;
		typedef map<double, LineCorr3> CorrSorter;
		class CorrComparator
		{
			public:
				bool operator() ( const LineCorr3& x, const LineCorr3& y )
				{
					return x.get<0>() < y.get<0>();
				}
		};
		typedef map<LineCorr3, double, CorrComparator> CorrsT;

		LinesFilteringMerger::LinesFilteringMerger( double thresh ) :
				threshold( thresh )
		{
		}

		void LinesFilteringMerger::merge( NViewMap& nviewmap )
		{
			NViewMap::EntryIterator eit = nviewmap.getEntriesBegin();
			vector<NView*> sets3;
			map< ViewBase*, NView* > firstViewSorted2;
			for ( ; eit != nviewmap.getEntriesEnd(); ++eit )
			{
				NView* nv = *eit;
				if ( nv->getNumViews() != 2 && nv->getNumViews() != 3 )
					continue;
				if ( nv->getNumViews() == 2 )
					firstViewSorted2[*( nv->getViewsBegin() )] = nv;
				else
					sets3.push_back( nv );
			}
			for ( uint i = 0; i < sets3.size(); ++i )
			{
				NView* set = sets3[i];
				NView::ViewIterator vit = set->getViewsBegin();
				ViewBase* vb = *vit;
				if ( firstViewSorted2.find( vb ) == firstViewSorted2.end() )
					continue;
				NView* nv1 = firstViewSorted2.find( vb )->second;
				++vit;
				vb = *vit;
				if ( vb != *( ++( nv1->getViewsBegin() ) ) )
					continue;
				if ( firstViewSorted2.find( vb ) == firstViewSorted2.end() )
					continue;
				NView* nv2 = firstViewSorted2.find( vb )->second;
				++vit;
				vb = *vit;
				if ( vb != *( ++( nv2->getViewsBegin() ) ) )
					continue;
				mergeLines( nv1, nv2, set );
			}
		}

		void LinesFilteringMerger::mergeLines( NView* nv1, NView* nv2, NView* set )
		{
			NViewInfo *info = set->getInfo();
			if ( isNullArray( info->getTTensor() ) )
			{
				DBGOUT( DBGNORMAL, "LinesFilteringMerger: No trifocal tensor.\n" );
				return;
			}
			ATensor3 ttensor( 3, 3, 3 );
			ttensor = info->getTTensor();

			NView::ViewIterator vit = set->getViewsBegin();
			vector<ViewBase*> views;
			for ( ; vit != set->getViewsEnd(); ++vit )
				views.push_back( *vit );
			vector<NView*> nviews;
			nviews.push_back( nv1 );
			nviews.push_back( nv2 );

			Corr1ToN * corr1( NULL ), * corr2( NULL );
			FeatureLine * fl1( NULL ), * fl2( NULL ), * fl3( NULL );
			AVector l1( 3 ), l2( 3 ), l3( 3 );
			double w = views[0]->getWidth(), h = views[0]->getHeight();
			double x = views[0]->getX(), y = views[0]->getY();
			double thresh = threshold;
			if ( views[0]->isNormalized() )
			{
				thresh = getNormalizedPixel( thresh, w, h );
			}

			NView::MultiCorrIterator mit = nviews[0]->getMultiLinesBegin();

			CorrSorter sorter;
			CorrsT corrs;
			for ( int i = 0 ; mit != nviews[0]->getMultiLinesEnd(); ++mit, ++i )
			{
				corr1 = *mit;
				fl1 = static_cast<FeatureLine*>( corr1->getFeature( views[0] ) );
				if ( NULL == fl1 )
					continue;

				sorter.clear();
				Corr1ToN::ConstIterator f2it = corr1->begin();
				for ( int j = 0 ; f2it != corr1->end(); ++f2it, ++j )
				{
					fl2 = static_cast<FeatureLine*>( *f2it );
					if ( NULL == fl2 )
						continue;

					corr2 = nviews[1]->getLineMultiCorr( static_cast<Feature*>( fl2 ) );
					if ( NULL == corr2 )
						continue;

					Corr1ToN::ConstIterator f3it = corr2->begin();
					for ( int k = 0 ; f3it != corr2->end(); ++f3it, ++k )
					{
						fl3 = static_cast<FeatureLine*>( *f3it );
						toLine( fl2, l2 ); toLine( fl3, l3 );
						AVector l1test( 3 );
						transfer23_1( l2, l3, ttensor, l1 );
						double x1, y1, x2, y2;
						bool ok = pointAsLine2( l1( 0 ), l1( 1 ), l1( 2 ), x1, y1, x2, y2, w, h, x, y );
						if ( ok )
						{
//							cout << "x1: " << x1 << ", y1: " << y1 << ", x2: " << x2 << ", y2: " << y2 << endl;
							const AVector& pt1 = fl1->getPt1();
							const AVector& pt2 = fl1->getPt2();
							double dist1 = distance( x1, y1, x2, y2, pt1( 0 ), pt1( 1 ) );
							double dist2 = distance( x1, y1, x2, y2, pt2( 0 ), pt2( 1 ) );
							LineCorr3 corr( fl1, fl2, fl3 );
							if ( dist1 < thresh && dist2 < thresh )
							{
								sorter.insert( make_pair( dist1 + dist2, corr ) );
//								cout << i << ", " << j << ", " << k << endl;
							}
						}
					}
				}
				if ( !sorter.empty() )
				{
					pair<double, LineCorr3> val = *( sorter.begin() );
					corrs.insert( make_pair( val.second, val.first ) );
				}
			}
			map<FeatureLine*, LineCorr3> checker2;
			CorrsT::iterator cit = corrs.begin();
			for ( ; cit != corrs.end(); ++cit )
			{
				LineCorr3 lc3 = cit->first;
				checker2[lc3.get<1>()] = lc3;
			}
			map<FeatureLine*, LineCorr3> checker3;
			map<FeatureLine*, LineCorr3>::iterator cit2 = checker2.begin();
			for ( ; cit2 != checker2.end(); ++cit2 )
			{
				LineCorr3 lc3 = cit2->second;
				checker3[lc3.get<2>()] = lc3;
			}

			vector<FeatureLine*> temp( 3 );
			cit2 = checker3.begin();
			for ( ; cit2 != checker3.end(); ++cit2 )
			{
				LineCorr3 lc3 = cit2->second;
				double badness = corrs.find( lc3 )->second;
				temp[0] = lc3.get<0>();
				temp[1] = lc3.get<1>();
				temp[2] = lc3.get<2>();
				set->addLineCorrNoCheck( temp.begin(), temp.end(), badness );
			}
		}

		void LinesFilteringMerger::toLine( FeatureLine* fl, AVector& l )
		{
			const AVector& pt1( fl->getPt1() );
			const AVector& pt2( fl->getPt2() );
			cross( pt1, pt2, l );
		}

		void LinesFilteringMerger::transfer23_1( const AVector& l2, const AVector& l3, const ATensor3& tt, AVector& l1 )
		{
			AMatrix Ti( 3, 3 );
			AVector result( 3 );
			for ( int i = 0; i < 3; ++i )
			{
				Ti = tt( i, Range::all(), Range::all() );
				matMul2_AT( Ti, l2, result );
				l1( i ) = dot( result, l3 );
			}
		}

		double LinesFilteringMerger::distance( double x1, double y1, double x2, double y2, double px, double py )
		{
			double x2x1 = x2 - x1;
			double y2y1 = y2 - y1;
			return abs( x2x1 * ( y1 - py ) - ( x1 - px ) * y2y1 ) / sqrt( x2x1 * x2x1 + y2y1 * y2y1 );
		}

	}
}
