

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include <adrcore/viewmatch/scaleharrisfinder.h>
#include <adrcore/types/view.h>
#include <adrcore/imgprocessing/kernelbinom_3_7__3_3.h>
#include <adrcore/imgprocessing/kernelbinom_delta_3_3.h>
#include <adrcore/imgprocessing/convolver2d.h>
#include <adrcore/imgprocessing/scalespacetraverser.h>
#include <adrcore/imgprocessing/rgbtogray.h>
#include <adrcore/imgprocessing/borderfiller.h>
#include <adrcore/features/featurefilter.h>
#include <memory>
#include <adrcore/util/debug.h>

#include <adrcore/imgprocessing/typeconverter.h>

namespace adore
{
	namespace core
	{
		void saveHarry( ImagePlane< ScaleHarrisFinder::harrisT > harrisImg, const char * name )
		{
			TypeConverter<ScaleHarrisFinder::harrisT, uint8_t> tc;
			ImagePlane<uint8_t> img( harrisImg.getWidth(), harrisImg.getHeight(), 1 );
			tc.process( &harrisImg, &img );
			img.save( name );
		}

		ScaleHarrisFinder::ScaleHarrisFinder ( uint ID, int _numSmooths, int _numImages, calcT minHarrisThresh, int maxPoints , float minDistance, float minFrac )
				: FeatureFinder(),
				fid ( ID ),
				patchImg ( 25, 25, 1 ),
				extrema ( minHarrisThresh ),
				candidates ( false ),//don't delete points
				numSmooths ( _numSmooths ),
				numImages ( _numImages ),
				descriptorcreator ( NULL ),
				ownCreator ( false ),
				maxPts ( maxPoints ),
				minDist ( minDistance ),
				minFraction ( minFrac ),
				currentView ( 0 ),
				patchFactor ( 2 ),
				snsRemove ( false ),
				snsMat ( 2, 2 )
		{
			assert ( numImages > 0 );
			currentView.setPointMode ( ViewBase::POINT_FORGET );//prevent Featurepoint deletion
			snsMat.setIdentity();
		}


		ScaleHarrisFinder::ScaleHarrisFinder ( const ScaleHarrisFinder &shf ) :FeatureFinder(),
				fid ( shf.fid ),
				patchImg ( 25, 25, 1 ),
				extrema ( shf.extrema.getThresh() ),
				candidates ( false ),
				numSmooths ( shf.numSmooths ),
				numImages ( shf.numImages ),
				descriptorcreator ( NULL ),
				ownCreator ( false ),
				maxPts ( shf.maxPts ),
				minDist ( shf.minDist ),
				minFraction ( shf.minFraction ),
				currentView ( 0 ),
				patchFactor ( shf.patchFactor ),
				snsRemove ( shf.snsRemove ),
				snsMat ( 2, 2 )
		{
			assert ( numImages > 0 );
			currentView.setPointMode ( ViewBase::POINT_FORGET );//prevent Featurepoint deletion
			snsMat.setIdentity();
			DescriptorCreator* dc = shf.getDescriptorCreator();
			if ( dc )
			{
				setDescriptorCreator ( dc->clone() );
				ownCreator = true;
			}
		}

		ScaleHarrisFinder::~ScaleHarrisFinder()
		{
			if ( ownCreator ) delete descriptorcreator;
		}

		unsigned int ScaleHarrisFinder::findFeaturesDerived ( ViewT *v )
		{

			ImagePlane<imgT> *chan = v->getImage();
			assert ( chan );

			assert ( !descriptorcreator || descriptorcreator->getFeatureID() == fid );

			candidates.init ( 0, 0, chan->getWidth(), chan->getHeight() );

			ImagePlane<imgT> *gray = NULL;
			if ( chan->getNumChannels() ==3 )
			{
				gray = new ImagePlane<imgT> ( ( int ) v->getImage()->getWidth(), ( int ) v->getImage()->getHeight(),1 );
				RGBtoGray<imgT> converter;
				Channel<imgT>::Set in ( true );
				chan->createChannels ( in );
				converter.process ( in, gray );
				chan=gray;
			};
			auto_ptr< ImagePlane<imgT> > grayDeletor ( gray ); //does nothing if gray is NULL

			//this convolver is used to generate the scale space
			typedef Convolver2D<KernelBinom_3_7__3_3<uint8_t> > Convolver2DT;
			Convolver2DT convolver2d;
			float delta = 3.0F / 7.0F;
			scale = getPreSmoothingAmount ( delta );

			//this convolver is used for a pre-smoothing
			typedef Convolver2D<KernelBinom_delta_3_3<uint8_t> > PreConvolver2DT;
			PreConvolver2DT preConvolver2d;
			preConvolver2d.getKernel().setDelta ( scale );

			//build the traverser
			typedef ScaleSpaceTraverser<Convolver2DT, PreConvolver2DT> ScaleSpaceTraverserT;
			ScaleSpaceTraverserT sstv ( chan, ScaleSpaceTraverserT::STATE_SMOOTH, //TODO PRESMOOTH?
			                            numSmooths /*smooths until the first subsampling*/,
			                            numSmooths /*smooths between subsamplings*/,
			                            2 /*number of images cached*/ );
			sstv.setConvolver ( &convolver2d );
			sstv.setPreConvolver ( &preConvolver2d );
			sstv.setSubSampleMode( ScaleSpaceTraverserT::STATE_SUBSAMPLE_AND_SMOOTH );

			BorderFiller<harrisT> bfill;

			ImagePlane<imgT> *plane = chan;

			scale = 1.0F;//TODO
//			ostringstream oss;
			for ( int i = 0; i < numImages; ++i )
			{
				plane = sstv.next();
//				DBGOUT( DBGNORMAL, "ScaleHarrisFinder: image %d/%d: (%d, %d)\n", i, numImages, plane->getWidth(), plane->getHeight() );

				spacing = sstv.getSpacing();

				harrisImg.init ( plane->getWidth(), plane->getHeight(), 1 );
				float scaleVal = scale / spacing;
				int hood = max ( 1, int ( scaleVal + 0.5F ) );
//				DBGOUT ( DBGNORMAL, "scaleVal = %f, hood: %d\n", scaleVal, hood );
				bfill.setWidth ( hood );
				bfill.BorderFiller<harrisT>::BaseT::process ( &harrisImg );
				HarrisFastHood<imgT> hafa ( hood, scaleVal );//TODO ceil or round?
				hafa.process ( plane, &harrisImg );

//				oss.str( "" );
//				oss << "harry_" << harrisImg.getWidth() << "_" << i << ".png";
//				saveHarry( harrisImg, oss.str().c_str() );

				//prepare the view for the descriptorcreator
				//and for affine matrix creation
				if ( descriptorcreator )
				{
					DBGOUT ( DBGSHF, "resetting view\n" );
					currentView.reset();
				}
				DBGOUT ( DBGSHF, "setting image\n" );
				currentView.setImage ( plane );
				DBGOUT ( DBGSHF, "adopting dimensions\n" );
				currentView.adoptDimensionsFromImage();
				DBGOUT ( DBGSHF, "initializing\n" );
				currentView.initialize();

				//adds the features to currentVIew and to candidates
				extrema ( harrisImg, *this );

				if ( descriptorcreator )
				{
					descriptorcreator->performAction ( &currentView );
					//move features and add to candidates TODO
					for ( ViewT::PointsIterator it = currentView.getPointsBegin ( fid );
					                it != currentView.getPointsEnd ( fid ); ++it )
					{
						//scale now
						( **it ) ( 0 ) = getPixelPosition( ( **it ) ( 0 ), spacing );
						( **it ) ( 1 ) = getPixelPosition( ( **it ) ( 1 ), spacing );
						candidates.add ( *it );
					}
				}

//				DBGOUT ( DBGNORMAL, "%d features at scale %f\n", candidates.getSize(), scale );
//				DBGOUT ( DBGNORMAL, "%d features at scale %f (maxStrength is %f at spacing %d)\n", candidates.getSize(), scale, ( *candidates.getBegin() )->getStrength(), spacing );

				scale += sstv.getScaleStep ( 1, delta );
			}
			currentView.reset();//clean currentview finally

			DBGOUT ( DBGSHF, "ScaleHarris: filtering features\n" );
			filterFeatures ( candidates, minDist, minFraction, maxPts );
			DBGOUT ( DBGSHF, "%d features after filtering\n", candidates.getSize() );

			FeatureGrid::Iterator it = candidates.getBegin();
			FeatureGrid::Iterator endIt = candidates.getEnd();
			while ( it != endIt )
			{
				FeaturePoint *fp = *it;
				fp->setView ( v );

				FeaturePoint *fpSame = v->pick ( ( *fp ) ( 0 ), ( *fp ) ( 1 ), fid );
				if ( fpSame )
				{
					DBGOUT ( DBGSHF, "point occupied by scale %f (tried %f)\n", fpSame->getScale(), fp->getScale() );
				}

				if ( v->add ( fp ) != GRID_SUCCESS )
				{
					++it;
					delete fp;
					continue;
				}

				++it;
			}

			DBGOUT ( DBGSHF, "view has %d feature points\n", v->getNumFeaturePoints() );
			return v->getNumFeaturePoints();
		}

	}

}
