

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Andrei Tchouev


#include "glcamera.h"

#include <cmath>
#include <GL/gl.h>

namespace adore
{
	namespace visual
	{
		void GLCamera::set(Vector3D eye, Vector3D aim, Vector3D up)
		{
			eyeloc = eye;
			spine = up;
			spine.normalize();
			n = aim - eye;
			u = up.cross(n);
			n.normalize();
			u.normalize();
			v = n.cross(u);
		}

		void GLCamera::set(Vector3D eye, Quaternion rot)
		{
			eyeloc = eye;
			u = rot * Vector3D(1.0f,0.0f,0.0f);
			v = rot * Vector3D(0.0f,1.0f,0.0f);
			n = rot * Vector3D(0.0f,0.0f,1.0f);
			spine = v;
		}

		void GLCamera::set(Vector3D _eyeloc, Vector3D _spine, Vector3D _u, Vector3D _v, Vector3D _n)
		{
			eyeloc = _eyeloc;
			spine  = _spine;
			u      = _u;
			v      = _v;
			n      = _n;
		}
		
		void GLCamera::roll(float angle)
		{
			u.rotate(angle,n);
			v.rotate(angle,n);
		}

		void GLCamera::pitch(float angle)
		{
			v.rotate(angle,u);
			n.rotate(angle,u);	
		}

		void GLCamera::yaw(float angle)
		{
			u.rotate(angle,v);
			n.rotate(angle,v);
		}

		void GLCamera::turn(float angle)
		{
			u.rotate(angle,spine);
			v.rotate(angle,spine);
			n.rotate(angle,spine);
		}

		void GLCamera::tilt(float angle, float gap)
		{
			float view_spine_angle = std::acos(n * spine);
			if(angle + view_spine_angle > PI - gap)
			{
				angle = PI - gap - view_spine_angle;
				if(angle < 0.0f)
					angle = 0.0f;
			}
			else if(angle + view_spine_angle < gap)
			{
				angle = view_spine_angle - gap;
				if(angle > 0.0f)
					angle = 0.0f;
			}

			v.rotate(angle,u);
			n.rotate(angle,u);
		}

		void GLCamera::rotate(Vector3D axis, float angle)
		{
			u.rotate(angle,axis);
			v.rotate(angle,axis);
			n.rotate(angle,axis);
		}

		void GLCamera::slide(Vector3D delta)
		{
			eyeloc.x += delta.x*u.x + delta.y*v.x + delta.z*n.x;
			eyeloc.y += delta.x*u.y + delta.y*v.y + delta.z*n.y;
			eyeloc.z += delta.x*u.z + delta.y*v.z + delta.z*n.z;
		}
	
		void GLCamera::perspective(float fov_, float aspect_, float near_dist_, float far_dist_)
		{
			fov = fov_;
			aspect = aspect_;
			near_dist = near_dist_;
			far_dist = far_dist_;
		}

		void GLCamera::setOpenGLModelView() const
		{
			float world[16];

			world[0] = u.x;		world[4] = u.y;		world[8] = u.z;		world[12] = -eyeloc*u;
			world[1] = v.x;		world[5] = v.y;		world[9] = v.z;		world[13] = -eyeloc*v;
			world[2] = n.x;		world[6] = n.y;		world[10] = n.z;	world[14] = -eyeloc*n;
			world[3] = 0.0f;	world[7] = 0.0f;	world[11] = 0.0f;	world[15] = 1.0f;

			glMatrixMode(GL_MODELVIEW);
			glLoadMatrixf(world);
		}

		void GLCamera::setOpenGLProjection() const
		{
			double top = near_dist * std::tan(fov*0.5);
			double bottom = -top;
			double right = top*aspect;
			double left = -right;
			if(top == bottom || left == right)
				return;

			glMatrixMode(GL_PROJECTION);
			glLoadIdentity();
			glFrustum(left,right,bottom,top,near_dist,far_dist);
		}
	
		void GLCamera::getMatrix(Matrix4x4& modelViewProj) const
		{
			Matrix4x4 world;

			world[0][0] = u.x;	world[1][0] = u.y;	world[2][0] = u.z;	world[3][0] = -eyeloc*u;
			world[0][1] = v.x;	world[1][1] = v.y;	world[2][1] = v.z;	world[3][1] = -eyeloc*v;
			world[0][2] = n.x;	world[1][2] = n.y;	world[2][2] = n.z;	world[3][2] = -eyeloc*n;
			world[0][3] = 0.0f;	world[1][3] = 0.0f;	world[2][3] = 0.0f;	world[3][3] = 1.0f;

			double top = near_dist * std::tan(fov*0.5);
			double bottom = -top;
			double right = top*aspect;
			double left = -right;

			double rpl = right + left;
			double rml = right - left;
			double tpb = top + bottom;
			double tmb = top - bottom;
			double fpn = far_dist + near_dist;
			double fmn = far_dist - near_dist;
			double near2 = 2.0 * near_dist;

			Matrix4x4 proj;

			proj[0][0] = near2/rml;	proj[1][0] = 0.0f;		proj[2][0] = rpl/rml;		proj[3][0] = 0.0f;
			proj[0][1] = 0.0f;		proj[1][1] = near2/tmb;	proj[2][1] = tpb/tmb;		proj[3][1] = 0.0f;
			proj[0][2] = 0.0f;		proj[1][2] = 0.0f;		proj[2][2] = -(fpn)/fmn;	proj[3][2] = -near2*far_dist/fmn;
			proj[0][3] = 0.0f;		proj[1][3] = 0.0f;		proj[2][3] = -1.0f;			proj[3][3] = 0.0f;

			modelViewProj = proj * world;
		}

		void GLCamera::projectVector(Vector3D src, Vector3D& dst) const
		{
			dst.x = src.x * u.x + src.y * u.y + src.z * u.z;
			dst.y = src.x * v.x + src.y * v.y + src.z * v.z;
			dst.z = src.x * n.x + src.y * n.y + src.z * n.z;
		}

		void GLCamera::unprojectVector(Vector3D src, Vector3D& dst) const
		{
			dst.x = src.x * u.x + src.y * v.x + src.z * n.x;
			dst.y = src.x * u.y + src.y * v.y + src.z * n.y;
			dst.z = src.x * u.z + src.y * v.z + src.z * n.z;
		}

		void GLCamera::buildRay(int x, int y, Vector3D& org, Vector3D& dir) const
		{
			float viewport[4];
			glGetFloatv(GL_VIEWPORT,viewport);
			float xf = (x-viewport[0]) / viewport[2] - 0.5f;
			float yf = (viewport[3]-(y-viewport[1])) / viewport[3] - 0.5f;
			double span = 2.0 * near_dist * std::tan(fov * 0.5);
			xf *= aspect * span;
			yf *= span;

			dir = -near_dist*n + xf*u + yf*v;
			org = eyeloc+dir;
			dir.normalize();
		}
	}
}
