

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Andrei Tchouev


#ifndef ADORE_GLMESH_H
#define ADORE_GLMESH_H

#include "glmath.h"

namespace adore
{
	namespace visual
	{
		/**	A structure, describing a single vertex of the mesh
		*/
		struct Vertex
		{
			Vector3D pos;			///< vertex position
			float tex_u;			///< texture coordinate u
			float tex_v;			///< texture coordinate v
			int extra;			///< private data index
		};

		/**	A structure, describing a single triangular face of the mesh
		*/
		struct Face
		{
			int v1;				///< index of the first vertex
			int v2;				///< index of the second vertex
			int v3;				///< index of the third vertex
			int normal;			///< face normal
		};

		/**	A mesh is a set of vertexes, faces and normals to faces.
		 *	GLMesh describes some 3-dimensional form.
		 *
		 *	 
		 */
		class GLMesh
		{
			/**	Copy constructor is closed
			 */
			GLMesh(const GLMesh& rv);

			/**	Assignment operator is closed
			 */
			GLMesh& operator=(const GLMesh& rv);

		public:
			/**	Constructor
			 */
			GLMesh();

			/**	Destructor
			 */
			~GLMesh();

			/**	Sets vertices of the mesh
			 *	@param vertexdata An array of vertices
			 *	@param startindex Starting index in buffer
			 *	@param count A number of vertices
			 *	@note GLMesh makes a copy of vertex data
			 */
			void setVertices(Vertex* vertexdata, int startindex, int count);

			/**	Sets normals of the mesh
			 *	@param normaldata An array of normals
			 *	@param startindex Starting index in buffer
			 *	@param count A number of normals
			 *	@note GLMesh makes a copy of normal data
			 */
			void setNormals(Vector3D* normaldata, int startindex, int count);

			/**	Sets faces of the mesh
			 *	@param facedata An array of facesaccording
			 *	@param startindex Starting index in buffer
			 *	@param count A number of faces
			 *	@note GLMesh makes a copy of face data
			 */
			void setFaces(Face* facedata, int startindex, int count);

			/**	Preallocate memory for storing a given number of vertices
			 */
			void createVertices(int count);

			/**	Preallocate memory for storing a given number of normals
			 */
			void createNormals(int count);

			/**	Preallocate memory for storing a given number of faces
			 */
			void createFaces(int count);

			/**	Get the number of vertices
			 */
			int getNumVertices();

			/**	Get the number of faces
			 */
			int getNumFaces();

			/**	Get the number of normals
			 */
			int getNumNormals();

			/**	Prepare the vertex buffer for writing and return the pointer to it
			 *	@param startindex Starting elements to write
			 *	@param count Number of elements to lock
			 *	@warning it is not safe to write past count elements in the buffer
			 */
			Vertex* lockVertices(int startindex, int count);

			/**	Prepare the normal buffer for writing and return the pointer to it
			 *	@param startindex Starting element to write
			 *	@param count Number of elements to lock
			 *	@warning it is not safe to write past count elements in the buffer
			 */
			Vector3D* lockNormals(int startindex, int count);

			/**	Prepare the face buffer for writing and return the pointer to it
			 *	@param startindex Starting element to write
			 *	@param count Number of elements to lock
			 *	@warning it is not safe to write past count elements in the buffer
			 */
			Face* lockFaces(int startindex, int count);

			/**	Unlock the vertex buffer after locking operation.
			 *	@warning Must eventually be called after locking to validate data
			 */
			void unlockVertices();

			/**	Unlock the normal buffer after locking operation.
			 *	@warning Must eventually be called after locking to validate data
			 */
			void unlockNormals();

			/**	Unlock the face buffer after locking operation.
			 *	@warning Must eventually be called after locking to validate data
			 */
			void unlockFaces();

			/**	Set the level of detail for the mesh.
			 *
			 *	This function changes the amount of visible faces. It is assumed that
			 *	the face list is sorted by the importance of triangles to the final view.
			 *	
			 *	@param value LOD value (0.0 - nothing is rendered, 1.0 -everything is rendered)
			 */
			void setLevelOfDetail(float value);

			/**	Build OpenGL display list and clear @c dirty flag
			 */
			void buildDisplayList();

			/**	Precalculate bounding volumes
			 */
			void buildBoundingVolumes();

			/**	Render this mesh using OpenGL API
			 */
			void draw();

			/**	Visualize normals during rendering
			 */
			void drawNormals();

			/**	Draw one or several vertices
			 */
			void drawVertexRange(int index, int count);

			/**	Draw one or several faces
			 */
			void drawFaceRange(int index, int count);

			/**	Draw one or several vertices visualized as meshed, that are always facing the camera
			 *	@param index Starting vertex
			 *	@param count Number of vertices
			 *	@param mesh Particle form
			 *	@param scl Particle scaling
			 */
			void drawBillboardRange(int index, int count, GLMesh& mesh, float scl);

			/**	Select triangle in the object space.
			 *	Radius is always defined in the screen space and is view-dependent.
			 *	@param org Origin of the picking ray
			 *	@param dir Direction vector of the picking ray
			 *	@param tret Intersection parameter
			 *	@return the first triangle which is hit by the ray or -1 if nothing is hit
			 */
			int select(Vector3D org, Vector3D dir, float* tret);

			/**	Select a vertex treating it as a sphere in the object space.
			 *	Radius is always defined in the screen space and is view-dependent.
			 *	@param org Origin of the picking ray
			 *	@param dir Direction vector of the picking ray
			 *	@param radius Radius of the sphere
			 *	@return the first vertex which is hit by the ray or -1 if nothing is hit
			 */
			int select(Vector3D org, Vector3D dir, float radius);

			/**	Get private data index of the vertex
			 *
			 *	Private data helps to associate extra information with the vertex
			 *
			 *	@param index Vertex index
			 */
			int extra(int index);

			/** Get the radius of the bounding sphere
			 */
			float getBoundingSphere();

		private:
			int displaylist;		///< OpenGL display list id

			Vertex* vertices;		///< vertex buffer
			Vector3D* normals;		///< normal buffer
			Face* faces;			///< face buffer

			int numvert;			///< number of vertices
			int numnorm;			///< number of normals
			int numface;			///< number of faces

			int allocvert;			///< allocated vertices
			int allocnorm;			///< allocated normals
			int allocface;			///< allocated faces

			bool lockvert;			///< vertex buffer is locked
			bool locknorm;			///< normal buffer is locked
			bool lockface;			///< face buffer is locked

			bool dirty;			///< dirty bit, need to update cache

			float lod;			///< level of detail
			float bsphere;			///< radius of the bounding sphere

			/**	Intersection test line-sphere
			 *
			 *	@param center Center of the Sphere
			 *	@param radius Radius of the Sphere
			 *	@param org Origin of the Line
			 *	@param dir Direction vector of the line
			 *	@param t Intersections
			 *	@return Number of intersections
			 */
			static int lineSphereIntersect(
				Vector3D& center,
				float& radius,
				Vector3D& org,
				Vector3D& dir,
				float t[2]);

			/**	Intersection test line-triangle
			 *
			 *	@param a First triangle vertex
			 *	@param b Second triangle vertex
			 *	@param c Third triangle vertex
			 *	@param org Origin of the Line
			 *	@param dir Direction vector of the line
			 *	@param t Intersection parameter
			 *	@return Line hits triangle
			 */
			static bool lineTriangleIntersect(
				Vector3D& a,
				Vector3D& b,
				Vector3D& c,
				Vector3D& org,
				Vector3D& dir,
				float* t);
		};
	}
}

#endif	// ADORE_GLMESH_H
