

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Andrei Tchouev


#include "glparser.h"

namespace adore
{
	namespace visual
	{
		std::map<std::string, int> GLParser::light_param_table;
		std::map<std::string, int> GLParser::mat_param_table;
		bool GLParser::table_init = false;

		GLParser::GLParser(std::istream& in)
		{
			source = &in;
			max_lights = 0;
			max_materials = 0;
			lights = 0;
			materials = 0;
			error = GL_REN_NO_ERROR;
		}

		bool GLParser::parse(GLLight* l_list, GLMaterial* m_list, int max_l, int max_m)
		{
			max_lights = max_l;
			max_materials = max_m;
			lights = l_list;
			materials = m_list;

			if(! (source && max_lights && max_materials && lights && materials) )
			{
				error = GL_REN_INTERNAL_ERROR;
				return false;
			}

			error = GL_REN_NO_ERROR;

			if(!table_init)
				buildParamTables();

			while(true)
			{
				switch(lexem())
				{
				case GL_REN_NAME:
					if(name=="light")
					{
						if(!readLight())
							return false;
					}
					else if(name=="material")
					{
						if(!readMaterial())
							return false;
					}
					break;
				case GL_REN_EOF:
					return true;
				}
			}
		}

		int GLParser::getError()
		{
			return error;
			error = GL_REN_NO_ERROR;
		}

		int GLParser::token()
		{
			char ch;
			//eof
			if(!source->get(ch))
				return GL_REN_EOF;

			//one or more whitespaces
			if(std::isspace(static_cast<unsigned char>(ch)))
			{
				name = ch;
				while(source->get(ch) && std::isspace(static_cast<unsigned char>(ch)))
					name.push_back(ch);
				source->putback(ch);
				return GL_REN_WHITESPACE;
			}

			//lexem or comment
			switch(ch)
			{
			case '#':
				source->ignore(std::numeric_limits<int>::max(),'\n');
				return GL_REN_WHITESPACE;

			case '=':	case '{':	case '}':
				name = ch;
				return recognize(ch);

			case '0':	case '1':	case '2':	case '3':	case '4':
			case '5':	case '6':	case '7':	case '8':	case '9':
			case '.':	case '-':
				source->putback(ch);
				(*source)>>value;
				return GL_REN_VALUE;

			case '\'':
				while(source->get(ch) && ch!='\'')
					name.push_back(ch);
				return GL_REN_STRING;

			default:
				if(std::isalpha(static_cast<unsigned char>(ch)) || ch=='_')
				{
					name = ch;
					while(source->get(ch) &&
						(std::isalnum(static_cast<unsigned char>(ch)) || ch=='_'))
							name.push_back(ch);
					source->putback(ch);
					return GL_REN_NAME;
				}
				else
				{
					source->unget();
					name.clear();
					return GL_REN_UNKNOWN;
				}
			}
		}

		int GLParser::lexem()
		{
			int ret = token();
			return ret != GL_REN_WHITESPACE ? ret : token();
		}

		int GLParser::recognize(char ch)
		{
			switch(ch)
			{
				case '=':
					return GL_REN_ASSIGN;
				case '{':
					return GL_REN_OPEN;
				case '}':
					return GL_REN_CLOSE;
				default:
					return GL_REN_UNKNOWN;
			}
		}

		bool GLParser::readLight()
		{
			int index;
			(*source)>>index;
			if(index<0 || index >= max_lights)
			{
				error = GL_REN_INVALID_INDEX;
				return false;
			}
			if(lexem()!=GL_REN_OPEN)
			{
				error = GL_REN_OPEN_EXPECTED;
				return false;
			}

			while(true)
			{
				switch(lexem())
				{
				case GL_REN_EOF:
					error = GL_REN_UNEXPECTED_EOF;
					return false;
				case GL_REN_NAME:
					if(!readLightParam(index))
						return false;
					break;
				case GL_REN_CLOSE:
					return true;
				default:
					{
						error = GL_REN_CLOSE_EXPECTED;
						return false;
					}
				}
			}

			return true;
		}

		bool GLParser::readMaterial()
		{
			int index;
			(*source)>>index;
			if(index<0 || index >= max_materials)
			{
				error = GL_REN_INVALID_INDEX;
				return false;
			}
			if(lexem()!=GL_REN_OPEN)
			{
				error = GL_REN_OPEN_EXPECTED;
				return false;
			}

			while(true)
			{
				switch(lexem())
				{
				case GL_REN_EOF:
					error = GL_REN_UNEXPECTED_EOF;
					return false;
				case GL_REN_NAME:
					if(!readMatParam(index))
						return false;
					break;
				case GL_REN_CLOSE:
					return true;
				default:
					{
						error = GL_REN_CLOSE_EXPECTED;
						return false;
					}
				}
			}

			return true;
		}

		bool GLParser::readLightParam(int index)
		{
			switch(recognizeLightParam(name))
			{
			case GL_REN_LIGHT_POSITION:
				return readVector(lights[index].position);
			case GL_REN_LIGHT_AMBIENT:
				return readVector(lights[index].ambient);
			case GL_REN_LIGHT_DIFFUSE:
				return readVector(lights[index].diffuse);
			case GL_REN_LIGHT_SPECULAR:
				return readVector(lights[index].specular);
			case GL_REN_LIGHT_SPOT:
				return readVector(lights[index].spot);
			case GL_REN_LIGHT_SPOT_EXP:
				return readValue(lights[index].spot_exp);
			case GL_REN_LIGHT_SPOT_CUTOFF:
				return readValue(lights[index].spot_cutoff);
			case GL_REN_LIGHT_ATT_CONST:
				return readValue(lights[index].att_const);
			case GL_REN_LIGHT_ATT_LINEAR:
				return readValue(lights[index].att_linear);
			case GL_REN_LIGHT_ATT_QUAD:
				return readValue(lights[index].att_quad);
			case GL_REN_LIGHT_ENABLED:
				return readBoolean(lights[index].enabled);
			default:
				{
					error = GL_REN_UNKNOWN_LIGHT_PARAM;
					return false;
				}
			}
		}

		bool GLParser::readMatParam(int index)
		{
			switch(recognizeMatParam(name))
			{
			case GL_REN_MAT_NAME:
				return readString(materials[index].name);
			case GL_REN_MAT_SMOOTH:
				return readBoolean(materials[index].smooth);
			case GL_REN_MAT_WIREFRAME:
				return readBoolean(materials[index].wireframe);
			case GL_REN_MAT_DEPTH:
				return readBoolean(materials[index].depth);
			case GL_REN_MAT_CULL:
				return readBoolean(materials[index].cull);
			case GL_REN_MAT_AMBIENT:
				return readVector(materials[index].ambient);
			case GL_REN_MAT_DIFFUSE:
				return readVector(materials[index].diffuse);
			case GL_REN_MAT_SPECULAR:
				return readVector(materials[index].specular);
			case GL_REN_MAT_SHININESS:
				return readValue(materials[index].shininess);
			default:
				{
					error = GL_REN_UNKNOWN_MAT_PARAM;
					return false;
				}
			}
			return true;
		}

		bool GLParser::readBoolean(bool& b)
		{
			if(lexem()!=GL_REN_ASSIGN)
			{
				error = GL_REN_ASSIGN_EXPECTED;
				return false;
			}
			(*source)>>b;
			return true;
		}

		bool GLParser::readValue(float& val)
		{
			if(lexem()!=GL_REN_ASSIGN)
			{
				error = GL_REN_ASSIGN_EXPECTED;
				return false;
			}

			if(lexem()!=GL_REN_VALUE)
			{
				error = GL_REN_VALUE_EXPECTED;
				return false;
			}
			val = value;
			return true;
		}

		bool GLParser::readVector(Vector4D& vec)
		{
			if(lexem()!=GL_REN_ASSIGN)
			{
				error = GL_REN_ASSIGN_EXPECTED;
				return false;
			}
			for(int i = 0; i<4;++i)
			{
				if(lexem()!=GL_REN_VALUE)
				{
					error = GL_REN_VALUE_EXPECTED;
					return false;
				}
				vec[i] = value;
			}
			return true;
		}

		bool GLParser::readString(std::string& st)
		{
			if(lexem()!=GL_REN_ASSIGN)
			{
				error = GL_REN_ASSIGN_EXPECTED;
				return false;
			}
			if(lexem()!=GL_REN_STRING)
			{
				error = GL_REN_STRING_EXPECTED;
				return false;
			}
			st = name;
			return true;
		}

		void GLParser::buildParamTables()
		{
			table_init = true;

			light_param_table["position"] = GL_REN_LIGHT_POSITION;
			light_param_table["ambient"] = GL_REN_LIGHT_AMBIENT;
			light_param_table["diffuse"] = GL_REN_LIGHT_DIFFUSE;
			light_param_table["specular"] = GL_REN_LIGHT_SPECULAR;
			light_param_table["spot"] = GL_REN_LIGHT_SPOT;
			light_param_table["exp"] = GL_REN_LIGHT_SPOT_EXP;
			light_param_table["cutoff"] = GL_REN_LIGHT_SPOT_CUTOFF;
			light_param_table["constant"] = GL_REN_LIGHT_ATT_CONST;
			light_param_table["linear"] = GL_REN_LIGHT_ATT_LINEAR;
			light_param_table["quadratic"] = GL_REN_LIGHT_ATT_QUAD;
			light_param_table["enabled"] = GL_REN_LIGHT_ENABLED;

			mat_param_table["name"] = GL_REN_MAT_NAME;
			mat_param_table["smooth"] = GL_REN_MAT_SMOOTH;
			mat_param_table["wireframe"] = GL_REN_MAT_WIREFRAME;
			mat_param_table["depth"] = GL_REN_MAT_DEPTH;
			mat_param_table["cull"] = GL_REN_MAT_CULL;
			mat_param_table["ambient"] = GL_REN_MAT_AMBIENT;
			mat_param_table["diffuse"] = GL_REN_MAT_DIFFUSE;
			mat_param_table["specular"] = GL_REN_MAT_SPECULAR;
			mat_param_table["shininess"] = GL_REN_MAT_SHININESS;
		}

		int GLParser::recognizeLightParam(std::string& param)
		{
			return light_param_table[param];
		}

		int GLParser::recognizeMatParam(std::string& param)
		{
			return mat_param_table[param];
		}
	}
}
