

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Andrei Tchouev


#ifndef ADORE_GLPARSER_H
#define ADORE_GLPARSER_H

#include <iostream>
#include <map>

#include "gllight.h"
#include "glmaterial.h"

namespace adore
{
	namespace visual
	{
		/**	Tokens of the settings parser (case insensitive)
		 */
		enum GL_REN_LEXEM
		{
			GL_REN_UNKNOWN,			///< unknown token
			GL_REN_WHITESPACE,		///< std::isspace == true
			GL_REN_ASSIGN,			///< assignment '='
			GL_REN_OPEN,			///< opening '{'
			GL_REN_CLOSE,			///< closing '}'
			GL_REN_VALUE,			///< floating point value
			GL_REN_NAME,			///< identifier
			GL_REN_STRING,			///< string literal in single quotes
			GL_REN_EOF				///< end of file
		};

		/**	Identifiers of the light source parameters
		*/
		enum GL_REN_LIGHT
		{
			GL_REN_LIGHT_UNKNOWN,		///< unknown parameter
			GL_REN_LIGHT_POSITION,		///< position
			GL_REN_LIGHT_AMBIENT,		///< ambient
			GL_REN_LIGHT_DIFFUSE,		///< diffuse
			GL_REN_LIGHT_SPECULAR,		///< specular
			GL_REN_LIGHT_SPOT,			///< spot position
			GL_REN_LIGHT_SPOT_EXP,		///< spot exponents
			GL_REN_LIGHT_SPOT_CUTOFF,	///< spot cutoff
			GL_REN_LIGHT_ATT_CONST,		///< constant attenuation
			GL_REN_LIGHT_ATT_LINEAR,	///< linear attenuation
			GL_REN_LIGHT_ATT_QUAD,		///< quadratic attenuation
			GL_REN_LIGHT_ENABLED		///< enabled
		};

		/**	Identifiers of the material parameters
		*/
		enum GL_REN_MAT
		{
			GL_REN_MAT_UNKNOWN,			///< unknown parameter
			GL_REN_MAT_NAME,			///< material name
			GL_REN_MAT_SMOOTH,			///< smooth shaded
			GL_REN_MAT_WIREFRAME,		///< wireframed
			GL_REN_MAT_DEPTH,			///< depth testing enabled
			GL_REN_MAT_CULL,			///< culling enabled
			GL_REN_MAT_AMBIENT,			///< ambient color
			GL_REN_MAT_DIFFUSE,			///< diffuse color
			GL_REN_MAT_SPECULAR,		///< specular color
			GL_REN_MAT_SHININESS		///< shininess
		};

		/**	Identifiers of the parser errors
		*/
		enum GL_REN_ERROR
		{
			GL_REN_NO_ERROR,			///< no errors
			GL_REN_INTERNAL_ERROR,		///< internal error
			GL_REN_INVALID_INDEX,		///< index is out of range
			GL_REN_OPEN_EXPECTED,		///< '{' is expected
			GL_REN_CLOSE_EXPECTED,		///< '}' is expected
			GL_REN_UNKNOWN_LIGHT_PARAM,	///< unknown light parameter
			GL_REN_UNKNOWN_MAT_PARAM,	///< unknown material parameter
			GL_REN_ASSIGN_EXPECTED,		///< '=' is expected
			GL_REN_VALUE_EXPECTED,		///< floating point value expeceted
			GL_REN_STRING_EXPECTED,		///< 'character string' expected
			GL_REN_UNEXPECTED_EOF		///< unexpected end of file
		};

		/**	Parses the settings file and builds material and light lists for rendering
		 *	
		 *	 
		 *
		 *	Parser grammar:
		 *
		 *	EOF = end of file
		 *	INDEX = any non-negative integer number
		 *	VALUE = any floating point number
		 *	STRING = character string inside single quotes
		 *
		 *	program:
		 *		EOF
		 *		def_list EOF
		 *
		 *	def_list:
		 *		define
		 *		define def_list
		 *
		 *	define:
		 *		light_define
		 *		mat_define
		 *
		 *	light_define:
		 *		light INDEX { expr_list }
		 *
		 *	mat_define:
		 *		material INDEX { expr_list }
		 *
		 *	expr_list
		 *		expression
		 *		expression expr_list
		 *
		 *	expression:
		 *		name = value_list
		 *
		 *	value_list:
		 *		value
		 *		value value_list
		 *
		 *	value:
		 *		STRING
		 *		REAL
		 */
		class GLParser
		{
		public:
			/**	Constructor
			 */
			GLParser(std::istream& in);

			/**	Parse the settings file and initialize provided structures
			 *
			 *	@param l_list Array of lights to initialize
			 *	@param m_list Array of materials to initialize
			 *	@param max_l Maximal number of lights
			 *	@param max_m Maximal number of materials
			 *
			 *	@return parsing successful
			 */
			bool parse(GLLight* l_list, GLMaterial* m_list, int max_l, int max_m);

			/**	Clear the last error and return it as a member of GL_REN_ERROR
			 */
			int getError();

		private:
			int max_lights;				///< maximum lights
			int max_materials;			///< maximum materials
			GLLight* lights;			///< light array
			GLMaterial* materials;		///< material array

			std::istream* source;		///< the source stream
			std::string name;			///< temporary name holder
			double value;				///< temporary value holder
			int error;					///< last error

			/**	Get the next token from the source stream
			 */
			int token();

			/**	Get the next lexem from the source stream
			 */
			int lexem();

			/**	Recognize one-element lexem
			 */
			int recognize(char ch);

			/**	Read light parametres from the source stream
			 */
			bool readLight();

			/**	Read light parametres from the source stream
			 */
			bool readMaterial();

			/**	Read a single light parameter from the source stream
			 */
			bool readLightParam(int index);

			/**	Read a single material parameter from the source stream
			 */
			bool readMatParam(int index);

			/**	Read boolean value from the source stream
			 */
			bool readBoolean(bool& b);

			/**	Read floating point value from the source stream
			 */
			bool readValue(float& val);

			/**	Read 4 floating point values from the source stream
			 */
			bool readVector(Vector4D& vec);

			/**	Read string in single quotes from the source stream
			 */
			bool readString(std::string& st);

			/**	table for light parameter's name conversion
			 */
			static std::map<std::string, int> light_param_table;

			/**	table for material parameter's name conversion
			 */
			static std::map<std::string, int> mat_param_table;

			/**	Conversion tables ready
			 */
			static bool table_init;

			/**	Build conversion tables
			 */
			static void buildParamTables();

			/** Get the integer representation of the light parameter
			 */
			static int recognizeLightParam(std::string& param);

			/** Get the integer representation of the material parameter
			 */
			static int recognizeMatParam(std::string& param);
		};
	}
}

#endif	// ADORE_GLPARSER_H
