

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Andrei Tchouev


#include "glrenderer.h"

#include <fstream>
#include <GL/gl.h>

#include "glparser.h"

namespace adore
{
	namespace visual
	{
		void GLRenderer::select(Vector3D& org, Vector3D& dir)
		{
			if((displayflags & DISPLAY_PICKING) && entity)
			{
				entity->highlighted = entity->select(org,dir);
			}
		}

		GLRenderer::GLRenderer()
			:background(0x00,0x00,0x40),
			pencolor(0xff,0xff,0xff),
			pensize(1.0f) 
		{
			entity = 0;
			displayflags = DISPLAY_MESH | DISPLAY_TEXTURE | DISPLAY_CAMERAS;
		}

		GLRenderer::~GLRenderer()
		{
		}

		bool GLRenderer::loadConfig(std::istream& in)
		{
			GLParser parser(in);
			return parser.parse(lights,materials,maxlights,maxmaterials);
		}

		bool GLRenderer::loadConfig(const char* filename)
		{
			std::ifstream in(filename);
			GLParser parser(in);
			return parser.parse(lights,materials,maxlights,maxmaterials);
		}

		int GLRenderer::settings()
		{
			return displayflags;
		}

		void GLRenderer::settings(int flags)
		{
			displayflags = flags;
			applyDisplayFlags();
		}

		void GLRenderer::init()
		{
			if(entity)
			{
				entity->mesh.buildDisplayList();
				entity->texture.update();
			}
		}

		void GLRenderer::deviation(float value)
		{
			if(entity)
				entity->mesh.setLevelOfDetail(value);
		}

		void GLRenderer::deviationTicks (int ticks, int max_ticks)
		{
			const float min_deviation = 0.7F;
			const float max_deviation = 1.1F;

			if(ticks<=0) {
				int df = this->settings();
				if(df & DISPLAY_MESH)
					this->settings(df ^ DISPLAY_MESH);
			} else {
				int df = this->settings();
				if(!(df & DISPLAY_MESH))
					this->settings(df | DISPLAY_MESH);

				if(ticks > max_ticks) ticks = max_ticks;
				float deviat = min_deviation + ((max_deviation-min_deviation) / float(max_ticks)) * ticks;
				this->deviation(deviat);
			}
		}
		
		void GLRenderer::focus()
		{
			applyDisplayFlags();
			applyPenStyle();
			applyColorStyle();
			applyLightSettings();
			camera.setOpenGLProjection();
		}

		void GLRenderer::unfocus()
		{
		}

		void GLRenderer::paint()
		{
			glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
			camera.setOpenGLModelView();
			if(entity)
			{
				if(displayflags | DISPLAY_LIGHTING)
					materials[entity->getMaterial()].setOpenGLMaterial();
	
				glPushMatrix();
				entity->render();
				glPopMatrix();
			}
		}

		void GLRenderer::resize(int width, int height)
		{
			const float spanfactor = 16.0f;
			aspect = static_cast<float>(width) / static_cast<float>(height);
			float nplane = 0.5f;
			float fplane = 100.0f;
			if(entity)
			{
				float span = entity->mesh.getBoundingSphere();
				nplane = span / spanfactor;
				fplane = span * spanfactor;

				if(span == 0.0f) {
					nplane = 0.5f;
					fplane = 100.0f;
				}
			}
			camera.perspective(fieldofview,aspect,nplane,fplane);
			camera.setOpenGLProjection();
			glViewport(0,0,width,height);
		}

		void GLRenderer::pick(int x, int y)
		{
			Vector3D pick_org;
			Vector3D pick_dir;
			camera.buildRay(x,y,pick_org,pick_dir);
			select(pick_org,pick_dir);
		}

		void GLRenderer::perspective(float fov)
		{
			const float spanfactor = 32.0f;
			float nplane = 0.5f;
			float fplane = 100.0f;
			if(entity)
			{
				float span = entity->mesh.getBoundingSphere();
				nplane = span / spanfactor;
				fplane = span * spanfactor;

				/// special case: it is no object here!
				if(span == 0.0f) {
					nplane = 0.5f;
					fplane = 100.0f;
				}
			}
			fieldofview = fov;
			camera.perspective(fieldofview,aspect,nplane,fplane);
			camera.setOpenGLProjection();
		}

		void GLRenderer::lookAt(const float eye[3], const float aim[3], const float up[3])
		{
			camera.set(Vector3D(eye[0],eye[1],eye[2]),Vector3D(aim[0],aim[1],aim[2]),Vector3D(up[0],up[1],up[2]));
		}

		void GLRenderer::roll(float angle)
		{
			camera.roll(angle);
		}

		void GLRenderer::yaw(float angle)
		{
			camera.yaw(angle);
		}

		void GLRenderer::pitch(float angle)
		{
			camera.pitch(angle);
		}

		void GLRenderer::turn(float angle)
		{
			camera.turn(angle);
		}

		void GLRenderer::tilt(float angle, float gap)
		{
			camera.tilt(angle,gap);
		}

		void GLRenderer::slide(const float delta[3])
		{
			camera.slide(Vector3D(delta[0],delta[1],delta[2]));
		}

		void GLRenderer::translate(const float delta[3], bool absolute)
		{
			if(entity)
			{
				if(absolute)
				{
					entity->world.setTranslation(Vector3D(delta[0],delta[1],delta[2]));
				}
				else
				{
					Vector3D position = entity->world.getTranslation();
					position.x += delta[0];
					position.y += delta[1];
					position.z += delta[2];
					entity->world.setTranslation(position);
				}
			}
		}

		void GLRenderer::rotate(const float axis[3], float angle, bool absolute)
		{
// 			DBGOUT(DBGNORMAL, "rotate\n");
			if(entity)
			{
// 				DBGOUT(DBGNORMAL, "rotate - entity\n");
				Vector3D world_axis(axis[0],axis[1],axis[2]);
// 				DBGOUT(DBGNORMAL, "c[%f, %f, %f]\n", world_axis[0],world_axis[1],world_axis[2]);
				camera.unprojectVector(world_axis,world_axis);
// 				DBGOUT(DBGNORMAL, "w[%f, %f, %f]\n", world_axis[0],world_axis[1],world_axis[2]);

				Quaternion rot;
				rot.axisAngle(world_axis,angle);
				if(absolute)
					entity->world.setRotation(rot);
				else
					entity->world.setRotation(rot * entity->world.getRotation());
			}
		}
		
		void GLRenderer::turnTable(float angle, bool absolute)
		{
			if(!entity) return;
				
			Vector3D world_axis = entity->getObjUp();
			world_axis = entity->world.getRotation() * world_axis;
			Quaternion rot;
			rot.axisAngle(world_axis,angle);
			if(absolute)
				entity->world.setRotation(rot);
			else
				entity->world.setRotation(rot * entity->world.getRotation());
			
		}


		void GLRenderer::scale(const float factor[3], bool absolute)
		{
			if(entity)
			{
				if(absolute)
				{
					entity->world.setScale(Vector3D(factor[0],factor[1],factor[2]));
				}
				else
				{
					Vector3D scale = entity->world.getScale();
					scale.x *= factor[0];
					scale.y *= factor[1];
					scale.z *= factor[2];
					entity->world.setScale(scale);
				}
			}
		}

		void GLRenderer::identity()
		{
			if(entity)
				entity->world.identity();
		}

		void GLRenderer::bgcolor(const unsigned char color[4])
		{
			background.red = color[0];
			background.green = color[1];
			background.blue = color[2];
			background.alpha = color[3];

			applyColorStyle();
		}

		void GLRenderer::pen(const unsigned char color[4], float size)
		{
			pencolor.red = color[0];
			pencolor.green = color[1];
			pencolor.blue = color[2];
			pencolor.alpha = color[3];

			pensize = size;

			applyColorStyle();
			applyPenStyle();
		}
		
		void GLRenderer::penTicks ( const unsigned char color[4], int ticks, int max_ticks )
		{
			const float min_point_size = 0.0002;
			const float max_point_size = 0.2;

			if(ticks<=0) {
				int df = this->settings();
				if(df & DISPLAY_VERTICES)
					this->settings(df ^ (DISPLAY_VERTICES | DISPLAY_PICKING));
				this->pen(color,0);
			} else {
				int df = this->settings();
				if(!(df & DISPLAY_VERTICES))
					this->settings(df | (DISPLAY_VERTICES | DISPLAY_PICKING));

				if(ticks > max_ticks) ticks = max_ticks;
				float point_size = min_point_size + ((max_point_size-min_point_size) / float(max_ticks)) * ticks;
				this->pen(color,point_size);
			}
		}

		void GLRenderer::applyDisplayFlags()
		{
			// texturing
			if(displayflags | DISPLAY_TEXTURE)
				glEnable(GL_TEXTURE_2D);
			else
				glDisable(GL_TEXTURE_2D);

			// lighting
			if(displayflags | DISPLAY_LIGHTING)
			{
				glEnable(GL_LIGHTING);
			}
			else
			{
				glDisable(GL_LIGHTING);
				// backface culling is disabled unless overriden by material properties
				glDisable(GL_CULL_FACE);
			}

			if(entity)
			{
				entity->drawmesh = displayflags & DISPLAY_MESH;
				entity->drawvertices = displayflags & DISPLAY_VERTICES;
				entity->drawnormals = displayflags & DISPLAY_NORMALS;
				entity->drawframe = displayflags & DISPLAY_LOCAL_FRAME;
			}
		}

		void GLRenderer::applyPenStyle()
		{
			if(entity)
			{
				entity->pointsize = pensize;
				entity->thickness = 1.0F;
			}
		}

		void GLRenderer::applyColorStyle()
		{
			glClearColor(
				static_cast<float>(background.red) * CLR2FLT,
				static_cast<float>(background.green) * CLR2FLT,
				static_cast<float>(background.blue) * CLR2FLT,
				static_cast<float>(background.alpha) * CLR2FLT);

			if(entity)
			{
				entity->pointcolor = pencolor;
				entity->linecolor = pencolor;
			}
		}

		void GLRenderer::applyLightSettings()
		{
			if(displayflags | DISPLAY_LIGHTING)
				for(int i = 0; i<maxlights;++i)
					lights[i].setOpenGLLight(i);
		}
		
		void GLRenderer::setTurnTableSpine ( const float axis[3] ) 
		{
			if(!entity) return;
			Vector3D vec(axis[0], axis[1], axis[2]);
			camera.unprojectVector(vec, vec);
// 			printf("u %f, %f, %f\n", vec[0], vec[1], vec[2]);
			Quaternion qu = entity->world.getRotation().inverse();
			vec = qu * vec;
// 			printf("r %f, %f, %f\n", vec[0], vec[1], vec[2]);
			entity->setObjUp(vec);
		}
	}
}
