

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Andrei Tchouev


#include "gltexture.h"

#include <memory>
#include <cassert>
#include <GL/gl.h>
#include <GL/glu.h>
#include <cstring>
#include <adrcore/util/debug.h>
 
namespace adore
{
	namespace visual
	{
		GLTexture::GLTexture()
		{
			glGenTextures(1,&texid);

			glBindTexture(GL_TEXTURE_2D,texid);

			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_NEAREST);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);

			envmode = TEX_MODULATE;
			envcolor = RGBA(0xff,0xff,0xff,0xff);

			filter = FILTER_LINEAR_MM_NEAREST;

			width = 0;
			height = 0;
			image = 0;
			dirty = false;

			glBindTexture(GL_TEXTURE_2D,0);
		}

		GLTexture::~GLTexture()
		{
// 			DBGOUT(DBGNORMAL, "GLTexture: Destructor\n");
			if(glIsTexture(texid))
				glDeleteTextures(1,&texid);
	
			delete[] image;
		}

		void GLTexture::create(int w, int h)
		{
			/** assuming min. OpenGL 2.0(ARB_texture_non_power_of_two) */
// 			assert(isPower2(w) && isPower2(h));
	
			width = w;
			height = h;
			image = new RGBA[width*height];
		}

		void GLTexture::set(int w, int h, RGBA* pixels)
		{
			/** assuming min. OpenGL 2.0(ARB_texture_non_power_of_two) */
/*			if(isPower2(w) && isPower2(h))
			{*/
				if(width!=w || height!=h)
				{
					width = w;
					height = h;
					delete[] image;
					image = new RGBA[width*height];
				}
				std::memcpy(image,pixels,width*height*sizeof(RGBA));
/*			}
			else
			{
				int p2w = GLTexture::ceilPower2(w);
				int p2h = GLTexture::ceilPower2(h);
				if(width!=p2w || height!=p2h)
				{
					width = p2w;
					height = p2h;
					image = new RGBA[width*height];
				}
				gluScaleImage(GL_RGBA,w,h,GL_UNSIGNED_BYTE,pixels,
					width,height,GL_UNSIGNED_BYTE,image);
			}*/
			dirty = true;
		}

		RGBA* GLTexture::lock()
		{
			return image;
		}

		void GLTexture::unlock()
		{
			dirty = true;
		}

		int GLTexture::getWidth() const
		{
			return width;
		}

		int GLTexture::getHeight() const
		{
			return height;
		}

		int GLTexture::getMode() const
		{
			return envmode;
		}

		void GLTexture::setMode(int mode)
		{
			envmode = mode;
		}

		RGBA GLTexture::getColorArgument() const
		{
			return envcolor;
		}

		void GLTexture::setColorArgument(RGBA arg)
		{
			envcolor = arg;
		}

		bool GLTexture::getHorizontalRepeat()
		{
			GLint repeat;
			glBindTexture(GL_TEXTURE_2D,texid);
			glGetTexParameteriv(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,&repeat);
			return repeat == GL_REPEAT;
		}

		void GLTexture::setHorizontalRepeat(bool rep)
		{
			glBindTexture(GL_TEXTURE_2D,texid);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,rep ? GL_REPEAT : GL_CLAMP);
		}

		bool GLTexture::getVerticalRepeat()
		{
			GLint repeat;
			glBindTexture(GL_TEXTURE_2D,texid);
			glGetTexParameteriv(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,&repeat);
			return repeat == GL_REPEAT;
		}

		void GLTexture::setVerticalRepeat(bool rep)
		{
			glBindTexture(GL_TEXTURE_2D,texid);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,rep ? GL_REPEAT : GL_CLAMP);
		}

		void GLTexture::setFilter(int filt)
		{
			filter = filt;
			static int convert[] = {GL_NEAREST,GL_LINEAR,GL_NEAREST_MIPMAP_NEAREST,
			GL_LINEAR_MIPMAP_NEAREST,GL_NEAREST_MIPMAP_LINEAR,GL_LINEAR_MIPMAP_LINEAR };

			filt = convert[filter];
			glBindTexture(GL_TEXTURE_2D,texid);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,filt);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,filt);
		}

		int GLTexture::getFilter()
		{
			return filter;
		}

		void GLTexture::update() const
		{
			glBindTexture(GL_TEXTURE_2D,texid);
			glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA,width,height,0,GL_RGBA,GL_UNSIGNED_BYTE,image);
			gluBuild2DMipmaps(GL_TEXTURE_2D,4,width,height,GL_RGBA,GL_UNSIGNED_BYTE,image);
			dirty = false;
		}

		void GLTexture::setOpenGLTexture() const
		{
			if(dirty)
				update();

			glBindTexture(GL_TEXTURE_2D,texid);

			GLint param = 0;
			switch(envmode)
			{
			case TEX_REPLACE:
				param = GL_REPLACE;
				break;
			case TEX_MODULATE:
				param = GL_MODULATE;
				break;
			case TEX_DECAL:
				param = GL_DECAL;
				break;
			case TEX_BLEND:
				param = GL_BLEND;
				break;
			}

			glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,param);

			const float CLR2FLT = 1.0f / 255.0f;
			float fColor[4] = {
				envcolor.red * CLR2FLT,
				envcolor.green * CLR2FLT,
				envcolor.blue * CLR2FLT,
				envcolor.alpha * CLR2FLT };

			glTexEnvfv(GL_TEXTURE_ENV,GL_TEXTURE_ENV_COLOR,fColor);
		}

		bool GLTexture::isPower2(int x)
		{
			return !(x & (x - 1)) && x;
		}

		int GLTexture::ceilPower2(int x)
		{
			x--;
			x |= x >> 1;
			x |= x >> 2;
			x |= x >> 4;
			x |= x >> 8;
			x |= x >> 16;
			x++;
			return x;
		}

		int GLTexture::floorPower2(int x)
		{
			return ceilPower2(x)-1;
		}
	}
}
