

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Sergey Cheremukhin


#include <QtGui/QMouseEvent>
#include <QtGui/QWheelEvent>

#include "featuresview.h"
#include "featuresscene.h"

#include "imagegl.h"
#include "slidergl.h"
#include "scrollbargl.h"

#include "qglhelper.h"

using namespace adore::core;

double FeaturesView::MinZoomFactor = 0.125;
double FeaturesView::MaxZoomFactor = 2.0;

FeaturesView::FeaturesView( QWidget* parent ) :
		QGLWidget( QGLFormat( QGL::SampleBuffers ), parent ),
		scene( NULL ),
		factor( 1.0 ),
		m_focused( NULL )
{
	setAttribute( Qt::WA_NoSystemBackground, true );
	scene = new FeaturesScene( this );
        setMinimumSize(200, 200);
}

void FeaturesView::reset()
{
	makeCurrent();
	scene->reset();
	resizeGL( width(), height() );
	updateGL();
}

void FeaturesView::resetPreservingViews()
{
	makeCurrent();
	scene->resetPreservingViews();
	updateGL();
}

void FeaturesView::initializeGL()
{
	glClearColor( 0.5, 0.5, 0.5, 0.0 );
	glClearDepth( 1.0 );
	glEnable( GL_TEXTURE_2D );

	QGLHelper::checkMultisampling();
	QGLHelper::prepareAATextures(this);

	m_image_zoom.reset( new ImageGL(":/images/zoom.png") );
	m_image_zoom->initialize(this);

	m_zoom_slider.reset( new SliderGL(SliderGL::Vertical, ZoomMin, ZoomMax) );
	m_zoom_slider->setValue(ZoomMin);
	m_zoom_slider->initialize(this);
	connect(m_zoom_slider.get(), SIGNAL(valueChanged(int)), this, SLOT(zoom(int)));

	m_hscroll.reset( new ScrollBarGL(ScrollBarGL::Horizontal) );
	m_hscroll->initialize(this);
	connect(m_hscroll.get(), SIGNAL(positionChanged(double)), this, SLOT(scrollHorizontal(double)));

	m_vscroll.reset( new ScrollBarGL(ScrollBarGL::Vertical) );
	m_vscroll->initialize(this);
	connect(m_vscroll.get(), SIGNAL(positionChanged(double)), this, SLOT(scrollVertical(double)));
}

void FeaturesView::resizeGL( int width, int height )
{
	glViewport( 0, 0, ( GLint )width, ( GLint )height );
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();

	double invFactor = 1.0 / factor;
	double bottom = height * invFactor;
	double top = 0;
	if ( 0 < bottom - scene->sceneRect().height() )
	{
		top -= 0.5 * ( bottom - scene->sceneRect().height() );
		bottom += top;
	}
	double right = width * invFactor;
	double left = 0;
	if ( 0 < right - scene->sceneRect().width() )
	{
		left -= 0.5 * ( right - scene->sceneRect().width() );
		right += left;
	}
	if(top == bottom || left == right)
		return;

	glOrtho( left, right, bottom, top, -1, 1 );

	resizeGui(width, height);
}

void FeaturesView::resizeGui( int width, int height )
{
	const int margin = 8;

	QRectF rect( scene->sceneRect() );
	rect = QRectF(rect.x() * factor, rect.y() * factor, rect.width() * factor, rect.height() * factor);
	m_hscroll->resize( QRectF( 0, 0, width, height ), rect );
	double maxY = m_hscroll->y();

	m_vscroll->resize( QRectF( 0, 0, width, height ), rect );

	double X = margin, Y = margin;
	double diff = 0.5 * ( m_zoom_slider->getSize().width() - m_image_zoom->getSize().width() );
	m_image_zoom->setPosition(QPointF(X + diff, Y));

	Y += m_image_zoom->getSize().height();
	double L = maxY - (Y + margin);
	m_zoom_slider->setPosition(QPointF(X, Y));
	m_zoom_slider->setLength(L);
}

void FeaturesView::paintGL()
{
	QGLHelper::setCurrentContext(this);

	boost::shared_ptr<GLEnabler> enabler;
	if(QGLHelper::isMultisamplingAvailable())
		enabler.reset(new GLEnabler(GL_MULTISAMPLE));

	glClearColor( 0.5, 0.5, 0.5, 0.0 );
	glClearDepth( 1.0 );
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	glMatrixMode( GL_MODELVIEW );

	glLoadIdentity();

	double invFactor = 1.0 / factor;
	glTranslatef( -m_hscroll->getPos() * invFactor, -m_vscroll->getPos() * invFactor, 0 );
	scene->paint( QRectF( m_hscroll->getPos() * invFactor, m_vscroll->getPos() * invFactor,
						  width() * invFactor, height() * invFactor ) );
	paintControls();
}

void FeaturesView::paintControls()
{
	glMatrixMode( GL_PROJECTION );
	glPushMatrix();
	glLoadIdentity();
	glOrtho( 0, width(), height(), 0, -1, 1 );
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();

	glColor4ub(255, 255, 255, 255);
	m_image_zoom->paint();

	m_zoom_slider->paint();

	m_hscroll->paint();
	m_vscroll->paint();

	glFlush();
	glMatrixMode( GL_PROJECTION );
	glPopMatrix();
}

void FeaturesView::mousePressEvent ( QMouseEvent * event )
{
	if(event->button() == Qt::LeftButton) {
		double x = event->x(), y = event->y();
		if(m_zoom_slider->rect().contains(x, y)) {
			m_zoom_slider->pressed(x, y);
			m_focused = m_zoom_slider.get();
			return;
		} else if(m_hscroll->rect().contains(x, y)) {
			m_hscroll->pressed(x, y);
			m_focused = m_hscroll.get();
			return;
		} else if(m_vscroll->rect().contains(x, y)) {
			m_vscroll->pressed(x, y);
			m_focused = m_vscroll.get();
			return;
		}
	}

	QPointF pos( event->posF() );

	Qt::KeyboardModifiers modif = event->modifiers();
	if(modif & Qt::ControlModifier) {
		scene->mouseCtrlClicked( toScene(pos) );
		updateGL();
		return;
	}

	scene->mouseClicked( toScene( pos ), event->button() );
}

void FeaturesView::mouseMoveEvent ( QMouseEvent * event )
{
	if(event->buttons() == Qt::LeftButton) {
		double x = event->x(), y = event->y();
		if(m_focused) {
			m_focused->moved(x, y);
			return;
		}
	}

	return QGLWidget::mouseMoveEvent( event );
}

void FeaturesView::mouseReleaseEvent ( QMouseEvent * event )
{
	if ( event->button() == Qt::LeftButton )
	{
		if(m_focused) {
			double x = event->x(), y = event->y();
			m_focused->released(x, y);
			m_focused = NULL;
			return;
		}
	}
	QGLWidget::mouseReleaseEvent( event );
}

void FeaturesView::wheelEvent ( QWheelEvent * event )
{
	double delta = -( double )event->delta() * 0.05;
	int sign = delta < 0 ? -1 : 1;
	delta = delta * delta * sign;
	if ( event->orientation() == Qt::Horizontal )
		m_hscroll->move( delta );
	else if( event->orientation() == Qt::Vertical ) {
		if(event->modifiers() & Qt::ControlModifier) {
			m_zoom_slider->move( delta );
			return;
		}
		double x = event->x(), y = event->y();
		if(m_hscroll->rect().contains(x, y))
			m_hscroll->move( delta );
		else if(m_zoom_slider->rect().contains(x, y))
			m_zoom_slider->move( delta );
		else
			m_vscroll->move( delta );
	}
	else
	{
		event->ignore();
		return;
	}
}

QPointF FeaturesView::toScene( const QPointF& p ) const
{
	double invFactor = 1.0 / factor;
	double bottom = height() * invFactor;
	double yd = 0;
	if ( 0 < bottom - scene->sceneRect().height() )
	{
		yd = 0.5 * ( bottom - scene->sceneRect().height() );
	}
	double right = width() * invFactor;
	double xd = 0;
	if ( 0 < right - scene->sceneRect().width() )
	{
		xd = 0.5 * ( right - scene->sceneRect().width() );
	}
	return p * invFactor - QPointF( xd, yd ) +
		   QPointF( m_hscroll->getPos() * invFactor, m_vscroll->getPos() * invFactor );
}

void FeaturesView::addView( ViewT* v )
{
	makeCurrent();
	scene->addView( v );
	resizeGL( width(), height() );
	updateGL();
}

void FeaturesView::getSelectedViews( std::vector<ViewT*>& views ) const
{
	return scene->getSelectedViews( views );
}

ViewT* FeaturesView::getFocusedView() const
{
	return scene->getFocusedView( );
}

void FeaturesView::highlightView ( const ViewBase * v )
{
	scene->highlightView( v );
	updateGL();
}

void FeaturesView::focusView ( const ViewBase* v )
{
	scene->focusView( v );
	updateGL();
}

void FeaturesView::selectView ( const ViewBase* v, bool on )
{
	scene->selectView( v, on );
	updateGL();
}

void FeaturesView::slotSelected( const ViewT* const view, bool on )
{
	emit sigViewSelected ( view, on );
}

void FeaturesView::slotFocused( const ViewT* const view )
{
	emit sigCameraSelected ( view );
}

void FeaturesView::removeEpipolarLines()
{
	scene->removeEpipolarLines();
	update();
}

void FeaturesView::calculateGoodZoom()
{
	int result = Zoom1;
	const QRectF& rect = scene->sceneRect();
	if ( rect.height() <= 0 ) return zoom(result);
	double val = double( height() ) / rect.height();
	val = std::max(val, MinZoomFactor);
	if(val <= 1.0) {
		result = int( round(val * Zoom1) );
	} else {
		result = (result - 1.0 ) / double(MaxZoomFactor);
		result = Zoom1 + int( round(val * Zoom1) );
	}
	m_zoom_slider->setValue(result);
	zoom(result);
}

void FeaturesView::zoom ( int n )
{
	n = std::max(n, int(ZoomMin));
	n = std::min(n, int(ZoomMax));
	if(n < Zoom1) {
		factor = double(n) / Zoom1;
		factor = std::max(factor, MinZoomFactor);
	} else {
		factor = ( double(n) - Zoom1 ) / Zoom1;
		factor = 1.0 + factor * MaxZoomFactor;
	}
	makeCurrent();
	resizeGL( width(), height() );
	updateGL();
}

void FeaturesView::scrollHorizontal(double)
{
	updateGL();
}

void FeaturesView::scrollVertical(double)
{
	updateGL();
}

void FeaturesView::featuresChanged()
{
	makeCurrent();
	scene->featuresChanged();
	updateGL();
}

void FeaturesView::nViewSelected( adore::core::NView* nv )
{
	makeCurrent();
	scene->nViewSelected( nv );
	updateGL();
}

void FeaturesView::highlightPointCorr( PointCorr* c )
{
	makeCurrent();
	scene->highlightPointCorr( c );
	updateGL();
}

void FeaturesView::setShowFeats( bool on )
{
	makeCurrent();
	scene->setShowFeats( on );
	updateGL();
}

void FeaturesView::setShowLines( bool on )
{
	makeCurrent();
	scene->setShowLines( on );
	updateGL();
}

void FeaturesView::setShowPointCorrs( bool on )
{
	makeCurrent();
	scene->setShowPointCorrs( on );
	updateGL();
}

void FeaturesView::setShowLineCorrs( bool on )
{
	makeCurrent();
	scene->setShowLineCorrs( on );
	updateGL();
}
