

/// FernUniversität in Hagen, Lehrgebiet Mensch-Computer-Interaktion
/// License: Creative Commons Attribution-Noncommercial-Share Alike 3.0  Germany, see README file that comes with this work
/// Author: Klaus Haeming


#include <iostream>
#include <cstdlib>
#include <vector>
#include <float.h>
#include <math.h>

/// The basic image class
#include <adrcore/imgprocessing/imageplane.h>

/// computes multi-channel gradients
#include <adrcore/imgprocessing/dizenzogradients.h>

/// converts images, e.g. 16-bit-float to 8-bit-unsigned
#include <adrcore/imgprocessing/typeconverter.h>

/// fills the border of an image
#include <adrcore/imgprocessing/borderfiller.h>

/// computes harris corners (there are alteratives)
#include <adrcore/imgprocessing/harris.h>

/// computes orientation and magnitudes form gradient images, especially dizenzo gradients
#include <adrcore/imgprocessing/orientationandmagnitude.h>

/// computes Canny-Edges
#include <adrcore/imgprocessing/canny.h>

/// convolves an image with a 1D-Kernel of size 3
#include <adrcore/imgprocessing/convolver1d3.h>

/// convolves an image with a 1D-Kernel of size 3 with the center=0
#include <adrcore/imgprocessing/convolver1d3notch.h>

/// a gaussian kernel
#include <adrcore/imgprocessing/kernelgauss_3_1.h>

/// a Prewitt kernel
#include <adrcore/imgprocessing/kernelprewitt_3_1.h>

/// applied like a kernel, but computes an extremum. Can be used for dilation/erosion
#include <adrcore/imgprocessing/kernelextremum_3_3.h>

/// Kernel for anisotropic smoothing (preserves borders, genereates comic-like appearance)
#include <adrcore/imgprocessing/kernelaniso_3_3.h>

/// convolves an image with a 2D-Kernel
#include <adrcore/imgprocessing/convolver2d.h>

/// a general non-symmetric 2D-Kernel
#include <adrcore/imgprocessing/kernel2d.h>

/// resamples an image around the center of the region of interest
#include <adrcore/imgprocessing/polarresampler.h>

/// normalizes colors by applying adiagonal transform
#include <adrcore/imgprocessing/colornormalizer.h>

/// methods for image sub-sampling by omitting pixels or combining neighboring ones
#include <adrcore/imgprocessing/subsampler.h>

/// converts an rgb image to a gray-scale image by applying some weightings
#include <adrcore/imgprocessing/rgbtoscalar.h>

/// to mark regions of maxima with their respective bounding box
#include <adrcore/imgprocessing/maxregions.h>

/// creates a thumbnail of an image. uses subsampler
#include <adrcore/imgprocessing/thumbnailer.h>

/// flips and flops images
#include <adrcore/imgprocessing/flipper.h>

using namespace std;
using namespace adore::core;

typedef int ggType;
typedef int magType;

/**
 * computes dizenzo gradients from an image
 * @param img the input image
 * @param gxx the squared gradient component in x-direction (pre-allocated space)
 * @param gyy the squared gradient component in y-direction (pre-allocated space)
 * @param gxy the product of the gradients in x- and y-direction (pre-allocated space)
 */
void dizenzo( ImagePlane<uint8_t> &img, ImagePlane<ggType> &gxx, ImagePlane<ggType> &gyy,ImagePlane<ggType> &gxy )
{
	/// create the Channelprocessor for this task
	DiZenzoGradients<uint8_t, ggType> grad;

	/// split the image into channels.
	/// since the second paramter is "true", the ChannelSet owns the Channel Structures
	ChannelSet<uint8_t> in( img, true );
	
	/// an empty ChannelSet 
	ChannelSet<ggType> out;

	/// populate the ChannelSet
	out.add( &gxx );
	out.add( &gyy );
	out.add( &gxy );

	/// compute the result
	grad.process( in, out );

	/// save the images for visual inspection after converting them to 8bit
	
	/// temporary space for conversion
	ImagePlane<uint8_t> ggS( gxx.getWidth(), gxx.getHeight(), 1 );

	/// The TypeConverter converts
	TypeConverter<ggType, uint8_t> tconvert;
	
	/// convert and save
	tconvert.process( &gxx, &ggS );
	ggS.save( "dizenzo_gxx.png" );
	tconvert.process( &gyy, &ggS );
	ggS.save( "dizenzo_gyy.png" );
	tconvert.process( &gxy, &ggS );
	ggS.save( "dizenzo_gxy.png" );
}


/**
 * computes canny edges from dizenzo gradients
 * 
 */
void canny( ImagePlane<ggType> &gxx, ImagePlane<ggType> &gyy,ImagePlane<ggType> &gxy )
{
	/// Create the Object that computes orientatinos and magnitudes
	OrientationAndMagnitude<ggType, uint8_t, magType> OM;
	
	/// magnitudes below this value are set to 0.
	OM.setThresh( 2 ); //integer

	/// create space to hold the magnitude information
	ImagePlane<magType> mag( gxx.getWidth(), gxx.getHeight(), 1 );
	
	/// create space to hold the sector information
	ImagePlane<uint8_t> sec( gxx.getWidth(), gxx.getHeight(), 1 );

	/// compute sector and magnitude images (sm = sector and magnitude)
	OM.sm( &gxx, &gyy, &gxy, &sec, &mag );

	/// save for visual inspection
	mag.save( "mag.png" );
	sec.save( "sec.png" );

	/// create space for canny edges
	ImagePlane<uint8_t> cImg( gxx.getWidth(), gxx.getHeight(), 1 );

	/// create object that performs the canny algorithm
	Canny<magType, uint8_t> canny;
	/// set thresholds for hysteresis thesholding
	canny.setMin( 5 ); //integer
	canny.setMax( 20 ); //integer

	/// perform the algorithm
	canny.process( &mag, &sec, &cImg );

	/// save for visual inspection
	cImg.save( "canny.png" );
}

/**
 * computes Harris corners from dizenzo gradients
 */
void harris( ImagePlane<ggType> &gxx, ImagePlane<ggType> &gyy,ImagePlane<ggType> &gxy )
{
	/// combine the gradient images into a ChannelSet
	ChannelSet<ggType> in;
	in.add( &gxx );
	in.add( &gyy );
	in.add( &gxy );

	/// create space for result
	ImagePlane< Harris<ggType>::ouT > h( gxx.getWidth(), gxx.getHeight(), 1 );

	/// instantiate an object that performs the harris algorithm
	Harris<ggType> harris;
	
	/// perform the algorithm
	harris.process( in, &h );

	/// save for visual inspection
	h.save( "harris.png" );
}

/**
 * fills the border of the first channel of an image with 0
 */
void border( ImagePlane<uint8_t> &img )
{
	/// temporarily set the channel mode of the image
	/// upon deletion, the ChannelMode is reset.
	/// CHANNEL_SPLIT=channels are treated seperately
	UsingChannelMode cm( &img, CHANNEL_SPLIT );

	/// make a working copy of the image
	/// true: copy contents, not only the paramters
	ImagePlane<uint8_t> copy( img, true );
	
	/// create an object that fills borders. The border is set to 10. The default value is 0.
	BorderFiller<uint8_t> bf( 10 );
	
	/// perform!
	bf.process( &copy, &copy );//in place
	
	/// save for visual inspection
	copy.save( "bordefiller.png" );
}

/**
 * demonstrates the application if image iterators
 */
void imageIterator( ImagePlane<uint8_t> &srcPlane )
{
	/// since we use the iterators to copy image regions
	/// we need to ensure the images have a descent size
	if ( srcPlane.getWidth() <= 280 ) return;
	if ( srcPlane.getHeight() <= 280 ) return;

	/// create an image to copy something into
	ImagePlane<uint8_t> dstPlane( srcPlane, false );
	
	/// make it black, "fillAll" means "fill all channels".
	dstPlane.fillAll( 0 );

	/// temporarily set the regions of interest.
	/// these are rest upon deletion
	UsingROI u1( &srcPlane, ROI( 80, 80, 200, 200 ) );
	UsingROI u2( &dstPlane, ROI( 40, 40, 200, 200 ) );
	
	/// temporarily set the ChannelMode. In combined mode, the iterator
	/// jumps from pixel to pixel
	UsingChannelMode cm1( &srcPlane, CHANNEL_COMBINED );
	UsingChannelMode cm2( &dstPlane, CHANNEL_COMBINED );

	/// get the iterators
	Channel<uint8_t>::Iterator2D srcIt = srcPlane.getIteratorBegin();
	Channel<uint8_t>::Iterator2D dstIt = dstPlane.getIteratorBegin();
	Channel<uint8_t>::Iterator2D srcEnd = srcPlane.getIteratorEnd();
	for ( ; srcIt != srcEnd; ++srcIt, ++dstIt )
	{
		/// the operator() is overloaded to copy a pixel
		/// *dstIt = *srcIt would copy the first channel only
		dstIt( srcIt );
	}
	
	/// save for visual inspection
	dstPlane.save( "iterator.png" );
}

/**
 * applies a 1D-Gaussian Kernel
 */
void gauss( ImagePlane<uint8_t> &srcPlane )
{
	/// create a convolver that convolves the image in x-direction
	/// the Kernel is therefore oriented vertically
	Convolver1D3<KernelGauss_3_1<uint8_t>, DIR_X> go;
	
	/// create space for the result, "false" means "copy parameters only"
	ImagePlane<uint8_t> dstPlane( srcPlane, false );
	
	/// process each channel separately
	for ( uint i = 0; i < srcPlane.getNumChannels(); ++i )
	{
		/// select the channel to process
		srcPlane.setChannelMode( i );
		dstPlane.setChannelMode( i );
		
		/// and process
		go.process( &srcPlane, &dstPlane );
	}
	
	/// save for visual inspection
	dstPlane.save( "gauss.png" );
}

/**
 * applies a Prewitt-Kernel
 */
void prewitt( ImagePlane<uint8_t> &srcPlane )
{
	/// create a suitable Convolver
	Convolver1D3Notch<KernelPrewitt_3_1<uint8_t>, DIR_Y> go;

	/// create space to store the 8bit-saveable result of the type-converter
	ImagePlane<uint8_t> savePlane( srcPlane, false );
	
	/// create space to store the signed result of the convolver
	ImagePlane<KernelPrewitt_3_1<uint8_t>::img_out_type> dstPlane( srcPlane.getWidth(), srcPlane.getHeight(), 1 );
	
	/// create a type-converter
	TypeConverter<short, uint8_t> conv;

	/// process each channel separately
	for ( uint i = 0; i < srcPlane.getNumChannels(); ++i )
	{
		/// select a channel
		srcPlane.setChannelMode( i );
		savePlane.setChannelMode( i );
		
		/// convolve
		go.process( &srcPlane, &dstPlane );
		
		/// convert to 8bit
		conv.process( &dstPlane, &savePlane );
	}
	
	/// save for visual inspection
	savePlane.save( "prewitt.png" );
}

/**
 * convolves an image with a self-made Sobel-Kernel
 */
void convolve2D( ImagePlane<uint8_t> &srcPlane )
{
	typedef int32_t kernelType;
	
	/// the types are "values of the kernel", "input image", "output image"
	typedef Kernel2D<kernelType, uint8_t, uint8_t> Kernel2DT;
	
	/// create a convolver that lets the image border spread out indefinitly
	/// this prevents artificial edges at image boundaries.
	Convolver2D<Kernel2DT> convolver( BORDER_COPY );
	
	/// the convolver instantiates the Kernel
	/// to modify it, we need to get a reference to it
	Kernel2DT &kernelOp = convolver.getKernel();

	/// our Sobel-Kernel is 3x3
	kernelOp.resize( 3, 3 );

	/// use an AArray to fill ther Kernel
	/// a normal array would be more efficient, but this looks nicer
	AArray<kernelType, 2> daKernel( 3, 3 ) ;
	daKernel = -1, 0, 1,
	           -2, 0, 2,
	           -1, 0, 1;
	std::copy( daKernel.begin(), daKernel.end(), kernelOp.begin() );

	/// let kernelOp be aware of the modified kernel
	kernelOp.init();

	/// space to store the result
	ImagePlane<uint8_t> convoPlane( srcPlane, false );

	/// convolve each channel separately
	for ( uint i = 0; i < srcPlane.getNumChannels(); ++i )
	{
		srcPlane.setChannelMode( i );
		convoPlane.setChannelMode( i );
		convolver.process( &srcPlane, &convoPlane );
	}

	/// save for visual inspection
	convoPlane.save( "convolved.png" );
}

/**
 * a variation of image convolution
 * this time we have a diagonal Prewitt-Kernel and the convolution is carried out in-place.
 * Also, the convolution area is restricted to a region of interest
 */
void convolve2DInPlace( ImagePlane<uint8_t> &srcPlane )
{
	/// define Kernel and convolver
	typedef int32_t kernelType;
	typedef Kernel2D<kernelType, uint8_t, uint8_t> Kernel2DT;
	Convolver2D<Kernel2DT> convolver( BORDER_COPY );
	Kernel2DT &kernelOp = convolver.getKernel();

	/// resize the internel kernel array
	kernelOp.resize( 3, 3 );

	/// create he actual Prewitt-Kernel
	AArray<kernelType, 2> daKernel( 3, 3 ) ;
	daKernel =  -1, -1, 0,
	            -1, 0, 1,
	            0, 1, 1;
	std::copy( daKernel.begin(), daKernel.end(), kernelOp.begin() );
	
	/// let kernelOp be aware of the modified kernel
	kernelOp.init();

	/// make a copy of the input image
	ImagePlane<uint8_t> convoPlane( srcPlane, true );

	/// save the current region of interest and restore it upon deletion
	UsingROI ur1( &srcPlane );
	UsingROI ur2( &convoPlane );
	int rX = 1;
	int rY = 1;
	int rW = 18;
	int rH = 18;
	srcPlane.setROI( rX, rY, rW, rH );
	convoPlane.setROI( rX, rY, rW, rH );

	/// convolve each channel separately
	for ( uint i = 0; i < srcPlane.getNumChannels(); ++i )
	{
		convoPlane.setChannelMode( i );
		///convolve in-place
		convolver.process( &convoPlane, &convoPlane );
	}

	/// save for visual inspection
	/// note, that the save-routine respects the region of interest
	convoPlane.save( "convolvedInPlace.png" );
}

/**
 * dilates an image in-place
 */
void dilate( ImagePlane<uint8_t> &srcPlane )
{
	/// Our Kernel is an extremum-Kernel that calculates the extremum in its area
	/// for dilation, our extremum shall be the maximum, therefore we use std::greater
	typedef KernelExtremum_3_3<uint8_t, std::greater<uint8_t> > Kernel;
	
	/// create a convolver
	Convolver2D<Kernel> convolver( BORDER_ZERO );

	/// copy the input image
	ImagePlane<uint8_t> convoPlane( srcPlane, true );

	/// convolve each channel separately
	for ( uint i = 0; i < srcPlane.getNumChannels(); ++i )
	{
		srcPlane.setChannelMode( i );
		convoPlane.setChannelMode( i );
		/// convolve in-place
		convolver.process( &convoPlane, &convoPlane );
	}

	/// save for visual inspection
	convoPlane.save( "dilate.png" );
}

/**
 * yet another convolution example, which uses a simple anisotropic Kernel,
 * which replaces the center pixel with the value of the closest average of
 * three adjacent neigbour pixel.
 * If this Kernel is applied many times, the result may look like taken from a comic book
 */
void aniso( ImagePlane<uint8_t> &srcPlane )
{
	/// create Kernel and convolver
	typedef KernelAniso_3_3<uint8_t > Kernel;
	Convolver2D<Kernel> convolver( BORDER_ZERO );

	/// copy input image
	ImagePlane<uint8_t> convoPlane( srcPlane, true );

	/// apply kernel k-times
	for ( int k = 0; k < 5; ++k )
	{
		/// apply for each channel separately
		for ( uint i = 0; i < srcPlane.getNumChannels(); ++i )
		{
			srcPlane.setChannelMode( i );
			convoPlane.setChannelMode( i );
			/// process in-place
			convolver.process( &convoPlane, &convoPlane );
		}
	}

	/// save for visual inspection
	convoPlane.save( "aniso.png" );
}

/**
 * resamples an image by mapping each pixel to its polar (rho, phi)-coordinate relative to the center of the region of interest
 */
void polar( ImagePlane<uint8_t> &srcPlane )
{
	/// allocate the destination image
	ImagePlane<uint8_t> polarPlane( srcPlane.getROI().getWidth(), srcPlane.getROI().getHeight(), srcPlane.getNumChans() );

	/// instantiate a resample
	PolarResampler<uint8_t> resample;

	/// resample each channel separately
	for ( uint i = 0; i < srcPlane.getNumChannels(); ++i )
	{
		/// select channel
		srcPlane.setChannelMode( i );
		polarPlane.setChannelMode( i );
		
		/// perform resampling
		resample.process( &srcPlane, &polarPlane );
	}

	/// save for visual inspection
	polarPlane.save( "polar.png" );
}

/**
 * remaps each channel of an image separately to the intervall [0;255]
 */
void colorNormalize( ImagePlane<uint8_t> &srcPlane )
{
	/// allocate the destination image
	ImagePlane<uint8_t> dstPlane( srcPlane.getROI().getWidth(), srcPlane.getROI().getHeight(), srcPlane.getNumChans() );

	/// set the upper bound of the range to 255, the lower bound is fixed to 0
	ColorNormalizer<uint8_t> cnorm( 255 );

	/// normalize each channel separately
	/// this is equivalent to a diagonal transform of the color space
	for ( uint i = 0; i < srcPlane.getNumChannels(); ++i )
	{
		/// select channel
		srcPlane.setChannelMode( i );
		dstPlane.setChannelMode( i );
		
		/// perform!
		cnorm.process( &srcPlane, &dstPlane );
	}

	/// save for visual inspection
	dstPlane.save( "cnorm.png" );
}

/**
 * showcases the subsampling of an image
 */
void subSample( ImagePlane<uint8_t> &srcPlane )
{
	/// create a sub-sampler, this one combines 4 pixels into one
	SubSamplerSmooth<uint8_t> sss;
	
	/// create a destination image of suitable dimaensions using the getSubSampleDim-function
	ImagePlane<uint8_t> dstPlane( getSubSampleDim( srcPlane.getROI().getWidth() ), getSubSampleDim( srcPlane.getROI().getHeight() ), srcPlane.getNumChans() );

	/// process each channel separately
	for ( uint i = 0; i < srcPlane.getNumChannels(); ++i )
	{
		srcPlane.setChannelMode( i );
		dstPlane.setChannelMode( i );
		sss.process( &srcPlane, &dstPlane );
	}

	/// save for visual inspection
	dstPlane.save( "subsampled.png" );

	/// process again, but this time in-place
	for ( uint i = 0; i < dstPlane.getNumChannels(); ++i )
	{
		dstPlane.setChannelMode( i );
		sss.process( &dstPlane, &dstPlane );
	}

	/// save for visual inspection
	dstPlane.save( "subsampledAgainInPlace.png" );
}

/**
 * conversion of a color image to a single channel image using a selection of operators
 */
void rgbToScalar( ImagePlane<uint8_t> &srcPlane )
{
	/// make sure we have 3 channels, since the operators assume this
	if ( srcPlane.getNumChans() < 3 ) return;
	
	/// convert the image to a channel-set, which own the channel structures (this is, wat the "true" means)
	ChannelSet<uint8_t> in( srcPlane, true );

	/// create a single channel image to store the result
	ImagePlane<uint8_t> dstPlane( srcPlane.getROI().getWidth(), srcPlane.getROI().getHeight(), 1 );

	/// apply the operators
	/// max( 0, r - ( g + b )/2 )
	RGBtoScalar< RGB2TunedRedOp<uint8_t, uint8_t> > c1;
	c1.process( in, &dstPlane );
	dstPlane.save( "rgb2red.png" );

	/// max( 0, g - ( r + b )/2 )
	RGBtoScalar< RGB2TunedGreenOp<uint8_t, uint8_t> > c2;
	c2.process( in, &dstPlane );
	dstPlane.save( "rgb2green.png" );

	/// max( 0, b - ( g + r )/2 )
	RGBtoScalar< RGB2TunedBlueOp<uint8_t, uint8_t> > c3;
	c3.process( in, &dstPlane );
	dstPlane.save( "rgb2blue.png" );

	/// max( 0, ( ( r + g ) - abs( r - g ) )/2 - b )
	RGBtoScalar< RGB2TunedYellowOp<uint8_t, uint8_t> > c4;
	c4.process( in, &dstPlane );
	dstPlane.save( "rgb2yellow.png" );
}


/**
 * flip and flop an image
 */
void flipper( ImagePlane<uint8_t> &srcPlane )
{
	/// create space for the destination image
	ImagePlane<uint8_t> dstPlane( srcPlane.getROI().getWidth(), srcPlane.getROI().getHeight(), srcPlane.getNumChans() );

	/// create the flipper
	Flipper<uint8_t> flipper;

	/// flip horizontally
	for ( uint i = 0; i < srcPlane.getNumChannels(); ++i )
	{
		srcPlane.setChannelMode( i );
		dstPlane.setChannelMode( i );
		flipper.mirrorH( &srcPlane, &dstPlane );
	}
	/// save for visual inspection
	dstPlane.save( "mirrorH.png" );

	/// flip horizontally in-place
	for ( uint i = 0; i < srcPlane.getNumChannels(); ++i )
	{
		dstPlane.setChannelMode( i );
		flipper.mirrorH( &dstPlane );
	}
	/// save for visual inspection
	dstPlane.save( "mirrorHinplace.png" );

	/// flip vertically
	for ( uint i = 0; i < srcPlane.getNumChannels(); ++i )
	{
		srcPlane.setChannelMode( i );
		dstPlane.setChannelMode( i );
		flipper.mirrorV( &srcPlane, &dstPlane );
	}
	/// save for visual inspection
	dstPlane.save( "mirrorV.png" );

	/// flip vertically in-place
	for ( uint i = 0; i < srcPlane.getNumChannels(); ++i )
	{
		dstPlane.setChannelMode( i );
		flipper.mirrorV( &dstPlane );
	}
	/// save for visual inspection
	dstPlane.save( "mirrorVinplace.png" );

	/// rotate 90deg clockwise
	for ( uint i = 0; i < srcPlane.getNumChannels(); ++i )
	{
		srcPlane.setChannelMode( i );
		dstPlane.setChannelMode( i );
		flipper.rotate90cw( &srcPlane, &dstPlane );
	}
	/// save for visual inspection
	dstPlane.save( "rotate90cw.png" );
}



/**
 * calls all the functions above on an image supplied on the command line
 */
int main( int argc, char *argv[] )
{
	if ( argc < 2 )
	{
		printf( "what image?\n" );
		exit( 0 );
	}

	///read an image from a file
	ImagePlane<uint8_t> srcPlane( argv[1] );

	flipper( srcPlane );

	aniso( srcPlane );

	rgbToScalar(srcPlane);

	subSample( srcPlane );

	colorNormalize( srcPlane );

	polar(srcPlane);

	gauss(srcPlane);

	prewitt(srcPlane);

	convolve2D(srcPlane);

	convolve2DInPlace(srcPlane);

	dilate(srcPlane);

	border(srcPlane);

	imageIterator(srcPlane);

	///pre-allocate space for dizenzo gradients, which are used in the last examples
	
	ImagePlane<ggType> gxx(srcPlane.getWidth(), srcPlane.getHeight(), 1);
	ImagePlane<ggType> gyy(srcPlane.getWidth(), srcPlane.getHeight(), 1);
	ImagePlane<ggType> gxy(srcPlane.getWidth(), srcPlane.getHeight(), 1);

	dizenzo(srcPlane, gxx, gyy, gxy);

	harris(gxx, gyy, gxy);

	canny(gxx, gyy, gxy);

	return EXIT_SUCCESS;
}


